// Scintilla source code edit control
// sciwrappers.cxx - Wrapper functions around Scintilla msg api.
// Copyright (c) 2013 Mentor Graphics Corporation
// The license.terms file describes the conditions under which this software may be distributed.

#include <string.h>

#include "sciwrappers.h"

#define _UNUSED_ 0

// macro to simplify calling scintilla message API
#define SSM(sci, m, w, l) scintilla_send_message(sci, m, w, l)

// macro to convert an XColor value for scintilla
#define _XRGB_(c) (c->red & 0xff) | (c->green & 0xff)<<8 | (c->blue & 0xff)<<16


sptr_t sci_cmd(ScintillaObject *sci, int msg, uptr_t wparam, sptr_t lparam)
{
   return SSM(sci, msg, wparam, lparam);
}


//#############################################################################
// ACTIONs
//#############################################################################

//=============================================================================
// sci_add_text
//
// Add new text at the specified document position
//-----------------------------------------------------------------------------
void sci_add_text(ScintillaObject *sci, sptr_t pos, const char *s)
{
    // check if the add is being done at the insertion point, and if it
    // is then we want the add to cause the insertion point to be moved
    // to the end of the added text.
    sptr_t currpos = sci_get_current_pos(sci);
    //fprintf(stderr,"add: currpos %d pos %d\n", currpos, pos);

    // inserting text clears the current selection, remember what the selection was
    int sel_beg = sci_cmd(sci, SCI_GETSELECTIONSTART, _UNUSED_, _UNUSED_);
    int sel_end = sci_cmd(sci, SCI_GETSELECTIONEND,   _UNUSED_, _UNUSED_);
    if (pos == currpos) {
        SSM(sci, SCI_ADDTEXT, strlen(s), (sptr_t)s); // moves insert pt after the add
    } else {
        SSM(sci, SCI_INSERTTEXT, pos, (sptr_t)s);
    }
    if (sel_beg != sel_end) {
        // restore selection if one existed before insert
        int len = strlen(s);
        if (pos < sel_beg) {
            // text inserted before selection
            // adjust selection offsets so they remain on same text
            sel_beg += len;
            sel_end += len;
        } else if ((pos >= sel_beg) && (pos <= sel_end)) {
            // text inserted within selection
            // "grow" selection by the added text
            sel_end += len;
        }
        sci_cmd(sci, SCI_SETSELECTIONSTART, sel_beg, _UNUSED_);
        sci_cmd(sci, SCI_SETSELECTIONEND,   sel_end, _UNUSED_);
    }
 }

//=============================================================================
// sci_delete_text
//
// Deletes text between two positions
//-----------------------------------------------------------------------------
void sci_delete_text(ScintillaObject *sci, sptr_t pos1, sptr_t pos2)
{
   if (pos1 == pos2) return;
   if (pos1 > pos2) {
      int tmp = pos2;
      pos2 = pos1;
      pos1 = tmp;
   }
   SSM(sci, SCI_SETTARGETSTART, pos1, 0);
   SSM(sci, SCI_SETTARGETEND,   pos2, 0);
   SSM(sci, SCI_REPLACETARGET, 0, (sptr_t)"");
}

//=============================================================================
// sci_enable_margin_click
//
// Set a margin to receive (or disable) click notifications
//-----------------------------------------------------------------------------
void sci_enable_margin_click(ScintillaObject *sci, int m, bool enable)
{
   SSM(sci, SCI_SETMARGINSENSITIVEN, m, (sptr_t)enable);
}

//=============================================================================
// sci_goto_pos
//
// Change current position and scroll(if needed) to see it
//-----------------------------------------------------------------------------
void sci_goto_pos(ScintillaObject *sci, sptr_t pos)
{
    int sel_beg = sci_cmd(sci, SCI_GETSELECTIONSTART, _UNUSED_, _UNUSED_);
    int sel_end = sci_cmd(sci, SCI_GETSELECTIONEND,   _UNUSED_, _UNUSED_);
    int end = sci_get_length(sci);

    if (pos && pos == end) pos--;

    //fprintf(stderr,"sci_goto_pos %d\n", pos);
    SSM(sci, SCI_GOTOPOS, pos, 0); // this will clear the selection!!!

    if (sel_beg != sel_end) {
        // restore selection
        SSM(sci, SCI_SETSELECTIONSTART, sel_beg, _UNUSED_);
        SSM(sci, SCI_SETSELECTIONEND,   sel_end, _UNUSED_);
    }
}

//=============================================================================
// sci_has_selection
//
// Return whether any text is selected
//-----------------------------------------------------------------------------
bool sci_has_selection(ScintillaObject *sci)
{
    sptr_t pos1 = SSM(sci, SCI_GETSELECTIONSTART, 0, 0);
    sptr_t pos2 = SSM(sci, SCI_GETSELECTIONEND, 0, 0);
    return (pos1 != pos2);
}

//=============================================================================
// sci_redo
//-----------------------------------------------------------------------------
void sci_redo(ScintillaObject *sci)
{
   SSM(sci, SCI_REDO, 0, 0);
}

//=============================================================================
// sci_reset_undo_history
//
// Clears the undo/redo history buffers
//-----------------------------------------------------------------------------
void sci_reset_undo_history(ScintillaObject *sci)
{
   SSM(sci, SCI_EMPTYUNDOBUFFER, 0, 0);
}

//=============================================================================
// sci_select
//
// Selects a range of text. If additive is TRUE then a "alternate"
// selection is created (preserving the "main" selection)
//-----------------------------------------------------------------------------
void sci_select(ScintillaObject *sci, sptr_t pos1, sptr_t pos2, bool additive)
{
   int end = sci_get_length(sci);
   if ((pos2 == end) && (pos2 > 0)) {
      int lastChar =  SSM(sci, SCI_GETCHARAT, pos2-1, 0);
      // don't include final end of line at end of document
      if (lastChar == '\n')
         --pos2;
   }
   if (additive) {
      // Need special handling to do additive tagging for the selection
      // tag since Scintilla will by default remove an existing selection
      // before setting a new one.
      SSM(sci, SCI_ADDSELECTION, pos1, pos2);
   } else {
      SSM(sci, SCI_SETANCHOR,     pos1, 0);
      SSM(sci, SCI_SETCURRENTPOS, pos2, 0);
   }
}


//=============================================================================
// sci_show_ws
//
// Control display of whitespace indicators
//-----------------------------------------------------------------------------
void sci_show_ws(ScintillaObject *sci, bool show)
{
   if (show)
      SSM(sci, SCI_SETVIEWWS, SCWS_VISIBLEALWAYS, 0);
   else
      SSM(sci, SCI_SETVIEWWS, SCWS_INVISIBLE, 0);
}

//=============================================================================
// sci_toggle_fold
//
// Toggle the fold level for the specfied line
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
void sci_toggle_fold(ScintillaObject *sci, int line)
{
   SSM(sci, SCI_TOGGLEFOLD, line, 0);
}

//=============================================================================
// sci_undo
//-----------------------------------------------------------------------------
void sci_undo(ScintillaObject *sci)
{
   SSM(sci, SCI_UNDO, 0, 0);
}

//=============================================================================
// sci_zoom_in/sci_zoom_out
//
// Causes the document font size to be increased/decreased by a set amount
//-----------------------------------------------------------------------------
void sci_zoom_in(ScintillaObject *sci)
{
   SSM(sci, SCI_ZOOMIN, 0, 0);
}
void sci_zoom_out(ScintillaObject *sci)
{
   SSM(sci, SCI_ZOOMOUT, 0, 0);
}

//#############################################################################
// GETs
//#############################################################################

//=============================================================================
// sci_get_column
//
// Return the column position from a document position
//-----------------------------------------------------------------------------
int sci_get_column(ScintillaObject *sci, sptr_t pos)
{
   int end = sci_get_length(sci);
   if (end && pos == end) pos--;
   return SSM(sci, SCI_GETCOLUMN, pos, 0);
}

//=============================================================================
// sci_get_current_pos
//
// Returns the current document position(ie where insertion point is)
//-----------------------------------------------------------------------------
sptr_t sci_get_current_pos(ScintillaObject *sci)
{
   return SSM(sci, SCI_GETCURRENTPOS, 0, 0);
}

//=============================================================================
// sci_get_foldlevel
//
// Return the fold level for the specified line number
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
int sci_get_foldlevel(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_GETFOLDLEVEL, line, 0);
}

//=============================================================================
// sci_get_is_modified
//
// Return if the document is currently modified(1) or not(0)
//-----------------------------------------------------------------------------
int sci_get_is_modified(ScintillaObject *sci)
{
   return SSM(sci, SCI_GETMODIFY, 0, 0);
}

//=============================================================================
// sci_get_length
//
// Return the number of characters in the document
//-----------------------------------------------------------------------------
sptr_t sci_get_length(ScintillaObject *sci)
{
   return SSM(sci, SCI_GETLENGTH, 0, 0);
}

//=============================================================================
// sci_get_line
//
// Return line number for a document position
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
sptr_t sci_get_line(ScintillaObject *sci, sptr_t pos)
{
   return SSM(sci, SCI_LINEFROMPOSITION, pos, 0);
}

//=============================================================================
// sci_get_line_count
//
// Return the number of lines in the document
//-----------------------------------------------------------------------------
sptr_t sci_get_line_count(ScintillaObject *sci)
{
   return SSM(sci, SCI_GETLINECOUNT, 0, 0);
}

//=============================================================================
// sci_get_line_length
//
// Return length(in characters) of a line
//-----------------------------------------------------------------------------
int sci_get_line_length(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_LINELENGTH, line, 0);
}

//=============================================================================
// sci_get_line_markers
//
// Return mask indicating which markers exist on the specified line
//-----------------------------------------------------------------------------
int sci_get_line_markers(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_MARKERGET, line, 0);
}

//=============================================================================
// sci_get_lineend_pos
//
// Return the document position for a line
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
sptr_t sci_get_lineend_pos(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_GETLINEENDPOSITION, line, 0);
}

//=============================================================================
// sci_get_margin_mask
//
// Return the margin mask for a specific margin
//-----------------------------------------------------------------------------
int sci_get_margin_mask(ScintillaObject *sci, int m)
{
   return SSM(sci, SCI_GETMARGINMASKN, m, 0);
}

//=============================================================================
// sci_get_margin_width
//
// Get the width(in pixels) of the specified margin
//-----------------------------------------------------------------------------
int sci_get_margin_width(ScintillaObject *sci, int m)
{
   return SSM(sci, SCI_GETMARGINWIDTHN, m, 0);
}

//=============================================================================
// sci_get_pos
//
// Return document position for specified line number
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
sptr_t sci_get_pos(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_POSITIONFROMLINE, line, 0);
}

//=============================================================================
// sci_get_pos_from_xy
//
// Return document position for specified X/Y screen position
// If "close" is true, then will only match if X/Y is actually determined
// to be "close enough" to a character.
//-----------------------------------------------------------------------------
sptr_t sci_get_pos_from_xy(ScintillaObject *sci, int x, int y, bool close)
{
    // in order to match the behavior of Tk's Text widget, we want to
    // ask for the character position at x/y; otherwise, what we will
    // get from Scintilla is the inter-character position.

    if (close)
        return SSM(sci, SCI_CHARPOSITIONFROMPOINTCLOSE, x, y);

    return SSM(sci, SCI_CHARPOSITIONFROMPOINT, x, y);
}

//=============================================================================
// sci_get_text_height
//
// Return the height(in pixels) of the specified line
//
// NOTE: LINE NUMBER IS 0-BASED
//-----------------------------------------------------------------------------
int sci_get_text_height(ScintillaObject *sci, int line)
{
   return SSM(sci, SCI_TEXTHEIGHT, line, 0);
}

//=============================================================================
// sci_get_text_width
//
// Return the text width, using current default style, of an arbitrary string
//-----------------------------------------------------------------------------
int sci_get_text_width(ScintillaObject *sci, const char *str)
{
   return (int)SSM(sci, SCI_TEXTWIDTH, STYLE_DEFAULT, (sptr_t)str);
}

//=============================================================================
// sci_get_x_from_pos
//
// Return the screen X location corresponding to a document position
//-----------------------------------------------------------------------------
int sci_get_x_from_pos(ScintillaObject *sci, sptr_t pos)
{
    return SSM(sci, SCI_POINTXFROMPOSITION, 0/*unused*/, pos);
}

//=============================================================================
// sci_get_y_from_pos
//
// Return the screen Y location corresponding to a document position
//-----------------------------------------------------------------------------
int sci_get_y_from_pos(ScintillaObject *sci, sptr_t pos)
{
    return SSM(sci, SCI_POINTYFROMPOSITION, 0/*unused*/, pos);
}

//=============================================================================
// sci_get_zoom
//
// Return the current zoom level
//-----------------------------------------------------------------------------
int sci_get_zoom(ScintillaObject *sci)
{
   return SSM(sci, SCI_GETZOOM, 0, 0);
}

//#############################################################################
// SETs
//#############################################################################

//=============================================================================
// sci_set_focus
//-----------------------------------------------------------------------------
void sci_set_focus(ScintillaObject *sci, bool turn_on)
{
   SSM(sci, SCI_SETFOCUS, turn_on, 0);
}

//=============================================================================
// sci_set_fold_flags
//
// The "flags" arg represents a bit-mask of various folding options
//-----------------------------------------------------------------------------
void sci_set_fold_flags(ScintillaObject *sci, int flags)
{
   SSM(sci, SCI_SETFOLDFLAGS, flags, 0);
}

//=============================================================================
// sci_set_font
//
// Propagate a font change to all known styles
//-----------------------------------------------------------------------------
void sci_set_font(ScintillaObject *sci, const char *family, int ptsize, bool bold, bool italic)
{
   for (int i=0; i < STYLE_MAX; i++) {
      SSM(sci, SCI_STYLESETFONT,   i, (sptr_t)family);
      SSM(sci, SCI_STYLESETSIZE,   i, ptsize);
      SSM(sci, SCI_STYLESETBOLD,   i, bold);
      SSM(sci, SCI_STYLESETITALIC, i, italic);
   }
}

//=============================================================================
// sci_set_margin_mask
//
// Set the mask of which markers a margin is allowed to display
//-----------------------------------------------------------------------------
void sci_set_margin_mask(ScintillaObject *sci, int m, int mask)
{
   SSM(sci, SCI_SETMARGINMASKN, m, mask);
}

//=============================================================================
// sci_set_margin_width
//
// Set the width(in pixels) of the specified margin. A value of zero
// will hide the margin.
//-----------------------------------------------------------------------------
void sci_set_margin_width(ScintillaObject *sci, int m, int width)
{
   SSM(sci, SCI_SETMARGINWIDTHN, m, width);
}

//=============================================================================
// sci_set_margin_type
//
// Set the type of a margin
//-----------------------------------------------------------------------------
void sci_set_margin_type(ScintillaObject *sci, int m, int type)
{
   SSM(sci, SCI_SETMARGINTYPEN, m, type);
}

//=============================================================================
// sci_set_padding
//
// Define how many pixels of padding exist between the left and right
// edges of the text (the default Scintilla uses is 1).
//-----------------------------------------------------------------------------
void sci_set_padding(ScintillaObject *sci, int left, int right)
{
   if (right < 0)
      right = left;
   SSM(sci, SCI_SETMARGINLEFT,  0/*unused*/, left);
   SSM(sci, SCI_SETMARGINRIGHT, 0/*unused*/, right);
}

//=============================================================================
// sci_set_readonly
//-----------------------------------------------------------------------------
void sci_set_readonly(ScintillaObject *sci, bool ro)
{
   SSM(sci, SCI_SETREADONLY, ro, 0);
}

//=============================================================================
// sci_set_style_bg
//
// Sets the background color for the specified style
//-----------------------------------------------------------------------------
void sci_set_style_bg(ScintillaObject *sci, int id, XColor *c)
{
   SSM(sci, SCI_STYLESETBACK, id, (c ? _XRGB_(c) : 0));
}

//=============================================================================
// sci_set_style_fg
//
// Sets the foreground color for the specified style
//-----------------------------------------------------------------------------
void sci_set_style_fg(ScintillaObject *sci, int id, XColor *c)
{
   SSM(sci, SCI_STYLESETFORE, id, (c ? _XRGB_(c) : 0));
}

//=============================================================================
// sci_set_tab_width
//-----------------------------------------------------------------------------
void sci_set_tab_width(ScintillaObject *sci, int width)
{
   SSM(sci, SCI_SETTABWIDTH, width, 0);
}
