//#####################################################################
//#
//# Implementation of a Tk widget that encapsulates the Scintilla
//# open-source text editor
//#
//# NOTE: DUE TO TECHNICAL CHALLENGES WITH MANAGING THE SCINTILLA
//#       DRAWING, WE DO __NOT__ SUPPORT A BORDER OR HIGHLIGHT
//#       RECTANGLE AS PART OF THIS NEW WIDGET.
//#
//#####################################################################
//
// Portions copied from tkText.c --
//
// Copyright (c) 1992-1994 The Regents of the University of California.
// Copyright (c) 1994-1996 Sun Microsystems, Inc.
// Copyright (c) 1999 by Scriptics Corporation.
//
// See the file "license.terms" for information on usage and redistribution of
// this file, and for a DISCLAIMER OF ALL WARRANTIES.
///
//
// Copyright (c) 2013 Mentor Graphics Corporation
// The license.terms file describes the conditions under which this software may be distributed.

//#define DEBUG_TAGLOCATIONS
#define CHECK_TAGLOCATIONS
//#define ENABLE_TAG_PRIORITIES

#include <assert.h>
#include <tcl.h>
#include <tk.h>
#include <tkInt.h> // for access to "region" features

#ifdef _WIN32
// We normally could get the F_OK constant by including the tclInt.h header
// but doing so ends up creating a bizarre link error complaining about some
// C++ exception symbols (eg std::_exception::*).
// So to avoid this grief I'm just getting the constant set explicitly.
//#include <tclInt.h>
#ifndef F_OK
#define F_OK 00
#endif
#endif

#include "sciwrappers.h"
#include "scintilla-ext.h"
#include "scintilla-cmd.h"

#include <algorithm>
#include <vector>

#ifndef F_OK
#define F_OK 0
#endif

Tcl_Interp *scintilla_interp = NULL;

// create our namespace
namespace TkSci {

#ifdef DEBUG
static void _DBG(const char *format, ...) {
   char buffer[2000];
   va_list pArguments;
   va_start(pArguments, format);
   vsprintf(buffer, format, pArguments);
   va_end(pArguments);
   fprintf(stderr, "%s", buffer);
}
#else
#define _DBG(a)
#endif

#ifdef CHECK_TAGLOCATIONS
static void checkLocations(std::vector<TagLocation*> &locations)
{
sptr_t lastloc = (sptr_t)-1;
TagLocation *tp;
    for (std::vector<TagLocation*>::iterator it = locations.begin(); it != locations.end(); ++it) {
        tp = *it;
        assert(tp->startpos() > lastloc);
        assert(tp->startpos() <= tp->endpos());
        lastloc = tp->endpos();
    }
}
#else
#define checkLocations(arg) /*nothing*/
#endif

//--------------------------------------------------------------------------
// local convenience macros
#define SSM(m, w, l) scintilla_send_message(sciwidgetPtr->editor, m, w, l)

// set color based on 0-255 values for RGB
#define _RGB_(r,g,b) r | (g << 8) | (b << 16)

// next one takes an XColor*, which has 16-bit values for RGB
#define _XRGB_(c) (c->red & 0xff) | (c->green & 0xff)<<8 | (c->blue & 0xff)<<16
//--------------------------------------------------------------------------

#define STYLEMASK(sciwidgetPtr) (~((~0U)<<sciwidgetPtr->styleBits))
#define SID2STYLE(sciwidgetPtr, sid) (sid&STYLEMASK(sciwidgetPtr))
#define SID2INDICATOR(sciwidgetPtr, sid) (sid>>sciwidgetPtr->styleBits)

// Tk doesn't have a TK_OPTION type defined for "state"
// so we need to do it ourselves
static const char *stateStrings[] = {
    "busy", "disabled", "normal", "readonly", NULL
};
enum {
    STATE_BUSY, STATE_DISABLED, STATE_NORMAL, STATE_READONLY
};

// constants for fold margin styles
static const char *foldStrings[] = {
   "disabled", "boxtree", "plusminus", "arrow", "circletree", NULL
};
enum FM_ID {
   foldDISABLED,   /* indicates folding is not used by client
                    * NOTE: This allows for a larger number of marker IDs to be
                    *       used by "symbol" margins
                    */
   foldBoxTree,    /* classic tree connections, w/ +- boxes and connection lines */
   foldPlusMinus,  /* uses just "+" and "-" symbols */
   foldArrow,      /* uses just "arrow right" and "arrow down" symbols */
   foldCircleTree, /* another classic style, using circles instead of boxes */
   foldMAX         /* not a style - indicates max identifier */
};

// mimic what's done in tkText.c to indicate when an option 
// is changed that _could_ impact pixel calculations
#define TK_SCI_LINE_GEOMETRY 1

// The next set of constants uniquely identify the config options that
// we want to be able to check for following a call to Tk_SetOptions.
// Note: the following need to be unique from the TK_SCI_LINE_GEOMETRY id.
#define OPT_BG      1 << 1
#define OPT_FG      1 << 2
#define OPT_FONT    1 << 3
#define OPT_LANG    1 << 4
#define OPT_FOLDM   1 << 5
#define OPT_WS      1 << 6
#define OPT_TABW    1 << 7
#define OPT_STATE   1 << 8
#define OPT_MBG     1 << 9
#define OPT_MFG     1 << 10
#define OPT_CURSOR  1 << 11

static Tk_OptionSpec optionSpecs[] = {
   {
      // THIS ONE IS ONLY FOR COMPATABILITY WITH TK TEXT WIDGET
      // This allows use of existing Text class binding code
      // without an error being thrown.
      TK_OPTION_BOOLEAN, "-autoseparators", NULL, NULL,
          "1", -1, Tk_Offset(SciWidget, autoseps), 0,0,0
      },
      {TK_OPTION_BORDER,                  /* type */
         "-background",                     /* switch name */
         "textbackground",                  /* database name */
         "TextBackground",                  /* database Class name */
         "#d9d9d9",                         /* default value */
         -1,                                /* object offset (rarely used) */
         Tk_Offset(SciWidget, background),  /* internal offset (used most) */
         TK_OPTION_NULL_OK,                 /* flags */
         0,                                 /* clientData */
         OPT_BG                             /* type mask */
         },
      {TK_OPTION_SYNONYM, "-bd", "borderwidth", NULL, 
            NULL, 0, -1, 0, (ClientData)"-borderwidth", 0
      },
      {TK_OPTION_SYNONYM, "-bg", "background", NULL, 
            NULL, 0, -1, 0, (ClientData)"-background", 0
      },
      {TK_OPTION_BOOLEAN, "-blink", NULL, NULL,
          "1", -1, Tk_Offset(SciWidget, blinkcaret), 0,0,0
      },
      {TK_OPTION_PIXELS,  "-borderwidth", "borderwidth", "BorderWidth",
            "5", -1, Tk_Offset(SciWidget, borderWidth), 0, 0, TK_SCI_LINE_GEOMETRY
      },
      {TK_OPTION_CURSOR,  "-busycursor", "cursor", "Cursor",
            "watch", -1, Tk_Offset(SciWidget, busycursor), TK_OPTION_NULL_OK, 0, OPT_CURSOR
      },
      {TK_OPTION_CURSOR,  "-cursor", "cursor", "Cursor",
            "xterm", -1, Tk_Offset(SciWidget, cursor), TK_OPTION_NULL_OK, 0, OPT_CURSOR
      },
      {TK_OPTION_BOOLEAN, "-exportselection", "exportSelection", "ExportSelection",
          "1", -1, Tk_Offset(SciWidget, exportSelection), 0,0,0
      },
      {TK_OPTION_SYNONYM, "-fg", "foreground", NULL, 
            NULL, 0, -1, 0, (ClientData)"-foreground", 0
      },
      {TK_OPTION_STRING_TABLE, "-foldstyle", "foldstyle", "FoldStyle",
            "boxtree", -1, Tk_Offset(SciWidget, foldstyle), 0, (ClientData) foldStrings, OPT_FOLDM
      },
      {TK_OPTION_FONT,    "-font", "textfont", "TextFont", /*<< avoid using generic "Font" resource name*/
            "courier", -1, Tk_Offset(SciWidget, tkfont), 0, 0, (OPT_FONT|TK_SCI_LINE_GEOMETRY)
      },
      {TK_OPTION_COLOR,   "-foreground", "foreground", "Foreground",
            "white", -1, Tk_Offset(SciWidget, foreground), 0, 0, OPT_FG
      },
      {TK_OPTION_INT,     "-height", "height", "Height",
            "0", -1, Tk_Offset(SciWidget, reqHeight), 0, 0, 0
      },
      {TK_OPTION_COLOR,   "-highlight", "highlight", "Highlight",
            "yellow", -1, Tk_Offset(SciWidget, highlight), 0, 0, 0
      },
      {TK_OPTION_INT,     "-highlightthickness", "highlightthickness", "HighlightThickness",
            "0", -1, Tk_Offset(SciWidget, highlightWidth), 0, 0, 0
      },
      {TK_OPTION_COLOR,   "-highlightbackground", "highlightbackground", "HighlightBackground",
            "pink", -1, Tk_Offset(SciWidget, highlightBg), 0, 0, 0
      },
      {TK_OPTION_STRING,  "-language", "language", "Language",
            "none", -1, Tk_Offset(SciWidget, language), 0, 0, OPT_LANG
      },
      {TK_OPTION_STRING,  "-linesaddedcommand", "linesAddedCommand", "LinesAddedCommand",
            "", -1, Tk_Offset(SciWidget, linesAddedCmd), TK_OPTION_NULL_OK, 0, 0
      },
      {TK_OPTION_COLOR,   "-marginbackground", "marginBackground", "MarginBackground",
            "#d3d3d3", -1, Tk_Offset(SciWidget, marginBg), 0, 0, OPT_MBG
      },
      {TK_OPTION_SYNONYM, "-marginbg", NULL, NULL, NULL, 0, -1, 0, (ClientData)"-marginbackground", 0},

      {TK_OPTION_COLOR,   "-marginforeground", "marginForeground", "MarginForeground",
            "black", -1, Tk_Offset(SciWidget, marginFg), 0, 0, OPT_MFG
      },
      {TK_OPTION_SYNONYM, "-marginfg", NULL, NULL, NULL, 0, -1, 0, (ClientData)"-marginforeground", 0},

      {TK_OPTION_STRING,  "-marginwidthcommand", NULL, NULL,
            "", -1, Tk_Offset(SciWidget, marginWidthCmd), TK_OPTION_NULL_OK, 0, 0
      },
      {TK_OPTION_STRING,  "-moveupdatecommand", "moveUpdateCommand", "MoveUpdateCommand",
            "", -1, Tk_Offset(SciWidget, moveUpdateCmd), TK_OPTION_NULL_OK, 0, 0
      },
      {TK_OPTION_RELIEF,  "-relief", "relief", "Relief", 
            "flat", -1, Tk_Offset(SciWidget, relief), 0, 0, 0
      },
      {TK_OPTION_BOOLEAN, "-showws", "showws", "ShowWs", 
            "0", -1, Tk_Offset(SciWidget, showws), 0, 0, OPT_WS
      },
      {TK_OPTION_STRING_TABLE, "-state", "state", "State",
            "normal", -1, Tk_Offset(SciWidget, state), 0, (ClientData) stateStrings, OPT_STATE
      },
      {TK_OPTION_INT,     "-tabwidth", "tabwidth", "TabWidth",
            "2", -1, Tk_Offset(SciWidget, tabwidth), 0, 0, OPT_TABW
      },
      {TK_OPTION_BOOLEAN, "-takefocus", "takeFocus", "TakeFocus", 
            "1", -1, Tk_Offset(SciWidget, takefocus), 0, 0, 0
      },
      {TK_OPTION_STRING,  "-xscrollcommand", "xScrollCommand", "ScrollCommand",
            "", -1, Tk_Offset(SciWidget, xScrollCmd), TK_OPTION_NULL_OK, 0, 0
      },
      {TK_OPTION_STRING,  "-yscrollcommand", "yScrollCommand", "ScrollCommand",
            "", -1, Tk_Offset(SciWidget, yScrollCmd), TK_OPTION_NULL_OK, 0, 0
      },
      {TK_OPTION_BOOLEAN, "-undo", NULL, NULL,
            "0", -1, Tk_Offset(SciWidget, undo), 0, 0, 0
      },
      {TK_OPTION_INT,     "-width", "width", "Width",
            "0", -1, Tk_Offset(SciWidget, reqWidth), 0, 0, 0
      },
      {TK_OPTION_END}
};


// The following code was taken from tk/generic/tkText.c
// It was written to be possibly shared outside the context of tkText.
// BSG - Sept. 2012
//Begin copy-

/*
 * These three typedefs, the structure and the SearchPerform, SearchCore
 * functions below are used for line-based searches of the text widget, and,
 * in particular, to handle multi-line matching even though the text widget is
 * a single-line based data structure. They are completely abstracted away
 * from the Text widget internals, however, so could easily be re-used with
 * any line-based entity to provide multi-line matching.
 *
 * We have abstracted this code away from the text widget to try to keep Tk as
 * modular as possible.
 */

struct SearchSpec;  /* Forward declaration. */

typedef ClientData  SearchAddLineProc(int lineNum,
			    struct SearchSpec *searchSpecPtr,
			    Tcl_Obj *theLine, int *lenPtr,
			    int *extraLinesPtr);
typedef int		SearchMatchProc(int lineNum,
			    struct SearchSpec *searchSpecPtr,
			    ClientData clientData, Tcl_Obj *theLine,
			    int matchOffset, int matchLength);
typedef int		SearchLineIndexProc(Tcl_Interp *interp,
			    Tcl_Obj *objPtr, struct SearchSpec *searchSpecPtr,
			    int *linePosPtr, int *offsetPosPtr);

typedef struct SearchSpec {
    int exact;			/* Whether search is exact or regexp. */
    int noCase;			/* Case-insenstivive? */
    int noLineStop;		/* If not set, a regexp search will use the
				 * TCL_REG_NLSTOP flag. */
    int overlap;		/* If set, results from multiple searches
				 * (-all) are allowed to overlap each
				 * other. */
    int strictLimits;		/* If set, matches must be completely inside
				 * the from,to range. Otherwise the limits
				 * only apply to the start of each match. */
    int all;			/* Whether all or the first match should be
				 * reported. */
    int startLine;		/* First line to examine. */
    int startOffset;		/* Index in first line to start at. */
    int stopLine;		/* Last line to examine, or -1 when we search
				 * all available text. */
    int stopOffset;		/* Index to stop at, provided stopLine is not
				 * -1. */
    int numLines;		/* Total lines which are available. */
    int backwards;		/* Searching forwards or backwards. */
    Tcl_Obj *varPtr;		/* If non-NULL, store length(s) of match(es)
				 * in this variable. */
    Tcl_Obj *countPtr;		/* Keeps track of currently found lengths. */
    Tcl_Obj *resPtr;		/* Keeps track of currently found locations */
    int searchElide;		/* Search in hidden text as well. */
    SearchAddLineProc *addLineProc;
				/* Function to call when we need to add
				 * another line to the search string so far */
    SearchMatchProc *foundMatchProc;
				/* Function to call when we have found a
				 * match. */
    SearchLineIndexProc *lineIndexProc;
				/* Function to call when we have found a
				 * match. */
    ClientData clientData;	/* Information about structure being searched,
				 * in this case a text widget. */
} SearchSpec;

static SearchMatchProc		SciSearchFoundMatch;
static SearchAddLineProc	SciSearchAddNextLine;
static SearchLineIndexProc	SciSearchGetLineIndex;

/*
 * The text-widget-independent functions which actually perform the search,
 * handling both regexp and exact searches.
 */

static int		SearchCore(Tcl_Interp *interp,
			    SearchSpec *searchSpecPtr, Tcl_Obj *patObj);
static int		SearchPerform(Tcl_Interp *interp,
			    SearchSpec *searchSpecPtr, Tcl_Obj *patObj,
			    Tcl_Obj *fromPtr, Tcl_Obj *toPtr);

//End Copy-

// return the ScintillaObject for the widget
#define SCIOBJ(sci) sci->editor

// map from a Scintilla position to Tk index and put it into the
// specified Tcl interpreter's result
static void PosToIndex(Tcl_Interp *interp, ScintillaObject *obj, int pos, Tcl_Obj *listObj = NULL, Tcl_DString *dstr = NULL) {

   int line = sci_get_line(obj, pos);
   int beg  = sci_get_pos(obj, line);
   int col  = pos - beg;
   char buf[20];
   sprintf(buf, "%d.%d", LINE2TK(line), col);
   if (listObj) {
       Tcl_ListObjAppendElement(interp, listObj, Tcl_NewStringObj(buf, strlen(buf)));
   } else if (dstr) {
       Tcl_DStringAppendElement(dstr, buf);
   } else if (interp) {
       Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, strlen(buf)));
   }
}

// convert line/column to Tk index. NOTE: Scintilla line numbers must have LINE2TK applied to them because they
// are zero based.
static void LineColumnToIndex(int line, int col, Tcl_DString *dstr)
{
   char buf[20];
   sprintf(buf, "%d.%d", line, col);
   Tcl_DStringAppendElement(dstr, buf);
}

// convenience function to allow consistent reporting of bad indices
static int BadIndex(Tcl_Interp *interp, const char *index_str) {

   Tcl_ResetResult(interp);
   Tcl_AppendResult(interp, "bad text index \"", index_str, "\"", NULL);
   return TCL_ERROR;
}

// return nmemonic name for indicator
static char* IndicatorName(int indicator)
{
    switch (indicator) {
        case INDIC_PLAIN:
            return "plain";
            break;
        case INDIC_SQUIGGLE:
            return "squiggle";
            break;
        case INDIC_TT:
            return "tt";
            break;
        case INDIC_DIAGONAL:
            return "diagonal";
            break;
        case INDIC_STRIKE:
            return "strike";
            break;
        case INDIC_HIDDEN:
            return "hidden";
            break;
        case INDIC_BOX:
            return "box";
            break;
        case INDIC_ROUNDBOX:
            return "roundbox";
            break;
        case INDIC_STRAIGHTBOX:
            return "straightbox";
            break;
        case INDIC_DASH:
            return "dash";
            break;
        case INDIC_DOTS:
            return "dots";
            break;
        case INDIC_SQUIGGLELOW:
            return "squigglelow";
            break;
        case INDIC_DOTBOX:
            return "dotbox";
            break;
        case INDIC_SQUIGGLEPIXMAP:
            return "squigglepixmap";
            break;
    }
    return "Unknown";
}

static int ParseIndex(Tcl_Interp *interp, ScintillaObject *sciobj, const char* idxStr, int &line, int &col)
{
    if (idxStr == NULL || !*idxStr) {
        Tcl_AppendResult(interp, "NULL or missing index", (char *)NULL);
        return TCL_ERROR;
    }
    if (isdigit(*idxStr)) {
        // index is line.something
        char *dot;
        line = strtol(idxStr, &dot, 0);
        if (!dot || (*dot != '.')) {
            Tcl_AppendResult(interp, "Index '", idxStr, "' does not contain a column designation", (char *)NULL);
            return TCL_ERROR;
        }
        ++dot;  // increment past dot
        if (isdigit(*dot)) {
            char *end;
            col = strtol(dot, &end, 0);
            if (end && *end) {
                Tcl_AppendResult(interp, "Index '", idxStr, "', unsupported format", (char *)NULL);
                return TCL_ERROR;
            }
        } else if (!strcmp(dot,"end")) {
            // get length of text in margin
            col = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line), 0);
        } else {
            Tcl_AppendResult(interp, "Index '", idxStr, "', unsupported format", (char *)NULL);
            return TCL_ERROR;
        }
    } else if (idxStr[0] == 'e' && (strncmp(idxStr, "end", 3) == 0)) {
        int pos = sci_get_length(sciobj);
        line = LINE2TK(sci_get_line(sciobj, pos));
        int beg  = sci_get_pos(sciobj, line);
        col  = pos - beg;
    } else {
        Tcl_AppendResult(interp, "Index '", idxStr, "' does not begin with line number", (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

// FORWARD DECLS
//--- these provide standard interface to a Tcl custom widget
static int  SciWidgetCmd         (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static int  SciWidgetConfigure   (Tcl_Interp*, SciWidget *, int objc, Tcl_Obj *CONST objv[]);
static void SciWidgetDestroy     (ClientData clientData);
static void SciWidgetGeometry    (SciWidget *);
static void SciWidgetBindProc    (ClientData clientData, XEvent *eventPtr);
static void SciWidgetEventProc   (ClientData clientData, XEvent *eventPtr);
static int  SciWidgetInstanceCmd (ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

//--- others
static int      GetPosition(SciWidget *, CONST char *, sptr_t *, bool);
static Tcl_Obj* GetTextRange(SciWidget *, int, int, bool);
static void     ExtractText(SciWidget *, Tcl_Obj *, int, int);
static void     SortTags(int numTags, TagData **tagArrayPtr);
static int      TagSortProc(const void *first, const void *second);

// search tag locations for range start to end. If add is true, create location(s) and return the first one
// return iterator through it_p that can be used to retrieve additional locations for range via nextLocation calls. 
// If add is false, then only create tags if start-end does not span existing set of tags, return NULL if none found
typedef enum {
    SEARCH_ADD, // searching for an add operation, locations may be created to represent start/end points
    SEARCH_REM, // searching for a remove operation, locations may be split to represent start/end points but new locations will not be added
    SEARCH_IDX  // searching for an index operation, no locations will be created or split
} SearchType;

static void dumpLocations(const char* ident, std::vector<TagLocation*> &tagLocations)
{
    fprintf(stderr,"%s\n", ident);
    int count = 0;
    for (std::vector<TagLocation*>::iterator it = tagLocations.begin(); it != tagLocations.end(); ++it) {
        TagLocation *t_p = *it;
        fprintf(stderr,"   %d %p range %ld-%ld\n", count++, (void*)t_p, t_p->startpos(), t_p->endpos());
    }
}

static TagLocation *nextLocation(std::vector<TagLocation*> &tagLocations, sptr_t start, sptr_t end, std::vector<TagLocation*>::iterator *it_p)
{
    // Iterator pointed to by it_p was returned by searchLocations and/or this routine, client can use this
    // routine to step through TagLocations until the end position is reached
    TagLocation *t_p;

    if ((*it_p) >= tagLocations.end())
        return NULL;
    ++(*it_p);  // advance iterator to next tag location
    if ((*it_p) < tagLocations.end()) {
        t_p = *(*it_p);
        if ((t_p->startpos() >= start) && (t_p->endpos() <= end)) {
            return t_p;
        }
    }
    return NULL;
}

static TagLocation *adjustLocations(const char* ident, std::vector<TagLocation*> &tagLocations, SearchType type, sptr_t start, sptr_t end, long *idx_p)
{
long idx = *idx_p;
std::vector<TagLocation*>::iterator tmp_it = tagLocations.begin() + idx;
TagLocation *t_p = *tmp_it;
TagLocation *newLoc;

    assert(tmp_it >= tagLocations.begin() && tmp_it <= tagLocations.end());
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"adjustLocations start %ld end %ld idx %d tp %p %ld-%ld\n", start, end, idx, t_p, t_p->startpos(), t_p->endpos());
#endif
    if (t_p->startpos() > start) {
        // we've hit a "hole" in the allocated locations.  If add is
        // true, add a missing location, otherwise adjust start
        // position.
        if (type == SEARCH_ADD) {
            sptr_t endpos;
            if (end < t_p->startpos() - 1)
                endpos = end;
            else
                endpos = t_p->startpos() -1;
            newLoc = new TagLocation(NULL, start, endpos);
            tagLocations.insert(tmp_it, newLoc);
            *idx_p = *idx_p + 1;
            ++tmp_it;
            ++idx;
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"adjustLocations1, index %d current location %ld-%ld adding location %p %ld-%ld\n", *idx_p, t_p->startpos(), t_p->endpos(), newLoc, newLoc->startpos(), newLoc->endpos());
            dumpLocations(ident, tagLocations);
#endif
            checkLocations(tagLocations);
            if (endpos < end) {
                adjustLocations(ident, tagLocations, type, newLoc->endpos() + 1, end, &idx);
            }
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"adjustLocations1.5 returning %p\n", newLoc);
#endif
            return newLoc;
        } else {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"adjustLocations2, adjusting start to %ld\n", t_p->startpos());
#endif
            if (t_p->startpos() >= end)
                return t_p;
            start = t_p->startpos();
        }
    }

#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"adjustLocations2.5 start %ld end %ld idx %d tp %p %ld-%ld\n", start, end, idx, t_p, t_p->startpos(), t_p->endpos());
#endif
    if (type != SEARCH_IDX) {
        // adjust existing tag locations so that locations begin/end
        // exactly on start/end locations iterator points to location
        // containing start location. The value of add determines if
        // new locations are created (we're adding tags so new
        // locations should be added as opposed to removing tags were
        // we don't creat locations for non-existant tags)
        
        // case 1: tmp_it points to location that starts before start,
        //         break location in two such that the first part ends
        //         right before start. Then recurse and process rest.
        if (t_p->startpos() < start) {
#ifdef DEBUG_TAGLOCATIONS
            sptr_t oldStart = t_p->startpos();
#endif
            // split the existing location so it ends where the desired range begins
            newLoc = new TagLocation(t_p, start-1);   // create new tag location, intialize it's tags and start position from old
            // replace the old locaton with the new
            *tmp_it = newLoc;
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"adjustLocations3, start 0x%d changing location %p %ld-%ld to %p %ld-%ld\n", start, t_p, oldStart, t_p->endpos(), newLoc, newLoc->startpos(), newLoc->endpos());
            dumpLocations(ident, tagLocations);
#endif
            checkLocations(tagLocations);
            t_p->setStartPos(start);
            ++tmp_it;  // advance iterator to next location in table and insert adjusted location
            tagLocations.insert(tmp_it, t_p);
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"adjustLocations4, adding location %p %ld-%ld\n", t_p, t_p->startpos(), t_p->endpos());
            dumpLocations(ident, tagLocations);
#endif
            checkLocations(tagLocations);
            // pass index of location containing start position
            *idx_p = *idx_p + 1;
            // use copy of index for recursion to prevent update
            idx = *idx_p;
            // recurse to handle rest
            t_p = adjustLocations(ident, tagLocations, type, start, end, &idx);
        } else {
            // case 2: t_p->startpos() == start, location ends before end position
            assert(t_p->startpos() == start);
            idx = *idx_p + 1;
            if (t_p->endpos() < end) {
                if (idx < tagLocations.size()) {
                    // recurse to handle rest
                    adjustLocations(ident, tagLocations, type, t_p->endpos() + 1, end, &idx);
                } else if (type == SEARCH_ADD) {
                    t_p = new TagLocation(NULL, t_p->endpos() + 1, end);
                    tagLocations.push_back(t_p);
#ifdef DEBUG_TAGLOCATIONS
                    fprintf(stderr,"adjustLocations5, adding location %p %ld-%ld\n", t_p, t_p->startpos(), t_p->endpos());
                    dumpLocations(ident, tagLocations);
#endif
                    checkLocations(tagLocations);
                }
            } else if (t_p->endpos() > end) {
                // case 3: t_p->startpos() == start, location ends
                //         after end position, break location in two
                //         such that the first part ends at end, then
                //         insert new location that spans the rest of
                //         the original location
                newLoc = new TagLocation(t_p, end);   // create new tag location, intialize it's tags and start position from old
                // replace the old locaton with the new
                *tmp_it = newLoc;
#ifdef DEBUG_TAGLOCATIONS
                fprintf(stderr,"adjustLocations6, changing idx %d %ld-%ld to %ld-%ld\n", tmp_it - tagLocations.begin(), t_p->startpos(), t_p->endpos(), newLoc->startpos(), newLoc->endpos());
                dumpLocations(ident, tagLocations);
#endif
                checkLocations(tagLocations);
                t_p->setStartPos(end+1);
                if (tmp_it < tagLocations.end()) {
                    ++tmp_it;  // advance iterator to next location in table and insert adjusted location
                    tagLocations.insert(tmp_it, t_p);
                } else {
                    tagLocations.push_back(t_p);
                }
#ifdef DEBUG_TAGLOCATIONS
                fprintf(stderr,"adjustLocations7, adding location %p %ld-%ld\n", t_p, t_p->startpos(), t_p->endpos());
                dumpLocations(ident, tagLocations);
#endif
                checkLocations(tagLocations);
                t_p = newLoc;   // return pointer to location containing startpos
            } else {
                // case 4: location is exact match, nothing needs to be done
            }
        }
    }
    return t_p;
}

static TagLocation *searchLocations(const char* ident, std::vector<TagLocation*> &tagLocations, SearchType type, sptr_t start, sptr_t end, std::vector<TagLocation*>::iterator *it_p)
{
TagLocation *t_p;
long low = 0, mid = 0;
long high = tagLocations.size() - 1;
std::vector<TagLocation*>::iterator tmp_it;

    // do binary search of tagLocations, note the start/end of a location may overlap that of desired location, return
    // first location that contains the start position; it may need to be broken up, return iterator that can be used by
    // the sister method, nextLocation(sptr_t start, sptr_t end, std::vector<TagLocation*>::iterator *it_p)
    // The iterator should point to the entry for the returned TagLocation pointer, or tagLocations.end() if returning NULL.
    if (tagLocations.size() > 0) {
        while (high >= low) {
            mid = low+((high-low)>>1);
            t_p = tagLocations[mid];
            if (t_p->endpos() < start) {
                // location too low, search upper half
                low = mid + 1;
            } else if (t_p->startpos() > start) {
                // location too high, search bottom half
                high = mid - 1;
            } else {
#ifdef DEBUG_TAGLOCATIONS
                fprintf(stderr,"type %d found tag span %d-%d for search %d-%d\n", type, t_p->startpos(), t_p->endpos(), start, end);
#endif
                if ((type != SEARCH_IDX) && ((t_p->startpos() != start) || (t_p->endpos() != end))) {
                    t_p = adjustLocations(ident, tagLocations, type, start, end, &mid);
#ifdef DEBUG_TAGLOCATIONS
                    dumpLocations(ident, tagLocations);
                    fprintf(stderr,"   1returning %p, setting iterator to element %d, size %ld\n", t_p, mid, tagLocations.size());
#endif
                    checkLocations(tagLocations);
                }
                *it_p = tagLocations.begin() + mid;
                return t_p;
            }
        }
    }
    tmp_it = tagLocations.begin() + mid;
    // advance iterator to first location that follows new location
    while (tmp_it < tagLocations.end()) {
        TagLocation *newLoc = *tmp_it;
        if (newLoc->endpos() < start) {
            ++tmp_it;
            ++mid;
        } else
            break;
    }
    // unable to find location containing start, if we're doing search-only, return failure with iterator at proper insertion point
    if (type == SEARCH_IDX) {
        if (tmp_it < tagLocations.end()) {
            TagLocation *newLoc = *tmp_it;
            *it_p = tmp_it;
            if (newLoc->startpos() >= start && newLoc->endpos() <= end) {
#ifdef DEBUG_TAGLOCATIONS
                fprintf(stderr,"found tag span %ld-%ld\n", newLoc->startpos(), newLoc->endpos());
#endif
                return newLoc;
            } else {
                return NULL;
            }
        } else {
            *it_p = tagLocations.end();
            return NULL;
        }
    }
    // couldn't find location containing start, create one being mindful that there may be locations in table
    // overlapping locations before the end locations
    if (tmp_it < tagLocations.end()) {
        TagLocation *newLoc = *tmp_it;
        long tmp_idx;
        if (newLoc->startpos() <= end) {
            // use a copy of mid so that adjustLocations does not alter mid value
            tmp_idx = mid;
            if (type == SEARCH_ADD) {
                // create new location to cover area from start to existing location
                t_p = new TagLocation(NULL, start, newLoc->startpos() - 1);
                tagLocations.insert(tmp_it, t_p);
#ifdef DEBUG_TAGLOCATIONS
                fprintf(stderr,"search1: adding location %ld-%ld ahead of existing location %p %ld-%ld\n", t_p->startpos(), t_p->endpos(), newLoc, newLoc->startpos(), newLoc->endpos());
                dumpLocations(ident, tagLocations);
#endif
                checkLocations(tagLocations);
                ++tmp_idx;
            } else {
                t_p = newLoc;
            }
            adjustLocations(ident, tagLocations, type, newLoc->startpos(), end, &tmp_idx);
#ifdef DEBUG_TAGLOCATIONS
            dumpLocations(ident, tagLocations);
            fprintf(stderr,"   2returning %p, setting iterator to element %d\n", t_p, mid);
#endif
            checkLocations(tagLocations);
            // inserting may have reallocated vector, re-init iterator to be sure it's valid
            *it_p = tagLocations.begin() + mid;
            return t_p;
        } else if (type == SEARCH_ADD) {
            t_p = new TagLocation(NULL, start, end);
            tagLocations.insert(tmp_it, t_p);
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"search2: unable to find location %ld-%ld, creating one\n", start, end);
            dumpLocations(ident, tagLocations);
            fprintf(stderr,"   3returning %p, setting iterator to element %d\n", t_p, mid);
#endif
            checkLocations(tagLocations);
            *it_p = tagLocations.begin() + mid;
            return t_p;
        }
    } else if (type == SEARCH_ADD) {
        // simple case, no existing locations overlap any portion of new range
        t_p = new TagLocation(NULL, start, end);
        tagLocations.push_back(t_p);
        mid = tagLocations.end() - tagLocations.begin() - 1;
#ifdef DEBUG_TAGLOCATIONS
        fprintf(stderr,"search3: unable to find location %ld-%ld, creating one\n", start, end);
        dumpLocations(ident, tagLocations);
        fprintf(stderr,"   returning %p, setting iterator to element %d\n", t_p, mid);
#endif
        checkLocations(tagLocations);
        *it_p = tagLocations.begin() + mid;
        return t_p;
    }
    *it_p = tagLocations.end();
    return NULL;
}


/***********************************************************
 * Scintillatk_Init is called when the package is loaded.
 ***********************************************************/
#define SCI_VERSION  PACKAGE_VERSION
#define SCI_VER_DATE PACKAGE_DATE

extern "C" int Scintillatk_Init(Tcl_Interp *interp) {

   if (Tcl_InitStubs(interp, "8.5", 0) == NULL) { return TCL_ERROR; }
   if (Tk_InitStubs (interp, "8.5", 0) == NULL) { return TCL_ERROR; }

   Tcl_CreateObjCommand(interp, "scintilla", SciWidgetCmd, (ClientData)NULL, SciWidgetDestroy);

   Tcl_PkgProvide(interp, "ScintillaTk", SCI_VERSION);
   scintilla_interp = interp;
   return TCL_OK;
}

static void SciWidgetDeleteCmd(void *data)
{
SciWidget *sciwidgetPtr = (SciWidget *)data;

    if (!(sciwidgetPtr->flags & DESTROYED)) {
        sciwidgetPtr->flags |= DESTROYED;
        Tk_DestroyWindow(sciwidgetPtr->tkwin);
    }
}

/* ScintillaFetchSelection 
 *   Called by Tk when the selection is requested by someone. It
 *   returns part or all of the selectionin a buffer provided by
 *   caller. 
 *
 * Results:
 *    The return value is the number of non-NULL bytes stored in
 *    buffer. Buffer is filled with NULL-terminated string containing
 *    part or all of selection, as specified by offset and maxBytes. 
 */
static int ScintillaFetchSelection(ClientData clientData, int offset, char* buffer, int maxBytes)
{
    SciWidget *sciwidgetPtr = (SciWidget*)clientData;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    if (!sciwidgetPtr->exportSelection)
        return -1;  // we're not exporting selections to others.
    int sel_beg = sci_cmd(obj, SCI_GETSELECTIONSTART);
    int sel_end = sci_cmd(obj, SCI_GETSELECTIONEND);
    if (sel_beg != sel_end) {
        Sci_TextRange tr;

        sel_beg += offset;
        if ((sel_end - sel_beg) > maxBytes)
            sel_end = sel_beg + maxBytes;
        tr.chrg.cpMin = sel_beg;
        tr.chrg.cpMax = sel_end;
        tr.lpstrText = buffer;
        SSM(SCI_GETTEXTRANGE, 0, (sptr_t)&tr);
        return (sel_end - sel_beg);
    }
    return 0;
}

static int
SciWidgetCmd(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   SciWidget *sciwidgetPtr;
   Tk_Window tkwin;
   int i;

   tkwin = Tk_CreateWindowFromPath(interp, Tk_MainWindow(interp), Tcl_GetString(objv[1]), NULL);
   if (tkwin == NULL) {
      return TCL_ERROR;
   }
   /*
    * Set resource class.
    */
   Tk_SetClass(tkwin, "Scintilla");
   /*
    * Allocate and initialize the widget record.
    */
   sciwidgetPtr             = (SciWidget *) Tcl_Alloc(sizeof(SciWidget));
   sciwidgetPtr->editor     = scintilla_new(interp, tkwin, sciwidgetPtr);
   sciwidgetPtr->tkwin      = tkwin;
   sciwidgetPtr->display    = Tk_Display(tkwin);
   sciwidgetPtr->interp     = interp;
   sciwidgetPtr->relief     = TK_RELIEF_FLAT;
   sciwidgetPtr->autoseps   = true;
   sciwidgetPtr->inUndoAction = false;
   sciwidgetPtr->undo       = false; // match Tk Text widget default
   sciwidgetPtr->background = NULL;
   sciwidgetPtr->foreground = NULL;
   sciwidgetPtr->highlight  = NULL;
   sciwidgetPtr->highlightBg = NULL;
   sciwidgetPtr->marginBg   = NULL;
   sciwidgetPtr->marginFg   = NULL;
   sciwidgetPtr->highlightWidth = 1;
   sciwidgetPtr->padX       = 0;
   sciwidgetPtr->padY       = 0;
   sciwidgetPtr->borderWidth = 2;
   sciwidgetPtr->tkfont      = NULL;
   sciwidgetPtr->copyGC      = None;
   sciwidgetPtr->mainPixmap  = None;
   sciwidgetPtr->width       = 0;
   sciwidgetPtr->height      = 0;
   sciwidgetPtr->flags       = 0;
   sciwidgetPtr->showws      = NULL;
   sciwidgetPtr->tabwidth    = 2;
   sciwidgetPtr->takefocus   = NULL;
   sciwidgetPtr->state       = STATE_NORMAL;
   sciwidgetPtr->readonly    = false;
   sciwidgetPtr->language    = NULL;
   sciwidgetPtr->langType    = SCLEX_NULL; // defined in include/SciLexer.h
   sciwidgetPtr->foldstyle   = foldBoxTree;
   sciwidgetPtr->busycursor  = None;
   sciwidgetPtr->cursor      = None;
   sciwidgetPtr->blinkcaret  = NULL;
   sciwidgetPtr->exportSelection = true;
   sciwidgetPtr->linesAddedCmd  = NULL;
   sciwidgetPtr->marginWidthCmd = NULL;
   sciwidgetPtr->moveUpdateCmd  = NULL;
   sciwidgetPtr->xScrollCmd     = NULL;
   sciwidgetPtr->yScrollCmd     = NULL;
   sciwidgetPtr->yscrollDelta = 0;
   sciwidgetPtr->styleBits = 8; // maximum is 8, value may change with different parsers
   for (i = 0; i < KEYWORDSET_MAX; ++i) {
      sciwidgetPtr->keywords[i] = NULL;
   }
   sciwidgetPtr->added_data = new SciWidgetLinesAdded;
   sciwidgetPtr->marginMgr  = new MarginMgr(sciwidgetPtr);
   sciwidgetPtr->tagMgr     = new TagMgr(sciwidgetPtr);
   sciwidgetPtr->markerMgr  = new MarkerMgr(sciwidgetPtr);
   sciwidgetPtr->markMgr    = new MarkMgr(sciwidgetPtr);
   sciwidgetPtr->bindingTable = NULL;

   // NOTE: IT WAS TOO DIFFICULT TO FIGURE OUT HOW TO MANAGE THE CLIPPED DRAWING
   //       OF TEXT WHEN A NON-ZERO PADDING WAS USED ON THE RIGHT SIDE. SO WE WILL
   //       LELAVE IT UP TO THE CLIENT WINDOW TO PROVIDE RHS PADDING.
   sci_set_padding(SCIOBJ(sciwidgetPtr), 4, 0);

   // turn off display of the extra carets which by default are
   // drawn at the edge of each "sub region" of the rectangle
   sci_cmd(SCIOBJ(sciwidgetPtr), SCI_SETADDITIONALCARETSVISIBLE, 0);

   // Set number of style bits to maximum supported, lexers may change this when they are loaded
   sci_cmd(SCIOBJ(sciwidgetPtr), SCI_SETSTYLEBITS, 8);

   // Register a handler for when the window is exposed or resized.
   //
   Tk_CreateEventHandler(sciwidgetPtr->tkwin,
      ExposureMask|StructureNotifyMask|FocusChangeMask,
      SciWidgetEventProc, (ClientData) sciwidgetPtr);

   // Register a handler for events that client could define bindings for
   Tk_CreateEventHandler(sciwidgetPtr->tkwin,
       ButtonPressMask|ButtonReleaseMask|
       EnterWindowMask|LeaveWindowMask|
       PointerMotionMask|VirtualEventMask,
      SciWidgetBindProc, (ClientData) sciwidgetPtr);

   // Register selection handler when other windows want our selection
   Tk_CreateSelHandler(sciwidgetPtr->tkwin, XA_PRIMARY, XA_STRING, ScintillaFetchSelection, (ClientData) sciwidgetPtr, XA_STRING);

   // Create a Tcl command that operates on the widget.
   //
   sciwidgetPtr->widgetCmd =
      Tcl_CreateObjCommand(interp, Tk_PathName(sciwidgetPtr->tkwin),
         SciWidgetInstanceCmd,
         (ClientData) sciwidgetPtr,
         SciWidgetDeleteCmd);

   // Parse the command line arguments.
   //
   sciwidgetPtr->optionTable = Tk_CreateOptionTable(interp, optionSpecs);

   if (Tk_InitOptions(interp, (char *)sciwidgetPtr,
         sciwidgetPtr->optionTable, sciwidgetPtr->tkwin) != TCL_OK) {
      Tk_DestroyWindow(sciwidgetPtr->tkwin);
      return TCL_ERROR;
   }
   if (SciWidgetConfigure(interp, sciwidgetPtr, objc-2, objv+2) != TCL_OK) {
      Tk_DestroyWindow(sciwidgetPtr->tkwin);
      return TCL_ERROR;
   }

   // Tag manager uses values from top-level config settings.
   // NOTE: All text-based margins, including the line numbers, use the same
   // style to control appearance. If we don't set it then it will default
   // to getting the background color from Platform::Chrome.
   //
   sciwidgetPtr->tagMgr->UpdateMarginColor(true, true);

   Tcl_SetObjResult(interp, Tcl_NewStringObj(Tk_PathName(sciwidgetPtr->tkwin), -1));
   return TCL_OK;
}

/****************************************************************
// Func: SciWidgetDestroy
// Desc: 
*****************************************************************/
static void
SciWidgetDestroy(ClientData clientData)
{
   register SciWidget *sciwidgetPtr = (SciWidget *) clientData;
   if (!sciwidgetPtr) {
      // nothing to do, we're already dead
      return;
   }
//printf("SciWidgetDestroy w/ sciwidgetPtr=%p\n",(void *)sciwidgetPtr);
   /*
    * Free up all the stuff that requires special handling,
    * then let Tk_FreeOptions handle resources associated
    * with the widget attributes.
    */
   if (sciwidgetPtr->copyGC != None) {
      Tk_FreeGC(sciwidgetPtr->display, sciwidgetPtr->copyGC);
   }
   if (sciwidgetPtr->flags & REDRAW_PENDING) {
      Tk_CancelIdleCall(SciWidgetDisplay, (ClientData) sciwidgetPtr);
   }
   /*
    * This frees up colors and fonts and any allocated
    * storage associated with the widget attributes.
    */
   if (sciwidgetPtr->tkwin) {
	Tk_FreeConfigOptions((char *)sciwidgetPtr, sciwidgetPtr->optionTable, sciwidgetPtr->tkwin);
   }
   Tk_DeleteOptionTable(sciwidgetPtr->optionTable);

   // other cleanup
   delete sciwidgetPtr->markerMgr;
   // marginMgr must be destroyed after markerMgr!
   delete sciwidgetPtr->marginMgr;
   delete sciwidgetPtr->markMgr;
   delete sciwidgetPtr->tagMgr;
   delete sciwidgetPtr->added_data;

   if (sciwidgetPtr->bindingTable) {
       Tk_DeleteBindingTable(sciwidgetPtr->bindingTable);
   }
   // do the following _AFTER_ the above "delete" calls, to ensure the
   // scintilla object will still be available for whatever messages the
   // delete handlers may need to send it.
   SSM(SCI_TKMSG_DESTROY,0,0);  // clean up...

   Tcl_Free((char *) sciwidgetPtr);
}

/*
 *----------------------------------------------------------------------
 * XScroller --
 *
 *    Called to handle scroll requests from bindings(moveto = false) as
 *    well as scrollbar-initiated scrolling(moveto = true).
 *
 * Results:
 *    None
 *
 * Side affects:
 *    The -xscrollcommand is called to update the scrollbar data. If
 *    moveto is true it also change's the first column displayed to
 *    correspond to a change from the scrollbar.
 *
 *----------------------------------------------------------------------
 */
static void XScroller(
    SciWidget *sciwidgetPtr, 
    bool report = true,  // are we reporting values to scrollbar
    bool moveto = false, // TRUE when scrollbar being dragged
    double frac = 0.0    // fractional amount of change from drag
) {
    if (sciwidgetPtr->xScrollCmd == NULL)
        return;

    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    int winw = Tk_Width(sciwidgetPtr->tkwin);   // width of screen
    winw -= sciwidgetPtr->marginMgr->Width();   // account for width of margins!!
    int xoff = sci_cmd(obj, SCI_GETXOFFSET);    // current hz scroll offset
    int sw   = sci_cmd(obj, SCI_GETSCROLLWIDTH);// max scrollable width

    int new_xoff;
    if (moveto) {
        // if the window width is larger than the scroll width then
        // there's no where to scroll to, so this would be a no-op
        if (sw <= winw) return;

        new_xoff = (int)((double)sw * frac);
        if (new_xoff < 0) {
            new_xoff = 0;
        } else if (new_xoff + winw > sw) {
            new_xoff = (sw - winw);
        }
        sci_cmd(obj, SCI_SETXOFFSET, new_xoff);
    } else {
        new_xoff = xoff;
    }

    // NOTE: This is same way that Tk's text widget does it(see
    // the GetXView function in tkTextDisp.c)
    char buf1[TCL_DOUBLE_SPACE+1];
    char buf2[TCL_DOUBLE_SPACE+1];
    int r_edge = new_xoff + winw;

    double first = (double)new_xoff / (double)sw;
    double last  = (double)r_edge  / (double)sw;

    buf1[0] = ' ';
    buf2[0] = ' ';
    Tcl_PrintDouble(NULL, first, buf1+1);
    Tcl_PrintDouble(NULL, last, buf2+1);

    if (!report) {
        Tcl_Obj *listObj = Tcl_NewListObj(0,NULL);
        Tcl_Interp *interp = sciwidgetPtr->interp;
        Tcl_ListObjAppendElement(interp, listObj, Tcl_NewDoubleObj(first));
        Tcl_ListObjAppendElement(interp, listObj, Tcl_NewDoubleObj(last));
        Tcl_SetObjResult(interp, listObj);
    } else {
        Tcl_VarEval(sciwidgetPtr->interp, sciwidgetPtr->xScrollCmd, buf1, buf2, NULL);
    }
}

/*
 *----------------------------------------------------------------------
 * YScroller --
 *
 *    Called to handle scroll requests from bindings(moveto = false) as
 *    well as scrollbar-initiated scrolling(moveto = true).
 *
 * Results:
 *    If report is false, then the interp's result is filled in with two
 *    real numbers, giving the position of the top and bottom of the
 *    window as fractions from 0 to 1, where 0 means the beginning of 
 *    the text and 1 means the end. If report is true, then the interp's
 *    result isn't modified directly, but a script is evaluated in
 *    interp to report the new scroll position to the scrollbar.
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static void YScroller(
    SciWidget *sciwidgetPtr, 
    bool report = true,  // are we reporting values to scrollbar
    bool moveto = false, // TRUE when scrollbar being dragged
    double frac = 0.0    // fractional amount of change from drag
) {
   if (sciwidgetPtr->yScrollCmd == NULL)
      return;

   int onscreen = SSM(SCI_LINESONSCREEN, 0,0);
   int nlines   = SSM(SCI_GETLINECOUNT, 0,0); // << #lines in document
   // have to account for folding!!
   nlines = SSM(SCI_VISIBLEFROMDOCLINE, nlines, 0);

   int topline;
   if (moveto) {
      bool keep_going = true;
      topline = (int)((double)nlines * frac);
      if (topline < 0) {
         topline = 0;
      }
      else if ( (topline + onscreen) > nlines ) {
         // we get here when the scrollbar "thumb" is being dragged down
         // and the view has hit the bottom of the document. if we don't
         // stop here the "thumb" height will shrink, giving the user a
         // false sense of where in the document they are.
          keep_going = false;
      }
      SSM(SCI_SETFIRSTVISIBLELINE, topline, 0);
      if (!keep_going) return;

      // Using the following seems to screw up the repaint -- also,
      // other editors don't move the caret during a scrollbar drag,
      // so not using it(for now) is ok 
      //SSM(SCI_MOVECARETINSIDEVIEW,0,0);

   } else {
      topline = SSM(SCI_GETFIRSTVISIBLELINE, 0,0);
   }
   // NOTE: This is same way that Tk's text widget does it(see
   // the GetYView function in tkTextDisp.c)
   char buf1[TCL_DOUBLE_SPACE+1];
   char buf2[TCL_DOUBLE_SPACE+1];
   int bottom = topline + onscreen;

   double first = (double)topline / (double)nlines;
   double last  = (double)bottom  / (double)nlines;

   buf1[0] = ' ';
   buf2[0] = ' ';
   Tcl_PrintDouble(NULL, first, buf1+1);
   Tcl_PrintDouble(NULL, last, buf2+1);

   if (!report) {
       Tcl_Obj *listObj = Tcl_NewListObj(0,NULL);
       Tcl_Interp *interp = sciwidgetPtr->interp;
       Tcl_ListObjAppendElement(interp, listObj, Tcl_NewDoubleObj(first));
       Tcl_ListObjAppendElement(interp, listObj, Tcl_NewDoubleObj(last));
       Tcl_SetObjResult(interp, listObj);
   } else {
       Tcl_VarEval(sciwidgetPtr->interp, sciwidgetPtr->yScrollCmd, buf1, buf2, NULL);
   }
}

// forward decl's for widget subcommand functions
static int CmdAnnotate (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdBbox     (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdCompare  (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdCount    (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdDecIndent(ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdDelete   (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdDump     (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdEdit     (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdFold     (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdGet      (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdIncIndent(ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdIndex    (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdInsert   (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdLoadLexer(ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdSearch   (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdSciSearch(ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdSee      (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdStyle    (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdXview    (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdYview    (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[]);
static int CmdZoom     (ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[], bool);

// other forward decl's of interest
static bool CheckAnnotationHit(SciWidget *w, int x, int y, int line);
static int ValidateLine(Tcl_Interp *, ClientData, Tcl_Obj *, int *);


/****************************************************************
// Func: SciWidgetInstanceCmd
// Desc: 
*****************************************************************/
static int
SciWidgetInstanceCmd(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc, 
   Tcl_Obj *CONST objv[] )
{
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;

   int result = TCL_OK;
   int index;

   static CONST char *optionStrings[] = {
       "annotate",       "bbox",           "cget",     "compare",  "configure", "count",
       "decreaseindent", "delete",         "dump",     "edit",     "fold",      "get",
       "index",          "increaseindent", "insert",   "keywords", "loadlexer", "margin",
       "mark",           "marker",         "property", "scisearch", "search",   "see",
       "style",          "tag",            "textwidth", "version",  "xview",    "yview",
       "zoomin",   "zoomout",
#ifdef DEBUG
      "dumpstyles",
#endif
      (char *)NULL
   };
   enum options {
       CMD_ANNOTATE,  CMD_BBOX,      CMD_CGET,      CMD_COMPARE,   CMD_CONFIGURE, CMD_COUNT,
       CMD_DECINDENT, CMD_DELETE,    CMD_DUMP,      CMD_EDIT,      CMD_FOLD,      CMD_GET,
       CMD_INDEX,     CMD_INCINDENT, CMD_INSERT,    CMD_KEYWORDS,  CMD_LOADLEXER, CMD_MARGIN,
       CMD_MARK,      CMD_MARKER,    CMD_PROPERTY,  CMD_SCISEARCH, CMD_SEARCH,    CMD_SEE,
       CMD_STYLE,     CMD_TAG,       CMD_TEXTWIDTH, CMD_VERSION,   CMD_XVIEW,     CMD_YVIEW,
       CMD_ZOOMIN,    CMD_ZOOMOUT
#ifdef DEBUG
      ,CMD_DUMPSTYLES
#endif
   };

   if (objc < 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "option ?arg arg ...?");
      return TCL_ERROR;
   }
   if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   switch ((enum options)index) {
#ifdef DEBUG
      case CMD_DUMPSTYLES: {
         SSM(SCI_TKMSG_DUMPSTYLES,1,0);
         break;
      }
#endif
      case CMD_ANNOTATE: {
         result = CmdAnnotate(clientData, interp, objc, objv);
         break;
      }
      case CMD_BBOX: {
          result = CmdBbox(clientData, interp, objc, objv);
          break;
      }
      case CMD_CGET: {

         if (objc != 3) {
            Tcl_WrongNumArgs(interp, 2, objv, "option");
            return TCL_ERROR;
         }
         Tcl_Obj *objPtr = Tk_GetOptionValue(interp, (char *)sciwidgetPtr,
            sciwidgetPtr->optionTable, objv[2], sciwidgetPtr->tkwin);
         if (objPtr == NULL) {
            return TCL_ERROR;
         }
         Tcl_SetObjResult(interp, objPtr);
         break;
      }
      case CMD_COMPARE: {
         result = CmdCompare(clientData, interp, objc, objv);
         break;
      }
      case CMD_CONFIGURE: {

         if (objc <= 3) {
            Tcl_Obj *objPtr = Tk_GetOptionInfo(interp, (char *)sciwidgetPtr,
               sciwidgetPtr->optionTable, ((objc == 3) ? objv[2] : NULL), sciwidgetPtr->tkwin);
            if (objPtr == NULL) {
               return TCL_ERROR;
            }
            Tcl_SetObjResult(interp, objPtr);
         } else {
            result = SciWidgetConfigure(interp, sciwidgetPtr, objc-2, objv+2);
         }
         break;
      }
      case CMD_COUNT: {
         result = CmdCount(clientData, interp, objc, objv);
         break;
      }
      case CMD_DECINDENT: {
         result = CmdDecIndent(clientData, interp, objc, objv);
         break;
      }
      case CMD_DELETE: {
         result = CmdDelete(clientData, interp, objc, objv);
         break;
      }
      case CMD_DUMP: {
         result = CmdDump(clientData, interp, objc, objv);
         break;
      }
      case CMD_EDIT: {
         result = CmdEdit(clientData, interp, objc, objv);
         break;
      }
      case CMD_FOLD: {
         result = CmdFold(clientData, interp, objc, objv);
         break;
      }
      case CMD_GET: {
         result = CmdGet(clientData, interp, objc, objv);
         break;
      }
      case CMD_INCINDENT: {
         result = CmdIncIndent(clientData, interp, objc, objv);
         break;
      }
      case CMD_INDEX: {
         result = CmdIndex(clientData, interp, objc, objv);
         break;
      }
      case CMD_INSERT: {
         result = CmdInsert(clientData, interp, objc, objv);
         break;
      }
      case CMD_KEYWORDS: {
         int keywordSet;
         /* 1st argument is keyword number 0-8, followed by a string defining keyword set */
         /* objv[2] == keyword number */
         /* objv[3] == cmd list (optional) */
         if ((objc != 3) && (objc != 4)) {
            Tcl_WrongNumArgs(interp, 1, objv, "number [keywords]");
            return TCL_ERROR;
         }
         if (Tcl_GetIntFromObj(interp, objv[2], &keywordSet) != TCL_OK) {
            return TCL_ERROR;
         }
         if (objc == 4) {
            if (sciwidgetPtr->keywords[keywordSet]) {
                Tcl_DecrRefCount(sciwidgetPtr->keywords[keywordSet]);
            }
            sciwidgetPtr->keywords[keywordSet] = objv[3];
            Tcl_IncrRefCount(sciwidgetPtr->keywords[keywordSet]);
            SSM(SCI_SETKEYWORDS, keywordSet, (sptr_t)Tcl_GetString(objv[3]));
         } else {
            // return keyword set
            if (sciwidgetPtr->keywords[keywordSet]) {
                Tcl_SetObjResult(interp, sciwidgetPtr->keywords[keywordSet]);
            } else {
                Tcl_ResetResult(interp);
            }
         }
         break;
      }
      case CMD_LOADLEXER: {
         result = CmdLoadLexer(clientData, interp, objc, objv);
         break;
      }
      case CMD_MARGIN: {
         result = sciwidgetPtr->marginMgr->InstanceCmd(objc, objv);
         break;
      }
      case CMD_MARK: {
         result = sciwidgetPtr->markMgr->InstanceCmd(objc, objv);
         break;
      }
      case CMD_MARKER: {
         result = sciwidgetPtr->markerMgr->InstanceCmd(objc, objv);
         break;
      }
      case CMD_PROPERTY: {
         /* 1st argument is property name if 2nd argument exists, set the property; otherwise return the property */
         if (objc == 2) {
             char *names;
             int len = SSM(SCI_PROPERTYNAMES, 0, 0);
             names = new char[len+1];
             len = SSM(SCI_PROPERTYNAMES, 0, (sptr_t)names);
             Tcl_SetObjResult(interp, Tcl_NewStringObj(names, len));
             delete[] names;
         } else if (objc == 3) {
            int len = SSM(SCI_GETPROPERTY,(uptr_t)Tcl_GetString(objv[2]),0);
            if (len > 0) {
                char *buf = new char[len+1];
                len = SSM(SCI_GETPROPERTY,(uptr_t)Tcl_GetString(objv[2]),(sptr_t)buf);
                Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, len));
                delete[] buf;
            }
         } else if (objc == 4) {
            SSM(SCI_SETPROPERTY,(uptr_t)Tcl_GetString(objv[2]),(sptr_t)Tcl_GetString(objv[3]));
         } else {
            Tcl_WrongNumArgs(interp, 1, objv, "property ?value?");
            return TCL_ERROR;
         }
         break;
      }
      case CMD_SEARCH: {
         result = CmdSearch(clientData, interp, objc, objv);
         break;
      }
      case CMD_SCISEARCH: {
         result = CmdSciSearch(clientData, interp, objc, objv);
         break;
      }
      case CMD_SEE: {
         result = CmdSee(clientData, interp, objc, objv);
         break;
      }
      case CMD_STYLE: {
         result = CmdStyle(clientData, interp, objc, objv);
         break;
      }
      case CMD_TAG: {
         result = sciwidgetPtr->tagMgr->InstanceCmd(objc, objv);
         break;
      }
      case CMD_TEXTWIDTH: {

         if (objc != 3) {
            Tcl_WrongNumArgs(interp, 2, objv, "string");
            return TCL_ERROR;
         }
         char *str = Tcl_GetString(objv[2]);
         int tw = SSM(SCI_TEXTWIDTH, STYLE_DEFAULT, (sptr_t)str);
         Tcl_SetObjResult(interp, Tcl_NewIntObj(tw));
         break;
      }
      case CMD_VERSION: {
          Tcl_Obj *listObj = Tcl_NewListObj(0,NULL);
          // version info(list of 3 elements)
          //    scintilla/tk widget version (matches ScintillaTk package)
          //    scintilla/tk widget build date
          //    scintilla core release version
          //
          Tcl_ListObjAppendElement(interp, listObj, Tcl_NewStringObj(SCI_VERSION, -1));
          Tcl_ListObjAppendElement(interp, listObj, Tcl_NewStringObj(SCI_VER_DATE, -1));
          Tcl_ListObjAppendElement(interp, listObj, Tcl_NewStringObj(SCI_CORE_VERSION, -1));
          Tcl_SetObjResult(interp, listObj);
          break;
      }
      case CMD_XVIEW: {
         result = CmdXview(clientData, interp, objc, objv);
         break;
      }
      case CMD_YVIEW: {
         result = CmdYview(clientData, interp, objc, objv);
         break;
      }
      case CMD_ZOOMIN: {
         result = CmdZoom(clientData, interp, objc, objv, true);
         break;
      }
      case CMD_ZOOMOUT: {
         result = CmdZoom(clientData, interp, objc, objv, false);
         break;
      }
   }
   return result;
}

/*
 *----------------------------------------------------------------------
 * ValidateLine --
 *
 *    Checks a line value that comes in via a command to determine
 *    if it's valid or not
 *
 * Results:
 *    Takes the line argument and validates it
 *
 * Side affects:
 *    For a valid line number, the variable pointed to by "line_out"
 *    is set with the correct Scintilla 0-based number
 *
 *----------------------------------------------------------------------
 */
static int ValidateLine(
   Tcl_Interp *interp,
   ClientData  clientData, 
   Tcl_Obj    *line_obj, 
   int        *line_out)
{
   int line;
   if (Tcl_GetIntFromObj(interp, line_obj, &line) != TCL_OK) {
      Tcl_AppendResult(interp, "expected integer but got \"",line_obj,"\"", (char *)NULL);
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // validate the line number
   int max = sci_get_line_count(obj);

   line--; // TK uses top line equal to 1 (Scintilla uses 0)
   if (line < 0 || line >= max) {
      Tcl_Obj *obj = Tcl_ObjPrintf("bad line value(%d): must be between 1 and %d", line, max);
      Tcl_AppendResult(interp, Tcl_GetString(obj));
      return TCL_ERROR;
   }
   *line_out = line;
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * ExpandTabs --
 *
 *    Modifies a text string to expand Tab characters to the
 *    equivalent number of spaces
 *
 * Results:
 *    Tab characters get replaced
 *
 * Side affects:
 *    Calling code responsible for freeing the allocated buffer; this
 *    is only required when the return value is >0.
 *
 *----------------------------------------------------------------------
 */
static int ExpandTabs(
   ClientData clientData,
   char      **_str)
{
   // We do Tab substitutions now since Scintilla doesn't
   // handle this for annotation strings
   char *src = *_str;
   int ntabs = 0;
   int col;
   while (*src != '\0') if (*src++ == '\t') ntabs++;

   if (ntabs == 0) return 0;

   src = *_str;

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   int tw = sciwidgetPtr->tabwidth; // current tab width
   if (tw <= 0) return 0;
   int newlen = (strlen(src) - ntabs) + (ntabs * tw) + 1;
   char *buf = (char *) Tcl_Alloc(newlen);
   char *dst = buf; 
   col = 0;
   while (*src != '\0') {
       if (*src != '\t') {
           *dst++ = *src++;
           col++;
       } else {
           int tabstop = col + tw - (col % tw);
           for (; col < tabstop; col++) {
               *dst++ = ' '; 
           }
           src++;
       }
   }
   *dst++ = '\0';
   *_str = buf;
   return ntabs;
}


/*
 *----------------------------------------------------------------------
 * CmdAnnotate --
 *
 *    Implements the ANNOTATE widget command
 *
 * Results:
 *    show    -- controlled visibility of annotations
 *    clearll -- removed annotations from all lines
 *    get     -- returned annotation text for a specific line
 *    set     -- set annotation text for a specific line
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdAnnotate(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option ?arg arg ...?");
      return TCL_ERROR;
   }
   static const char *optionStrings[] = {"clearall", "get", "set", "show"};
   enum options {CMD_CLEARALL, CMD_GET, CMD_SET, CMD_SHOW};
   int index;
   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings,
         "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

   switch ((enum options)index) {
      case CMD_SHOW: {

         if (objc != 4) {
            Tcl_WrongNumArgs(interp, 3, objv, "mode (0|1|2)");
            return TCL_ERROR;
         }
         int mode;
         if (Tcl_GetIntFromObj(interp, objv[3], &mode) != TCL_OK) {
            Tcl_AppendResult(interp, "expected integer but got \"",objv[3],"\"", (char *)NULL);
            return TCL_ERROR;
         }
         switch (mode) {
            case 0: sci_cmd(sciobj, SCI_ANNOTATIONSETVISIBLE, ANNOTATION_HIDDEN); break;
            case 1: sci_cmd(sciobj, SCI_ANNOTATIONSETVISIBLE, ANNOTATION_STANDARD); break;
            case 2: sci_cmd(sciobj, SCI_ANNOTATIONSETVISIBLE, ANNOTATION_BOXED); break;
            default:
               Tcl_AppendResult(interp, "bad show option: must be 0, 1 or 2", (char *)NULL);
               return TCL_ERROR;
         }
         break;
      }
      case CMD_CLEARALL: {

         sci_cmd(sciobj, SCI_ANNOTATIONCLEARALL);
         break;
      }
      case CMD_GET: {

         if (objc != 4) {
            Tcl_WrongNumArgs(interp, 3, objv, "line");
            return TCL_ERROR;
         }
         int line;
         if (ValidateLine(interp, clientData, objv[3], &line) != TCL_OK) {
            return TCL_ERROR;
         }
         // find out how long the text is, since we need to pass in an
         // allocated buffer for Scintilla to fill
         int len = SSM(SCI_TKMSG_ANNOTATIONLENGTH, line, 0);
         if (len > 0) {
            char *buf = new char[len+1];
            int slen = sci_cmd(sciobj, SCI_ANNOTATIONGETTEXT, line, (sptr_t)buf);
            Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, slen));
            delete[] buf;
         }
         break;
      }
      case CMD_SET: {

         if (objc < 5) {
            Tcl_WrongNumArgs(interp, 3, objv, "line text ?tagName text tagName ...?");
            return TCL_ERROR;
         }
         int line;
         if (ValidateLine(interp, clientData, objv[3], &line) != TCL_OK) {
            return TCL_ERROR;
         }
         // handle the case where an annotation needs to be 
         // cleared, which is done by specifying an empty string
         char *str = Tcl_GetString(objv[4]);
         int slen  = strlen(str);
         if (slen == 0) {
            sci_cmd(sciobj, SCI_ANNOTATIONSETTEXT, line, 0);
            return TCL_OK;
         }
         // now loop through the text/tag args so we can figure out how
         // the style information needs to be assembled. if just a single
         // tag is specified, it is used for the whole text string. when
         // multiple pieces of text and multiple tags are used it's a lot
         // more complicated.
         // REMINDER: Scintilla only allows assigning a single text string
         //           to a line. So if multiple styles are needed, the style
         //           information needs to be embedded into the text in such
         //           a way acceptable to SCI_ANNOTATIONSETSTYLES.

         // a simple case would be either having no tag or just a single tag
         if (objc <= 6) {
            TagData *tag = NULL;
            if (objc == 6) {
               char *tagName = Tcl_GetString(objv[5]);
               tag = sciwidgetPtr->tagMgr->FindTag(tagName);
               // Tk's Text widget ignores bogus tag names, we'll match that behavior
            }
            // assign the text to the line
            int ntabs = ExpandTabs(clientData, &str);

            sci_cmd(sciobj, SCI_ANNOTATIONSETTEXT,  line, (sptr_t)str);
            if (ntabs)
               Tcl_Free(str);

            // if we have a valid tag, use it to set the style
            int offset = sci_cmd(sciobj, SCI_ANNOTATIONGETSTYLEOFFSET);
            if (tag && (tag->Type() & TT_ANNO)) {
               sci_cmd(sciobj, SCI_ANNOTATIONSETSTYLE, line, tag->StyleID() - offset);

               // as per Neil H. on 6/26/2012, the lack of support for the
               // following setting for attribute styles is a "limitation"
               //SSM(SCI_STYLESETEOLFILLED, tag->StyleID(), 0);
            }
            return TCL_OK;
         }
         // we have multiple pieces of text and possibly multiple tags

//TODO: CmdAnnotate >> support use of more than one tag

         break;
      }
   }
   return TCL_OK;
}

#if 0
//***************************************************************************
// EXAMPLE CODE FOR HOW TO SPECIFY MULTIPLE STYLES FOR ANNOTATION TEXT
//***************************************************************************
static void StyleAnnotation(ScintillaObject *sciobj, int line, char *str) {

   if (!str) return;

   int slen = strlen(str);
   char *stylebuf = Tcl_Alloc(slen+1);
   // check if there are multiple lines
   char *p = strchr(str, '\n');

   static int style_id  = 200;
   static int style_id2 = 201;
   static int style_id3 = 202;

   if (!p) {
      ::memset(stylebuf, style_id2, slen-1);
   } else {
      int n=0;
      int offset=0;
      char *last = str;
      do {
         int len = (p - last)+1;
         ::memset(stylebuf+offset, (++n == 3) ? style_id : ((n==2)?style_id3:style_id2), len);
         offset += len;
         last = p + 1;
         p = strchr(last, '\n');
      } while (p);
      // check if the last line had a newline, and if not then
      // it still needs to be styled
      if (offset < slen) {
         ::memset(stylebuf+(last-str), style_id2, (slen-(last-str)));
      }
   }
   stylebuf[strlen(str)] = '\0'; // not required for the next call, but play it safe!
   sci_cmd(sciobj, SCI_ANNOTATIONSETSTYLES, line, (sptr_t)stylebuf);

   Tcl_Free(stylebuf);
   return;
}
#endif


/*
 *----------------------------------------------------------------------
 * CmdBbox --
 *
 *    Implements the BBOX widget command
 *
 *    Returns a list of four elements(x,y,width,height) describing the
 *    screen area of the character given by the "index" option.
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdBbox(
    ClientData     clientData,
    Tcl_Interp    *interp,
    int            objc,
    Tcl_Obj *CONST objv[])
{
    sptr_t pos, endpos;
    int x, y, width, line, line_ht, r_edge;
    char *index_str;

    if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "index");
        return TCL_ERROR;
    }
    SciWidget *sciwidgetPtr = (SciWidget *)clientData;
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

    index_str = Tcl_GetString(objv[2]);
    if (GetPosition(sciwidgetPtr, index_str, &pos, false) != TCL_OK) {
        return BadIndex(interp, index_str);
    }
    line    = sci_get_line(sciobj, pos);
    endpos  = sci_get_lineend_pos(sciobj, line);
    x       = sci_get_x_from_pos(sciobj, pos);
    y       = sci_get_y_from_pos(sciobj, pos);
    line_ht = sci_get_text_height(sciobj, line);
    // check if the position is at the end of the line
    r_edge = Tk_Width(sciwidgetPtr->tkwin);
    if (pos == endpos) {
        // set width to space from end of line to edge off window
        width = (r_edge - x);
        if (width < 0) width = 0;
    } else if (x > r_edge) {
        // not at end of line, but outside visible area
        width = 0; //<< same as Tk Text widget behavior
    } else {
        char buf[2];
        unsigned char sid = sci_cmd(sciobj, SCI_GETSTYLEAT, pos);
        buf[0]  = sci_cmd(sciobj, SCI_GETCHARAT, pos);
        buf[1]  = '\0';
        width = sci_cmd(sciobj, SCI_TEXTWIDTH, sid, (sptr_t)buf);
    }
    Tcl_Obj *listObj = Tcl_NewListObj(0, NULL);

    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(x));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(y));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(width));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(line_ht));

    Tcl_SetObjResult(interp, listObj);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdCompare --
 *
 *    Implements the COMPARE widget command
 *
 *    Compares the indices given by index1 and index2 according to the
 *    relational operator given by op, and returns 1 if the relationship
 *    is satisfied and 0 if it is not. Op must be one of the operators:
 *        <, <=, ==, >=, >, or !=
 *    If op is == then 1 is returned if the two indices refer to the 
 *    same character, if op is < then 1 is returned if index1 refers to
 *    an earlier character in the text than index2, and so on.
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdCompare(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   char *index_str, *op;
   int slen;
   sptr_t pos1 = -1;
   sptr_t pos2 = -1;

   if (objc != 5) {
      Tcl_WrongNumArgs(interp, 2, objv, "index1 op index2");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   
   index_str = Tcl_GetString(objv[2]);
   if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
      goto error;
   }
   index_str = Tcl_GetString(objv[4]);
   if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
      goto error;
   }
   op = Tcl_GetString(objv[3]);
   slen = strlen(op);
   if (slen == 1) {
      if (*op == '<') {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 < pos2));
         goto done;
      } else if (*op == '>') {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 > pos2));
         goto done;
      }
   } else if (slen == 2) {
      if (!strcmp(op, "<=")) {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 <= pos2));
         goto done;
      } else if (!strcmp(op, "==")) {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 == pos2));
         goto done;
      } else if (!strcmp(op, ">=")) {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 >= pos2));
         goto done;
      } else if (!strcmp(op, "!=")) {
         Tcl_SetObjResult(interp, Tcl_NewBooleanObj(pos1 != pos2));
         goto done;
      }
   }
   Tcl_AppendResult(interp, "bad comparison operator \"", op, 
      "\": must be <, <=, ==, >=, >, or !=", (char *)NULL);
   return TCL_ERROR;

 done:
   return TCL_OK;
 error:
   return BadIndex(interp, index_str);
}
/*
 *----------------------------------------------------------------------
 * CmdCount --
 *
 *    Implements the COUNT widget command
 *
 *    Counts the number of relevant things between the two indices.
 *
 * Results:
 *    A standard Tcl result holding the count results. If multiple
 *    count options are requested, then the return value will be a
 *    list of numbers, one for each option.
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdCount(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   char *index_str, *option;
   sptr_t pos1 = -1;
   sptr_t pos2 = -1;
   int i = 0, found = 0, value = 0;
   Tcl_Obj *objPtr = NULL;

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 index2");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // get the two indices
   index_str = Tcl_GetString(objv[objc-2]);
   if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
       goto error;
   }
   index_str = Tcl_GetString(objv[objc-1]);
   if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
       goto error;
   }
   // now figure out what we're counting
   //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   // NOTE: for now we're only implementing the option(s) needed to 
   //       support calls from the Text class bindings
   //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   for (i = 2; i < objc-2; i++) {
       int length;
       option = Tcl_GetStringFromObj(objv[i], &length);
       char c;

       if (length < 2 || option[0] != '-') {
       badOption:
           Tcl_ResetResult(interp);
           Tcl_AppendResult(interp, "bad option \"",
               Tcl_GetString(objv[i]),
               "\" must be -chars, -displaylines, -lines", NULL);

// #ifdef THIS_IS_FULL_LIST_FOR_TK_TEXT_WIDGET
//                "\" must be -chars, -displaychars, -displayindices, ",
//                "-displaylines, -indices, -lines, -update, ",
//                "-xpixels, or -ypixels", NULL);
// #endif
           return TCL_ERROR;
       }
       c = option[1];
       if (c == 'c' && !strncmp("-chars", option, (unsigned) length)) {
           value = pos2 - pos1;
       } else if (c == 'd' && (length > 8)
           && !strncmp("-displaylines", option, (unsigned) length)) {
           // NOTE: the count doesn't include the line with the second index
           int line1 = sci_get_line(obj, pos1);
           int line2 = sci_get_line(obj, pos2);
           value = 0;
           if (line1 < line2) {
               for (int n = line1; n < line2; n++) {
                   // ignore folded lines
                   if (sci_cmd(obj, SCI_GETLINEVISIBLE, n)) value++;
               }
           } else if (line2 < line1) {
               for (int n = line2; n < line1; n++) {
                   if (sci_cmd(obj, SCI_GETLINEVISIBLE, n)) value++;
               }
               value = -value; // make it a negative value since line2 was <line1
           }
       } else if (c == 'l' && !strncmp("-lines", option, (unsigned) length)) {
           // NOTE: the count doesn't include the line with the second index
           int line1 = sci_get_line(obj, pos1);
           int line2 = sci_get_line(obj, pos2);
           value = (line2 - line1);

       } else {
           goto badOption;
       }
   countDone:
       found++;
       if (found == 1) {
           Tcl_SetObjResult(interp, Tcl_NewIntObj(value));
       } else {
           if (found == 2) {
               /*
                * Move the first item we put into the result into
                * the first element of the list object.
                */
               objPtr = Tcl_NewObj();
               Tcl_ListObjAppendElement(NULL, objPtr, Tcl_GetObjResult(interp));
           }
           Tcl_ListObjAppendElement(NULL, objPtr, Tcl_NewIntObj(value));
       }
   } // end of FOR loop
   if (found == 0) {
       // use the default '-chars'
       value = (pos2 - pos1);
       goto countDone;
   }
// done:
   return TCL_OK;
 error:
   return BadIndex(interp, index_str);
}

/*
 *----------------------------------------------------------------------
 * CmdDecIndent --
 *
 *    Implements the decreaseindent widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Text is shifted left
 *
 *----------------------------------------------------------------------
 */
static int CmdDecIndent(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   char *lineNumber, *end;
   int firstLine = -1;
   int lastLine = -1;

   if ((objc < 3) || (objc > 4)) {
      Tcl_WrongNumArgs(interp, 2, objv, "line1 ?line2?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   if (sciwidgetPtr->state == STATE_NORMAL) {
      int i;

      // simple case: either one or two indices
      lineNumber = Tcl_GetString(objv[2]);
      firstLine = strtol(lineNumber, &end, 0);
      if (*end)
         goto error;
      if (objc == 4) {
         lineNumber = Tcl_GetString(objv[3]);
         lastLine = strtol(lineNumber, &end, 0);
         if (*end)
            goto error;
      } else {
         lastLine = firstLine;
      }
      // start atomic undo operation
      sci_cmd(obj, SCI_BEGINUNDOACTION);
      // set state where backspace will unindent document
      sci_cmd(obj, SCI_SETBACKSPACEUNINDENTS, 1);
      // for each specified line, if first character is whitespace, delete it
      for (i = firstLine; i <= lastLine; ++i) {
         int pos = sci_get_pos(obj, LINE2SCI(i));
         if (pos != -1) {
            char firstChar = (char)sci_cmd(obj, SCI_GETCHARAT, pos);
            if (isspace(firstChar)) {
               sci_delete_text(SCIOBJ(sciwidgetPtr), pos, pos+1);
            }
         }
      }
      // turn off state where backspace will unindent document
      sci_cmd(obj, SCI_SETBACKSPACEUNINDENTS, 0);
      // end atomic undo operation
      sci_cmd(obj, SCI_ENDUNDOACTION);
   }
   return TCL_OK;
 error:
   Tcl_ResetResult(interp);
   Tcl_AppendResult(interp, "bad line number \"", lineNumber, "\"", NULL);
   return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 * CmdDelete --
 *
 *    Implements the DELETE widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Text is removed from the underlying document
 *
 *----------------------------------------------------------------------
 */
static int CmdDelete(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   char *index_str;
   sptr_t pos1 = -1;
   sptr_t pos2 = -1;

   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "index1 ?index2? ...?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   if (sciwidgetPtr->state == STATE_NORMAL) {
      if (objc < 5) {
         // simple case: either one or two indices
         index_str = Tcl_GetString(objv[2]);
         if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
            goto error;
         }
         if (pos1 == -1) goto done; // index is valid, but not within document
         if (objc == 4) {
            index_str = Tcl_GetString(objv[3]);
            if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
               goto error;
            }
            if (pos2 == -1) goto done;
         } else {
            pos2 = pos1+1; // just delete one character
         }
      } else {
//TODO: CmdDelete >> support more than two indices...
         Tcl_ResetResult(interp);
         Tcl_AppendResult(interp, "more than two indices is not supported", NULL);
         return TCL_ERROR;
      }
      if (pos1 > pos2) {
         // according to the Tk documenation this is a no-op
      } else {
         sci_delete_text(SCIOBJ(sciwidgetPtr), pos1, pos2);
      }
   }
 done:
   return TCL_OK;
 error:
   return BadIndex(interp, index_str);
}

static void DumpText(SciWidget *sciwidgetPtr, Tcl_DString *dstr, int pos1, int pos2, const char *command)
{
    if (pos2 >= pos1) {
        ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
        Tcl_DString cmdStr;
        Sci_TextRange tr;

        // read desired text
        tr.chrg.cpMin = pos1;
        tr.chrg.cpMax = pos2;
        tr.lpstrText = new char[pos2-pos1+1];
        SSM(SCI_GETTEXTRANGE, 0, (sptr_t)&tr);

        if (command) {
            Tcl_DStringInit(&cmdStr);
            Tcl_DStringAppendElement(&cmdStr, command);
            dstr = &cmdStr;
        }
        // append text element(s) (one per line) to dstr
        char *newline, *cp = tr.lpstrText;
        while (cp && *cp) {
            Tcl_DStringAppendElement(dstr, "text");
            newline = strchr(cp, '\n');
            if (newline++) {
                char savechar = *newline;
                *newline= 0;
                Tcl_DStringAppendElement(dstr, cp);
                PosToIndex(NULL, sciobj, pos1, NULL, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
                *newline = savechar;
                pos1 += (newline - cp);
                cp = newline;
            } else {
                Tcl_DStringAppendElement(dstr, cp);
                PosToIndex(NULL, sciobj, pos1, NULL, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
                break;
            }
        }
        if (command) {
            Tcl_DStringFree(&cmdStr);
        }
        delete[] tr.lpstrText;
    }
}

/*
 *----------------------------------------------------------------------
 * CmdDump --
 *
 *    Implements the DUMP widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdDump(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
    if (objc < 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }
    // all the Tk text widget options plus -indicator and -style for Scintilla
    static CONST char *optionStrings[] = {
        "-all", "-command", "-image", "-indicator", "-mark", "-style", "-tag", "-text", "-window", NULL
    };
    enum options {
        CMD_ALL, CMD_COMMAND, CMD_IMAGE, CMD_INDICATOR, CMD_MARK, CMD_STYLE, CMD_TAG, 
        CMD_TEXT, CMD_WINDOW
    };
    enum optionMask {
        OPTION_COMMAND = 1, OPTION_IMAGE = 2, OPTION_MARK = 4, OPTION_TAG = 8, OPTION_TEXT = 16,
        OPTION_WINDOW = 32, OPTION_INDICATOR=64, OPTION_STYLE=128
    };
    // Parse the switch options
    int i, options = 0;
    char *command = NULL;
    for (i=2; i < objc; i++) {
        int index;

        if (Tcl_GetString(objv[i])[0] != '-') break;

        if (Tcl_GetIndexFromObj(interp, objv[i], optionStrings, "switch", 0, &index) != TCL_OK) {
            return TCL_ERROR;
        }
        switch ((enum options)index) {
            case CMD_ALL:
                options = OPTION_IMAGE|OPTION_MARK|OPTION_TAG|OPTION_TEXT|OPTION_WINDOW;
                /* add scintilla-specific option bits */
                options |= OPTION_INDICATOR|OPTION_STYLE;
                break;
            case CMD_COMMAND:
                options |= OPTION_COMMAND;
                if (++i < objc) {
                    command = Tcl_GetString(objv[i]);
                } else {
                    Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
                    return TCL_ERROR;
                }
                break;
            case CMD_IMAGE:
                options |= OPTION_IMAGE;
                break;
            case CMD_INDICATOR:
                options |= OPTION_INDICATOR;
                break;
            case CMD_MARK:
                options |= OPTION_MARK;
                break;
            case CMD_STYLE:
                options |= OPTION_STYLE;
                break;
            case CMD_TAG:
                options |= OPTION_TAG;
                break;
            case CMD_TEXT:
                options |= OPTION_TEXT;
                break;
            case CMD_WINDOW:
                options |= OPTION_WINDOW;
                break;
        }
    }
    if (i >= objc) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    SciWidget *sciwidgetPtr = (SciWidget *)clientData;
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    sptr_t pos1 = -1;
    sptr_t pos2 = -1;
    char *index_str = Tcl_GetString(objv[i++]);
    if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
        goto error;
    }
    if (i < objc) {
        index_str = Tcl_GetString(objv[i++]);
        if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
            goto error;
        }
    }
    if (i != objc) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    // default is to dump everything if specific option not specified
    if (options == 0) {
        options |= OPTION_IMAGE|OPTION_MARK|OPTION_TAG|OPTION_TEXT|OPTION_WINDOW;
    }

    // now go through the various selected options and dump them
    Tcl_DString dstr;
    Tcl_DStringInit(&dstr);

    if (options & OPTION_IMAGE) {
        // images are currently unsupported
        // dump image(command, &dstr)
    }
    if (options & OPTION_MARK) {
        // marks are on a per-margin and per-line basis, not in the text.
        // convert positions to line numbers & and send -1 as margin_id to signal all margins
        // dump marks between pos1 and pos2, marks are line-based so convert pos1/pos2 to lines
        int line1 = sci_get_line(sciobj, pos1);
        int line2 = sci_get_line(sciobj, pos2);
        sciwidgetPtr->markMgr->Dump(&dstr, -1, line1, line2, command);
    }
    if (options & (OPTION_TAG|OPTION_STYLE|OPTION_INDICATOR)) {
        int mask = 0;
        if (options&OPTION_TAG)
            mask |= 1;
        if (options&OPTION_STYLE)
            mask |= 2;
        if (options&OPTION_INDICATOR)
            mask |= 4;
        sciwidgetPtr->tagMgr->DumpText(mask, &dstr, pos1, pos2, command);
    }
    if (options & OPTION_TEXT) {
        // dump text(command, &dstr)
        DumpText(sciwidgetPtr, &dstr, pos1, pos2, command);
    }
    if (options & OPTION_WINDOW) {
        // embedded windows are currently unsupported
        // dump window(command, &dstr)
    }

    Tcl_SetResult(interp, Tcl_DStringValue(&dstr), TCL_VOLATILE);
    Tcl_DStringFree(&dstr);
    return TCL_OK;

 error:
    return BadIndex(interp, index_str);
}

/*
 *----------------------------------------------------------------------
 * CmdEdit --
 *
 *    Implements the EDIT widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Text is removed from the underlying document
 *
 *----------------------------------------------------------------------
 */
static int CmdEdit(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   int index;

   static CONST char *optionStrings[] = {
      "canredo", "canundo", "modified", "redo", "reset", "separator", "undo", (char *)NULL
   };
   enum options {
      CMD_CANREDO, CMD_CANUNDO, CMD_MODIFIED, CMD_REDO, CMD_RESET, CMD_SEP, CMD_UNDO
   };

   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option ?arg arg ...?");
      return TCL_ERROR;
   }
   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings, "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   switch ((enum options)index) {

      case CMD_CANREDO:
            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(sci_cmd(obj, SCI_CANREDO)));
        break;

      case CMD_CANUNDO:
            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(sci_cmd(obj, SCI_CANUNDO)));
        break;

      case CMD_MODIFIED: {
         //---  MODIFIED  ------------------------------------------
         if (objc == 3) {
            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(sci_get_is_modified(obj)));
         } else if (objc != 4) {
            Tcl_WrongNumArgs(interp, 3, objv, "?boolean?");
            return TCL_ERROR;
         } else {
            int setModified;
            if (Tcl_GetBooleanFromObj(interp, objv[3], &setModified) != TCL_OK) {
               return TCL_ERROR;
            }
            if (setModified) {
               /************************************************************
                * This one doesn't have an equivalent in Scintilla. So what
                * we'll do is to add a character, then immediately remove it.
                * This will force the "dirty" flag to get set, but we'll be
                * careful not to allow the add/delete actions into the undo
                * history.
                ************************************************************/

               // create an atomic undo-able action from the add/delete
               sci_cmd(obj, SCI_BEGINUNDOACTION);
               sci_add_text(obj, 0, " ");
               sci_delete_text(obj, 0, 1);
               sci_cmd(obj, SCI_ENDUNDOACTION);
               sci_undo(obj); // THEN GET RID OF IT FROM THE UNDO HISTORY!!!

            } else {
               sci_cmd(obj, SCI_SETSAVEPOINT);
            }
         }
         break;
      }
      case CMD_REDO: {
         //---  REDO  ----------------------------------------------
         if (sci_cmd(obj, SCI_CANREDO)) {
             sci_redo(obj);
         } else if (sciwidgetPtr->undo) {
             // match Tk Text widget behavior
             Tcl_AppendResult(interp, "nothing to redo", NULL);
             return TCL_ERROR;
         }
         break;
      }
      case CMD_RESET: {
         //---  RESET  ---------------------------------------------
         sci_reset_undo_history(obj);
         break;
      }
      case CMD_SEP: {
         //---  SEPARATOR  -----------------------------------------
         // toggle undo action state
         if (sci_cmd(obj, SCI_GETUNDOCOLLECTION)) {
            if (sciwidgetPtr->inUndoAction) {
                // end an undo action
                sci_cmd(obj, SCI_ENDUNDOACTION);
                sciwidgetPtr->inUndoAction = 0;
            } else {
                // start an undo action
                sci_cmd(obj, SCI_BEGINUNDOACTION);
                sciwidgetPtr->inUndoAction = 1;
            }
         }
#if 0
         if (sci_cmd(obj, SCI_GETUNDOCOLLECTION)) {
             // Force the next undo-able action so it's _not_
             // grouped with the preceding ones
             sci_cmd(obj, SCI_BEGINUNDOACTION);
         }
#endif
         break;
      }
      case CMD_UNDO: {
         //---  UNDO  ----------------------------------------------
         if (sci_cmd(obj, SCI_CANUNDO)) {
             sci_undo(obj);
         } else if (sciwidgetPtr->undo) {
             // match Tk Text widget behavior
             Tcl_AppendResult(interp, "nothing to undo", NULL);
             return TCL_ERROR;
         }
         break;
      }
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdFold --
 *
 *    Implements the FOLD widget command
 *
 * Results:
 *    A standard Tcl result - if the operation can change the fold
 *    structure and it does, then a result of "1" is set; otherwise
 *    the result is set to "0".
 *
 * Side affects:
 *    Fold structure may change based on which command is performed
 *
 *----------------------------------------------------------------------
 */
static int CmdFold(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   int index;

   static CONST char *optionStrings[] = {
      "foldall", "foldbelow", "foldparent", "foldroot", "isfolded", "show",
      "toggle", "topunfolded", "unfoldall", "unfoldbelow", (char *)NULL
   };
   enum options {
      CMD_FOLDALL,   CMD_FOLDBELOW, CMD_FOLDPARENT, CMD_FOLDROOT, 
      CMD_ISFOLDED,  CMD_SHOW,      CMD_TOGGLE,     CMD_TOPUNFOLDED,
      CMD_UNFOLDALL, CMD_UNFOLDBELOW
   };

   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option ?arg arg ...?");
      return TCL_ERROR;
   }
   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings, "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // multiple commands need the line argument, so consolidate that code here
   int line = -1;
   switch ((enum options)index) {
      case CMD_FOLDPARENT:
      case CMD_FOLDROOT:
      case CMD_ISFOLDED:
      case CMD_SHOW:
      case CMD_FOLDBELOW:
      case CMD_UNFOLDBELOW:
      case CMD_TOGGLE:
      case CMD_TOPUNFOLDED: {
         if (objc != 4) {
            Tcl_WrongNumArgs(interp, 3, objv, "line");
            return TCL_ERROR;
         }
         if (ValidateLine(interp, clientData, objv[3], &line) != TCL_OK) {
            return TCL_ERROR;
         }
         break;
      }
      default: break;
   }

   switch ((enum options)index) {
      case CMD_FOLDALL:
      case CMD_FOLDBELOW:
      case CMD_UNFOLDALL:
      case CMD_UNFOLDBELOW: {
         bool do_fold = (index == CMD_FOLDALL || index == CMD_FOLDBELOW);
         if (do_fold) {
            // this is needed to ensure the fold points have been all calculated.
            // without it we run the risk of only getting fold regions in the
            // visible part of the window folded.
            sci_cmd(obj, SCI_COLOURISE, 0, -1);
         }
         int lines, first;
         int start = (index == CMD_FOLDBELOW || index == CMD_UNFOLDBELOW) ? line : 0;
         lines = sci_get_line_count(obj);
         first = sci_cmd(obj, SCI_GETFIRSTVISIBLELINE);
         bool changed = false;
         for (int i=start; i < lines; i++) {
            int level = sci_cmd(obj, SCI_GETFOLDLEVEL, i);
            if (level & SC_FOLDLEVELHEADERFLAG) {
               if ((bool)sci_cmd(obj, SCI_GETFOLDEXPANDED, i) == do_fold) {
                  sci_toggle_fold(obj, i);
                  changed = true;
               }
            }
         }
         if (changed) {
            // make it so top line prior to expansion doesn't change
            sci_cmd(obj, SCI_SETFIRSTVISIBLELINE, first);

            // update the Y-scrollbar data
            YScroller(sciwidgetPtr);
            Tcl_SetObjResult(interp, Tcl_NewStringObj("1", -1));
         } else {
            Tcl_SetObjResult(interp, Tcl_NewStringObj("0", -1));
         }
         break;
      }
      case CMD_TOGGLE: {
         // limit toggle to when the click is actually over the fold
         // marker; otherwise, a click within a foldable region will
         // end up folding the parent!

         int level = sci_cmd(obj, SCI_GETFOLDLEVEL, line);
         if (level & SC_FOLDLEVELHEADERFLAG) {
            sci_toggle_fold(obj, line);

            // update the Y-scrollbar data
            YScroller(sciwidgetPtr);

            Tcl_SetObjResult(interp, Tcl_NewStringObj("1", -1));
         } else {
            Tcl_SetObjResult(interp, Tcl_NewStringObj("0", -1));
         }
         break;
      }
      case CMD_ISFOLDED: {
         // return if the specified line is contained under a folded parent
         int vis = sci_cmd(obj, SCI_GETLINEVISIBLE, line);
         Tcl_SetObjResult(interp, Tcl_NewIntObj(vis ? 0 : 1));
         break;
      }
      case CMD_FOLDPARENT: {
         // return the immediate foldable parent of the specified line
         int fp = sci_cmd(obj, SCI_GETFOLDPARENT, line);
         Tcl_SetObjResult(interp, Tcl_NewIntObj(LINE2TK(fp)));
         break;
      }
      case CMD_FOLDROOT: {
         // return the top-most foldable parent of the specified line
         int fp, _fp;
         fp = _fp = sci_cmd(obj, SCI_GETFOLDPARENT, line);
         while (_fp != -1) {
            fp = _fp;
            _fp = sci_cmd(obj, SCI_GETFOLDPARENT, _fp);
         }
         Tcl_SetObjResult(interp, Tcl_NewIntObj(LINE2TK(fp)));
         break;
      }
      case CMD_SHOW: {
         // if the specified line is hidden within a fold, unfold enough
         // of its hiearchy that it is visible
         sci_cmd(obj, SCI_ENSUREVISIBLE, line);
         break;
      }
      case CMD_TOPUNFOLDED: {
         // return the highest fold parent for the target line that is unfolded
         int fp, _fp;
         fp = _fp = sci_cmd(obj, SCI_GETFOLDPARENT, line);
         int vis =  sci_cmd(obj, SCI_GETLINEVISIBLE, fp);
         while (!vis) {
            fp = _fp;
            _fp = sci_cmd(obj, SCI_GETFOLDPARENT, _fp);
            vis = sci_cmd(obj, SCI_GETLINEVISIBLE, _fp);
            if (vis) fp = _fp;
         }
         Tcl_SetObjResult(interp, Tcl_NewIntObj(LINE2TK(fp)));

         break;
      }
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdGet --
 *
 *    Implements the GET widget command
 *
 * Results:
 *    Returns a requested range(or ranges) of text in the result object
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
// look at the TEXT_GET switch-branch in the TextWidgetObjCmd function,
// located in the following file to see how the core Text code 
// implements the "get" subcommand
//   tk/generic/tkText.c

static int CmdGet(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   Tcl_Obj *objPtr = NULL;
   int i, length, found = 0;
   bool visible = false;
   char *index_str = NULL;


   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, 
         "?-displaychars? ?--? index1 ?index2 ...?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;

   // Follow parsing strategy of TextWidgetObjCmd
   i = 2;
   if (objc > 3) {
      char *name = Tcl_GetStringFromObj(objv[i], &length);
      if (length > 1 && name[0] == '-') {
         if (strncmp("-displaychars", name, (unsigned)length)==0) {
            i++;
            visible = true;

            // check the arg after "-displaychars"
            name = Tcl_GetStringFromObj(objv[i], &length);
         }
         if ((i < objc-1) && (length == 2) && !strcmp("--", name)) {
            i++;
         }
      }
   }
   // At this point, the value of "i" will be an argument
   // which should correspond with an index

   for (; i < objc; i += 2) {
      index_str = Tcl_GetString(objv[i]);

      sptr_t pos1 = -1;
      sptr_t pos2 = -1;
      if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
         goto error;
      }
      if (pos1 == -1) goto done; // index is valid, but not within document

      if (i+1 == objc) {
         // just a single index, either alone or as a single index
         // in the last pair of a sequence of indices
         pos2 = pos1 + 1;

      } else {
         index_str = Tcl_GetString(objv[i+1]);
         if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
            goto error;
         }
         if (pos2 == -1) goto done;
      }
      if (pos2 > pos1) {

         // retrieve text
         Tcl_Obj *get = GetTextRange(sciwidgetPtr, pos1, pos2, visible);

         found++;
         if (found == 1) {
            // this is either a single index or the first pair
            Tcl_SetObjResult(interp, get);
         } else {
            if (found == 2) {
               // This indicates we're processing the first index or
               // pair of indices after the first. This means we need
               // to prepair for returning a list of results.
               // Move the first item we put into the result into the
               // first element of the list object.
               //
               objPtr = Tcl_NewObj();
               Tcl_ListObjAppendElement(NULL, objPtr, Tcl_GetObjResult(interp));
            }
            Tcl_ListObjAppendElement(NULL, objPtr, get);
         }
      }
   }
   if (found > 1) {
      Tcl_SetObjResult(interp, objPtr);
   }
 done:
   return TCL_OK;

 error:
   if (objPtr) {
      Tcl_DecrRefCount(objPtr);
   }
   return BadIndex(interp, index_str);
}

/*
 *----------------------------------------------------------------------
 * CmdIncIndent --
 *
 *    Implements the increaseindent widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Text is shifted right
 *
 *----------------------------------------------------------------------
 */
static int CmdIncIndent(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   char *lineNumber, *end;
   int firstLine = -1;
   int lastLine = -1;

   if ((objc < 3) || (objc > 4)) {
      Tcl_WrongNumArgs(interp, 2, objv, "line1 ?line2?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   if (sciwidgetPtr->state == STATE_NORMAL) {
      int i;

      // simple case: either one or two indices
      lineNumber = Tcl_GetString(objv[2]);
      firstLine = strtol(lineNumber, &end, 0);
      if (*end)
         goto error;
      if (objc == 4) {
         lineNumber = Tcl_GetString(objv[3]);
         lastLine = strtol(lineNumber, &end, 0);
         if (*end)
            goto error;
      } else {
         lastLine = firstLine;
      }
      // start atomic undo operation
      sci_cmd(obj, SCI_BEGINUNDOACTION);
      // set state where tab will indent document
      sci_cmd(obj, SCI_SETTABINDENTS, 1);
      // for each specified line, insert a tab as the first character
      for (i = firstLine; i <= lastLine; ++i) {
         int pos = sci_get_pos(obj, LINE2SCI(i));
         if (pos != -1) {
            sci_add_text(obj, pos, "\t");
         }
      }
      // turn off state where tab will indent document
      sci_cmd(obj, SCI_SETTABINDENTS, 0);
      // end atomic undo operation
      sci_cmd(obj, SCI_ENDUNDOACTION);
   }
   return TCL_OK;
 error:
   Tcl_ResetResult(interp);
   Tcl_AppendResult(interp, "bad line number \"", lineNumber, "\"", NULL);
   return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 * CmdIndex --
 *
 *    Implements the INDEX widget command
 *
 * Results:
 *    The location corresponding to the request index is returned in
 *    in the form ROW.COLUMN if a match exists; otherwise, the return
 *    value is left as an empty string.
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdIndex(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   if (objc != 3 && objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "index ?-textonly?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // figure out which index is being asked for
   char *ix = Tcl_GetString(objv[2]);

   // this is only used with "@x,y" - if true, it means don't return
   // an index when over an annotation
   bool textonly = (objc == 3) 
       ? false
       : ((strcmp("-textonly", Tcl_GetString(objv[3])) == 0));

   if (!strcmp(ix, "insert")) {
      //----------------------
      //---  INSERT
      //----------------------
      int pos = static_cast<int>(sci_get_current_pos(obj));
      PosToIndex(interp, obj, pos);
      goto done;
   } else if (!strcmp(ix, "current")) {
      //----------------------
      //---  CURRENT
      //----------------------
      int pos = sciwidgetPtr->markMgr->GetCurrentPos(); // "current" mark's pos
      PosToIndex(interp, obj, pos);
      goto done;

   } else if (ix[0] == '@') {
      //----------------------
      //---  @X,Y
      //----------------------
      int x, y;
      char *cp = ix+1;
      char *end;
      x = strtol(cp, &end, 0);
      if ((end > cp) && (*end == ',')) {
         // the X location has to made relative to the text area
         int marginwidth = sciwidgetPtr->marginMgr->Width();
         marginwidth += SSM(SCI_GETMARGINLEFT,0,0); // internal text area padding
         if (x < marginwidth) {
            x = marginwidth;
         }
         cp = end+1;
         y = strtol(cp, &end, 0);
         if (end > cp) {
            int pos  = sci_get_pos_from_xy(obj, x, y);
            int line = sci_get_line(obj, pos);

            if (textonly && CheckAnnotationHit(sciwidgetPtr,line,x,y)) {
               // only return index for a line and _NOT_ its annotation text
               goto done;
            }
#if 0
// here's an example of retrieving the annotation text for a line
            char *tmp = new char[1000];
            int slen = SSM(SCI_ANNOTATIONGETTEXT, line, (sptr_t)tmp);
            if (slen) {
               tmp[slen] = '\0';
               printf(" STR(%d): %s\n",slen,tmp);
            }
            delete tmp;
#endif
            PosToIndex(interp, obj, pos);
            goto done;
#if 0
// here's an example of how to retrieve a "word" at a specific POSition
            printf("%d.%d\n",line,col);;
            Sci_TextRange tr;
            int beg = SSM(SCI_WORDSTARTPOSITION, pos, 1);
            int end = SSM(SCI_WORDENDPOSITION, pos, 1);
            tr.chrg.cpMin = beg;
            tr.chrg.cpMax = end;
            tr.lpstrText = new char[end-beg+1];
            SSM(SCI_GETTEXTRANGE,0,(sptr_t)&tr);
            //printf(">>>> FOUND (%d -> %d) <%s>\n",beg,end,tr.lpstrText);
            delete tr.lpstrText;
#endif
         }
      }
   } else {
      sptr_t pos = -1;
      if (GetPosition(sciwidgetPtr, ix, &pos, false) == TCL_OK) {
         if (pos == -1) {
             // index was a valid format, but couldn't be found
             goto done;
         }
         PosToIndex(interp, obj, pos);
         goto done;
      }
   }
   Tcl_AppendResult(interp, "bad index name: ", ix, (char*)NULL);
   return TCL_ERROR;
 done:
   return TCL_OK;
}

/*
 *---------------------------------------------------------------------------
 *
 * GetCharOffset --
 *
 *    Given a string representing an index, return the number of characters
 *    to move(forward or back) to account for the "modifier" portion of the
 *    index. For example, an index string of "1.0 + 3 chars" would mean the
 *    the returned offset is "3". Similarly, an index of "1.5+2l" would 
 *    return the number of characters needed to move two lines down from
 *    the first line, 5th character.
 *
 * Results:
 *    Returns the number of characters indicated by the index modifier
 *    portion of the input string, or zero if a value couldn't be determined
 *    in the "offset" argument. The return value from the function call
 *    is a standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *---------------------------------------------------------------------------
 */
#define CHARS_UNIT_TYPE(u,len) \
   ((*u == 'c') && (strncmp(u, "chars", len) == 0)) ||\
   ((*u == 'i') && (strncmp(u, "indices", len) == 0)) ||\
   ((*u == 'd') && (strncmp(u, "displayindices", len) == 0))

static int
GetCharOffset(
    SciWidget  *sciwidgetPtr, 
    int         start_pos, 
    CONST char *index_str, 
    char      **rp, // will point to the remaining part of modifier (if any)
    int        *offset
) {
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   *rp = '\0';
   if (index_str == NULL) {
      *offset = 0;
      return TCL_OK;
   }
   CONST char *cp = index_str;
   while (isspace(*cp)) cp++; // remove leading space
   if (*cp == '\0') {
      *offset = 0;
      return TCL_OK;
   }

   // check for +/-
 beginParse:
   if ((*cp == '+') || (*cp == '-')) {
      bool plus = (*cp == '+');
      cp++;
      while (isspace(*cp)) cp++;
      if ((*cp == '+') || (*cp == '-')) {
          // we've hit something like "... 5.0 + +3 chars" (the 2nd '+' char)
          goto beginParse;
      }
      // get the number of units
      if (isdigit(*cp)) {
         char *end;
         int num = strtol(cp, &end, 0);

         if (end == cp) return TCL_ERROR;

         while (isspace(*end)) end++;

         // check for those funky submodifiers (ie "display" and "any")
         if ((*end == 'd') && (strncmp(end, "display", 7) == 0)) {
             end += 7;

         } else if ((*end == 'a') && (strncmp(end, "any", 3) == 0)) {
             end += 3;
         }
         while (isspace(*end)) end++;

         // find out how long the units text is(ie "c", "ch", "chars", etc)
         char *units = end;
         while (*end!='\0' && !isspace(*end) && *end!='+' && *end!='-') {
            end++;
         }
         int len = end - units;

         *rp = units + len;

         // Note: the Tk Text widget's unit type of "indices" is the same
         // as "chars" except it considers embedded images and windows.
         // Since neither of those is supported in our(ie Scintilla) widget
         // we can consider those two types equivalent.

         if (CHARS_UNIT_TYPE(units,len)) {
            *offset = (plus ? num : -num);
            return TCL_OK;
         }
         if ((*units == 'l') && (strncmp(units, "lines", len) == 0)) {
            // this one is trickier because the offset needs to consider
            // the column of the starting position so that the offset end
            // up in the same column(if possible) of the target line

            int start_col = sci_get_column(obj, start_pos);
            int start_ln  = sci_get_line(obj, start_pos);
            int new_ln = (plus ? (start_ln + num) : (start_ln - num));
            if (new_ln > sci_get_line_count(obj)) {
               // beyond last line, so use offset to end of document
               *offset = (sci_get_length(obj) - start_pos);
               return TCL_OK;
            }
            if (new_ln < 0) {
              *offset = (plus ? start_pos : -start_pos);
               return TCL_OK;
            }
            // ok, so we know we'll land on a valid line. check that it
            // has a column to match the starting position, and if not
            // then use its end of line
            int endcol = sci_get_line_length(obj, new_ln);
            endcol--; // ignore eol char
            if (start_col > endcol) {
               start_col = endcol;
            }
            int ln_start = sci_get_pos(obj, new_ln);
            *offset = (ln_start + start_col - start_pos);
            return TCL_OK;
         }
      }
   } else if ((*cp == 'l') && (strncmp(cp, "lineend", 7) == 0)) {
      int ln  = sci_get_line(obj, start_pos);
      int pos = sci_get_lineend_pos(obj, ln);
      *offset = (pos - start_pos);
      *rp = const_cast<char *>(cp)+7;
      return TCL_OK;

   } else if ((*cp == 'l') && (strncmp(cp, "linestart", 9) == 0)) {
      int ln  = sci_get_line(obj, start_pos);
      int pos = sci_get_pos(obj, ln);
      *offset = (pos - start_pos);
      *rp = const_cast<char *>(cp)+9;
      return TCL_OK;

   } else if ((*cp == 'w') && (strncmp(cp, "wordstart", 9) == 0)) {
       // NOTE: We don't use the behavior of Scintilla's SCI_WORDSTARTPOSITION
       // command because it differs from what the Tk Text widget does. Specifically,
       // if the current character(ie the one to the right of
       // of start_pos) is a non-word character then we do _NOT_ move any
       // further along in the word.
       int pos = start_pos;
       if (Tcl_UniCharIsWordChar((char)sci_cmd(obj, SCI_GETCHARAT, start_pos))) {
           pos = sci_cmd(obj, SCI_WORDSTARTPOSITION, start_pos, 1);
       }
       *offset = (pos - start_pos);
       *rp = const_cast<char *>(cp)+9;
       return TCL_OK;

   } else if ((*cp == 'w') && (strncmp(cp, "wordend", 7) == 0)) {
       int pos = sci_cmd(obj, SCI_WORDENDPOSITION, start_pos, 1);
       *offset = (pos - start_pos);
       *rp = const_cast<char *>(cp)+7;
       return TCL_OK;
   } else {
       // another check for those funky submodifiers (ie "display" and "any")
       if ((*cp == 'd') && (strncmp(cp, "display", 7) == 0)) {
           *rp = const_cast<char *>(cp)+7;
           return TCL_OK;
       } else if ((*cp == 'a') && (strncmp(cp, "any", 3) == 0)) {
           *rp = const_cast<char *>(cp)+3;
           return TCL_OK;
       }
   }
   return TCL_ERROR;
}

/*
 *---------------------------------------------------------------------------
 *
 * GetPosition --
 *
 *    Given a string, return the document position that is described.
 *
 * Results:
 *    The return value is a standard Tcl return result. If TCL_OK is
 *    returned, then everything went well and the position is assigned
 *    to *posPtr; otherwise TCL_ERROR is returned and an error message is
 *    filled in left in the interp's result. If the "visible" flag is
 *    on, then a position that is currently not visible due to folding
 *    will end up not setting posPtr.
 *
 * Side effects:
 *    None.
 *
 *---------------------------------------------------------------------------
 */
static int GetPosition(
   SciWidget  *sciwidgetPtr,
   CONST char *index_str,
   sptr_t     *posPtr,
   bool        visible)
{
    char *cp, *endp;
    sptr_t pos, curr_pos;
    int offset, line;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    if (index_str[0] == '@') {
        //----------------------
        //---  @X,Y
        //----------------------
        int x, y;
        cp = const_cast<char *>(index_str)+1;
        x = strtol(cp, &endp, 0);
        if ((endp > cp) && (*endp == ',')) {
            // the X location has to made relative to the text area
            int marginwidth = sciwidgetPtr->marginMgr->Width();
            marginwidth += SSM(SCI_GETMARGINLEFT,0,0); // internal text area padding
            if (x < marginwidth) {
                x = marginwidth;
            }
            cp = endp+1;
            y = strtol(cp, &endp, 0);
            if (endp > cp) {
                pos  = sci_get_pos_from_xy(obj, x, y);
                line = sci_get_line(obj, pos);

                if (CheckAnnotationHit(sciwidgetPtr,line,x,y)) {
                    // only return index for a line and _NOT_ its annotation text
                    goto done;
                }
                *posPtr = pos;
                cp = endp;
                goto done;
            }
        }
    } else if (isdigit(index_str[0])) {
        //----------------------
        //--- LINE.COLUMN
        //----------------------
        int lineIndex, charIndex;
        lineIndex = strtol(index_str, &endp, 0);
        // Special check to allow client to use a Tk index of "0.0". When this
        // happens we want to treat it the same as "1.0"
        if (lineIndex==0 && (index_str[0] == '0')) {
            lineIndex = 1;
        }
        if (lineIndex) {
            if (visible && !SSM(SCI_GETLINEVISIBLE, lineIndex, 0)) {
                // a valid line but it's not currently visible
                cp = NULL;
                goto done;
            }
            if (endp == index_str) {
                // missing the ".column" part of index!!
                goto error;
            }
            cp = endp+1;
            if ((*cp == 'e') && (strncmp(cp, "end", 3) == 0)) {
                // this is "line.end" format
                *posPtr = sci_get_lineend_pos(obj, LINE2SCI(lineIndex));
                cp = NULL;
                goto done;
            } else {
                charIndex = strtol(cp, &endp, 0);
                if (endp == cp) {
                    // uh oh, remaining part isn't a number
                    goto error;
                }
                pos = sci_get_pos(obj, LINE2SCI(lineIndex));
                if (pos != -1) {
                    pos += charIndex;
                    // sanity check here to prevent a column which was greater than
                    // the last one on the line from causing us to generate a
                    // position which is actually on the next line
                    int endpos = sci_get_lineend_pos(obj, LINE2SCI(lineIndex));
                    if (pos > endpos) {
                        int nlines   = SSM(SCI_GETLINECOUNT, 0,0); // << #lines in document
                        // last line of document is special case, no end of line!
                        if (nlines > LINE2SCI(lineIndex))
                            pos = endpos;
                    }
                }
                *posPtr = pos;
                cp = endp;
                goto done;
            }
        }
    } else if ((cp = const_cast<char *>(strrchr(index_str, '.'))) != NULL) {
        // Check for form "tag.first" or "tag.last" where "tag" is the name of
        // an existing tag.

        int tagLen = cp-index_str;
        char *tagName = ckalloc(tagLen+1);
        strncpy(tagName, index_str, tagLen);
        tagName[tagLen] = 0;
        // check for the selection tag since it's handled special (it's
        // not kept in the TagMgr hash table)
        //
        TagData *tagPtr = sciwidgetPtr->tagMgr->FindTag(tagName);
        ckfree(tagName);
        if (!tagPtr) {
            // unable to find tag
            goto error;
        }
        if (tagPtr == sciwidgetPtr->tagMgr->GetSelectTag()) {
            if ((cp[1] == 'f') && (strncmp(cp+1, "first", 5) == 0)) {
                if (sci_has_selection(obj)) {
                    *posPtr = sci_cmd(obj, SCI_GETSELECTIONSTART);
                }
                cp = NULL;
                goto done;
            } else if ((cp[1] == 'l') && (strncmp(cp+1, "last", 4) == 0)) {
                if (sci_has_selection(obj)) {
                    *posPtr = sci_cmd(obj, SCI_GETSELECTIONEND);
                }
                cp = NULL;
                goto done;
            } else {
                goto error;
            }
        } else {
            if (sciwidgetPtr->tagMgr->IsIndicator(tagPtr)) {
                // find first or last instance of given tag in document
                int first;
                if ((cp[1] == 'f') && (strncmp(cp+1, "first", 5) == 0)) {
                    first = 1;
                } else if ((cp[1] == 'l') && (strncmp(cp+1, "last", 4) == 0)) {
                    first = 0;
                } else {
                    goto error;
                }
                int id = tagPtr->IndicID();
                int found = 0;
                for (int i=0; i < sci_get_length(obj); i++) {
                    int beg = sci_cmd(obj, SCI_INDICATORSTART, id, i);
                    int end = sci_cmd(obj, SCI_INDICATOREND,   id, beg);
                    // check if the indicator exists at all within the document
                    if ((beg == end) && (beg == 0)) {
                        if (found)
                            break;
                        else
                            goto error;
                    }
                    i = end;
                    int mask = sci_cmd(obj, SCI_INDICATORALLONFOR, beg);
                    if (mask & (1 << id)) {
                        found = 1;
                        if (first) {
                            *posPtr = beg;
                            cp = NULL;
                            goto done;
                        } else {
                            *posPtr = end;
                        }
                    } else {
                        --i;
                    }
                }
                cp = NULL;
                goto done;
            } else {
                // unable to handle non-selection non-indicator type tag
                goto error;
            }
        }
    } else if (index_str[0] == 'c' && (strncmp(index_str, "current", 7) == 0)) {
        //----------------------
        //--- CURRENT
        //----------------------
        *posPtr = sciwidgetPtr->markMgr->GetCurrentPos(); // "current" mark's pos
        cp = const_cast<char *>(index_str) + 7;
        goto done;
    } else if (index_str[0] == 'e' && (strncmp(index_str, "end", 3) == 0)) {
        //----------------------
        //--- END
        //----------------------
        *posPtr = sci_get_length(obj);
        cp = const_cast<char *>(index_str) + 3;
        goto done;

    } else if (index_str[0] == 'i' && (strncmp(index_str, "insert", 6) == 0)) {
        //----------------------
        //--- INSERT
        //----------------------
        *posPtr = sci_get_current_pos(obj);
        cp = const_cast<char *>(index_str) + 6;
        goto done;
    } else {
        //----------------------
        //--- MARK
        //----------------------
        // nothing else matched, so let's see if it's a mark name
        int _offset = 0;
        MarkData *markPtr = sciwidgetPtr->markMgr->FindMark(index_str, &_offset);

        if (markPtr != NULL) {
            *posPtr = markPtr->pos;
            cp = const_cast<char *>(index_str) + (_offset ? _offset : strlen(index_str));
            goto done;
        } else {
            // clear out the error that the mark mgr will have left
            Tcl_ResetResult(sciwidgetPtr->interp);
        }
    }
 error:
    return TCL_ERROR;
 done:
    if (cp != '\0' && strlen(cp)) {
        // handle modifier portion(s) for the remainder of the index string
        curr_pos = *posPtr;
        offset = 0;
        while (1) {
            int off = 0;
            char *rp = cp;
            if (GetCharOffset(sciwidgetPtr, curr_pos, cp, &rp, &off) != TCL_OK)
                return TCL_ERROR;
            //printf(">> CP=%s  REMAINDER=%s  OFF=%d  OFFSET=%d\n",cp,rp,off,offset);
            curr_pos += off;
            if (curr_pos < 0) {
                // when moving into negative space, adjust the offset
                // so we end up "clamping" it to the beginning
                off += -curr_pos;
                curr_pos = 0;
            }
            offset += off;
            if (!rp || strlen(rp) == 0) break; // nothing left to process
            cp = rp;
        }
    } else {
        offset = 0;
    }
    *posPtr += offset;
    //printf(">>>> POS %d\n",*posPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetTextRange --
 *
 *    Returns the text from pos1 to pos2, placing that text in a
 *    string object which is returned with a refCount of zero.
 *
 * Results:
 *    Tcl_Obj of string type containing the specified text. If the
 *    "visibleOnly" flag is set to 1, then only those characters which
 *    are not folded will be returned. Otherwise (flag is 0) all 
 *    characters in the given range are returned.
 *
 * Side effects:
 *    Memory will be allocated for the new object. Remember to free it
 *    if it isn't going to be stored appropriately.
 *
 *----------------------------------------------------------------------
 */
static Tcl_Obj *GetTextRange(
   SciWidget *sciwidgetPtr,
   int        pos1, 
   int        pos2,
   bool       visibleOnly)
{
   Tcl_Obj *resultPtr = Tcl_NewObj();
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   int line1, line2, max;

   // just ignore bogus positions
   if (pos2 <= pos1) goto done;

   // safety check to catch a situation that throws an assertion
   // down in the Scintilla code (in Editor.cxx)
   max = sci_get_length(obj);
   if (pos2 > max) goto done;

   line1 = sci_get_line(obj, pos1);
   line2 = sci_get_line(obj, pos2);

   if (visibleOnly && (line1 != line2)) {
      for (int ln = line1; ln <= line2; ln++) {
         if (! SSM(SCI_GETLINEVISIBLE, ln, 0)) continue;
         int endpos = sci_get_lineend_pos(obj, ln);
         endpos++; // include EOL character
         if (ln == line1) {
            // get from pos1 to end of the line
            ExtractText(sciwidgetPtr, resultPtr, pos1, endpos);
            continue;
         }
         int beginpos = sci_get_pos(obj, ln);
         if (ln == line2) {
            // get from start of line to pos2
            ExtractText(sciwidgetPtr, resultPtr, beginpos, pos2);
         } else {
            // get the whole line
            ExtractText(sciwidgetPtr, resultPtr, beginpos, endpos);
         }
      }
   } else {
      ExtractText(sciwidgetPtr, resultPtr, pos1, pos2);
   }
 done:
   return resultPtr;
}
static void ExtractText(SciWidget *sciwidgetPtr, Tcl_Obj *objPtr, int pos1, int pos2) {

    PLATFORM_ASSERT(pos2 >= pos1);
    Sci_TextRange tr;
    tr.chrg.cpMin = pos1;
    tr.chrg.cpMax = pos2;
    tr.lpstrText = new char[pos2-pos1+1];
    SSM(SCI_GETTEXTRANGE, 0, (sptr_t)&tr);

    Tcl_AppendToObj(objPtr, tr.lpstrText, strlen(tr.lpstrText));
    delete[] tr.lpstrText;
}

/*
 *----------------------------------------------------------------------
 * CheckAnnotationHit --
 *
 *    Determine if an X/Y screen location is over the annotation
 *    text associated with a specific line
 *
 * Results:
 *    The location corresponding to the requested index is returned in
 *    in the form ROW.COLUMN if a match exists; otherwise, the return
 *    value is left as an empty string.
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static bool CheckAnnotationHit(
   SciWidget *sciwidgetPtr,
   int        line,
   int        x,
   int        y)
{
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // confirm if annotations are being displayed
   int astyle = SSM(SCI_ANNOTATIONGETVISIBLE, 0,0);
   if (astyle == ANNOTATION_HIDDEN) {
      return false; // nothing to do
   }
   // check if this specific line has an annotation
   int annlines = SSM(SCI_ANNOTATIONGETLINES, line, 0);
   if (annlines == 0) {
      return false;
   }
   // we know the line has an annotation, check if the x/y screen
   // location is over the line or the annotation

   // get the document position of the line
   int pos = sci_get_pos(obj, line);

   // determine the extents for the line
   int top_y = sci_get_y_from_pos(obj, pos);
   int lh    = sci_get_text_height(obj, line);
   int bot_y = top_y + lh;

   if (y <= bot_y) {
#if 0
      printf("OVER LINE %d %d\n",line, LINE2TK(line));
#endif
      return false;
   } else {
      // we must be over the annotation area, but we aren't
      // sure yet if we're actually over the annotation text

      // get the amount of indentation that precedes the text
      int indent     = SSM(SCI_TKMSG_ANNOTATIONINDENT, line, 1);
      if (x >= indent) {
#if 0
         int p    = sci_get_pos_from_xy(obj, x, (y-lh));
         int col  = sci_get_column(obj, p);
         printf("ANNO FOR LINE %d  COL=%d\n",LINE2TK(line),col);
#endif
         return true;
      }
   }
   return false;
}

/*
 *----------------------------------------------------------------------
 * CmdInsert --
 *
 *    Implements the INSERT widget command
 *
 * Results:
 *    The location corresponding to the request index is returned in
 *    in the form ROW.COLUMN if a match exists; otherwise, the return
 *    value is left as an empty string.
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdInsert(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "index text");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   if (sciwidgetPtr->state == STATE_NORMAL) {

      char *index_str = Tcl_GetString(objv[2]);
      char *to_add    = Tcl_GetString(objv[3]);

      // get the position where we'll do the add
      sptr_t pos;
      if (GetPosition(sciwidgetPtr, index_str, &pos, false) != TCL_OK) {
         return TCL_ERROR;
      }
      sci_add_text(obj, pos, to_add);
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdLoadLexer --
 *
 *    Implements the LOADLEXER widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    The lexer for given language is demand-loaded
 *
 *----------------------------------------------------------------------
 */
static int CmdLoadLexer(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;

   if (objc != 3) {
      Tcl_WrongNumArgs(interp, 1, objv, "library");
      return TCL_ERROR;
   }

   char *library    = Tcl_GetString(objv[2]);
   // always returns 0, so can't check return value
   SSM(SCI_LOADLEXERLIBRARY, 0, (sptr_t)library);
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdSciSearch --
 *
 *    Implements the SCISEARCH widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    The display may be scrolled and/or the selection changed to
 *    match the requested search options.
 *
 *----------------------------------------------------------------------
 */
static int CmdSciSearch(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // NOTE: The list of supported switches is designed to match the standard
   // Tk Text widget, but _ONLY_ those options that can be easily implemented
   // using the Scintilla widget.
   static CONST char *optionStrings[] = {
      "--", /*"-all",*/ "-backwards", "-count", /*"-elide",*/ "-exact",
      "-forwards", /*"-hidden",*/ "-nocase", "-regexp", "-word", NULL
   };
   enum options {
      CMD_END, /*CMD_ALL,*/ CMD_BACK, CMD_COUNT, /*CMD_ELIDE,*/ CMD_EXACT, 
      CMD_FWD, /*CMD_HIDDEN,*/ CMD_NOCASE, CMD_REGEXP, CMD_WORD
   };

   // Parse the switch options
   int i, argsLeft, found;
   bool back = false;
   char *pattern, *index_str;
   Tcl_Obj *count_var = NULL;
   sptr_t pos1 = -1;
   sptr_t pos2 = -1;
   int flags = SCFIND_MATCHCASE; // default search flags

   for (i=2; i < objc; i++) {
      int index;

      if (Tcl_GetString(objv[i])[0] != '-') break;

      if (Tcl_GetIndexFromObj(interp, objv[i], optionStrings, "switch", 0, &index) != TCL_OK) {
         return TCL_ERROR;
      }
      switch ((enum options)index) {
         case CMD_END:
            i++;
            goto endOfSwitchProcessing;
         case CMD_BACK:
            back = true;
            break;
         case CMD_NOCASE:
            flags &= ~SCFIND_MATCHCASE;
            break;
         case CMD_REGEXP:
            flags |= SCFIND_REGEXP;
            break;
         case CMD_WORD:
            flags |= SCFIND_WHOLEWORD;
            break;
         case CMD_COUNT:
            count_var = objv[++i];
            break;
         case CMD_EXACT:
         case CMD_FWD:
            // provide compatible API with Tk Text widget
            break;
         default:
            // this is the same response found in tkText.c(TextSearchCmd)
             Tcl_Panic("unexpected switch fallthrough >> %s\n", optionStrings[index]);
      }
   }
 endOfSwitchProcessing:
   argsLeft = objc - (i+2);
   if (argsLeft != 0 && argsLeft != 1) {
      Tcl_WrongNumArgs(interp, 2, objv, "?switches? pattern index ?stopIndex?");
      return TCL_ERROR;
   }
   // at this point we know we have the pattern and at least one index
   pattern   = Tcl_GetString(objv[i]);
   index_str = Tcl_GetString(objv[i+1]);
   if (GetPosition(sciwidgetPtr, index_str, &pos1, false) == TCL_OK) {
      if (pos1 == -1) {
         goto badIndex;
      }
   }
   if (argsLeft == 1) {
      // we were given the second index
      index_str = Tcl_GetString(objv[i+2]);
      if (GetPosition(sciwidgetPtr, index_str, &pos2, false) == TCL_OK) {
         if (pos2 == -1) {
            goto badIndex;
         }
         int tmp = pos1;
         if (back) {
            pos1 = Platform::Minimum(pos1, pos2);
            pos2 = Platform::Maximum(tmp,  pos2);
         } else {
            pos1 = Platform::Maximum(pos1, pos2);
            pos2 = Platform::Minimum(tmp,  pos2);
         }
      }
   } else {
      // only one index, so figure out how to limit the range
      if (back) {
         pos2 = pos1;
         pos1 = 0; // search until start of document
      } else {
         pos2 = sci_get_length(obj); // search until end of document
      }
   }

   // Ok, now we can try the search
   sci_cmd(obj, SCI_SETSEARCHFLAGS, flags);
   if (back) {
      sci_cmd(obj, SCI_SETTARGETSTART, pos2);
      sci_cmd(obj, SCI_SETTARGETEND,   pos1);
   } else {
      sci_cmd(obj, SCI_SETTARGETSTART, pos1);
      sci_cmd(obj, SCI_SETTARGETEND,   pos2);
   }
   found = sci_cmd(obj, SCI_SEARCHINTARGET, strlen(pattern), (sptr_t)pattern);

   if ((found == -1) && (argsLeft != 1)) {
       // We're not doing a "mark all" and we didn't find a match. So since
       // we were only given the start index, perform a "wrapped" search
       if (back) {
           sci_cmd(obj, SCI_SETTARGETSTART, sci_get_length(obj));
           sci_cmd(obj, SCI_SETTARGETEND,   pos1+1);
       } else {
           sci_cmd(obj, SCI_SETTARGETSTART, 0);
           sci_cmd(obj, SCI_SETTARGETEND,   pos1);
       }
       found = sci_cmd(obj, SCI_SEARCHINTARGET, strlen(pattern), (sptr_t)pattern);
   }

   if (found == -1) {
      Tcl_ResetResult(interp); // return an empty string
   } else {
      PosToIndex(interp, obj, found);
      if (count_var) {
          int p1 = sci_cmd(obj, SCI_GETTARGETSTART);
          int p2 = sci_cmd(obj, SCI_GETTARGETEND);

          Tcl_ObjSetVar2(interp, count_var, NULL, Tcl_NewIntObj(p2-p1), TCL_LEAVE_ERR_MSG);
      }
   }
   return TCL_OK;

 badIndex:
   return BadIndex(interp, index_str);
}


// The following code was taken from tk/generic/tkText.c
// It was written to be possibly shared outside the context of tkText.
// BSG - Sept. 2012
//Begin copy-
/*
 *----------------------------------------------------------------------
 *
 * SearchPerform --
 *
 *	Overall control of search process. Is given a pattern, a starting
 *	index and an ending index, and attempts to perform a search. This
 *	function is actually completely independent of Tk, and could in the
 *	future be split off.
 *
 * Results:
 *	Standard Tcl result code. In particular, if fromPtr or toPtr are not
 *	considered valid by the 'lineIndexProc', an error will be thrown and
 *	no search performed.
 *
 * Side effects:
 *	See 'SearchCore'.
 *
 *----------------------------------------------------------------------
 */

static int
SearchPerform(
    Tcl_Interp *interp,         /* For error messages. */
    SearchSpec *searchSpecPtr,  /* Search parameters. */
    Tcl_Obj *patObj,            /* Contains an exact string or a regexp
                                 * pattern. Must have a refCount > 0. */
    Tcl_Obj *fromPtr,           /* Contains information describing the first
                                 * index. */
    Tcl_Obj *toPtr)             /* NULL or information describing the last
                                 * index. */
{
   /*
	* Find the starting line and starting offset (measured in Unicode chars
	* for regexp search, utf-8 bytes for exact search).
	*/

   if ((*searchSpecPtr->lineIndexProc)(interp, fromPtr, searchSpecPtr,
									   &searchSpecPtr->startLine,
									   &searchSpecPtr->startOffset) != TCL_OK) {
		return TCL_ERROR;
   }

   /*
	* Find the optional end location, similarly.
	*/

   if (toPtr != NULL) {
		SciWidget *sciwidgetPtr = (SciWidget *)searchSpecPtr->clientData;
		sptr_t posTo, posFrom;

		if (GetPosition(sciwidgetPtr, Tcl_GetString(toPtr), &posTo, false) != TCL_OK) {
			 return BadIndex(interp, Tcl_GetString(toPtr));
		}

		if (GetPosition(sciwidgetPtr, Tcl_GetString(fromPtr), &posFrom, false) != TCL_OK) {
			 return BadIndex(interp, Tcl_GetString(fromPtr));
		}

		/*
		 * Check for any empty search range here. It might be better in the
		 * future to embed that in SearchCore (whose default behaviour is to
		 * wrap when given a negative search range).
		 */

		if (searchSpecPtr->backwards) {
			 if (posFrom - posTo < 0) {
				  return TCL_OK;
			 }
		} else {
			 if (posFrom - posTo > 0) {
				  return TCL_OK;
			 }
		}

		if ((*searchSpecPtr->lineIndexProc)(interp, toPtr, searchSpecPtr,
											&searchSpecPtr->stopLine,
											&searchSpecPtr->stopOffset) != TCL_OK) {
			 return TCL_ERROR;
		}
   } else {
		searchSpecPtr->stopLine = -1;
   }

   /*
	* Scan through all of the lines of the text circularly, starting at the
	* given index. 'patObj' is the pattern which may be an exact string or a
	* regexp pattern depending on the flags in searchSpecPtr.
	*/

   return SearchCore(interp, searchSpecPtr, patObj);
}

/*
 *----------------------------------------------------------------------
 *
 * SearchCore --
 *
 *	The core of the search function. This function is actually completely
 *	independent of Tk, and could in the future be split off.
 *
 *	The function assumes regexp-based searches operate on Unicode strings,
 *	and exact searches on utf-8 strings. Therefore the 'foundMatchProc'
 *	and 'addLineProc' need to be aware of this distinction.
 *
 * Results:
 *	Standard Tcl result code.
 *
 * Side effects:
 *	Only those of the 'searchSpecPtr->foundMatchProc' which is called
 *	whenever a match is found.
 *
 *	Note that the way matching across multiple lines is implemented, we
 *	start afresh with each line we have available, even though we may
 *	already have examined the contents of that line (and further ones) if
 *	we were attempting a multi-line match using the previous line. This
 *	means there may be ways to speed this up a lot by not throwing away
 *	all the multi-line information one has accumulated. Profiling should
 *	be done to see where the bottlenecks lie before attempting this,
 *	however. We would also need to be very careful such optimisation keep
 *	within the specified search bounds.
 *
 *----------------------------------------------------------------------
 */

static int
SearchCore(
    Tcl_Interp *interp,         /* For error messages. */
    SearchSpec *searchSpecPtr,  /* Search parameters. */
    Tcl_Obj *patObj)            /* Contains an exact string or a regexp
                                 * pattern. Must have a refCount > 0. */
{
   /*
	* For exact searches these are utf-8 char* offsets, for regexp searches
	* they are Unicode char offsets.
	*/

   int firstOffset, lastOffset, matchOffset, matchLength;
   int passes;
   int lineNum = searchSpecPtr->startLine;
   int code = TCL_OK;
   Tcl_Obj *theLine;
   int alreadySearchOffset = -1;

   CONST char *pattern = NULL;	/* For exact searches only. */
   int firstNewLine = -1; 	/* For exact searches only. */
   Tcl_RegExp regexp = NULL;	/* For regexp searches only. */

   /*
	* These items are for backward regexp searches only. They are for two
	* purposes: to allow us to report backwards matches in the correct order,
	* even though the implementation uses repeated forward searches; and to
	* provide for overlap checking between backwards matches on different
	* text lines.
	*/

#define LOTS_OF_MATCHES 20
   int matchNum = LOTS_OF_MATCHES;
   int smArray[2 * LOTS_OF_MATCHES];
   int *storeMatch = smArray;
   int *storeLength = smArray + LOTS_OF_MATCHES;
   int lastBackwardsLineMatch = -1;
   int lastBackwardsMatchOffset = -1;

   if (searchSpecPtr->exact) {
		/*
		 * Convert the pattern to lower-case if we're supposed to ignore case.
		 */

		if (searchSpecPtr->noCase) {
			 patObj = Tcl_DuplicateObj(patObj);

			 /*
			  * This can change the length of the string behind the object's
			  * back, so ensure it is correctly synchronised.
			  */

			 Tcl_SetObjLength(patObj, Tcl_UtfToLower(Tcl_GetString(patObj)));
		}
   } else {
		/*
		 * Compile the regular expression. We want '^$' to match after and
		 * before \n respectively, so use the TCL_REG_NLANCH flag.
		 */

		regexp = Tcl_GetRegExpFromObj(interp, patObj,
									  (searchSpecPtr->noCase ? TCL_REG_NOCASE : 0)
									  | (searchSpecPtr->noLineStop ? 0 : TCL_REG_NLSTOP)
									  | TCL_REG_ADVANCED | TCL_REG_CANMATCH | TCL_REG_NLANCH);
		if (regexp == NULL) {
			 return TCL_ERROR;
		}
   }

   /*
	* For exact strings, we want to know where the first newline is, and we
	* will also use this as a flag to test whether it is even possible to
	* match the pattern on a single line. If not we will have to search
	* across multiple lines.
	*/

   if (searchSpecPtr->exact) {
		CONST char *nl;

		/*
		 * We only need to set the matchLength once for exact searches, and we
		 * do it here. It is also used below as the actual pattern length, so
		 * it has dual purpose.
		 */

		pattern = Tcl_GetStringFromObj(patObj, &matchLength);
		nl = strchr(pattern, '\n');

		/*
		 * If there is no newline, or it is the very end of the string, then
		 * we don't need any special treatment, since single-line matching
		 * will work fine.
		 */

		if (nl != NULL && nl[1] != '\0') {
			 firstNewLine = (nl - pattern);
		}
   } else {
		matchLength = 0;	/* Only needed to prevent compiler warnings. */
   }

   /*
	* Keep a reference here, so that we can be sure the object doesn't
	* disappear behind our backs and invalidate its contents which we are
	* using.
	*/

   Tcl_IncrRefCount(patObj);

   /*
	* For building up the current line being checked.
	*/

   theLine = Tcl_NewObj();
   Tcl_IncrRefCount(theLine);

   for (passes = 0; passes < 2; ) {
		ClientData lineInfo;
		int linesSearched = 1;
		int extraLinesSearched = 0;

		if (lineNum >= searchSpecPtr->numLines) {
			 /*
			  * Don't search the dummy last line of the text.
			  */

			 goto nextLine;
		}

		/*
		 * Extract the text from the line, storing its length in 'lastOffset'
		 * (in bytes if exact, chars if regexp), since obviously the length is
		 * the maximum offset at which it is possible to find something on
		 * this line, which is what 'lastOffset' represents.
		 */

		lineInfo = (*searchSpecPtr->addLineProc)(lineNum, searchSpecPtr,
												 theLine, &lastOffset, &linesSearched);
		
		if (lineInfo == NULL) {
			 /*
			  * This should not happen, since 'lineNum' should be valid in the
			  * call above. However, let's try to be flexible and not cause a
			  * crash below.
			  */

			 goto nextLine;
		}

		if (lineNum == searchSpecPtr->stopLine && searchSpecPtr->backwards) {
			 firstOffset = searchSpecPtr->stopOffset;
		} else {
			 firstOffset = 0;
		}

		if (alreadySearchOffset != -1) {
			 if (searchSpecPtr->backwards) {
				  if (alreadySearchOffset < lastOffset) {
					   lastOffset = alreadySearchOffset;
				  }
			 } else {
				  if (alreadySearchOffset > firstOffset) {
					   firstOffset = alreadySearchOffset;
				  }
			 }
			 alreadySearchOffset = -1;
		}
		
		if (lineNum == searchSpecPtr->startLine) {
			 /*
			  * The starting line is tricky: the first time we see it we check
			  * one part of the line, and the second pass through we check the
			  * other part of the line.
			  */
			 
			 passes++;
			 if ((passes == 1) ^ searchSpecPtr->backwards) {
				  /*
				   * Forward search and first pass, or backward search and
				   * second pass.
				   *
				   * Only use the last part of the line.
				   */
				  
				  if (searchSpecPtr->startOffset > firstOffset) {
					   firstOffset = searchSpecPtr->startOffset;
				  }
				  if ((firstOffset >= lastOffset)
					  && ((lastOffset != 0) || searchSpecPtr->exact)) {
					   goto nextLine;
				  }
			 } else {
				  /*
				   * Use only the first part of the line.
				   */
				  
				  if (searchSpecPtr->startOffset < lastOffset) {
					   lastOffset = searchSpecPtr->startOffset;
				  }
			 }
		}
		
		/*
		 * Check for matches within the current line 'lineNum'. If so, and if
		 * we're searching backwards or for all matches, repeat the search
		 * until we find the last match in the line. The 'lastOffset' is one
		 * beyond the last position in the line at which a match is allowed to
		 * begin.
		 */
		
		matchOffset = -1;
		
		if (searchSpecPtr->exact) {
			 int maxExtraLines = 0;
			 CONST char *startOfLine = Tcl_GetString(theLine);
			 
			 do {
				  Tcl_UniChar ch;
				  CONST char *p;
				  int lastFullLine = lastOffset;
				  
				  if (firstNewLine == -1) {
					   if (searchSpecPtr->strictLimits
						   && (firstOffset + matchLength > lastOffset)) {
							/*
							 * Not enough characters to match.
							 */
							
							break;
					   }
					   
					   /*
						* Single line matching. We want to scan forwards or
						* backwards as appropriate.
						*/
					   
					   if (searchSpecPtr->backwards) {
							/*
							 * Search back either from the previous match or from
							 * 'startOfLine + lastOffset - 1' until we find a
							 * match.
							 */
							
							CONST char c = pattern[0];
							
							if (alreadySearchOffset != -1) {
								 p = startOfLine + alreadySearchOffset;
								 alreadySearchOffset = -1;
							} else {
								 p = startOfLine + lastOffset -1;
							}
							while (p >= startOfLine + firstOffset) {
								 if (p[0] == c && !strncmp(p, pattern,
														   (unsigned)matchLength)) {
									  goto backwardsMatch;
								 }
								 p--;
							}
							break;
					   } else {
							p = strstr(startOfLine + firstOffset, pattern);
					   }
					   if (p == NULL) {
							/*
							 * Single line match failed.
							 */
							
							break;
					   }
				  } else if (firstNewLine >= (lastOffset - firstOffset)) {
					   /*
						* Multi-line match, but not enough characters to match.
						*/
					   
					   break;
				  } else {
					   /*
						* Multi-line match has only one possible match position,
						* because we know where the '\n' is.
						*/
					   
					   p = startOfLine + lastOffset - firstNewLine - 1;
					   if (strncmp(p, pattern, (unsigned)(firstNewLine + 1))) {
							/*
							 * No match.
							 */
							
							break;
					   } else {
							int extraLines = 1;
							
							/*
							 * If we find a match that overlaps more than one
							 * line, we will use this value to determine the first
							 * allowed starting offset for the following search
							 * (to avoid overlapping results).
							 */
							
							int lastTotal = lastOffset;
							int skipFirst = lastOffset - firstNewLine -1;
							
							/*
							 * We may be able to match if given more text. The
							 * following 'while' block handles multi-line exact
							 * searches.
							 */
							
							while (1) {
								 lastFullLine = lastTotal;
								 
								 if (lineNum+extraLines>=searchSpecPtr->numLines) {
									  p = NULL;
									  break;
								 }
								 
								 /*
								  * Only add the line if we haven't already done so
								  * already.
								  */
								 
								 if (extraLines > maxExtraLines) {
									  if ((*searchSpecPtr->addLineProc)(lineNum
												+ extraLines, searchSpecPtr, theLine,
												&lastTotal, &extraLines) == NULL) {
										   p = NULL;
										   if (!searchSpecPtr->backwards) {
												extraLinesSearched = extraLines;
										   }
										   break;
									  }
									  maxExtraLines = extraLines;
								 }
								 
								 startOfLine = Tcl_GetString(theLine);
								 p = startOfLine + skipFirst;
								 
								 /*
								  * Use the fact that 'matchLength = patLength' for
								  * exact searches.
								  */
								 
								 if ((lastTotal - skipFirst) >= matchLength) {
									  /*
									   * We now have enough text to match, so we
									   * make a final test and break whatever the
									   * result.
									   */
									  
									  if (strncmp(p,pattern,(unsigned)matchLength)) {
										   p = NULL;
									  }
									  break;
								 } else {
									  /*
									   * Not enough text yet, but check the prefix.
									   */
									  
									  if (strncmp(p, pattern,
												  (unsigned)(lastTotal - skipFirst))) {
										   p = NULL;
										   break;
									  }
									  
									  /*
									   * The prefix matches, so keep looking.
									   */
								 }
								 extraLines++;
							}
							/*
							 * If we reach here, with p != NULL, we've found a
							 * multi-line match, else we started a multi-match but
							 * didn't finish it off, so we go to the next line.
							 */
							
							if (p == NULL) {
								 break;
							}
							
							/*
							 * We've found a multi-line match.
							 */
							
							if (extraLines > 0) {
								 extraLinesSearched = extraLines - 1;
							}
					   }
				  }
				  
			 backwardsMatch:
				  if ((p - startOfLine) >= lastOffset) {
					   break;
				  }
				  
				  /*
				   * Remember the match.
				   */
				  
				  matchOffset = p - startOfLine;
				  
				  if (searchSpecPtr->all &&
					  !(*searchSpecPtr->foundMatchProc)(lineNum,
										searchSpecPtr, lineInfo, theLine, matchOffset,
										matchLength)) {
					   /*
						* We reached the end of the search.
						*/
					   
					   goto searchDone;
				  }
				  
				  if (!searchSpecPtr->overlap) {
					   if (searchSpecPtr->backwards) {
							alreadySearchOffset = p - startOfLine;
							if (firstNewLine != -1) {
								 break;
							} else {
								 alreadySearchOffset -= matchLength;
							}
					   } else {
							firstOffset = p - startOfLine + matchLength;
							if (firstOffset >= lastOffset) {
								 /*
								  * Now, we have to be careful not to find
								  * overlapping matches either on the same or
								  * following lines. Assume that if we did find
								  * something, it goes until the last extra line we
								  * added.
								  *
								  * We can break out of the loop, since we know no
								  * more will be found.
								  */
								 
								 if (!searchSpecPtr->backwards) {
									  alreadySearchOffset =
										   firstOffset - lastFullLine;
									  break;
								 }
							}
					   }
				  } else {
					   if (searchSpecPtr->backwards) {
							alreadySearchOffset = p - startOfLine - 1;
							if (alreadySearchOffset < 0) {
								 break;
							}
					   } else {
							firstOffset = p - startOfLine +
								 Tcl_UtfToUniChar(startOfLine+matchOffset,&ch);
					   }
				  }
			 } while (searchSpecPtr->all);
		} else {
			 int maxExtraLines = 0;
			 int matches = 0;
			 int lastNonOverlap = -1;
			 
			 do {
				  Tcl_RegExpInfo info;
				  int match;
				  int lastFullLine = lastOffset;
				  
				  match = Tcl_RegExpExecObj(interp, regexp, theLine,
							firstOffset, 1, (firstOffset>0 ? TCL_REG_NOTBOL : 0));
				  if (match < 0) {
					   code = TCL_ERROR;
					   goto searchDone;
				  }
				  Tcl_RegExpGetInfo(regexp, &info);
				  
				  /*
				   * If we don't have a match, or if we do, but it extends to
				   * the end of the line, we must try to add more lines to get a
				   * full greedy match.
				   */
				  
				  if (!match ||
					  ((info.extendStart == info.matches[0].start)
					   && (info.matches[0].end == lastOffset-firstOffset))) {
					   int extraLines = 0;
					   int prevFullLine;
					   
					   /*
						* If we find a match that overlaps more than one line, we
						* will use this value to determine the first allowed
						* starting offset for the following search (to avoid
						* overlapping results).
						*/
					   
					   int lastTotal = lastOffset;
					   
					   if ((lastBackwardsLineMatch != -1)
						   && (lastBackwardsLineMatch == (lineNum + 1))) {
							lastNonOverlap = lastTotal;
					   }
					   
					   if (info.extendStart < 0) {
							/*
							 * No multi-line match is possible.
							 */
							
							break;
					   }
					   
					   /*
						* We may be able to match if given more text. The
						* following 'while' block handles multi-line regexp
						* searches.
						*/
					   
					   while (1) {
							prevFullLine = lastTotal;
							
							/*
							 * Move firstOffset to first possible start.
							 */
							
							if (!match) {
								 firstOffset += info.extendStart;
							}
							if (firstOffset >= lastOffset) {
								 /*
								  * We're being told that the only possible new
								  * match is starting after the end of the line.
								  * But, that is the next line which we will handle
								  * when we look at that line.
								  */
								 
								 if (!match && !searchSpecPtr->backwards
									 && (firstOffset == 0)) {
									  extraLinesSearched = extraLines;
								 }
								 break;
							}
							
							if (lineNum + extraLines >= searchSpecPtr->numLines) {
								 break;
							}
							
							/*
							 * Add next line, provided we haven't already done so.
							 */
							
							if (extraLines > maxExtraLines) {
								 if ((*searchSpecPtr->addLineProc)(lineNum
											   + extraLines, searchSpecPtr, theLine,
											   &lastTotal, &extraLines) == NULL) {
									  /*
									   * There are no more acceptable lines, so we
									   * can say we have searched all of these.
									   */
									  
									  if (!match && !searchSpecPtr->backwards) {
										   extraLinesSearched = extraLines;
									  }
									  break;
								 }
								 
								 maxExtraLines = extraLines;
								 if ((lastBackwardsLineMatch != -1)
									 && (lastBackwardsLineMatch
										 == (lineNum + extraLines + 1))) {
									  lastNonOverlap = lastTotal;
								 }
							}
							
							match = Tcl_RegExpExecObj(interp, regexp, theLine,
											  firstOffset, 1,
											  ((firstOffset > 0) ? TCL_REG_NOTBOL : 0));
							if (match < 0) {
								 code = TCL_ERROR;
								 goto searchDone;
							}
							Tcl_RegExpGetInfo(regexp, &info);
							
							/*
							 * Unfortunately there are bugs in Tcl's regexp
							 * library, which tells us that info.extendStart is
							 * zero when it should not be (should be -1), which
							 * makes our task a bit more complicated here. We
							 * check if there was a match, and the end of the
							 * match leaves an entire extra line unmatched, then
							 * we stop searching. Clearly it still might sometimes
							 * be possible to add more text and match again, but
							 * Tcl's regexp library doesn't tell us that.
							 *
							 * This means we often add and search one more line
							 * than might be necessary if Tcl were able to give us
							 * a correct value of info.extendStart under all
							 * circumstances.
							 */
							
							if ((match &&
								 firstOffset+info.matches[0].end != lastTotal &&
								 firstOffset+info.matches[0].end < prevFullLine)
								|| info.extendStart < 0) {
								 break;
							}
							
							/*
							 * If there is a match, but that match starts after
							 * the end of the first line, then we'll handle that
							 * next time around, when we're actually looking at
							 * that line.
							 */
							
							if (match && (info.matches[0].start >= lastOffset)) {
								 break;
							}
							if (match && ((firstOffset + info.matches[0].end)
										  >= prevFullLine)) {
								 if (extraLines > 0) {
									  extraLinesSearched = extraLines - 1;
								 }
								 lastFullLine = prevFullLine;
							}
							
							/*
							 * The prefix matches, so keep looking.
							 */
							
							extraLines++;
					   }
					   
					   /*
						* If we reach here with 'match == 1', we've found a
						* multi-line match, which we will record in the code
						* which follows directly else we started a multi-line
						* match but didn't finish it off, so we go to the next
						* line.
						*/
					   
					   if (!match) {
							/*
							 * Here is where we could perform an optimisation,
							 * since we have already retrieved the contents of the
							 * next line (perhaps many more), so we shouldn't
							 * really throw it all away and start again. This
							 * could be particularly important for complex regexp
							 * searches.
							 *
							 * This 'break' will take us to just before the
							 * 'nextLine:' below.
							 */
							
							break;
					   }
					   
					   if (lastBackwardsLineMatch != -1) {
							if ((lineNum + linesSearched + extraLinesSearched)
								== lastBackwardsLineMatch) {
								 /*
								  * Possible overlap or inclusion.
								  */
								 
								 int thisOffset = firstOffset + info.matches[0].end
									  - info.matches[0].start;
								 
								 if (lastNonOverlap != -1) {
									  /*
									   * Possible overlap or enclosure.
									   */
									  
									  if (thisOffset-lastNonOverlap >=
										  lastBackwardsMatchOffset+matchLength){
										   /*
											* Totally encloses previous match, so
											* forget the previous match.
											*/
										   
										   lastBackwardsLineMatch = -1;
									  } else if ((thisOffset - lastNonOverlap)
												 > lastBackwardsMatchOffset) {
										   /*
											* Overlap. Previous match is ok, and the
											* current match is only ok if we are
											* searching with -overlap.
											*/
										   
										   if (searchSpecPtr->overlap) {
												goto recordBackwardsMatch;
										   } else {
												match = 0;
												break;
										   }
									  } else {
										   /*
											* No overlap, although the same line was
											* reached.
											*/
										   
										   goto recordBackwardsMatch;
									  }
								 } else {
									  /*
									   * No overlap.
									   */
									  
									  goto recordBackwardsMatch;
								 }
							} else if (lineNum+linesSearched+extraLinesSearched
									   < lastBackwardsLineMatch) {
								 /*
								  * No overlap.
								  */
								 
								 goto recordBackwardsMatch;
							} else {
								 /*
								  * Totally enclosed.
								  */
								 
								 lastBackwardsLineMatch = -1;
							}
					   }
					   
				  } else {
					   /*
						* Matched in a single line.
						*/
					   
					   if (lastBackwardsLineMatch != -1) {
					   recordBackwardsMatch:
							(*searchSpecPtr->foundMatchProc)(
								 lastBackwardsLineMatch, searchSpecPtr, NULL,
								 NULL, lastBackwardsMatchOffset, matchLength);
							lastBackwardsLineMatch = -1;
							if (!searchSpecPtr->all) {
								 goto searchDone;
							}
					   }
				  }
				  
				  firstOffset += info.matches[0].start;
				  if (firstOffset >= lastOffset) {
					   break;
				  }
				  
				  /*
				   * Update our local variables with the match, if we haven't
				   * yet found anything, or if we're doing '-all' or
				   * '-backwards' _and_ this match isn't fully enclosed in the
				   * previous match.
				   */
				  
				  if (matchOffset == -1 ||
					  ((searchSpecPtr->all || searchSpecPtr->backwards)
					   && ((firstOffset < matchOffset)
						   || ((firstOffset + info.matches[0].end
								- info.matches[0].start)
							   > (matchOffset + matchLength))))) {
					   
					   matchOffset = firstOffset;
					   matchLength = info.matches[0].end - info.matches[0].start;
					   
					   if (searchSpecPtr->backwards) {
							/*
							 * To get backwards searches in the correct order, we
							 * must store them away here.
							 */
							
							if (matches == matchNum) {
								 /*
								  * We've run out of space in our normal store, so
								  * we must allocate space for these backwards
								  * matches on the heap.
								  */
								 
								 int *newArray = (int *)
									  ckalloc(4 * matchNum * sizeof(int));
								 memcpy(newArray, storeMatch, matchNum*sizeof(int));
								 memcpy(newArray + 2*matchNum, storeLength,
										matchNum * sizeof(int));
								 if (storeMatch != smArray) {
									  ckfree((char *) storeMatch);
								 }
								 matchNum *= 2;
								 storeMatch = newArray;
								 storeLength = newArray + matchNum;
							}
							storeMatch[matches] = matchOffset;
							storeLength[matches] = matchLength;
							matches++;
					   } else {
							/*
							 * Now actually record the match, but only if we are
							 * doing an '-all' search.
							 */
							
							if (searchSpecPtr->all &&
								!(*searchSpecPtr->foundMatchProc)(lineNum,
											  searchSpecPtr, lineInfo, theLine, matchOffset,
											  matchLength)) {
								 /*
								  * We reached the end of the search.
								  */
								 
								 goto searchDone;
							}
					   }
					   
					   /*
						* For forward matches, unless we allow overlaps, we move
						* this on by the length of the current match so that we
						* explicitly disallow overlapping matches.
						*/
					   
					   if (matchLength > 0 && !searchSpecPtr->overlap
						   && !searchSpecPtr->backwards) {
							firstOffset += matchLength;
							if (firstOffset >= lastOffset) {
								 /*
								  * Now, we have to be careful not to find
								  * overlapping matches either on the same or
								  * following lines. Assume that if we did find
								  * something, it goes until the last extra line we
								  * added.
								  *
								  * We can break out of the loop, since we know no
								  * more will be found.
								  */
								 
								 alreadySearchOffset = firstOffset - lastFullLine;
								 break;
							}
							
							/*
							 * We'll add this on again just below.
							 */
							
							firstOffset --;
					   }
				  }
				  
				  /*
				   * Move the starting point on, in case we are doing repeated
				   * or backwards searches (for the latter, we actually do
				   * repeated forward searches).
				   */
				  
				  firstOffset++;
			 } while (searchSpecPtr->backwards || searchSpecPtr->all);
			 
			 if (matches > 0) {
				  /*
				   * Now we have all the matches in our array, but not stored
				   * with 'foundMatchProc' yet.
				   */
				  
				  matches--;
				  matchOffset = storeMatch[matches];
				  matchLength = storeLength[matches];
				  while (--matches >= 0) {
					   if (lineNum == searchSpecPtr->stopLine) {
							/*
							 * It appears as if a condition like:
							 *
							 * if (storeMatch[matches]<searchSpecPtr->stopOffset)
							 *	break;
							 *
							 * might be needed here, but no test case has been
							 * found which would exercise such a problem.
							 */
					   }
					   if (storeMatch[matches] + storeLength[matches]
						   >= matchOffset + matchLength) {
							/*
							 * The new match totally encloses the previous one, so
							 * we overwrite the previous one.
							 */
							
							matchOffset = storeMatch[matches];
							matchLength = storeLength[matches];
							continue;
					   }
					   if (!searchSpecPtr->overlap) {
							if (storeMatch[matches] + storeLength[matches]
								> matchOffset) {
								 continue;
							}
					   }
					   (*searchSpecPtr->foundMatchProc)(lineNum, searchSpecPtr,
										lineInfo, theLine, matchOffset, matchLength);
					   if (!searchSpecPtr->all) {
							goto searchDone;
					   }
					   matchOffset = storeMatch[matches];
					   matchLength = storeLength[matches];
				  }
				  if (searchSpecPtr->all && matches > 0) {
					   /*
						* We only need to do this for the '-all' case, because
						* just below we will call the foundMatchProc for the
						* non-all case.
						*/
					   
					   (*searchSpecPtr->foundMatchProc)(lineNum, searchSpecPtr,
														lineInfo, theLine, matchOffset, matchLength);
				  } else {
					   lastBackwardsLineMatch = lineNum;
					   lastBackwardsMatchOffset = matchOffset;
				  }
			 }
		}
		
		/*
		 * If the 'all' flag is set, we will already have stored all matches,
		 * so we just proceed to the next line.
		 *
		 * If not, and there is a match we need to store that information and
		 * we are done.
		 */
		
		if ((lastBackwardsLineMatch == -1) && (matchOffset >= 0)
			&& !searchSpecPtr->all) {
			 (*searchSpecPtr->foundMatchProc)(lineNum, searchSpecPtr, lineInfo,
											  theLine, matchOffset, matchLength);
			 goto searchDone;
		}
		
		/*
		 * Go to the next (or previous) line;
		 */
		
   nextLine:
		linesSearched += extraLinesSearched;
		
		while (linesSearched-- > 0) {
			 /*
			  * If we have just completed the 'stopLine', we are done.
			  */
			 
			 if (lineNum == searchSpecPtr->stopLine) {
				  goto searchDone;
			 }
			 
			 if (searchSpecPtr->backwards) {
				  lineNum--;
				  
				  if (lastBackwardsLineMatch != -1
					  && ((lineNum < 0)
						  || (lineNum + 2 < lastBackwardsLineMatch))) {
					   (*searchSpecPtr->foundMatchProc)(lastBackwardsLineMatch,
														searchSpecPtr, NULL, NULL,
														lastBackwardsMatchOffset, matchLength);
					   lastBackwardsLineMatch = -1;
					   if (!searchSpecPtr->all) {
							goto searchDone;
					   }
				  }
				  
				  if (lineNum < 0) {
					   lineNum = searchSpecPtr->numLines-1;
				  }
				  if (!searchSpecPtr->exact) {
					   /*
						* The 'exact' search loops above are designed to give us
						* an accurate picture of the number of lines which we can
						* skip here. For 'regexp' searches, on the other hand,
						* which can match potentially variable lengths, we cannot
						* skip multiple lines when searching backwards. Therefore
						* we only allow one line to be skipped here.
						*/
					   
					   break;
				  }
			 } else {
				  lineNum++;
				  if (lineNum >= searchSpecPtr->numLines) {
					   lineNum = 0;
				  }
			 }
			 if (lineNum == searchSpecPtr->startLine && linesSearched > 0) {
				  /*
				   * We've just searched all the way round and have gone right
				   * through the start line without finding anything in the last
				   * attempt.
				   */
				  
				  break;
			 }
		}
		
		Tcl_SetObjLength(theLine, 0);
   }
searchDone:
   
   if (lastBackwardsLineMatch != -1) {
		(*searchSpecPtr->foundMatchProc)(lastBackwardsLineMatch, searchSpecPtr,
										 NULL, NULL, lastBackwardsMatchOffset, matchLength);
   }
   
   /*
	* Free up the cached line and pattern.
	*/
   
   Tcl_DecrRefCount(theLine);
   Tcl_DecrRefCount(patObj);
   
   /*
	* Free up any extra space we allocated.
	*/
   
   if (storeMatch != smArray) {
		ckfree((char *) storeMatch);
   }
   
   return code;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSearch --
 *
 *	This function is invoked to process the "search" widget command for
 *	text widgets. See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
CmdSearch(
    ClientData clientData,   /* Information about text widget. */
    Tcl_Interp *interp,      /* Current interpreter. */
    int objc,                /* Number of arguments. */
    Tcl_Obj *CONST objv[])   /* Argument objects. */
{
   SciWidget *sciwidgetPtr = (SciWidget*)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   int i, argsLeft, code;
   SearchSpec searchSpec;

   static CONST char *switchStrings[] = {
		"--", "-all", "-backwards", "-count", "-elide", "-exact", "-forwards",
		"-hidden", "-nocase", "-nolinestop", "-overlap", "-regexp",
		"-strictlimits", NULL
   };
   enum SearchSwitches {
		SEARCH_END, SEARCH_ALL, SEARCH_BACK, SEARCH_COUNT, SEARCH_ELIDE,
		SEARCH_EXACT, SEARCH_FWD, SEARCH_HIDDEN, SEARCH_NOCASE,
		SEARCH_NOLINESTOP, SEARCH_OVERLAP, SEARCH_REGEXP, SEARCH_STRICTLIMITS
   };

   /*
	* Set up the search specification, including the last 4 fields which are
	* text widget specific.
	*/

   searchSpec.exact = 1;
   searchSpec.noCase = 0;
   searchSpec.all = 0;
   searchSpec.backwards = 0;
   searchSpec.varPtr = NULL;
   searchSpec.countPtr = NULL;
   searchSpec.resPtr = NULL;
   searchSpec.searchElide = 0;
   searchSpec.noLineStop = 0;
   searchSpec.overlap = 0;
   searchSpec.strictLimits = 0;
   searchSpec.numLines = sci_get_line_count(obj);
   searchSpec.clientData = (ClientData)sciwidgetPtr;
   searchSpec.addLineProc = &SciSearchAddNextLine;
   searchSpec.foundMatchProc = &SciSearchFoundMatch;
   searchSpec.lineIndexProc = &SciSearchGetLineIndex;

   /*
	* Parse switches and other arguments.
	*/

   for (i=2 ; i<objc ; i++) {
		int index;
		if (Tcl_GetString(objv[i])[0] != '-') {
			 break;
		}

		if (Tcl_GetIndexFromObj(interp, objv[i], switchStrings, "switch", 0,
								&index) != TCL_OK) {
			 /*
			  * Hide the -hidden option.
			  */

			 Tcl_ResetResult(interp);
			 Tcl_AppendResult(interp, "bad switch \"", Tcl_GetString(objv[i]),
							  "\": must be --, -all, -backward, -count, -elide, ",
							  "-exact, -forward, -nocase, -nolinestop, -overlap, ",
							  "-regexp, or -strictlimits", NULL);
			 return TCL_ERROR;
		}

		switch ((enum SearchSwitches) index) {
		case SEARCH_END:
			 i++;
			 goto endOfSwitchProcessing;
		case SEARCH_ALL:
			 searchSpec.all = 1;
			 break;
		case SEARCH_BACK:
			 searchSpec.backwards = 1;
			 break;
		case SEARCH_COUNT:
			 if (i >= objc-1) {
				  Tcl_SetResult(interp, "no value given for \"-count\" option",
								TCL_STATIC);
				  return TCL_ERROR;
			 }
			 i++;

			 /*
			  * Assumption objv[i] isn't going to disappear on us during this
			  * function, which is fair.
			  */

			 searchSpec.varPtr = objv[i];
			 break;
		case SEARCH_ELIDE:
		case SEARCH_HIDDEN:
			 searchSpec.searchElide = 1;
			 break;
		case SEARCH_EXACT:
			 searchSpec.exact = 1;
			 break;
		case SEARCH_FWD:
			 searchSpec.backwards = 0;
			 break;
		case SEARCH_NOCASE:
			 searchSpec.noCase = 1;
			 break;
		case SEARCH_NOLINESTOP:
			 searchSpec.noLineStop = 1;
			 break;
		case SEARCH_OVERLAP:
			 searchSpec.overlap = 1;
			 break;
		case SEARCH_STRICTLIMITS:
			 searchSpec.strictLimits = 1;
			 break;
		case SEARCH_REGEXP:
			 searchSpec.exact = 0;
			 break;
		default:
			 Tcl_Panic("unexpected switch fallthrough");
		}
   }
endOfSwitchProcessing:

   argsLeft = objc - (i+2);
   if ((argsLeft != 0) && (argsLeft != 1)) {
		Tcl_WrongNumArgs(interp, 2, objv,
						 "?switches? pattern index ?stopIndex?");
		return TCL_ERROR;
   }

   if (searchSpec.noLineStop && searchSpec.exact) {
		Tcl_SetResult(interp, "the \"-nolinestop\" option requires the "
					  "\"-regexp\" option to be present", TCL_STATIC);
		return TCL_ERROR;
   }

   if (searchSpec.overlap && !searchSpec.all) {
		Tcl_SetResult(interp, "the \"-overlap\" option requires the "
					  "\"-all\" option to be present", TCL_STATIC);
		return TCL_ERROR;
   }

   /*
	* Scan through all of the lines of the text circularly, starting at the
	* given index. 'objv[i]' is the pattern which may be an exact string or a
	* regexp pattern depending on the flags set above.
	*/

   code = SearchPerform(interp, &searchSpec, objv[i], objv[i+1],
						(argsLeft == 1 ? objv[i+2] : NULL));
   if (code != TCL_OK) {
		goto cleanup;
   }

   /*
	* Set the '-count' variable, if given.
	*/

   if (searchSpec.varPtr != NULL && searchSpec.countPtr != NULL) {
		Tcl_IncrRefCount(searchSpec.countPtr);
		if (Tcl_ObjSetVar2(interp, searchSpec.varPtr, NULL,
						   searchSpec.countPtr, TCL_LEAVE_ERR_MSG) == NULL) {
			 code = TCL_ERROR;
			 goto cleanup;
		}
   }

   /*
	* Set the result.
	*/

   if (searchSpec.resPtr != NULL) {
		Tcl_SetObjResult(interp, searchSpec.resPtr);
		searchSpec.resPtr = NULL;
   }

cleanup:
   if (searchSpec.countPtr != NULL) {
		Tcl_DecrRefCount(searchSpec.countPtr);
   }
   if (searchSpec.resPtr != NULL) {
		Tcl_DecrRefCount(searchSpec.resPtr);
   }
   return code;
}

/*
 *----------------------------------------------------------------------
 *
 * SciSearchGetLineIndex --
 *
 * Extract a row, text offset index position from an objPtr
 *
 * This means we ignore any embedded windows/images and elidden text
 * (unless we are searching that).
 *
 * Results:
 * Standard Tcl error code (with a message in the interpreter on error
 * conditions).
 *
 * The offset placed in offsetPosPtr is a utf-8 char* byte index for
 * exact searches, and a Unicode character index for regexp searches.
 *
 * The line number should start at zero (searches which wrap around
 * assume the first line is numbered 0).
 *
 * Side effects:
 * None.
 *
 *----------------------------------------------------------------------
 */

static int
SciSearchGetLineIndex(
    Tcl_Interp *interp,         /* For error messages. */
    Tcl_Obj *objPtr,            /* Contains a textual index like "1.2" */
    SearchSpec *searchSpecPtr,  /* Contains other search parameters. */
    int *linePosPtr,            /* For returning the line number. */
    int *offsetPosPtr)          /* For returning the text offset in the
                                 * line. */
{
   sptr_t pos1;
   int line;
   SciWidget *sciwidgetPtr = (SciWidget *)searchSpecPtr->clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   if (GetPosition(sciwidgetPtr, Tcl_GetString(objPtr), &pos1, false) != TCL_OK) {
      return BadIndex(interp, Tcl_GetString(objPtr));
   }

   line = sci_get_line(obj, pos1);
   if (line >= searchSpecPtr->numLines) {
      int posstart;

      /*
       * Count the number of bytes in this line.
       */

      line = searchSpecPtr->numLines-1;
      posstart = sci_get_pos(obj, line);
      *offsetPosPtr = pos1 - posstart;
   } else {
      int posstart = sci_get_pos(obj, line);
      *offsetPosPtr = pos1 - posstart;
   }

   *linePosPtr = line; /* should be 1 based? */

   return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * SciSearchAddNextLine --
 *
 *	Adds a line from the text widget to the object 'theLine'.
 *
 * Results:
 *	A pointer to the TkTextLine corresponding to the given line, or NULL
 *	if there was no available line.
 *
 *	Also 'lenPtr' (if non-NULL) is filled in with the total length of
 *	'theLine' (not just what we added to it, but the length including what
 *	was already in there). This is in bytes for an exact search and in
 *	chars for a regexp search.
 *
 *	Also 'extraLinesPtr' (if non-NULL) will have its value incremented by
 *	1 for each additional logical line we have added because a newline is
 *	elided (this will only ever happen if we have chosen not to search
 *	elided text, of course).
 *
 * Side effects:
 *	Memory may be allocated or re-allocated for theLine's string
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static ClientData
SciSearchAddNextLine(
    int lineNum,                /* Line we must add. */
    SearchSpec *searchSpecPtr,  /* Search parameters. */
    Tcl_Obj *theLine,           /* Object to append to. */
    int *lenPtr,                /* For returning the total length. */
    int *extraLinesPtr)         /* If non-NULL, will have its value
                                 * incremented by the number of additional
                                 * logical lines which are merged into this
                                 * one by newlines being elided. */
{
   SciWidget *sciwidgetPtr = (SciWidget *)searchSpecPtr->clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   int pos1, pos2;
   int visible = 1; /* must deal with folded regions? */
   Tcl_Obj *get;
   int nothingyet = 1;
   int startline = lineNum+1;

   /*
	* Extract the text from the line.
	*/

   while (nothingyet && lineNum < sci_get_line_count(obj)) {
		//int elideWraps = 0;
		pos1 = sci_get_pos(obj, lineNum);
		pos2 = sci_get_lineend_pos(obj, lineNum);
		get = GetTextRange(sciwidgetPtr, pos1, pos2, visible);
		if (get) {
//			 if (Tcl_GetCharLength(theLine) > 0) {
//				  Tcl_AppendToObj(theLine, "\n", 1);
//			 }
			 Tcl_AppendToObj(get, "\n", 1);
			 Tcl_AppendObjToObj(theLine, get);
			 Tcl_DecrRefCount(get);
			 nothingyet = 0;
			 lineNum++;
			 if (lineNum >= searchSpecPtr->numLines) {
				  break;
			 }
			 if (0 && extraLinesPtr) { /* for folded cases */
				  *extraLinesPtr = (*extraLinesPtr) + 1;
			 }
		} else {
			 break;
		}
   }
   
   /*
	* If we're ignoring case, convert the line to lower case. There is no
	* need to do this for regexp searches, since they handle a flag for this
	* purpose.
	*/
   
   if (searchSpecPtr->exact && searchSpecPtr->noCase) {
		Tcl_SetObjLength(theLine, Tcl_UtfToLower(Tcl_GetString(theLine)));
   }
   
   if (lenPtr != NULL) {
		if (searchSpecPtr->exact) {
			 Tcl_GetStringFromObj(theLine, lenPtr);
		} else {
			 *lenPtr = Tcl_GetCharLength(theLine);
		}
   }
   return (ClientData)startline;
}

/*
 *----------------------------------------------------------------------
 *
 * SciSearchFoundMatch --
 *
 *	Stores information from a successful search.
 *
 * Results:
 *	1 if the information was stored, 0 if the position at which the match
 *	was found actually falls outside the allowable search region (and
 *	therefore the search is actually complete).
 *
 * Side effects:
 *	Memory may be allocated in the 'countPtr' and 'resPtr' fields of
 *	'searchSpecPtr'. Each of those objects will have refCount zero and
 *	must eventually be freed or stored elsewhere as appropriate.
 *
 *----------------------------------------------------------------------
 */

static int
SciSearchFoundMatch(
    int lineNum,                /* Line on which match was found. */
    SearchSpec *searchSpecPtr,  /* Search parameters. */
    ClientData clientData,      /* Token returned by the 'addNextLineProc',
                                 * TextSearchAddNextLine. May be NULL, in
                                 * which we case we must generate it (from
                                 * lineNum). */
    Tcl_Obj *theLine,           /* Text from current line, only accessed for
                                 * exact searches, and is allowed to be NULL
                                 * for regexp searches. */
    int matchOffset,            /* Offset of found item in utf-8 bytes for
                                 * exact search, Unicode chars for regexp. */
    int matchLength)            /* Length also in bytes/chars as per search
                                 * type. */
{
   SciWidget *sciwidgetPtr = (SciWidget *)searchSpecPtr->clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   int numChars;
   int foundPos;
   Tcl_Obj *foundIndex;

   if (lineNum == searchSpecPtr->stopLine) {
		/*
		 * If the current index is on the wrong side of the stopIndex, then
		 * the item we just found is actually outside the acceptable range,
		 * and the search is over.
		 */

		if (searchSpecPtr->backwards ^
			(matchOffset >= searchSpecPtr->stopOffset)) {
			 return 0;
		}
   }

   /*
	* Calculate the character count, which may need augmenting if there are
	* embedded windows or elidden text.
	*/

   if (searchSpecPtr->exact) {
		CONST char *startOfLine = Tcl_GetString(theLine);

		numChars = Tcl_NumUtfChars(startOfLine + matchOffset, matchLength);
   } else {
		numChars = matchLength;
   }

   /*
	* If we're using strict limits checking, ensure that the match with its
	* full length fits inside the given range.
	*/

   if (searchSpecPtr->strictLimits && lineNum == searchSpecPtr->stopLine) {
		if (searchSpecPtr->backwards ^
			((matchOffset + numChars) > searchSpecPtr->stopOffset)) {
			 return 0;
		}
   }

   /*
	* The index information returned by the regular expression parser only
	* considers textual information: it doesn't account for elided text 
	* (when we are not searching elided text) or any other non-textual info. 
	* Scan through the line's segments again to adjust both matchChar and 
	* matchCount.
	*
	* We will walk through the segments of this line until we have either
	* reached the end of the match or we have reached the end of the line.
	*/

   /* TODO: account for folded text */


   /*
	* Calculate and store the found index in the result.
	*/

   /* not sure there's any difference here for Scintilla */
   if (searchSpecPtr->exact) {
		int start = sci_get_pos(obj, lineNum);
		foundPos = start + matchOffset;
   } else {
		int start = sci_get_pos(obj, lineNum);
		foundPos = start + matchOffset;
   }

   foundIndex = Tcl_NewObj();
   PosToIndex(NULL, obj, foundPos, foundIndex);

   if (searchSpecPtr->all) {
		if (searchSpecPtr->resPtr == NULL) {
			 searchSpecPtr->resPtr = Tcl_NewObj();
		}
		Tcl_ListObjAppendElement(NULL, searchSpecPtr->resPtr, foundIndex);
   } else {
		searchSpecPtr->resPtr = foundIndex;
   }

   /*
	* Find the end point. Here 
	*/

   /* TODO: account for folded text */

   /*
	* Now store the count result, if it is wanted.
	*/
   
   if (searchSpecPtr->varPtr != NULL) {
		Tcl_Obj *tmpPtr = Tcl_NewIntObj(numChars);
		if (searchSpecPtr->all) {
			 if (searchSpecPtr->countPtr == NULL) {
				  searchSpecPtr->countPtr = Tcl_NewObj();
			 }
			 Tcl_ListObjAppendElement(NULL, searchSpecPtr->countPtr, tmpPtr);
		} else {
			 searchSpecPtr->countPtr = tmpPtr;
		}
   }
   return 1;
}

/*
 *----------------------------------------------------------------------
 * CmdSee --
 *
 *    Implements the SEE widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    The display may be scrolled to ensure that the requested index
 *    location is visible on the screen.
 *
 *----------------------------------------------------------------------
 */
static int CmdSee(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   if (objc != 3 && objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "index ?-center?");
      return TCL_ERROR;
   }
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   char *index_str = Tcl_GetString(objv[2]);
   bool force_center = false;
   if (objc == 4) {
      char *arg = Tcl_GetString(objv[3]);
      if ((arg[0] == '-') && (strcmp(arg, "-center") == 0)) {
         force_center = true;
      }
   }
   sptr_t pos = -1;
   if (GetPosition(sciwidgetPtr, index_str, &pos, false) == TCL_OK) {
      if (pos == -1) {
         // assume index was specified beyond end of doc
         pos = sci_get_length(obj);
      }
      // if the widget isn't mapped yet, there's no point in continuing
      if ( ! Tk_IsMapped(sciwidgetPtr->tkwin)) {
          return TCL_OK;
      }
      // find out the top and bottom visible lines
      int topline  = sci_cmd(obj, SCI_GETFIRSTVISIBLELINE);
      // You can't use SCI_LINESONSCREEN because the returned value includes any annotation lines, not just document lines
      //int vislines = sci_cmd(obj, SCI_LINESONSCREEN);
      int x = 0;
      int y = Tk_Height(sciwidgetPtr->tkwin);
      int marginwidth = sciwidgetPtr->marginMgr->Width();
      marginwidth += SSM(SCI_GETMARGINLEFT,0,0); // internal text area padding
      if (x < marginwidth) {
         x = marginwidth;
      }
      int pos2  = sci_get_pos_from_xy(obj, x, y);
      int vislines = sci_get_line(obj, pos2) - topline;
      int botline  = topline + vislines;
      int midline  = topline + (vislines / 2);

      // get the line containing the target
      int line = sci_get_line(obj, pos);

      // get the last line
      int last = LINE2SCI(sci_get_line_count(obj));

      // check if the line containing the target index is already visible
      if (line >= topline && line <= botline) {

         if (!force_center) {
            // it's already visible, nothing more to do

            // check if the bottom line was the target, and if so, check if
            // it's bottom edge is completely in the visible region
            int line_ht    = sci_get_text_height(obj, line);
            int boty       = Tk_Height(sciwidgetPtr->tkwin);
            int botedgepos = sci_cmd(obj, SCI_POSITIONFROMPOINT,0, boty);
            int botedgey   = sci_get_y_from_pos(obj, botedgepos);

            if ((line == botline) && ((botedgey + line_ht) > boty)) {
                // scroll one more line to get it into view
                sci_cmd(obj, SCI_LINESCROLL, 0/*columns*/, 1);
            }
            return TCL_OK;
         }
         // center it (if possible)
         if (line > midline) {

             // we need to move down...
             int remaining = last - botline; // #lines below bottom edge
             int delta = (line - midline);   // #lines we want to move

             if (delta > remaining) {
                 // IMPORTANT:
                 // reset the target line/pos to avoid scrolling last line
                 // up above the bottom edge of the window
                 pos = sci_get_pos(obj, (midline + remaining));
             }
         }
         sci_goto_pos(obj, pos); // need this to set the new current pos
         sci_cmd(obj, SCI_VERTICALCENTRECARET);
      } else {
         // It's not visible, figure out how much to scroll. The rule is
         // that if we're not forcing the line to be centered, then check
         // if it's "close" to the visible region; if it is then it gets
         // moved to the top/bottom edge, otherwise it gets centered.

         if (force_center) {
             int remaining = last - botline;
             int delta = (line - midline);
             if (delta > remaining) {
                 pos = sci_get_pos(obj, (midline + remaining));
             }
            sci_goto_pos(obj, pos);
            sci_cmd(obj, SCI_VERTICALCENTRECARET);
         } else {
            sci_goto_pos(obj, pos);
            int delta = (line < topline) ? (topline - line) : (line - botline);
            if (delta > 3) {
               sci_cmd(obj, SCI_VERTICALCENTRECARET);
            }
         }
      }
   } else {
      return BadIndex(interp, index_str);
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdStyle --
 *
 *    Implements the STYLE widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    The given style ID is modified according to the options provided.
 *
 *----------------------------------------------------------------------
 */
// data structure for marker configuration options
typedef struct {
    XColor *bgColor; //(COLOR) bgd color of marker
    XColor *fgColor; //(COLOR) fgd color
    int            bold;       // boolean to indicate if text is bold
    int            italic   ;  // boolean to indicate if text is italicized
    int            underline;  // boolean to indicate if text is underlined
    int            eolfilled;  // boolean to indicate if background should extend to EOL
} StyleOptions;

#define STYLE_OPT_BG         1
#define STYLE_OPT_FG         1 << 1
#define STYLE_OPT_UNDERLINE  1 << 2
#define STYLE_OPT_BOLD       1 << 3
#define STYLE_OPT_ITALIC     1 << 4
#define STYLE_OPT_EOLFILLED  1 << 5

static Tk_OptionSpec styleOptionSpecs[] = {
   {TK_OPTION_COLOR,  "-background", NULL, NULL,
         NULL, -1, Tk_Offset(StyleOptions, bgColor), TK_OPTION_NULL_OK, 0, STYLE_OPT_BG
   },
   {TK_OPTION_SYNONYM, "-bg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-background", 0},

   {TK_OPTION_BOOLEAN, "-bold", NULL, NULL,
         "0", -1, Tk_Offset(StyleOptions, bold), 0, 0, STYLE_OPT_BOLD
   },
   {TK_OPTION_BOOLEAN, "-eolfilled", NULL, NULL,
         "0", -1, Tk_Offset(StyleOptions, eolfilled), 0, 0, STYLE_OPT_EOLFILLED
   },
   {TK_OPTION_COLOR,  "-foreground", NULL, NULL,
         NULL, -1, Tk_Offset(StyleOptions, fgColor), TK_OPTION_NULL_OK, 0, STYLE_OPT_FG
   },
   {TK_OPTION_SYNONYM, "-fg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-foreground", 0},

   {TK_OPTION_BOOLEAN, "-italic", NULL, NULL,
         "0", -1, Tk_Offset(StyleOptions, italic), 0, 0, STYLE_OPT_ITALIC
   },
   {TK_OPTION_BOOLEAN, "-underline", NULL, NULL,
         "0", -1, Tk_Offset(StyleOptions, underline), 0, 0, STYLE_OPT_UNDERLINE
   },
   {TK_OPTION_END}
};

static Tk_OptionTable styleOptionTable;
static int styleOptionTableInitialized;

static int CmdStyle(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
int styleID;
StyleOptions opts;

    SciWidget *sciwidgetPtr = (SciWidget *)clientData;
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

    if (objc <= 3) {
        Tcl_WrongNumArgs(interp, 1, objv, "styleID ?-fg? ?-bold?");
        return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj(interp, objv[2], &styleID) != TCL_OK) {
        return TCL_ERROR;
    }

    if (!styleOptionTableInitialized) {
        styleOptionTable = Tk_CreateOptionTable(interp, styleOptionSpecs);
    }
    if (Tk_InitOptions(interp, (char *)&opts, styleOptionTable, sciwidgetPtr->tkwin) != TCL_OK) {
        fprintf(stderr, "ERROR in Tk_InitOptions: %s\n", Tcl_GetStringResult(interp));
        return TCL_ERROR;
    }

    Tk_SavedOptions savedOptions;
    int mask = 0;
    if (Tk_SetOptions(interp, (char *)&opts, styleOptionTable,
            objc-3, objv+3, sciwidgetPtr->tkwin, &savedOptions, &mask) != TCL_OK) {
        return TCL_ERROR;
    }
    if (mask & STYLE_OPT_BG) {
        sci_set_style_bg(sciobj, styleID, opts.bgColor);
    }
    if (mask & STYLE_OPT_FG) {
        sci_set_style_fg(sciobj, styleID, opts.fgColor);
    }
    if (mask & STYLE_OPT_UNDERLINE) {
        SSM(SCI_STYLESETUNDERLINE, styleID, opts.underline);
    }
    if (mask & STYLE_OPT_BOLD) {
        SSM(SCI_STYLESETBOLD, styleID, opts.bold);
    }
    if (mask & STYLE_OPT_ITALIC) {
        SSM(SCI_STYLESETITALIC, styleID, opts.italic);
    }
    if (mask & STYLE_OPT_EOLFILLED) {
        SSM(SCI_STYLESETEOLFILLED, styleID, opts.eolfilled);
    }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdXview --
 *
 *    Implements the XVIEW widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdXview(
    ClientData     clientData,
    Tcl_Interp    *interp,
    int            objc,
    Tcl_Obj *CONST objv[])
{
    SciWidget *sciwidgetPtr = (SciWidget *)clientData;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    if (objc == 2) {
        // return current xview info
        XScroller(sciwidgetPtr, false);
        return TCL_OK;
    }
    char *subcmd = Tcl_GetString(objv[2]);
    if (strcmp(subcmd, "scroll") == 0) {
        //##########
        //# SCROLL
        //##########
        if (objc != 5) {
            Tcl_WrongNumArgs(interp, 2, objv, "scroll number what");
            return TCL_ERROR;
        }
        int num;
        if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
            Tcl_AppendResult(interp, "expected integer but got \"",objv[3],"\"", (char *)NULL);
            return TCL_ERROR;
        }
        char *what = Tcl_GetString(objv[4]);

        // get some info about current conditions
        int winw = Tk_Width(sciwidgetPtr->tkwin);   // width of screen
        winw -= sciwidgetPtr->marginMgr->Width();   // account for width of margins!!
        int xoff = sci_cmd(obj, SCI_GETXOFFSET);    // current hz scroll offset
        int sw   = sci_cmd(obj, SCI_GETSCROLLWIDTH);// max scrollable width

        if (strcmp(what, "units") == 0) {
            // each unit is ~1 character's width
            if (num >= 1) {
                // To match Tk scroll behavior, which prevents unit scrolling
                // beyond the end of the content, we do the following check
                if (xoff + winw > sw)
                    return TCL_OK;
            }
            sci_cmd(obj, SCI_LINESCROLL, num, 0/*lines*/);

        } else if (strcmp(what, "pages") == 0) {
            // each "page" represents a screen width
            if (num >= 1) {
                int maxoff = (sw - winw);
                sci_cmd(obj, SCI_SETXOFFSET, Platform::Minimum(maxoff, (xoff + winw)));
            } else if (num < 0) {
                sci_cmd(obj, SCI_SETXOFFSET, Platform::Maximum(0, (xoff - winw)));
            }
        }
        // The above call updates Scintilla's view within the document. For the
        // scrollbar we'll get a callback from ScintillaTK::SetHorizontalScollPos
        // function which will allow us to get the scrollbar(if one exists)
        // updated to reflect the change(see MessageFromScintilla).

    } else if (strcmp(subcmd, "moveto") == 0) {
        //##########
        //# MOVETO
        //##########
        // we're handling an update from a scrollbar being dragged
        double frac;
        if (Tcl_GetDoubleFromObj(interp, objv[3], &frac) != TCL_OK) {
            Tcl_AppendResult(interp, "expected double but got \"",objv[3],"\"", (char *)NULL);
            return TCL_ERROR;
        }
        XScroller(sciwidgetPtr, false, true, frac);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdYview --
 *
 *    Implements the YVIEW widget command
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
static int CmdYview(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   if (objc == 2) {
      // return current yview info
      YScroller(sciwidgetPtr, false);
      return TCL_OK;
   }
   char *subcmd = Tcl_GetString(objv[2]);
   if (strcmp(subcmd, "scroll") == 0) {
      //##########
      //# SCROLL
      //##########
      if (objc != 5) {
         Tcl_WrongNumArgs(interp, 2, objv, "scroll number what");
         return TCL_ERROR;
      }
      int num;
      if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
         Tcl_AppendResult(interp, "expected integer but got \"",objv[3],"\"", (char *)NULL);
         return TCL_ERROR;
      }
      char *what = Tcl_GetString(objv[4]);
      if (strcmp(what, "units") == 0) {
         sci_cmd(obj, SCI_LINESCROLL, 0/*columns*/, num);
      } else if (strcmp(what, "pages") == 0) {
         if (num >= 1) {
            sci_cmd(obj, SCI_PAGEDOWN);
         } else if (num < 0) {
            sci_cmd(obj, SCI_PAGEUP);
         }
      } else if (strcmp(what, "pixels") == 0) {
          // Tk Text widget supports pixel-level scrolling ... we don't
          Tcl_AppendResult(interp, "yview pixel scroll requested", (char *)NULL);
          return TCL_ERROR;
      }
      // The above call updates Scintilla's view within the document. For the
      // scrollbar we'll get a callback from ScintillaTK::SetVerticalScollPos
      // function which will allow us to get the scrollbar(if one exists)
      // updated to reflect the change(see MessageFromScintilla).

   } else if (strcmp(subcmd, "moveto") == 0) {
      //##########
      //# MOVETO
      //##########
      // we're handling an update from a scrollbar being dragged
      double frac;
      if (Tcl_GetDoubleFromObj(interp, objv[3], &frac) != TCL_OK) {
         Tcl_AppendResult(interp, "expected double but got \"",objv[3],"\"", (char *)NULL);
         return TCL_ERROR;
      }
      YScroller(sciwidgetPtr, false, true, frac);
   } else {
      if (objc == 3) {
	     int line_number,len;
		 const char* string_val = Tcl_GetStringFromObj(objv[2], &len);
		 sscanf(string_val, "%d", &line_number);
	   	 line_number--;
		 if (line_number <= 0) {
			 line_number = 1;
		 }
         SSM(SCI_SETFIRSTVISIBLELINE, line_number, 0);
	  }
   }
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * CmdZoom --
 *
 *    Implements the ZOOMIN/ZOOMOUT widget commands.
 *    This called to request a change in the size of the text in
 *    support of doing a "zoom in" or "zoom out".
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Margins which show text may have their width changed
 *
 *----------------------------------------------------------------------
 */
static int CmdZoom(
   ClientData     clientData,
   Tcl_Interp    *interp,
   int            objc,
   Tcl_Obj *CONST objv[],
   bool           zoom_in)
{
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   // this will let the margin manager know how much
   // to adjust its margin widths by
   sciwidgetPtr->marginMgr->ZoomPre();

   // do the appropriate zoom
   int currz = sci_get_zoom(obj);
   if (zoom_in) {
      sci_zoom_in(obj);
   } else {
      sci_zoom_out(obj);
   }
   // check if the zoom actually changed (it may not if scintilla detected
   // the request was outside the limit of what can be done)
   int newz = sci_get_zoom(obj);

   if (newz != currz) {
      // update the scrollbar data
      XScroller(sciwidgetPtr);
      YScroller(sciwidgetPtr);

      // tell the margin manager
      sciwidgetPtr->marginMgr->ZoomPost();
   }
   return TCL_OK;
}



/****************************************************************
// Func: SciWidgetUpdateLang
// Desc: 
*****************************************************************/
static int
SciWidgetUpdateLang(Tcl_Interp *interp, SciWidget *sciwidgetPtr) {
    int newtype, i;

    if (!sciwidgetPtr->language) return TCL_OK;

    if( strcmp(sciwidgetPtr->language,"none") == 0) {
        newtype = SCLEX_NULL;
        SSM(SCI_SETLEXER, newtype, 0);
    } else {
        // try to load lexer by name
        SSM(SCI_SETLEXERLANGUAGE, 0, (sptr_t)sciwidgetPtr->language);
        newtype =  SSM(SCI_GETLEXER,0,0);
        if (newtype == SCLEX_NULL) {
            Tcl_AppendResult(interp, "bad language name: ", sciwidgetPtr->language, (char*)NULL);
            return TCL_ERROR;
        }
    }
    if ((sciwidgetPtr->langType != SCLEX_NULL) && (sciwidgetPtr->langType != newtype)) {
        // determine style bits used by new lexer before reallocating styles for existant tags
        if (newtype == SCLEX_NULL) {
            // set number of style bits to maximum
            SSM(SCI_SETSTYLEBITS, 8, 0);
            sciwidgetPtr->styleBits = 8;
        } else {
            sciwidgetPtr->styleBits = SSM(SCI_GETSTYLEBITSNEEDED, 0, 0);
        }
        if (sciwidgetPtr->tagMgr->Reset(interp, newtype) != TCL_OK)
            return TCL_ERROR;
    }
    sciwidgetPtr->langType = newtype;

    // Not all of the lexers support this property, but the majority do
    SSM(SCI_SETPROPERTY, (uptr_t)"fold",         (sptr_t)"1");
    SSM(SCI_SETPROPERTY, (uptr_t)"fold.comment", (sptr_t)"1"); // notepad++ allows this

    // clear out any saved keywords for previous lexer
    for (i = 0; i < KEYWORDSET_MAX; ++i) {
        if (sciwidgetPtr->keywords[i]) {
           Tcl_DecrRefCount(sciwidgetPtr->keywords[i]);
            sciwidgetPtr->keywords[i] = NULL;
        }
    }
    return TCL_OK;
}

static void
SciWidgetCheckFoldMargin(SciWidget *sciwidgetPtr) {

   if (sciwidgetPtr->foldstyle < 0 || sciwidgetPtr->foldstyle >= foldMAX) {
      sciwidgetPtr->foldstyle = foldBoxTree; // force to valid value
   }
   switch (sciwidgetPtr->foldstyle) {
      case foldBoxTree: {
         //#####  BOX TREE  ##########
         //-- define the symbols to use
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPEN,     SC_MARK_BOXMINUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDER,         SC_MARK_BOXPLUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERSUB,      SC_MARK_VLINE);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERTAIL,     SC_MARK_LCORNER);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEREND,      SC_MARK_BOXPLUSCONNECTED);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPENMID,  SC_MARK_BOXMINUSCONNECTED);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERMIDTAIL,  SC_MARK_TCORNER);

         //-- define the colors
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPEN,    _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPEN,    _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDER,        _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDER,        _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERSUB,     _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERTAIL,    _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEREND,     _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEREND,     _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPENMID, _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPENMID, _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERMIDTAIL, _RGB_(0x88,0x88,0x88));
         break;
      }
      case foldPlusMinus: {
         //#####  PLUS/MINUS  ##########
         //-- define the symbols to use
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPEN,     SC_MARK_MINUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDER,         SC_MARK_PLUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERSUB,      SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERTAIL,     SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEREND,      SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPENMID,  SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERMIDTAIL,  SC_MARK_EMPTY);

         //-- define the colors
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPEN, _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPEN, _RGB_(0x00,0x00,0x00));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDER,     _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDER,     _RGB_(0x00,0x00,0x00));
         break;
      }
      case foldArrow: {
         //#####  ARROW  ##########
         //-- define the symbols to use
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPEN,     SC_MARK_ARROWDOWN);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDER,         SC_MARK_ARROW);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERSUB,      SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERTAIL,     SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEREND,      SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPENMID,  SC_MARK_EMPTY);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERMIDTAIL,  SC_MARK_EMPTY);

         //-- define the colors
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPEN, _RGB_(0x00,0xff,0x00));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPEN, _RGB_(0x00,0x00,0x00));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDER,     _RGB_(0xff,0x00,0x00));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDER,     _RGB_(0x00,0x00,0x00));
         break;
      }
      case foldCircleTree: {
         //#####  CIRCLE TREE  #########
         //-- define the symbols to use
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPEN,     SC_MARK_CIRCLEMINUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDER,         SC_MARK_CIRCLEPLUS);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERSUB,      SC_MARK_VLINE);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERTAIL,     SC_MARK_LCORNERCURVE);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEREND,      SC_MARK_CIRCLEPLUSCONNECTED);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDEROPENMID,  SC_MARK_CIRCLEMINUSCONNECTED);
         SSM(SCI_MARKERDEFINE, SC_MARKNUM_FOLDERMIDTAIL,  SC_MARK_TCORNERCURVE);

         //-- define the colors
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPEN,    _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPEN,    _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDER,        _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDER,        _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERSUB,     _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERTAIL,    _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEREND,     _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEREND,     _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDEROPENMID, _RGB_(0x88,0x88,0x88));
         SSM(SCI_MARKERSETFORE, SC_MARKNUM_FOLDEROPENMID, _RGB_(0xff,0xff,0xff));
         SSM(SCI_MARKERSETBACK, SC_MARKNUM_FOLDERMIDTAIL, _RGB_(0x88,0x88,0x88));

         break;
      }
      case foldDISABLED: {
//TODO: SciWidgetCheckFoldMargin >> what should we do here?? the margin mgr should be told so that
//      additional marker symbol IDs are available in the masks
         break;
      }
   }
}

/****************************************************************
// Func: SciWidgetUpdateCursor
// Desc: Handles getting the widget cursor set based on the
//       current "state" setting, as well as the optional X/Y
*****************************************************************/
static void
SciWidgetUpdateCursor(SciWidget *sci, int x=-1, int y=-1) {

    Tcl_Interp *interp = sci->interp;
    Tk_Window   tkwin  = sci->tkwin;

    if (sci->flags & BUSY_MODE) {
        //### BUSY
        if (sci->busycursor == None) {
            Tk_DefineCursor(tkwin,
                Tk_AllocCursorFromObj(interp, tkwin, Tcl_NewStringObj("watch", -1))
            );
        } else {
            Tk_DefineCursor(tkwin, sci->busycursor);
        }
        return;
    }
    //### NOT BUSY

    // if are given an X/Y location then we can check if the margin
    // manager needs to get involved
    if (x != -1 && y != -1) {
        MarginMgr *mmgr = sci->marginMgr;
        if (mmgr->SetCursor(x, y)) return;
    }
    // over the text area (or we didn't get x/y)
    if (sci->cursor == None) {
        Tk_UndefineCursor(tkwin);
    } else {
        Tk_DefineCursor(tkwin, sci->cursor);
    }
}

/****************************************************************
// Func: ScintillaLostSelection
// Desc: Called by Tk when selection is grabbed away from a Scintilla widget.
*****************************************************************/
static void ScintillaLostSelection(ClientData clientData)
{
    SciWidget *sciwidgetPtr = (SciWidget*)clientData;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    if (!sciwidgetPtr->exportSelection)
        return;
    int sel_beg = sci_cmd(obj, SCI_GETSELECTIONSTART);
    int sel_end = sci_cmd(obj, SCI_GETSELECTIONEND);
    if (sel_beg != sel_end) {
        sci_cmd(obj, SCI_SETEMPTYSELECTION, sci_get_current_pos(obj));
        sciwidgetPtr->tagMgr->FireSelectionVirtual(); // signal selection changed.
    }
    sciwidgetPtr->flags &= ~GOT_SELECTION;
}

/****************************************************************
// Func: SciWidgetConfigure
// Desc: 
*****************************************************************/
static int
SciWidgetConfigure(
   Tcl_Interp    *interp,
   SciWidget     *sciwidgetPtr,
   int            objc,
   Tcl_Obj *CONST objv[])
{
   Tk_SavedOptions savedOptions;
   int mask = 0;
   // check if this is the first time doing this
   bool firsttime = false;
   bool oldExport = sciwidgetPtr->exportSelection;

   if (!(sciwidgetPtr->flags & INIT_CFG_DONE)) {
      sciwidgetPtr->flags |= INIT_CFG_DONE;
      firsttime = true;
   }

   /*
    * Tk_SetOptions does all the work of parsing the
    * arguments and looking for things in the resource database.
    */
   if (Tk_SetOptions(interp, (char *)sciwidgetPtr, sciwidgetPtr->optionTable,
         objc, objv, sciwidgetPtr->tkwin, &savedOptions, &mask) != TCL_OK) {
      return TCL_ERROR;
   }
   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

   if (mask == OPT_CURSOR) {
       //----------------
       //----- CURSOR
       //----------------
       // only changing the cursor, so nothing else is needed - this is important
       // so we avoid scheduling a repaint
       return TCL_OK;
   }

   // Now check for changes to options that might generate an error, in which
   // case we'll want to restore option values

   if (firsttime || mask & OPT_FONT ) {
      //--------------------------
      //----- FONT
      //--------------------------
      // Note: the call to Tk_SetOptions will handle catching any poorly formatted
      //       font strings, so we can safely call "font actual" without doing any
      //       error checking. IMPORTANT: The call to Tk_SetOptions _WILL NOT_ detect
      //       a bogus family name, which is why we let "font actual" tell us what
      //       the system will actually use for the font that was specified.
      Tcl_DString dstr;
      Tcl_DStringInit(&dstr);
      Tcl_DStringAppend(&dstr, "font actual ", -1);
      Tcl_DStringAppendElement(&dstr, Tk_NameOfFont(sciwidgetPtr->tkfont));

      Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(&dstr));

      Tcl_Obj *result = Tcl_GetObjResult(sciwidgetPtr->interp); // this will have the font specification
//printf("ACTUAL=%s\n", Tcl_GetStringFromObj(Tcl_GetObjResult(sciwidgetPtr->interp), NULL));

      // The format of the returned list is:
      //   -family <name> -size <size> -weight <normal|bold> -slant <roman|italic> -underline <0|1> -overstrike <0|1>
      enum {FAMILY=1, SIZE=3, WEIGHT=5, SLANT=7};

      int len;
      Tcl_Obj *arg;
      char *family = NULL;
      int ptsize   = 0;
      bool bold    = false;
      bool italic  = false;
      // Extract out the font data
      Tcl_ListObjLength(sciwidgetPtr->interp, result, &len);
      for (int i=1; i < len; i += 2) {
         Tcl_ListObjIndex(sciwidgetPtr->interp, result, i, &arg);
         char *val = Tcl_GetString(arg);
         switch (i) {
            case FAMILY: family = val; break;
            case WEIGHT: bold   = !strcasecmp(val, "bold"); break;
            case SLANT:  italic = !strcasecmp(val, "italic"); break;
            case SIZE:   
                ptsize = atoi(val);
                if (ptsize < 0) {
                    // encode the value for use in Font::Create (in PlatTK.cxx)
                    ptsize = MAGIC_FONT_OFFSET + abs(ptsize);
                }
         }
      }
      sci_set_font(sciobj, family, ptsize, bold, italic);
   }
   if ( mask & OPT_BG ) {
      //--------------------------
      //----- BACKGROUND
      //--------------------------
      XColor *bc = Tk_3DBorderColor(sciwidgetPtr->background);
//      printf("BG COLOR=%d %d %d\n",bc->red,bc->green,bc->blue);
      sci_set_style_bg(sciobj, STYLE_DEFAULT, bc);
      mask |= OPT_LANG; // reapply lexer
   }
   if ( mask & OPT_LANG ) {
      //--------------------------
      //----- LANGUAGE
      //--------------------------
      if ( SciWidgetUpdateLang(interp, sciwidgetPtr) != TCL_OK ) {
         Tk_RestoreSavedOptions(&savedOptions);
         return TCL_ERROR;
      }
   }
   static bool init_bg = true;
   if ( init_bg ) {
      // Without the following call, the background of our parent
      // widget will "show through" any areas that aren't immediately
      // painted by our code. An example would be when the widget is 
      // resized to be larger
      Tk_SetWindowBackground(sciwidgetPtr->tkwin,
         Tk_3DBorderColor(sciwidgetPtr->background)->pixel);
      init_bg = false;
   }
   if ( mask & OPT_WS ) {
      //--------------------------
      //----- WHITE SPACE
      //--------------------------
      sci_show_ws(sciobj, sciwidgetPtr->showws);
   }
   if ( firsttime || mask & OPT_TABW ) {
      //--------------------------
      //----- TAB WIDTH
      //--------------------------
      sci_set_tab_width(sciobj, sciwidgetPtr->tabwidth);
   }
   if ( firsttime || mask & OPT_FOLDM ) {
      //--------------------------
      //----- FOLD MARGIN
      //--------------------------
      SciWidgetCheckFoldMargin(sciwidgetPtr);
   }
   if ( mask & OPT_STATE ) {
      //--------------------------
      //----- STATE
      //--------------------------
      switch (sciwidgetPtr->state) {
         case STATE_DISABLED:
         case STATE_NORMAL:
            // check if we need to turn off readonly
            if (sciwidgetPtr->readonly) {
               sci_set_readonly(sciobj, false);
               sciwidgetPtr->readonly = 0;
            }
            sciwidgetPtr->flags &= ~BUSY_MODE;
            break;
         case STATE_READONLY:
            if (!sciwidgetPtr->readonly) {
               sci_set_readonly(sciobj, true);
               sciwidgetPtr->readonly = 1;
            }
            sciwidgetPtr->flags &= ~BUSY_MODE;
            break;
          case STATE_BUSY:
              sciwidgetPtr->flags |= BUSY_MODE;
              break;
      }
      SciWidgetUpdateCursor(sciwidgetPtr);
   }
   if ( sciwidgetPtr->tagMgr && (mask & OPT_MBG) ) {
      //--------------------------
      //----- MARGIN BACKGROUND
      //--------------------------
      sciwidgetPtr->tagMgr->UpdateMarginColor(true, false);
   }
   if ( sciwidgetPtr->tagMgr && (mask & OPT_MFG) ) {
      //--------------------------
      //----- MARGIN FOREGROUND
      //--------------------------
      sciwidgetPtr->tagMgr->UpdateMarginColor(false, true);
   }

   if ( sciwidgetPtr->undo != sci_cmd(sciobj, SCI_GETUNDOCOLLECTION) ) {
       //----------------
       //----- UNDO
       //----------------
       sci_cmd(sciobj, SCI_SETUNDOCOLLECTION, sciwidgetPtr->undo);
       if ( ! sciwidgetPtr->undo ) {
           sci_cmd(sciobj, SCI_EMPTYUNDOBUFFER);
       }
   }
   //-----------------
   //----- BLINK -----
   //-----------------
   static int blinkrate = sci_cmd(sciobj, SCI_GETCARETPERIOD);
   if ( sciwidgetPtr->blinkcaret ) {
        sci_cmd(sciobj, SCI_SETCARETPERIOD, blinkrate);
   } else {
       sci_cmd(sciobj, SCI_SETCARETPERIOD, 0); //<<<TURNS OFF CARET BLINKING
   }

   if (sciwidgetPtr->xScrollCmd != NULL) {
       // In order to make use of the "track widest line" functionality
       // we have to tell Scintilla we're using a horizontal scrollbar and
       // then turn on the tracking mode. We set the initial scrollwidth
       // to some small value since the Scintilla behavior is to increase
       // the value whenever it encounters a line which is longer.
       sci_cmd(sciobj, SCI_SETHSCROLLBAR, 1);
       sci_cmd(sciobj, SCI_SETSCROLLWIDTH, 100);
       sci_cmd(sciobj, SCI_SETSCROLLWIDTHTRACKING, 1);
   }
   SciWidgetGeometry(sciwidgetPtr);

   if (sciwidgetPtr->exportSelection && !oldExport) {
      /* If we've turned on exporting selection and there are selected characters,
       * claim the global selection.
       */
      int sel_beg = sci_cmd(sciobj, SCI_GETSELECTIONSTART);
      int sel_end = sci_cmd(sciobj, SCI_GETSELECTIONEND);
      if (sel_beg != sel_end) {
         Tk_OwnSelection(sciwidgetPtr->tkwin, XA_PRIMARY, ScintillaLostSelection, sciwidgetPtr);
         sciwidgetPtr->flags |= GOT_SELECTION;
      }
   }

   /*
    * Set up a call to display ourself.
    */
   if ((sciwidgetPtr->tkwin != NULL) &&
      Tk_IsMapped(sciwidgetPtr->tkwin) &&
      !(sciwidgetPtr->flags & REDRAW_PENDING)) {
	  Tk_CancelIdleCall(SciWidgetDisplay, (ClientData) sciwidgetPtr);
      Tk_DoWhenIdle(SciWidgetDisplay, (ClientData) sciwidgetPtr);
      sciwidgetPtr->flags |= REDRAW_PENDING;
   }
   Tk_FreeSavedOptions(&savedOptions);

   return TCL_OK;
}

/******************************************************************************
//  Func:   treew_Xevent2String
//  Desc:   Returns a string representing a X event.  Only the most
//			common ones are enumerated, others are "event #" string.
//			Used for debug messages.
******************************************************************************/
char *treew_Xevent2String(
	int event,
	char *buf )
{
	/*  Just take care of the common ones. */
	switch (event) {
		case EnterNotify:
			strcpy(buf, "EnterNotify"); break;
		case LeaveNotify:
			strcpy(buf, "LeaveNotify"); break;
		case Expose:
			strcpy(buf, "Expose"); break;
		case VisibilityNotify:
			strcpy(buf, "VisibilityNotify"); break;
		case DestroyNotify:
			strcpy(buf, "DestroyNotify"); break;
		case UnmapNotify:
			strcpy(buf, "UnmapNotify"); break;
		case MapNotify:
			strcpy(buf, "MapNotify"); break;
		case MapRequest:
			strcpy(buf, "MapRequest"); break;
		case ConfigureNotify:
			strcpy(buf, "ConfigureNotify"); break;
		case ConfigureRequest:
			strcpy(buf, "ConfigureRequest"); break;
		case MappingNotify:
			strcpy(buf, "MappingNotify"); break;
	    case MotionNotify:
		    strcpy(buf,"MotionNotify"); break;
	    case FocusIn:
		    strcpy(buf,"FocusIn"); break;
	    case FocusOut:
		    strcpy(buf,"FocusOut"); break;
		default:
			sprintf( buf, "unknown event %d", event ); break;
	}
	return( buf );
}

/****************************************************************
// Func: SciWidgetEventProc
// Desc: 
*****************************************************************/
static void
SciWidgetEventProc(ClientData clientData, XEvent *eventPtr)
{
   int redraw = 0;
   char buf[32];

   SciWidget *sciwidgetPtr = (SciWidget *) clientData;

   if (0) {
     fprintf(stderr, "[XEVENT] wave_EventProc: %s (type=%s (%d))\n", 
			Tk_PathName(sciwidgetPtr->tkwin), 
			treew_Xevent2String( eventPtr->type, buf ),
			eventPtr->type );
	}

   switch (eventPtr->type) {
      case MapNotify:
         //####################################
         //# MAP
         //####################################
         // we get the scrollbar updated here instead of in response to the
         // ModifyScrollBars call from Editor because the size info of the
         // client rectangle isn't available until first mapped
         XScroller(sciwidgetPtr);
         YScroller(sciwidgetPtr);
	     break;

	  case GraphicsExpose: {
         XGraphicsExposeEvent e = eventPtr->xgraphicsexpose;
//printf("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX G_EXPOSED: +%02d+%02dx%03dx%03d  COUNT=%d\n", e.x,e.y,e.width,e.height,e.count);
         if (sciwidgetPtr->flags & REDRAW_EXPOSE) {
            sciwidgetPtr->expose_x = Platform::Minimum(sciwidgetPtr->expose_x, e.x);
            sciwidgetPtr->expose_y = Platform::Minimum(sciwidgetPtr->expose_y, e.y);
            sciwidgetPtr->expose_w = Platform::Maximum(sciwidgetPtr->expose_w, e.width);
            sciwidgetPtr->expose_h = Platform::Maximum(sciwidgetPtr->expose_h, e.height);
         } else {
            sciwidgetPtr->expose_x = e.x;
            sciwidgetPtr->expose_y = e.y;
            sciwidgetPtr->expose_w = e.width;
            sciwidgetPtr->expose_h = e.height;
         }
         // tell the paint code to use the expose rectangle
         sciwidgetPtr->flags |= REDRAW_EXPOSE;
         redraw = 1;
	  }
      break;

	  case Expose: {
         XExposeEvent e = eventPtr->xexpose;
//printf("EXPOSED: +%02d+%02dx%03dx%03d  COUNT=%d\n", e.x,e.y,e.width,e.height,e.count);
         if ((sciwidgetPtr->flags & REDRAW_EXPOSE)) {
		    int max_right, max_bottom;

            max_right  = Platform::Maximum((sciwidgetPtr->expose_x + sciwidgetPtr->expose_w), (e.x + e.width));
            max_bottom = Platform::Maximum((sciwidgetPtr->expose_y + sciwidgetPtr->expose_h), (e.y + e.height));

            sciwidgetPtr->expose_x = Platform::Minimum(sciwidgetPtr->expose_x, e.x);
            sciwidgetPtr->expose_y = Platform::Minimum(sciwidgetPtr->expose_y, e.y);
            sciwidgetPtr->expose_w = max_right - sciwidgetPtr->expose_x;
            sciwidgetPtr->expose_h = max_bottom - sciwidgetPtr->expose_y;
//printf("expose additive +%02d+%02dx%03dx%03d\n", sciwidgetPtr->expose_x,sciwidgetPtr->expose_y,sciwidgetPtr->expose_w,sciwidgetPtr->expose_h);
		 } else { 
            sciwidgetPtr->expose_x = e.x;
            sciwidgetPtr->expose_y = e.y;
            sciwidgetPtr->expose_w = e.width;
            sciwidgetPtr->expose_h = e.height;
            sciwidgetPtr->flags |= REDRAW_EXPOSE;
//printf("expose and render +%02d+%02dx%03dx%03d\n", sciwidgetPtr->expose_x,sciwidgetPtr->expose_y,sciwidgetPtr->expose_w,sciwidgetPtr->expose_h);
		 }
         if (eventPtr->xexpose.count < 1) {
//printf(" expose done +%02d+%02dx%03dx%03d\n", sciwidgetPtr->expose_x,sciwidgetPtr->expose_y,sciwidgetPtr->expose_w,sciwidgetPtr->expose_h);
			redraw = 1;
		 }
	     break;
     }
      
	 case DestroyNotify:
         //####################################
         //# DESTROY
         //####################################
         Tcl_DeleteCommandFromToken(sciwidgetPtr->interp,
            sciwidgetPtr->widgetCmd);
         /*
          * Zapping the tkwin lets the other procedures
          * know we are being destroyed.
          */
         sciwidgetPtr->tkwin = NULL;

         if (sciwidgetPtr->flags & REDRAW_PENDING) {
            Tk_CancelIdleCall(SciWidgetDisplay, (ClientData) sciwidgetPtr);
            sciwidgetPtr->flags &= ~REDRAW_PENDING;
         }
         sciwidgetPtr->flags |= DESTROYED;
         /*
          * This results in a call to SciWidgetDestroy.
          */
         Tcl_EventuallyFree((ClientData) sciwidgetPtr, (Tcl_FreeProc *)SciWidgetDestroy);
		 break;
      
	 case FocusIn :
         //####################################
         //# FOCUS IN
         //####################################
         if (eventPtr->xfocus.detail != NotifyPointer) {
            sci_set_focus(SCIOBJ(sciwidgetPtr), true);
            sciwidgetPtr->flags |= GOT_FOCUS;
            if (sciwidgetPtr->highlightWidth > 0) {
//            redraw = 1;
            }
		 }
		 break;

	 case FocusOut :
         //####################################
         //# FOCUS OUT
         //####################################
         if (eventPtr->xfocus.detail != NotifyPointer) {
            sci_set_focus(SCIOBJ(sciwidgetPtr), false);
            sciwidgetPtr->flags &= ~GOT_FOCUS;
            if (sciwidgetPtr->highlightWidth > 0) {
//            redraw = 1;
            }
         }
		 break;

	 case ConfigureNotify : {
            //####################################
            //# CONFIGURE
            //####################################

            if (sciwidgetPtr->width != eventPtr->xconfigure.width ) {
	  		   sciwidgetPtr->flags |= RESIZE;
			   sciwidgetPtr->width = eventPtr->xconfigure.width;
		    }
		    if (sciwidgetPtr->width != eventPtr->xconfigure.height ) { 
               sciwidgetPtr->flags |= RESIZE;
 			   sciwidgetPtr->height = eventPtr->xconfigure.height;
            }
           
//printf("ConfigureNotify %d %d\n",Tk_Width(sciwidgetPtr->tkwin),Tk_Height(sciwidgetPtr->tkwin));

            // filter out the expose bit to make sure the repaint is for
            // the whole window and not just what the most recently queued
            // expose region is
            sciwidgetPtr->flags &= ~REDRAW_EXPOSE;
            redraw = 1;
         }
		 break;
   }
   

   if (redraw) {
      if ((sciwidgetPtr->tkwin != NULL) && !(sciwidgetPtr->flags & REDRAW_PENDING)) {
         // schedule a call to the display routine
		 Tk_CancelIdleCall(SciWidgetDisplay, (ClientData) sciwidgetPtr);
		 Tk_DoWhenIdle(SciWidgetDisplay, (ClientData) sciwidgetPtr);
         sciwidgetPtr->flags |= REDRAW_PENDING;
      }
   }
}

/*
 *--------------------------------------------------------------
 *
 * SciWidgetBindProc --
 *
 *    This function is invoked by the Tk dispatcher to handle 
 *    events associated with bindings on items.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Depends on the command invoked as part of the binding (if
 *    there was any).
 *
 *--------------------------------------------------------------
 */
static void
SciWidgetBindProc(ClientData clientData, XEvent *eventPtr)
{
// HERE'S TK CORE CODE:
// /u/prod_dev/tcltk/src/8.5.a5/tk/generic/tkTextTag.c  (see TkTextBindProc)

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   int repick = 0; // do we repick a new "current" mark?

#define AnyButtonMask \
   (Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask)

   int etype = eventPtr->type;

   if ((etype == EnterNotify) || (etype == LeaveNotify)) {
       //####################################
       //# ENTER / LEAVE
       //####################################
       if (sciwidgetPtr->flags & BUSY_MODE) return;

       if (eventPtr->xcrossing.state & AnyButtonMask)  {
           sciwidgetPtr->flags |= BUTTON_DOWN;
       } else {
           sciwidgetPtr->flags &= ~BUTTON_DOWN;
       }
       sciwidgetPtr->tagMgr->PickCurrent(eventPtr);
       return;
   }
   if ((etype == ButtonPress) || (etype == ButtonRelease)) {
       //####################################
       //# BUTTON PRESS/RELEASE
       //####################################
       if (sciwidgetPtr->flags & BUSY_MODE) return;

       MarginMgr *mgr = sciwidgetPtr->marginMgr;
       XButtonEvent be = eventPtr->xbutton;
       if ((mgr->Width() > 0) && (be.x < mgr->Width())) {
           // in the margin area
           mgr->Click(be);
       } else {
           // in the text area
           /*
            * This code simulates grabs for mouse buttons by keeping track of whether
            * a button is pressed and refusing to pick a new current character while
            * a button is pressed.
            */
           if (etype == ButtonPress) {
               //PRESS
               sciwidgetPtr->flags |= BUTTON_DOWN;
           } else {
               //RELEASE
               int mask;
               switch (be.button) {
                   case Button1: mask = Button1Mask; break;
                   case Button2: mask = Button2Mask; break;
                   case Button3: mask = Button3Mask; break;
                   case Button4: mask = Button4Mask; break;
                   case Button5: mask = Button5Mask; break;
                   default:      mask = 0; break;
               }
               if ((be.state & AnyButtonMask) == (unsigned) mask) {
                   sciwidgetPtr->flags &= ~BUTTON_DOWN;
                   repick = 1;
               }
           }
       }

   } else if (eventPtr->type == MotionNotify) {
       //####################################
       //# MOUSE MOTION
       //####################################
       XMotionEvent me = eventPtr->xmotion;

       if (me.state != 0) return; //only handle motion w/o any modifiers

       SciWidgetUpdateCursor(sciwidgetPtr, me.x, me.y);
       if (sciwidgetPtr->flags & BUSY_MODE) {
           return;
       }
       if (me.state & AnyButtonMask)  {
           sciwidgetPtr->flags |= BUTTON_DOWN;
       } else {
           sciwidgetPtr->flags &= ~BUTTON_DOWN;
       }
       sciwidgetPtr->tagMgr->PickCurrent(eventPtr);
   } else {
       // any other type of event
       if (sciwidgetPtr->flags & BUSY_MODE) return;
   }
   // let tag manager have a chance at the event
   sciwidgetPtr->tagMgr->HandleBindEvent(eventPtr);

   // check if we have a delayed request for picking the current location
   if (repick) {
       unsigned int oldState;

       oldState = eventPtr->xbutton.state;
       eventPtr->xbutton.state &= ~AnyButtonMask;

       sciwidgetPtr->tagMgr->PickCurrent(eventPtr);

       eventPtr->xbutton.state = oldState;
   }
}

/****************************************************************
// Func: SortTags
//
// This function sorts an array of tag pointers in increasing
// order of priority, optimizing for the common case where the
// array is small.
*****************************************************************/
static void
SortTags(int numTags, TagData **tagArrayPtr) {
//     int i, j, prio;
//     register TagData **tagPtrPtr;
//     TagData **maxPtrPtr, *tmp;

    if (numTags < 2) {
        return;
    }
    if (numTags < 20) {
//TODO: SortTags >> implement this if tag priorities are to be supported
//         for (i = numTags-1; i > 0; i--, tagArrayPtr++) {
//             maxPtrPtr = tagPtrPtr = tagArrayPtr;
//             prio = tagPtrPtr[0]->priority;
//             for (j = i, tagPtrPtr++; j > 0; j--, tagPtrPtr++) {
//                 if (tagPtrPtr[0]->priority < prio) {
//                     prio = tagPtrPtr[0]->priority;
//                     maxPtrPtr = tagPtrPtr;
//                 }
//             }
//             tmp = *maxPtrPtr;
//             *maxPtrPtr = *tagArrayPtr;
//             *tagArrayPtr = tmp;
//         }
    } else {
        ::qsort((void *) tagArrayPtr, (unsigned) numTags,
            sizeof (TagData *), TagSortProc);
    }
}
static int
TagSortProc(const void *first, const void *second) {
//TODO: TagSortProc >> implement this if tag priorities are to be supported
//     TagData *tagPtr1, *tagPtr2;

//     tagPtr1 = * (TagData **) first;
//     tagPtr2 = * (TagData **) second;
//     return tagPtr1->priority - tagPtr2->priority;
    return 0; // quiet compiler
}


/****************************************************************
// Func: SciWidgetGeometry
// Desc: 
*****************************************************************/
static void
SciWidgetGeometry(SciWidget *sciwidgetPtr)
{
   int width, height, tk_w, tk_h;
   int bd;                    /* Padding from borders */
   bd = sciwidgetPtr->highlightWidth + sciwidgetPtr->borderWidth;
   tk_w = Tk_Width(sciwidgetPtr->tkwin);
   tk_h = Tk_Height(sciwidgetPtr->tkwin);

   height = (tk_h == 1) ? 600 : tk_h;
   width  = (tk_w == 1) ? 800 : tk_w;

   Tk_GeometryRequest(sciwidgetPtr->tkwin, width, height);
   Tk_SetInternalBorder(sciwidgetPtr->tkwin, bd);
}

/****************************************************************
// Func: SciWidgetDisplay
// Desc: 
*****************************************************************/
void SciWidgetDisplay(ClientData clientData)
{
   PRectangle rc;
   PRectangle cursor_rc;
   int efficient_blt = 0;
   ScintillaObject *sciobj;
   SciWidget *sciwidgetPtr = (SciWidget *)clientData;
   Tk_Window tkwin = sciwidgetPtr->tkwin;
   rc = PRectangle(0, 0, Tk_Width(tkwin), Tk_Height(tkwin)); 
   sciobj = SCIOBJ(sciwidgetPtr);
   cursor_rc = PRectangle(0,0,0,0);

#if 0
printf("#############################################\n");
printf("SciWidgetDisplay w/ FLAGS = ");
if (sciwidgetPtr->flags & REDRAW_PENDING) printf("|REDRAW_PENDING");
if (sciwidgetPtr->flags & REDRAW_YSCROLL) printf("|REDRAW_YSCROLL");
if (sciwidgetPtr->flags & REDRAW_EXPOSE)  printf("|REDRAW_EXPOSE");
printf("\n");
printf("=============================================\n");
#endif
   /*
    * Make sure the widget still exists
    * and is mapped onto the display before painting.
    */
   if ((sciwidgetPtr->tkwin == NULL) || !Tk_IsMapped(tkwin)) {
      sciwidgetPtr->flags &= ~REDRAW_PENDING;
      return;
   }

   /*
    * If the resize flag is set it means the window size has changed 
	* and we need to send a message to get the pixmap reallocated to the 
	* correct size
	*/
   if ( (sciwidgetPtr->flags & RESIZE) || (sciwidgetPtr->mainPixmap == 0) ) {
      sciwidgetPtr->flags &= ~(RESIZE); /* ONLY place RESIZE turned off */
      scintilla_send_message(sciwidgetPtr->editor, SCI_TKMSG_RESIZE, 0, 0);
   }

   /* 
    * Yscroll tries to minimize the work.  It blits the retainable region
	* and it invalidate (asks scintilla to draw the newly exposed region
	*/
   if (sciwidgetPtr->flags & REDRAW_YSCROLL) {
      
//printf("SciWidgetDisplay -- YSCROLL  sciwidgetPtr->mainPixmap=%d\n",(int)sciwidgetPtr->mainPixmap);
      // we're just doing a vertical "bit blit"
      sciwidgetPtr->flags &= ~(REDRAW_PENDING|REDRAW_YSCROLL);

      // the number of pixels to shift up/down will be held in yscrollDelta.
      // a positive value means we move pixels up, a negative value means move them down
      Display *display = sciwidgetPtr->display;
      Drawable draw    = Tk_WindowId(tkwin);
      int diff = sciwidgetPtr->yscrollDelta;
      sciwidgetPtr->yscrollDelta = 0;
      if (sciwidgetPtr->copyGC == None) {
         XGCValues gcValues;
         gcValues.graphics_exposures = True;
         sciwidgetPtr->copyGC = Tk_GetGC(tkwin, GCGraphicsExposures, &gcValues);
      }
	  /* test to see if the scroll region is less than the window hieght
	   * if it isn't, nosense in blitting as its all new text
	   */
	  if (sciwidgetPtr->mainPixmap && sciwidgetPtr->copyGC) {
	   if (diff > 0 && diff < Tk_Height(tkwin)) {
	     /* if the cursor is blinking, we need to determine if it's rectangle also needs to be invalidated and redrawn */
	     
		if ( sciwidgetPtr->blinkcaret) {
			int cursorpos, currentline, firstline,lines;
			cursorpos = sci_cmd(sciobj, SCI_GETCURRENTPOS);
			firstline = sci_cmd(sciobj, SCI_GETFIRSTVISIBLELINE);
			lines = sci_cmd(sciobj, SCI_LINESONSCREEN);
			currentline = sci_cmd(sciobj, SCI_LINEFROMPOSITION,cursorpos);
			if (currentline > firstline && currentline <= (firstline + lines)) { 
				int y, height;				
				y = sci_cmd(sciobj, SCI_POINTYFROMPOSITION,0,cursorpos);
				height = sci_cmd(sciobj, SCI_TEXTHEIGHT,currentline);
				if ((y + height - diff) > 0)
                   cursor_rc = PRectangle(0,y- diff,Tk_Width(tkwin),y + height - diff);
			}
		}
         // shift pixels UP

         XCopyArea(
            display, sciwidgetPtr->mainPixmap, sciwidgetPtr->mainPixmap, sciwidgetPtr->copyGC,
            0, diff,
            Tk_Width(tkwin), Tk_Height(tkwin) - diff,
            0, 0
         );
         XCopyArea(
            display, sciwidgetPtr->mainPixmap, draw, sciwidgetPtr->copyGC,
            0, 0,
            Tk_Width(tkwin), Tk_Height(tkwin),
            0, 0
         );
		 rc = PRectangle(0,Tk_Height(tkwin) - diff ,
                         Tk_Width(tkwin), Tk_Height(tkwin));
		 efficient_blt = 1;
	  }
      
      if (diff < 0 && abs(diff) < Tk_Height(tkwin)) {
	     /* if the cursor is blinking, we need to determine if it's rectangle also needs to be invalidated and redrawn */
         if ( sciwidgetPtr->blinkcaret) {
			int cursorpos, currentline, firstline,lines;
			cursorpos = sci_cmd(sciobj, SCI_GETCURRENTPOS);
			firstline = sci_cmd(sciobj, SCI_GETFIRSTVISIBLELINE);
			lines = sci_cmd(sciobj, SCI_LINESONSCREEN);
			currentline = sci_cmd(sciobj, SCI_LINEFROMPOSITION,cursorpos);
			if (currentline >= firstline && currentline < (firstline + lines)) { 
				int y, height;
				y = sci_cmd(sciobj, SCI_POINTYFROMPOSITION,0,cursorpos);
				height = sci_cmd(sciobj, SCI_TEXTHEIGHT,currentline);
                cursor_rc = PRectangle(0,y+ abs(diff),Tk_Width(tkwin),y + height + abs(diff));
			}
		 }
          // shift pixels DOWN
         XCopyArea(
            display, sciwidgetPtr->mainPixmap, sciwidgetPtr->mainPixmap, sciwidgetPtr->copyGC,
            0, 0,
            Tk_Width(tkwin), Tk_Height(tkwin) + diff,
            0, -diff
         );
          XCopyArea(
            display, sciwidgetPtr->mainPixmap, draw, sciwidgetPtr->copyGC,
            0, 0,
            Tk_Width(tkwin), Tk_Height(tkwin),
            0, 0
         );
		 rc = PRectangle(0,0,
                         Tk_Width(tkwin), abs(diff));
 		 efficient_blt = 1;
      }	
    }
   }
   sciwidgetPtr->flags &= ~(REDRAW_PENDING|REDRAW_YSCROLL);
 
   if (sciwidgetPtr->flags & REDRAW_EXPOSE) {
      /*
	   * This is a work around for a scintilla bug.  If an expose occurs and the expose is entirerly over the 
	   * the margin (ie not text area exposed) Editor.cxx paint repaints the margin into the off screen pixmap.
	   * However, it doesn't call NotifyPainted.  
	   *
	   * The bug is that Editor::Paint first handles the margin.  Next the following if statement determines if
       * any of the text area is effected.  If not, it just returns.  If the text area is effected, a NotifyPaint gets
	   * called and the exposed area is redrawn.
	   *
       *  	if (rcArea.right > vs.fixedColumnWidth - leftTextOverlap) {
       *       ...
	   *       NotifyPainted();  <-- Must get called to so the offscreen bitmaps gets copied to the screen
	   *
	   * The work around below determines if the exposed area is entirely in the margin and if it is it increaase
	   * the exposure to include a couple of pixels of the text area.
	   *
	   * This issue should be revisited once we move to 3.30, they changed how margins are handled.
	   */
	  int margin = sciwidgetPtr->marginMgr->Width() + 5;
	  if ((sciwidgetPtr->expose_x + sciwidgetPtr->expose_w) < margin) { 
			sciwidgetPtr->expose_w = margin;
	  }
	  /* end of work around */

	  if ( !efficient_blt  ) {
		  rc = PRectangle(
			sciwidgetPtr->expose_x,
			sciwidgetPtr->expose_y,
			(sciwidgetPtr->expose_x + sciwidgetPtr->expose_w),
			(sciwidgetPtr->expose_y + sciwidgetPtr->expose_h)
		);
	  }

//printf("REDRAW_EXPOSE: +%02d+%02dx%03dx%03d\n", sciwidgetPtr->expose_x, sciwidgetPtr->expose_y, sciwidgetPtr->expose_w, sciwidgetPtr->expose_h);
      sciwidgetPtr->flags &= ~REDRAW_EXPOSE;
   } 
   if ((efficient_blt) && (cursor_rc.Width() > 0)) {
	  	uptr_t cursor_rc_ptr = reinterpret_cast<uptr_t>(&cursor_rc);
	    scintilla_send_message(sciwidgetPtr->editor, SCI_TKMSG_PAINT, cursor_rc_ptr, 0);
   }
   uptr_t rc_ptr = reinterpret_cast<uptr_t>(&rc);
   scintilla_send_message(sciwidgetPtr->editor, SCI_TKMSG_PAINT, rc_ptr, 0);
}

/////////////////////////////////////////////////////////////////////////////////
// Handle message notifications from Scintilla code
//
void MessageFromScintilla(ClientData clientData, int msg_id, uptr_t data) {

   SciWidget *sciwidgetPtr = (SciWidget *)clientData;

   switch (msg_id) {

      case SCIW_SYNC_XSCROLL: {
         // get the horizontal scrollbar sync'd with the current view
         XScroller(sciwidgetPtr);
         break;
      }
      case SCIW_SYNC_YSCROLL: {
         // get the vertical scrollbar sync'd with the current view
         YScroller(sciwidgetPtr);
         break;
      }
      case SCIW_MOVE_UPDATE: {
         // called to indicate the insertion point location changed
         if (sciwidgetPtr->moveUpdateCmd) {
            Tcl_Eval(sciwidgetPtr->interp, sciwidgetPtr->moveUpdateCmd);
         }
         break;
      }
      case SCIW_LINESADDED_UPDATE: {
         if (sciwidgetPtr->linesAddedCmd) {
//            printf("SCIW_LINESADDED_UPDATE: %d %d\n",
//               sciwidgetPtr->added_data->line,sciwidgetPtr->added_data->added);

            Tcl_Interp *i = sciwidgetPtr->interp;
            Tcl_Obj *obj = Tcl_ObjPrintf("%s %d %d", 
               sciwidgetPtr->linesAddedCmd,
               sciwidgetPtr->added_data->line,
               sciwidgetPtr->added_data->added);
            if (Tcl_Eval(i, Tcl_GetString(obj)) != TCL_OK) {
               fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(i));
            }
         }
         break;
      }
      case SCIW_MODIFIED: {
         /* notify TCL that a modification event has occurred, patterned after GenerateModifiedEvent in tkText.c */
         XEvent event;
         memset(&event, 0, sizeof(event));
         event.xany.type = VirtualEvent;
         event.xany.serial = NextRequest(Tk_Display(sciwidgetPtr->tkwin));
         event.xany.send_event = False;
         event.xany.window = Tk_WindowId(sciwidgetPtr->tkwin);
         event.xany.display = Tk_Display(sciwidgetPtr->tkwin);
         ((XVirtualEvent *) &event)->name = Tk_GetUid("Modified");
         Tk_HandleEvent(&event);
         break;
      }
      default:
         fprintf(stderr, "Unhandled message received: %d\n", msg_id);
         break;
   }
//printf("MESSAGE FROM SCINTILLA: sciwidget=%p  %d\n",(void *)sciwidgetPtr,code);
}
void MessageFromScintilla(ClientData clientData, int msg_id) {
   TkSci::MessageFromScintilla(clientData, msg_id, 0);
}


/////////////////////////////////////////////////////////////////////////////////
// Implementation of MarginData class
//
/////////////////////////////////////////////////////////////////////////////////

// define IDs for options we either need to error check or respond to
#define MOPT_CLICK  1
#define MOPT_WIDTH  1 << 1
#define MOPT_TYPE   1 << 2
#define MOPT_CURSOR 1 << 3

// define configuration options
static Tk_OptionSpec marginOptionSpecs[] = {
   {TK_OPTION_STRING,  "-alias", "alias", "Alias",
         "", -1, Tk_Offset(MarginOptions, alias), TK_OPTION_NULL_OK, 0, 0
   },
   {TK_OPTION_STRING, "-clickcommand", "clickcommand", "ClickCommand", 
         "", -1, Tk_Offset(MarginOptions, clickcmd), TK_OPTION_NULL_OK, 0, MOPT_CLICK
   },
   {TK_OPTION_CURSOR,  "-cursor", "cursor", "Cursor",
         "left_ptr", -1, Tk_Offset(MarginOptions, cursor), TK_OPTION_NULL_OK, 0, MOPT_CURSOR
   },
   {TK_OPTION_STRING_TABLE, "-type", "type", "Type",
	 (char*)MarginTypeNames[NONE], -1, Tk_Offset(MarginOptions, type), 0, (ClientData)MarginTypeNames, MOPT_TYPE
   },
   {TK_OPTION_INT,     "-width", "width", "Width",
         "0", -1, Tk_Offset(MarginOptions, width), 0, 0, MOPT_WIDTH
	 },
   {TK_OPTION_END}
};
//### CTOR ###
MarginData::MarginData(MarginMgr *mgr, int c, int w) :
   col_id(c),
   viswidth(0),
   mgr(mgr)
{
   mopts = (MarginOptions *) Tcl_Alloc(sizeof(MarginOptions));

   // must initialize all pointer/token values as per the Tk doc page:
   //    "...is responsible for storing NULL or None in all pointer 
   //    and token fields before invoking Tk_InitOptions."
   mopts->alias     = NULL;
   mopts->clickcmd  = NULL;
   mopts->cursor    = None;
   mopts->type      = NONE;
   mopts->width     = w;

   optionTable = Tk_CreateOptionTable(mgr->interp, marginOptionSpecs);

   if (Tk_InitOptions(mgr->interp, (char *)mopts, optionTable, mgr->tkwin) != TCL_OK) {
      fprintf(stderr, "ERROR in Tk_InitOptions: %s\n", Tcl_GetStringResult(mgr->interp));
   }
}
//### DTOR ###
MarginData::~MarginData() {

   // clean up objects used for margin options
   Tk_FreeConfigOptions((char *)mopts, optionTable, mgr->tkwin);
   Tk_DeleteOptionTable(optionTable);
   Tcl_Free((char *)mopts);
}

//==========================================================================
int MarginData::Configure(int objc, Tcl_Obj *CONST objv[]) {

   Tk_SavedOptions savedOptions;
   int mask = 0;
//printf("----MARGIND(%d):\n",col_id);for (int ix=0; ix < objc; ix++) printf(">> %s\n",Tcl_GetString(objv[ix]));
   /*
    * Tk_SetOptions does all the work of parsing the
    * arguments and looking for things in the resource database.
    */
   if (Tk_SetOptions(mgr->interp, (char *)mopts, optionTable,
         objc, objv, mgr->tkwin, &savedOptions, &mask) != TCL_OK) {
      return TCL_ERROR;
   }
   if (mask & MOPT_TYPE) {
      if (mgr->ValidateTypeChange(col_id) != TCL_OK) {
         Tk_RestoreSavedOptions(&savedOptions);
         return TCL_ERROR;
      }
   }
   // check for options that will require an update by the manager
   bool update = false;
   MarginType t = GetType();

   if (mask & MOPT_CLICK) {
      // a change to whether the margin has a click command callback
      update = true;
   }

   if (mask & MOPT_WIDTH) {
      update = true;
      // for line numbers, a negative width value will cause the width
      // to be calculated based on the current line count in the document.
      if (IS_LINE_NUMS(t)) {
         int width = GetWidth();
         if (width < 0) {
            SetWidth(mgr->GetLineNWidth());
         } else if (width > 0) {
            SetWidth(width);
         }
      } else {
          // the following call is made because the call to Tk_SetOptions will only
          // take care of setting the "width" value, but not "viswidth"
          SetWidth(GetWidth());
      }
   }
   if (update) {
      mgr->UpdateMargin(col_id);
   }
   Tk_FreeSavedOptions(&savedOptions);
   return TCL_OK;
}

//==========================================================================
const char *MarginData::GetCursorStr() {

    if (mopts->cursor == None) {
        return "None";
    }
    return Tk_NameOfCursor(mgr->display, mopts->cursor);
}



/////////////////////////////////////////////////////////////////////////////////
// Implementation of MarginMgr class
//
/////////////////////////////////////////////////////////////////////////////////

//### CTOR ###
MarginMgr::MarginMgr(SciWidget *w) : 
   sciwidgetPtr(w),
   interp(w->interp),
   display(w->display),
   tkwin(w->tkwin),
   currLnum(-1), currFold(-1)
{
   // create the margin objects, default all to zero width; also disable
   // click handling by default until client explicitly enables it
   for (int i=0; i < MAXM; i++) {
      margins[i] = new MarginData(this, i, 0);
      sci_set_margin_width(SCIOBJ(w), i, 0);
      sci_enable_margin_click(SCIOBJ(w), i, false);
   }
}
//### DTOR ###
MarginMgr::~MarginMgr() {

   // clean up the margin objects
   for (int i=0; i < MAXM; i++) {
      delete margins[i];
   }
}

/*
 *----------------------------------------------------------------------
 * Click --
 *
 *    Processes a button event. This function handles extracting the
 *    details needed by the _click function.
 *
 * Results:
 *    None
 *
 * Side affects:
 *    Depends on supported behavior by margin where click occurred
 *
 *----------------------------------------------------------------------
 */
void MarginMgr::Click(XButtonEvent be) {

    // check if the event is over a margin
    int id;
    if ((id = _atX(be.x)) == -1) return;

#ifndef _WIN32
    // we only want to handle "real" button events, not mouse wheel events
    if (be.button == 4 || be.button == 5) return;
#endif
    // is it a press or release
    if (be.type != ButtonPress) return;

    bool shift = ((be.state & ShiftMask) != 0);
    bool ctrl  = ((be.state & ControlMask) != 0);
#ifdef _WIN32
    // ya, weird mask value, I know
    bool alt   = ((be.state & (1 << 17)) != 0);
#else
    bool alt   = ((be.state & (1 << 3)) != 0);
#endif
    int pos = sci_get_pos_from_xy(SCIOBJ(sciwidgetPtr), be.x, be.y);

    _click(id, pos, be.x, be.y, be.button, shift, ctrl, alt);
}

/*
 *----------------------------------------------------------------------
 * GetLineNWidth --
 *
 *    Calculates the width of the line-number margin required to display
 *    the current number of lines in the document. If the "ndigits"
 *    option is specified, it identifies how many digits the width
 *    should accomodate.
 *
 * Results:
 *    Returns the margin width, in pixels
 *
 * Side affects:
 *    None
 *
 *----------------------------------------------------------------------
 */
int MarginMgr::GetLineNWidth(int ndigits) {

   static const char* NUMS = "9999999999";
   static int NLEN = strlen(NUMS);
   const char *cp = NUMS + NLEN; // point at end

   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

   if (ndigits > 0) {
      if (ndigits > NLEN) ndigits = NLEN;
      cp -= ndigits;

   } else {
      int nchars;
      int nlines = sci_get_line_count(obj); // << #lines in document
      if (nlines == 1) {
         // this means the doc is empty -- we'll force a reasonable default width
         nlines = 9999;
      }
      nchars = (int)ceil(log10((double)nlines)) + 1;
      cp -= nchars;
   }
   return sci_get_text_width(obj, cp);
}

/*
 *----------------------------------------------------------------------
 * InstanceCmd --
 *
 *    Implements the MARGIN sub-command for the widget
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    Some margins may have their visibility change
 *
 *----------------------------------------------------------------------
 */
int MarginMgr::InstanceCmd(int objc, Tcl_Obj *CONST objv[]) {

   int result = TCL_OK;
   int index;

   CONST char *optionStrings[] = {
       "bbox", "cget", "configure", "count", "dump", "fillnumbers",
       "hide", "index", "marker", "names", "show", "tag", "text", "visible", "width",
      (char *)NULL
   };
   enum options {
       CMD_BBOX, CMD_CGET, CMD_CONFIGURE, CMD_COUNT, CMD_DUMP, 
       CMD_FILLNUMS, CMD_HIDE, CMD_INDEX, CMD_MARKER, CMD_NAMES, 
       CMD_SHOW, CMD_TAG, CMD_TEXT, CMD_VISIBLE, CMD_WIDTH
   };
//printf("----MARGIN:\n");for (int ix=0; ix < objc; ix++) printf(">> %s\n",Tcl_GetString(objv[ix]));
   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "command ?option? ...?");
      return TCL_ERROR;
   }

   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings, "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }
   switch ((enum options)index) {

      case CMD_BBOX: {
         result = CmdBbox(objc, objv);
         break;
      }
      case CMD_CGET: {
         result = CmdCget(objc, objv);
         break;
      }
      case CMD_CONFIGURE: {
         result = CmdConfig(objc, objv);
         break;
      }
      case CMD_COUNT: {
         Tcl_SetObjResult(interp, Tcl_NewIntObj(MAXM));
         break;
      }
      case CMD_DUMP: {
         result = CmdDump(objc, objv);
         break;
      }
      case CMD_FILLNUMS: {
         result = CmdFillNums(objc, objv);
         break;
      }
      case CMD_HIDE: {
         result = CmdShow(objc, objv, false);
         break;
      }
      case CMD_INDEX: {
         result = CmdIndex(objc, objv);
         break;
      }
      case CMD_MARKER: {
         result = CmdMarker(objc, objv);
         break;
      }
      case CMD_NAMES: {

         Tcl_Obj *listObj = Tcl_NewListObj(0, NULL);
         for (int i=0; i < MAXM; i++) {
            MarginData *m = margins[i];
            Tcl_Obj *sublist = Tcl_NewListObj(0, NULL);
            Tcl_ListObjAppendElement(interp, sublist, Tcl_NewIntObj(i));
            Tcl_ListObjAppendElement(interp, sublist, Tcl_NewStringObj(m->GetTypeStr(), -1));
            Tcl_ListObjAppendElement(interp, sublist, Tcl_NewStringObj(m->GetAlias(), -1));

            Tcl_ListObjAppendElement(interp, listObj, sublist);
         }
         // FORMAT: {{col# type/name alias} ...}
         Tcl_SetObjResult(interp, listObj);
         break;
      }
      case CMD_SHOW: {
         result = CmdShow(objc, objv, true);
         break;
      }
      case CMD_TAG: {
         result = CmdTag(objc, objv);
         break;
      }
      case CMD_TEXT: {
         result = CmdText(objc, objv);
         break;
      }
      case CMD_VISIBLE: {
          result = CmdVisible(objc, objv);
          break;
      }
      case CMD_WIDTH: {
         // return the full width of the margin area
         Tcl_SetObjResult(interp, Tcl_NewIntObj(Width()));
         break;
      }
   }
   return result;
}

//==========================================================================
int MarginMgr::CmdBbox(int objc, Tcl_Obj *CONST objv[]) {

    if (objc != 5) {
        Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias line");
        return TCL_ERROR;
    }
    int id;
    if (_get(objv[3], &id) != TCL_OK) {
        return TCL_ERROR;
    }
    MarginData *m = margins[id];

    // verify margin is showing
    if (m->GetVisibleWidth() <= 0) {
        // it's not
        Tcl_ResetResult(interp); // force return of an empty string
        return TCL_OK;
    }
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    int x, y, width, line_ht;
    int line; // <<< filled as 0-based

    if (ValidateLine(interp, (ClientData)sciwidgetPtr, objv[4], &line) != TCL_OK) {
        return TCL_ERROR;
    }
    // to get Y location we'll use position within the text area
    int pos = sci_get_pos(sciobj, line);

    x = 0;
    for (int i=0; i < id; i++) x += margins[i]->GetVisibleWidth();

    y       = sci_get_y_from_pos(sciobj, pos);
    line_ht = sci_get_text_height(sciobj, line);
    width = m->GetWidth();

    Tcl_Obj *listObj = Tcl_NewListObj(0, NULL);

    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(x));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(y));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(width));
    Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(line_ht));

    Tcl_SetObjResult(interp, listObj);
    return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdCget(int objc, Tcl_Obj *CONST objv[]) {

   if (objc != 5) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];

   Tcl_Obj *objPtr = Tk_GetOptionValue(interp, (char *)m->mopts,
      m->optionTable, objv[4], tkwin);
   if (objPtr == NULL) {
      return TCL_ERROR;
   }
   Tcl_SetObjResult(interp, objPtr);
   return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdConfig(int objc, Tcl_Obj *CONST objv[]) {

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias ?option? ?value option value...?");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];

   int result = TCL_OK;
   if (objc <= 5) {
      Tcl_Obj *objPtr = Tk_GetOptionInfo(interp, (char *)m->mopts,
         m->optionTable, ((objc == 5) ? objv[4] : NULL), tkwin);
      if (objPtr == NULL) {
         return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, objPtr);
   } else {
      result = m->Configure(objc-4, objv+4);
   }
   return result;
}

//==========================================================================
int MarginMgr::CmdDump(int objc, Tcl_Obj *CONST objv[]) {
    if (objc < 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    // check to see if user wants internal dump style
    if (!strcmp(Tcl_GetString(objv[3]), "-internals")) {
        Tcl_Obj *buf = Tcl_NewStringObj("Margin DUMP:\n", -1);
        for (int i=0; i < MAXM; i++) {
            MarginData *m = margins[i];
            if (m->GetType() != SYMBOL && m->GetType() != FOLD)
                Tcl_AppendPrintfToObj(buf, 
                    "%d - WIDTH= %2d/%2d, CURSOR= %-10s, TYPE= %-10s, MASK=      <n/a>, ALIAS= %-12s, CLICKCMD= %s\n",
                    i,
                    m->GetVisibleWidth(), 
                    m->GetWidth(), 
                    m->GetCursorStr(),
                    m->GetTypeStr(),
                    m->GetAlias()? m->GetAlias() : "<null>",
                    m->GetClickCmd()? m->GetClickCmd() : "<null>"
                );
            else {
                // Need to use sprintf because Tcl_AppendPrintfToObj doesn't do the
                // hex formatting correctly
                char buf2[1024*10];
                sprintf(buf2, "%d - WIDTH= %2d/%2d, CURSOR= %-10s, TYPE= %-10s, MASK= 0x%08X, ALIAS= %-12s, CLICKCMD= %s\n",
                    i,
                    m->GetVisibleWidth(), 
                    m->GetWidth(), 
                    m->GetCursorStr(),
                    m->GetTypeStr(),
                    sci_get_margin_mask(SCIOBJ(sciwidgetPtr),i),
                    m->GetAlias()? m->GetAlias() : "<null>",
                    m->GetClickCmd()? m->GetClickCmd() : "<null>"
                );
                Tcl_AppendToObj(buf, buf2, -1);
            }
        }
        Tcl_SetObjResult(interp, buf);
        return TCL_OK;
    }

    if (objc < 6) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    // get selected margin
    int id;
    if (_get(objv[3], &id) != TCL_OK) {
        return TCL_ERROR;
    }

    // Tk text widget dump options supported within margins
    static CONST char *optionStrings[] = {
        "-all", "-command", "-indicator", "-mark", "-style", "-tag", "-text", NULL
    };
    enum options {
        CMD_ALL, CMD_COMMAND, CMD_INDICATOR, CMD_MARK, CMD_STYLE, CMD_TAG, CMD_TEXT
    };
    enum optionMask {
        OPTION_COMMAND = 1, OPTION_MARK = 2, OPTION_TAG = 4, OPTION_TEXT = 8,
        OPTION_INDICATOR=16, OPTION_STYLE=32
    };

    // Parse the switch options
    int i, options = 0;
    char *command = NULL;
    for (i=4; i < objc; i++) {
        int index;

        if (Tcl_GetString(objv[i])[0] != '-') break;

        if (Tcl_GetIndexFromObj(interp, objv[i], optionStrings, "switch", 0, &index) != TCL_OK) {
            return TCL_ERROR;
        }
        switch ((enum options)index) {
            case CMD_ALL:
                options = OPTION_MARK|OPTION_TAG|OPTION_TEXT|OPTION_INDICATOR|OPTION_STYLE;
                break;
            case CMD_COMMAND:
                options |= OPTION_COMMAND;
                if (++i < objc) {
                    command = Tcl_GetString(objv[i]);
                } else {
                    Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
                    return TCL_ERROR;
                }
                break;
            case CMD_INDICATOR:
                options |= OPTION_INDICATOR;
                break;
            case CMD_MARK:
                options |= OPTION_MARK;
                break;
            case CMD_STYLE:
                options |= OPTION_STYLE;
                break;
            case CMD_TAG:
                options |= OPTION_TAG;
                break;
            case CMD_TEXT:
                options |= OPTION_TEXT;
                break;
        }
    }
    if (i >= objc) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    int idx1 = -1;
    int idx2 = -1;
    int line1 = -1;
    int line2 = -1;

    // parse index into line + offset
    if (ParseIndex(interp, sciobj, Tcl_GetString(objv[i++]), line1, idx1) != TCL_OK) {
        // the previous call will leave an error message in the Tcl interp
        return TCL_ERROR;
    }
    if (i < objc) {
        if (ParseIndex(interp, sciobj, Tcl_GetString(objv[i++]), line2, idx2) != TCL_OK) {
            // the previous call will leave an error message in the Tcl interp
            return TCL_ERROR;
        }
    } else {
        line2 = line1;
        idx2 = idx1 + 1;
    }
    if (i != objc) {
        Tcl_WrongNumArgs(interp, 2, objv, "?options? index1 ?index2?");
        return TCL_ERROR;
    }

    // default is to dump everything if specific option not specified
    if (options == 0) {
        options |= OPTION_INDICATOR|OPTION_MARK|OPTION_STYLE|OPTION_TAG|OPTION_TEXT;
    }

    // now go through the various selected options and dump them
    Tcl_DString dstr;
    Tcl_DStringInit(&dstr);

    if (options & OPTION_MARK) {
        // dump marks for margin specified by id
        sciwidgetPtr->markMgr->Dump(&dstr, id, line1, line2, command);
    }
    if (options & (OPTION_TAG|OPTION_STYLE|OPTION_INDICATOR)) {
        int mask = 8; // bit 4 signals margin, not text tags
        if (options&OPTION_TAG)
            mask |= 1;
        if (options&OPTION_STYLE)
            mask |= 2;
        if (options&OPTION_INDICATOR)
            mask |= 4;
        sciwidgetPtr->tagMgr->DumpMargin(mask, &dstr, line1, idx1, line2, idx2, command);
    }
    if (options & OPTION_TEXT) {
        // dump text(command, &dstr)
        DumpText(&dstr, line1, idx1, line2, idx2, command);
    }

    Tcl_SetResult(interp, Tcl_DStringValue(&dstr), TCL_VOLATILE);
    Tcl_DStringFree(&dstr);
    return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdFillNums(int objc, Tcl_Obj *CONST objv[]) {

   if (objc != 6 && objc != 8) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias start count ?tagName {line ...}?");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];

   if (!IS_LINE_NUMS(m->GetType())) {
      char buf[10];
      sprintf(buf, "%d", id);
      Tcl_AppendResult(interp, "margin ", buf, " does not support line numbers", (char *)NULL);
      return TCL_ERROR;
   }
   // the default line-number margin type is managed by Scintilla
   if (m->GetType() == LINE_NUMS) {
      char buf[10];
      sprintf(buf, "%d", id);
      Tcl_AppendResult(interp, "margin ", buf, " does not support custom line filling", (char *)NULL);
      return TCL_ERROR;
   }

   int start; // <<< filled as 0-based

   if (ValidateLine(interp, (ClientData)sciwidgetPtr, objv[4], &start) != TCL_OK) {
      return TCL_ERROR;
   }
   int count;
   if (Tcl_GetIntFromObj(interp, objv[5], &count) != TCL_OK) {
      Tcl_AppendResult(interp, "expected integer but got \"",objv[5],"\"", (char *)NULL);
      return TCL_ERROR;
   }
   // get the line number strings assigned
   int last = sci_get_line_count(SCIOBJ(sciwidgetPtr));

   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

   int offset = sci_cmd(sciobj, SCI_MARGINGETSTYLEOFFSET);
   int sid = sciwidgetPtr->tagMgr->GetMarginTag()->StyleID();

   // Because of how Scintilla looks up styles, we have to make the
   // the style id reflect the fact that the offset will get added
   // to it during the look-up.
   sid -= offset;

   for (int n=0, ln=start; ln <= last; ln++) {
      if ((count > 0) && (++n > count)) break;
      char num[100];
      sprintf(num, "%d", LINE2TK(ln));
      SSM(SCI_MARGINSETTEXT,  ln, (sptr_t)num);
      SSM(SCI_MARGINSETSTYLE, ln, sid);
   }

   // Check if we're styling any of the line numbers
   if (objc == 8) {
      Tcl_Obj *_objv[5];
      _objv[0] = NULL;    // widget path (not needed)
      _objv[1] = NULL;    // "margin" command name (not needed)
      _objv[2] = NULL;    // "tag" sub-command name (not needed)
      _objv[3] = objv[7]; // list of line numbers
      _objv[4] = objv[6]; // tagName

      CmdTag(5, _objv);
   }
   return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdIndex(int objc, Tcl_Obj *CONST objv[]) {

   if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "@x,y");
      return TCL_ERROR;
   }
   char *ix = Tcl_GetString(objv[3]);
   char *cp;
   if (ix[0] != '@') {
      goto badindex;
   }
   int x, y;
   cp = ix+1;
   char *end;
   x = strtol(cp, &end, 0);
   if ((end > cp) && (*end == ',')) {
      // find out if that X location is over a visible margin
      int marginwidths = Width();
      if ((marginwidths == 0) || (x >= marginwidths) || (x < 0)) {
         // we either have no visible margins, or the requested
         // X location is to the right side of the margins
         return TCL_OK;
      }
      // at this point we know the X location must be over a margin
      int id = _atX(x);

      if (id == -1) {
         // should _never_ be here!!
         Tcl_AppendResult(interp, "unable to identify the margin", (char *)NULL);
         return TCL_ERROR;
      }
      // now find out the line corresponding to the Y value
      cp = end+1;
      y = strtol(cp, &end, 0);
      if ((y >= 0) && (end > cp)) {
         int pos  = sci_get_pos_from_xy(SCIOBJ(sciwidgetPtr), x, y);
         int line = sci_get_line(SCIOBJ(sciwidgetPtr), pos);

         Tcl_DString dstr;
         Tcl_DStringInit(&dstr);
         // return {margin# line alias}
         Tcl_DStringAppendElement(&dstr, Tcl_GetString(Tcl_NewIntObj(id)));
         Tcl_DStringAppendElement(&dstr, Tcl_GetString(Tcl_NewIntObj(LINE2TK(line))));
         Tcl_DStringAppendElement(&dstr, margins[id]->GetAlias());

         Tcl_DStringResult(interp, &dstr);
         goto done;
      } else {
         goto badindex;
      }
   } else {
      goto badindex;
   }
 badindex:
   Tcl_AppendResult(interp, "bad index: ", ix, " expected @x,y", (char *)NULL);
   return TCL_ERROR;
 done:
   return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdMarker(int objc, Tcl_Obj *CONST objv[]) {

   if (objc < 5) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option ?arg ...arg?");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];

   if (m->GetType() != SYMBOL) {
      char buf[10];
      sprintf(buf, "%d", id);
      Tcl_AppendResult(interp, "margin ", buf, " does not support markers", (char *)NULL);
      return TCL_ERROR;
   }

   static const char *optionStrings[] = { "add", "clear", "list", (char *)NULL };
   enum options { CMD_ADD, CMD_CLEAR, CMD_LIST };

   int index;
   if (Tcl_GetIndexFromObj(interp, objv[4], optionStrings,
         "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   switch ((enum options)index) {
      case CMD_ADD: {
         if (objc != 7) {
            Tcl_WrongNumArgs(interp, 5, objv, "markerName {line ...}");
            return TCL_ERROR;
         }
         // check for a valid marker name
         char *name = Tcl_GetString(objv[5]);
         MarkerData *mPtr = sciwidgetPtr->markerMgr->FindMarker(id, name);
         if (mPtr == NULL) {
            return TCL_ERROR;
         }
         // check that we were given line number(s)
         Tcl_Obj **numPtrs;
         int cnt;
         if (Tcl_ListObjGetElements(interp, objv[6], &cnt, &numPtrs) != TCL_OK) {
            return TCL_ERROR;
         }
         // now add the marker to each of the lines
         for (int i=0; i < cnt; i++) {
            int n;
            if (Tcl_GetIntFromObj(interp, numPtrs[i], &n) != TCL_OK) {
               return TCL_ERROR;
            }
            // Get the marker added
            if (sciwidgetPtr->markerMgr->AddMarker(id, mPtr, n) != TCL_OK) {
                return TCL_ERROR;
            }
         }
         break;
      }
      case CMD_CLEAR: {
         if (objc != 7) {
            Tcl_WrongNumArgs(interp, 5, objv, "markerName {?line ...?}");
            return TCL_ERROR;
         }
         // check for a valid marker name (note: using "*" deletes all markers from line)
         char *name = Tcl_GetString(objv[5]);
         MarkerData *mPtr = NULL;
         int marker_id = -1;
         if (strcmp(name, "*") != 0) {
            mPtr = sciwidgetPtr->markerMgr->FindMarker(id, name);
            if (mPtr == NULL) {
               return TCL_ERROR;
            }
            marker_id = mPtr->id;
         }
         // check if we were given line number(s)
         Tcl_Obj **numPtrs;
         int cnt;
         if (Tcl_ListObjGetElements(interp, objv[6], &cnt, &numPtrs) != TCL_OK) {
            return TCL_ERROR;
         }
         ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
         if (cnt == 0) {
             //
             // Remove the marker(or all markers if "*") from all lines
             //
             if (strcmp(name, "*") == 0) {
                 // Remove all markers from all lines for this margin
                 RemoveAllMarkers(id);

             } else {
                 // Removes just the specified marker from all lines 
                 RemoveMarker(id, marker_id);
             }
         } else {
            for (int i=0; i < cnt; i++) {
               int n;
               if (Tcl_GetIntFromObj(interp, numPtrs[i], &n) != TCL_OK) {
                  return TCL_ERROR;
               }
               //
               // Get the marker(or all markers if "*") removed from the line
               //
               sci_cmd(sciobj, SCI_MARKERDELETE, LINE2SCI(n), marker_id);
            }
         }
         break;
      }
      case CMD_LIST: {
         // return list of all marks of given type
         if (objc != 6 && objc != 7) {
            Tcl_WrongNumArgs(interp, 5, objv, "markerName ?line?");
            return TCL_ERROR;
         }
         ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
         char *name = Tcl_GetString(objv[5]);

         if (strcmp(name, "*") == 0) {
             // the marker name is "*", verify we got a line number
             if (objc != 7) {
                 Tcl_AppendResult(interp, "must specify line number with marker name \"*\"", (char *)NULL);
                 return TCL_ERROR;
             }
             int line;
             if (Tcl_GetIntFromObj(interp,objv[6], &line) != TCL_OK) {
                 return TCL_ERROR;
             }
             // only get the markers set for the specified line
             int mask = sci_get_margin_mask(sciobj, id);
             int markers_mask = (mask & sci_cmd(sciobj, SCI_MARKERGET, LINE2SCI(line)));
             Tcl_Obj *listObj = Tcl_NewListObj(0,NULL);
             for (int i=0; i < MARKER_MAX; i++) {
                 if (markers_mask & (1 << i)) {
                     MarkerData *mPtr = sciwidgetPtr->markerMgr->FindMarker(i);
                     if (mPtr)
                         Tcl_ListObjAppendElement(interp, listObj, Tcl_NewStringObj(mPtr->name, -1));
                 }
             }
             Tcl_SetObjResult(interp, listObj);

         } else {
             // check for a valid marker name
             MarkerData *mPtr = sciwidgetPtr->markerMgr->FindMarker(name);
             if (mPtr == NULL) {
                 return TCL_ERROR;
             }
             Tcl_Obj *listObj = Tcl_NewListObj(0,NULL);
             int found = 1;
             int line = 0;
             int mask = 1<<mPtr->id;
             do {
                 line = sci_cmd(sciobj, SCI_MARKERNEXT, line, mask);
                 if (line == -1)
                     found = 0;
                 else
                     Tcl_ListObjAppendElement(interp, listObj, Tcl_NewIntObj(LINE2TK(line)));
                 ++line; // go past current line
             } while (found);
             Tcl_SetObjResult(interp, listObj);
         }
         break;
      }
   }
   return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdShow(int objc, Tcl_Obj *CONST objv[], bool show) {

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias ?col#|name|alias...?");
      return TCL_ERROR;
   }
   int id;
   for (int i=3; i < objc; i++) {
      if (_get(objv[i], &id) != TCL_OK) {
         return TCL_ERROR;
      }
      if (_show(id, show) != TCL_OK) {
         return TCL_ERROR;
      }
   }
   return TCL_OK;
}

//==========================================================================
// Command options are:
//   1. ADD a tag to margin
//          >>> $sci margin tag add tagName index1 ?index2
//          >>>    If only one index specified, just the one character position is modified.
//   2. QUERY a location for tag names
//          >>> $sci margin tag names ?index1
//          >>>    Returns tag names for given index, if index missing returns all margin tag names
//   3. REMOVE a tag from margin
//          >>> $sci margin tag remove tagName index1 ?index2
//          >>>    If only one index specified, just the one character position is modified.
//
int MarginMgr::CmdTag(int objc, Tcl_Obj *CONST objv[]) {
    // margin tag cmd tagName ?idx...
    if ((objc < 4) || (objc > 7)) {
        Tcl_WrongNumArgs(interp, 3, objv, "cmd ?tagName ?index1 ?index2");
        return TCL_ERROR;
    }
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    static const char *optionStrings[] = { "add", "names", "remove", (char *)NULL };
    enum options { CMD_ADD, CMD_NAMES, CMD_REMOVE };

    int index;
    if (Tcl_GetIndexFromObj(interp, objv[3], optionStrings,
         "option", 0, &index) != TCL_OK) {
        return TCL_ERROR;
    }

    TagData *tagPtr = NULL;
    int sid = 0;
    // flag for style for entire margin string
    int everything = 0;
    int line, idx1, idx2, len;

    // get style id offset for margin styles
    int offset = sci_cmd(sciobj, SCI_MARGINGETSTYLEOFFSET);

    // names command does not have tagName and may be missing index value
    if (CMD_NAMES == (enum options)index) {
        if (objc > 5) {
            Tcl_WrongNumArgs(interp, 3, objv, "names ?index");
            return TCL_ERROR;
        }
        if (objc == 5) {
            // index provided
            if (ParseIndex(interp, sciobj, Tcl_GetString(objv[4]), line, idx1) != TCL_OK) {
                // the previous call will leave an error message in the Tcl interp
                return TCL_ERROR;
            }
        } else {
            everything = 1;
        }
        // calling with NULL pointer gets the length of margin text string
        len = sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line), 0);
    } else {
        // get tag name
        char *tagName = Tcl_GetString(objv[4]);
        tagPtr = sciwidgetPtr->tagMgr->ValidateForMargin(tagName);
        if (tagPtr == NULL) {
            // the previous call will leave an error message in the Tcl interp
            return TCL_ERROR;
        }


        // get index range for tag
        if (ParseIndex(interp, sciobj, Tcl_GetString(objv[5]), line, idx1) != TCL_OK) {
            // the previous call will leave an error message in the Tcl interp
            return TCL_ERROR;
        }
        if (objc == 7) {
            int line2;
            if (ParseIndex(interp, sciobj, Tcl_GetString(objv[6]), line2, idx2) != TCL_OK) {
                // the previous call will leave an error message in the Tcl interp
                return TCL_ERROR;
            }
            if (line != line2) {
                // tag can not span multiple lines
                _reportError("margin tag may not span multiple lines");
                return TCL_ERROR;
            }
        } else {
            // only one index, single char gets tag
            idx2 = idx1 + 1;
        }

        // calling with NULL pointer gets the length of margin text string
        len = sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line), 0);
        if ((idx1 == 0) && (idx2 == len))
            everything = 1; // entire string selected

        // determine desired style id
        sid = tagPtr->StyleID();
        // adjust for call to SCI_MARGINSETSTYLE
        sid -= offset;
    }

    // removing a tag style means putting it back to that of the margin "background" style
    int background_sid = sciwidgetPtr->tagMgr->GetMarginTag()->StyleID();
    background_sid -= offset;

    switch ((enum options)index) {
        case CMD_ADD: {
                if (len == 0) {
                    if (everything) {
                        // simple command to set style for entire margin
                        sci_cmd(sciobj, SCI_MARGINSETSTYLE, LINE2SCI(line), sid);
                    } else {
                        // original style was for whole line, now we need to set different styles
                        // for different characters, first get length of text
                        len = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line), 0);
                        // allocate buffer for style array
                        char *buf = new char[len+1];
                        // set codes for line
                        for (int i = 0; i < len; ++i) {
                            if ((i >= idx1) && (i < idx2)) {
                                buf[i] = sid;
                            } else {
                                // sid is unset, set it to background sid
                                buf[i] = background_sid;
                            }
                        }
                        sci_cmd(sciobj, SCI_MARGINSETSTYLES, LINE2SCI(line), (sptr_t)buf);
                        delete[] buf;
                    }
                } else {
                    // allocate buffer for style array
                    char *buf = new char[len+1];
                    // get style codes for line
                    sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line), (sptr_t)buf);
                    // set style for selected indexes
                    for (int i = 0; i < len; ++i) {
                        if ((i >= idx1) && (i < idx2)) {
                            buf[i] = sid;
                        }
                    }
                    // set codes for line
                    sci_cmd(sciobj, SCI_MARGINSETSTYLES, LINE2SCI(line), (sptr_t)buf);
                    delete[] buf;
                }
            }
            break;
        case CMD_NAMES: {
                // return tag name(s) in effect for given index or all margin tag names
                if (everything) {
                    // return all margin tags
                    sciwidgetPtr->tagMgr->GetMarginTags(interp);
                } else if (len == 0) {
                    // no style tags, return style for entire line
                    int curr_sid = sci_cmd(sciobj, SCI_MARGINGETSTYLE, LINE2SCI(line));
                    tagPtr = sciwidgetPtr->tagMgr->FindTag(curr_sid+offset);
                    if (tagPtr && (tagPtr->Type() & TT_MARG)) {
                        Tcl_SetObjResult(interp, Tcl_NewStringObj(tagPtr->Name(), -1));
                    } else {
                        Tcl_ResetResult(interp); // return nothing
                    }
                } else {
                    // return list of magin tags at given index, Scintilla only supports one style for a given character
                    // so we only return a single tag
                    // allocate buffer for style array
                    unsigned char *buf = new unsigned char[len+1];
                    // get style codes for line
                    sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line), (sptr_t)buf);
                    if (idx1 < len) {
                        sid = buf[idx1];
                        tagPtr = sciwidgetPtr->tagMgr->FindTag(sid+offset);
                        if (tagPtr && (tagPtr->Type() & TT_MARG)) {
                            Tcl_SetObjResult(interp, Tcl_NewStringObj(tagPtr->Name(), -1));
                        }
                    } else {
                        // index outside of text, return empty string (i.e. nothing)
                        Tcl_ResetResult(interp); // clear error info from GetIntFromObj call
                    }
                    delete[] buf;
                }
            }
            break;
        case CMD_REMOVE: {
                if (sid == background_sid) {
                    _reportError("You may not remove margin tag");
                    return TCL_ERROR;
                }
                if (len == 0) {
                    int curr_sid = sci_cmd(sciobj, SCI_MARGINGETSTYLE, LINE2SCI(line));
                    if (curr_sid == sid) {
                        if (everything) {
                            // simple command to set style for entire margin
                            sci_cmd(sciobj, SCI_MARGINSETSTYLE, LINE2SCI(line), background_sid);
                        } else {
                            // original style was for whole line, now we need to set different styles
                            // for different characters, first get length of text
                            len = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line), 0);
                            // allocate buffer for style array
                            char *buf = new char[len+1];
                            // set codes for line
                            for (int i = 0; i < len; ++i) {
                                if ((i >= idx1) && (i < idx2)) {
                                    // remove style
                                    buf[i] = background_sid;
                                } else {
                                    // leave style alone
                                    buf[i] = curr_sid;
                                }
                            }
                            sci_cmd(sciobj, SCI_MARGINSETSTYLES, LINE2SCI(line), (sptr_t)buf);
                            delete[] buf;
                        }
                    }
                } else {
                    // allocate buffer for style array
                    char *buf = new char[len+1];
                    // get style codes for line
                    sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line), (sptr_t)buf);
                    // set style for selected indexes
                    for (int i = 0; i < len; ++i) {
                        if ((i >= idx1) && (i < idx2) && (buf[i] == sid)) {
                            buf[i] = background_sid;
                        }
                    }
                    // set codes for line
                    sci_cmd(sciobj, SCI_MARGINSETSTYLES, LINE2SCI(line), (sptr_t)buf);
                    delete[] buf;
                }
            }
            break;
    }
    
    return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdText(int objc, Tcl_Obj *CONST objv[]) {

   if (objc < 5 || objc > 7) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option ?arg ...arg?");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];

   switch (m->GetType()) {
      case TEXT:
      case RTEXT:
         break;
      default: {
         char buf[10];
         sprintf(buf, "%d", id);
         Tcl_AppendResult(interp, "margin ", buf, " does not support text", (char *)NULL);
         return TCL_ERROR;
         }
         break;
   }

   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
   static const char *optionStrings[] = { "clear", "get", "set", (char *)NULL };
   enum options { CMD_CLEAR, CMD_GET, CMD_SET };

   int index;
   if (Tcl_GetIndexFromObj(interp, objv[4], optionStrings,
         "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   switch ((enum options)index) {
      case CMD_CLEAR: {
        if (objc == 5) {
            /* clear all lines */
            sci_cmd(sciobj, SCI_MARGINTEXTCLEARALL);
        } else if (objc == 6) {
            /* clear specific lines */
            Tcl_Obj **numPtrs;
            int cnt;
            if (Tcl_ListObjGetElements(interp, objv[5], &cnt, &numPtrs) != TCL_OK) {
               return TCL_ERROR;
            }
            for (int i=0; i < cnt; i++) {
               int n;
               if (Tcl_GetIntFromObj(interp, numPtrs[i], &n) != TCL_OK) {
                  return TCL_ERROR;
               }
               // clear text
               sci_cmd(sciobj, SCI_MARGINSETTEXT, LINE2SCI(n), (sptr_t)"");
            }
        } else {
           Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option ?arg ...arg?");
           return TCL_ERROR;
        }
         break;
      }
      case CMD_GET: {
         // get text for designated line
         if (objc != 6) {
           Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option ?arg ...arg?");
           return TCL_ERROR;
         }
         int line;
         if (Tcl_GetIntFromObj(interp, objv[5], &line) != TCL_OK) {
            return TCL_ERROR;
         }
         // calling with NULL pointer gets the length of string
         int len = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line), 0);
         if (len > 0) {
            char *buf = new char[len+1];
            sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line), (sptr_t)buf);
            buf[len] = 0;   // NUL terminate
            Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, len));
            delete[] buf;
         } else
            Tcl_SetObjResult(interp, Tcl_NewStringObj("", 0));
         break;
      }
      case CMD_SET: {
         if (objc != 7) {
           Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias option ?arg ...arg?");
           return TCL_ERROR;
         }
         char *str = Tcl_GetString(objv[5]);
         Tcl_Obj **numPtrs;
         int cnt;
         if (Tcl_ListObjGetElements(interp, objv[6], &cnt, &numPtrs) != TCL_OK) {
            return TCL_ERROR;
         }

         // Get styyle id of margin so the text gets the same style as the rest of the margin
         int offset = sci_cmd(sciobj, SCI_MARGINGETSTYLEOFFSET);
         int sid = sciwidgetPtr->tagMgr->GetMarginTag()->StyleID();

         // Because of how Scintilla looks up styles, we have to make the
         // the style id reflect the fact that the offset will get added
         // to it during the look-up.
         sid -= offset;
         for (int i=0; i < cnt; i++) {
            int n, ln;
            if (Tcl_GetIntFromObj(interp, numPtrs[i], &n) != TCL_OK) {
               return TCL_ERROR;
            }
            // set text
            ln = LINE2SCI(n);
            sci_cmd(sciobj, SCI_MARGINSETTEXT, ln, (sptr_t)str);
            SSM(SCI_MARGINSETSTYLE, ln, sid);
         }
         break;
      }
   }
   return TCL_OK;
}

//==========================================================================
int MarginMgr::CmdVisible(int objc, Tcl_Obj *CONST objv[]) {

   if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "column#|name|alias");
      return TCL_ERROR;
   }
   int id;
   if (_get(objv[3], &id) != TCL_OK) {
      return TCL_ERROR;
   }
   MarginData *m = margins[id];
   bool is_vis = (m->GetVisibleWidth() > 0) ? true : false;
   Tcl_SetObjResult(interp, Tcl_NewBooleanObj(is_vis));

   return TCL_OK;
}

//==========================================================================
void MarginMgr::DumpText(Tcl_DString *dstr, int line1, int col1, int line2, int col2, const char* command)
{
    Tcl_DString cmdStr;
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    if (command) {
        Tcl_DStringInit(&cmdStr);
        Tcl_DStringAppendElement(&cmdStr, command);
        dstr = &cmdStr;
    }
    int len;
    int idx1 = col1, idx2;
    while (line1 <= line2) {
        // calling with NULL pointer gets the length of margin text string
        len = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line1), 0);
        char *buf = new char[len+1];
        // get text for line
        sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line1), (sptr_t)buf);
        if (line1 < line2) {
            // set idx2 to end of current line if we haven't reached the end line yet
            idx2 = len;
        } else {
            // else set end position of last line
            idx2 = col2;
        }
        // append text element(s) from idx1 to idx2 to dstr
        char *newline, *cp = &buf[idx1];
        buf[idx2] = 0;  // terminate text at idx2
        while (cp && *cp) {
            Tcl_DStringAppendElement(dstr, "text");
            newline = strchr(cp, '\n');
            if (newline++) {
                char savechar = *newline;
                *newline= 0;
                Tcl_DStringAppendElement(dstr, cp);
                LineColumnToIndex(line1, idx1, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
                *newline = savechar;
                idx1 += (newline - cp);
                cp = newline;
            } else {
                Tcl_DStringAppendElement(dstr, cp);
                LineColumnToIndex(line1, idx1, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
                break;
            }
        }
        ++line1;
        idx1 = 0;
        delete[] buf;
    }
    if (command) {
        Tcl_DStringFree(&cmdStr);
    }
}

//==========================================================================
// If the specified marker id is set for any of the margins, it is removed
//
void MarginMgr::RemoveMarker(int marker_id) {

    for (int i=0; i < MAXM; i++) {
        if (_hasMarker(i, marker_id)) {
            RemoveMarker(i, marker_id);
        }
    }
}
//==========================================================================
// Remove just the specified id for the specified margin
// NOTE: This DOES NOT cause Scintilla to remove the marker's definition.
//
void MarginMgr::RemoveMarker(int marginID, int marker_id) {

    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

    // first we update the margin mask to take out the tag
    int curr_mask = sci_get_margin_mask(sciobj, marginID);
    sci_set_margin_mask(sciobj, marginID, (curr_mask & ~(1 << marker_id)));
//fprintf(stderr,"REMOVEMARKER: marginID=%d  markerID=%d\n",marginID,marker_id);

    // then we tell Scintilla to do the same (this updates 
    // Scintilla's internal data management of the marker)
    sci_cmd(sciobj, SCI_MARKERDELETEALL, marker_id);
}
//==========================================================================
// This one scans for any markers added for the specified margin
// and removes them
//
void MarginMgr::RemoveAllMarkers(int marginID) {

    for (int i=0; i < sciwidgetPtr->markerMgr->GetMaxId(); i++) {
        if (_hasMarker(marginID, i)) {
            RemoveMarker(marginID, i);
        }
    }
}

//==========================================================================
bool MarginMgr::SetCursor(int x, int y) {

    int id = _atX(x);
    if (id == -1) return false;

    MarginData *mPtr = margins[id];
    if (mPtr->GetCursor() != None) {
        Tk_DefineCursor(tkwin, mPtr->GetCursor());
    } else {
        // when a margin doesn't define it's own cursor we inherit
        // the cursor used by the main text area

        if (sciwidgetPtr->cursor != None) {
            Tk_DefineCursor(tkwin, sciwidgetPtr->cursor);
        } else {
            Tk_UndefineCursor(tkwin);
        }
    }
    return true;
}

//==========================================================================
void MarginMgr::UpdateMargin(int id) {

    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    MarginData *m = margins[id];

    // special handling of fold margin
    bool clickable = false;
    if (m->GetType() == FOLD) {
        clickable = true;
    } else {
        clickable = (m->GetClickCmd() != NULL);
    }
    sci_enable_margin_click(obj, id, clickable);

    int curr = sci_get_margin_width(obj, id);
    if (curr != m->GetWidth()) {
        sci_set_margin_width(obj, id, m->GetWidth());

        // check if client has requested to hear about width changes
        if (sciwidgetPtr->marginWidthCmd) {

            Tcl_Obj *obj = Tcl_ObjPrintf("%s %d %d %d",
                sciwidgetPtr->marginWidthCmd, id, curr, m->GetWidth()
            );
            if (Tcl_Eval(interp, Tcl_GetString(obj)) != TCL_OK) {
               fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(interp));
            }
        }
    }
}

//==========================================================================
// Called following a change to a margin's type
//
int MarginMgr::ValidateTypeChange(int id) {

   MarginData *m = margins[id];
   MarginType type = m->GetType();

   // the margin types for linenumbers and folding are special in that there
   // should only be one for each type

   // look for duplicate assignments of a line number margin
   if (IS_LINE_NUMS(type)) {
      if ((currLnum != -1) && (margins[currLnum]->col_id != id)) {
         char buf[100];
         sprintf(buf, "cannot assign multiple margins for line numbers(margin %d is already set for showing them)", currLnum);
         _reportError(buf);

         return TCL_ERROR;
      }
      currLnum = id;
   } else if (currLnum == id) {
      currLnum = -1; // reset since type is being changed!
   }
   // look for duplicate assignments of a fold margin
   if (type == FOLD) {
      if ((currFold != -1) && (margins[currFold]->col_id != id)) {
         char buf[100];
         sprintf(buf, "cannot assign multiple margins for folding(margin %d is already set for that)", currFold);
         _reportError(buf);

         return TCL_ERROR;
      }
      currFold = id;
   } else if (currFold == id) {
      currFold = -1; // reset
   }
   // now we can tell Scintilla the type of the margin
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
   switch (type) {
      case NONE: 
         break;
      case LINE_NUMS:
         sci_set_margin_type(obj, id, SC_MARGIN_NUMBER); 
         break;
      case LINE_NUMS_ALT:
         sci_set_margin_type(obj, id, SC_MARGIN_RTEXT); 
         break;
      case TEXT:
         sci_set_margin_type(obj, id, SC_MARGIN_TEXT); 
         break;
      case RTEXT:
         sci_set_margin_type(obj, id, SC_MARGIN_RTEXT); 
         break;
      case SYMBOL:
         sci_set_margin_type(obj, id, SC_MARGIN_SYMBOL);

         // By default set the mask so _NO_ markers are enabled.
         // The mask is dynamically adjusted based on which
         // markers are requested by the client(see MarkerMgr::AddMarker)
         //
         sci_set_margin_mask(obj, id, 0);
         break;
      case FOLD:
         sci_set_margin_type(obj, id, SC_MARGIN_SYMBOL); 
         sci_set_margin_mask(obj, id, SC_MASK_FOLDERS); 

         // set appearance of margin column background
         SSM(SCI_SETFOLDMARGINCOLOUR,  id,_RGB_(0xec,0xec,0xec));
         SSM(SCI_SETFOLDMARGINHICOLOUR,id,_RGB_(0xff,0xff,0xff)); // use same color to avoid stipple

         // show a horizontal line under folded levels
         sci_set_fold_flags(obj, SC_FOLDFLAG_LINEAFTER_CONTRACTED);

         // always enable fold margin to be clickable
         sci_enable_margin_click(SCIOBJ(sciwidgetPtr), id, true);
         break;
   }
   return TCL_OK;
}

//==========================================================================
// These next two work together to allow "scaling" of margin widths
// following a zoom operation
//
void MarginMgr::ZoomPre() {

   // save the current character width
   preZoomWidth = sci_get_text_width(SCIOBJ(sciwidgetPtr), "5");
}
void MarginMgr::ZoomPost() {

   // see how much the character width has changed, if any
   int postZoomWidth = sci_get_text_width(SCIOBJ(sciwidgetPtr), "5");
   int delta = postZoomWidth - preZoomWidth;

   if (delta == 0) return;

   // ok, the plan is we'll scale only margins that are text based
   for (int i=0; i < MAXM; i++) {
      MarginData *m = margins[i];
      MarginType t  = m->GetType();
      int width     = m->GetWidth();
      if ((width > 0) && IS_TEXT_TYPE(t)) {
         double numchars = (double)width / (double)preZoomWidth;

         m->SetWidth(width + (int)((ceil)(delta * numchars))); // "ceil" for extra space
         UpdateMargin(i);
      }
   }
}

//==========================================================================
// Return back the total width(in pixels) for all visible margins
//
int MarginMgr::Width() {

   int w = 0;
   for (int i=0; i < MAXM; i++) {
      w += margins[i]->GetVisibleWidth();
   }
   return w;
}


//---------------------------------------------------------
// Private functions
//---------------------------------------------------------

//==========================================================================
// Return the margin id at the specified X screen location, or -1 if
// not over a margin
int MarginMgr::_atX(int xloc) {

    int id = -1;
    int left = 0;
    for (int i=0; i < MAXM; i++) {
        MarginData *m = margins[i];
        int w = m->GetVisibleWidth();
        if (w == 0) continue;
        left += w;
        if (xloc <= left) {
            id = i;
            break;
        }
    }
    return id;
}

//==========================================================================
void MarginMgr::_click(
   int  clicked_margin, /* the margin the click was in */
   int  pos,            /* document position where it occurred */
   int  x,              /* X screen location of the click */
   int  y,              /* Y screen location of the click */
   int  btn_id,         /* button id (1=LMB, 3=RMB) */
   bool shift,          /* was SHIFT key modifier down */
   bool ctrl,           /* was CONTROL key modifier down */
   bool alt             /* was ALT key modifier down */
) {

   if (_validId(clicked_margin) != TCL_OK) {
      // __should__ never get here
      return;
   }
   MarginData *m = margins[clicked_margin];
   ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

//printf("MARGINMGR->clicked in margin %d at pos=%d w/ type=%s\n",clicked_margin,pos,MarginTypeNames[m->GetType()]);
   if (m->GetType() == FOLD) {
      if (btn_id == 1) {
         int line = sci_get_line(obj, pos);

         // call "fold" command so we get consistent behavior
         Tcl_Obj *objv[4];
         objv[0] = NULL; // widget path (not needed)
         objv[1] = NULL; // "fold" command name (not needed)
         objv[2] = Tcl_NewStringObj("toggle", -1);
         objv[3] = Tcl_NewIntObj(LINE2TK(line));
         CmdFold((ClientData)sciwidgetPtr, interp, 4, objv);
      }
   } else if (m->GetType() == SYMBOL) {

//#define DEMO_INDICATORS
#ifdef DEMO_INDICATORS
      int line = sci_get_line(obj, pos);
      // CLICK in 1st line to get all the indicators drawn on subsequent lines
      if (line == 0) {
         for (int i=0; i<=12; i++) {
            SSM(SCI_INDICSETSTYLE, i, i);
            SSM(SCI_SETINDICATORCURRENT, i, 0);
            int pos1 = sci_get_pos(obj, i);
            int pos2 = sci_get_lineend_pos(obj, i);
            SSM(SCI_INDICATORFILLRANGE, pos1, (pos2-pos1));
         }
      }
#endif
   }
   // check if client wants to hear about the click
   if (m->GetClickCmd()) {
      int line = sci_get_line(obj, pos);
      Tcl_Obj *objPtr = Tcl_ObjPrintf("%s %d %d %d %d %d %d %d %d", 
         m->GetClickCmd(), 
         clicked_margin,
         LINE2TK(line),
         x, y,
         btn_id,
         shift, ctrl, alt
      );
      if (Tcl_Eval(interp, Tcl_GetString(objPtr)) != TCL_OK) {
         fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(interp));
      }
   }
}

//==========================================================================
int MarginMgr::Get(Tcl_Obj *obj, int *id) {
    return _get(obj, id);
}
int MarginMgr::_get(Tcl_Obj *obj, int *id) {

   int num = -1;

   // Check if we were given a column number/id...

   if (Tcl_GetIntFromObj(interp, obj, &num) != TCL_OK) {
      // not an id, then we assume this is a name

      Tcl_ResetResult(interp); // clear error info from GetIntFromObj call

      if (_validName(Tcl_GetString(obj), &num) != TCL_OK) {
         return TCL_ERROR;
      }
   } else {
      // validate number
      if (_validId(num) != TCL_OK) {
         return TCL_ERROR;
      }
   }
   *id = num; // return the found margin id
   return TCL_OK;
}

//==========================================================================
// Return if the specified margin has the specified marker id set
bool MarginMgr::_hasMarker(int marginID, int marker_id) {

    if (marginID < 0 || marginID > MAXM) return false;

    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

    int curr_mask = sci_get_margin_mask(sciobj, marginID);
    return (curr_mask & (1 << marker_id));
}

//==========================================================================
void MarginMgr::_reportError(const char *msg) {
   if (msg) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, msg, (char *)NULL);
   }
}
void MarginMgr::_reportError(Tcl_Obj *obj) {
   if (obj) {
      _reportError(Tcl_GetString(obj));
   }
}

//==========================================================================
int MarginMgr::_show(int id, bool show) {

   if (_validId(id) != TCL_OK) {
      return TCL_ERROR;
   }
   ScintillaObject *sci = SCIOBJ(sciwidgetPtr);
   MarginData *m = margins[id];
//printf("MarginMgr::_show id=%d type=%d typestr=%s show=%d width=%d\n",id,m->GetType(),m->GetTypeStr(),show,m->GetWidth());
   if (show) {
      // some margins have pre-defined widths, or ways
      // of setting them
      if (m->GetWidth() == 0) {
         switch (m->GetType()) {
            case LINE_NUMS:
            case LINE_NUMS_ALT:
               m->SetWidth(GetLineNWidth());
               break;
            case FOLD:
               m->SetWidth(14); // same as SciTE
               break;
            default:
               if (m->viswidth > 0) {
                  // reuse the last width value that was set when
                  // the margin was showing
                  m->SetWidth(m->viswidth);
               }
               break;
         }
      }
      if (m->GetWidth() == 0) {
         _reportError(Tcl_ObjPrintf("cannot show margin %d, it has zero width", m->col_id));
         return TCL_ERROR;
      }
      sci_set_margin_width(sci, m->col_id, m->GetWidth());

      m->viswidth = m->GetWidth();
   } else {
      sci_set_margin_width(sci, m->col_id, 0);
      m->viswidth = 0;
   }
   return TCL_OK;
}

//==========================================================================
int MarginMgr::_validAlias(const char *alias) {

   // check for a duplicate
   for (int i=0; i < MAXM; i++) {
      MarginData *m = margins[i];
      if (m->GetAlias() && (strcasecmp(alias, m->GetAlias()) == 0)) {
         char buf[100];
         sprintf(buf, "alias \"%s\" is already in use by margin %d", alias, i);
         _reportError(buf);

         return TCL_ERROR;
      }
   }
   return TCL_OK;
}

//==========================================================================
int MarginMgr::_validId(int id) {

   if (id >=0 && id < MAXM) {
      return TCL_OK;
   }
   char buf[100];
   sprintf(buf, "bad margin column(%d): must be in range 0 %d", id, MAXM-1);
   _reportError(buf);

   return TCL_ERROR;
}

//==========================================================================
int MarginMgr::_validName(const char *name, int *id) {

   // check if the specified name matches an existing margin, and if
   // not then treat the name as an alias and see if that matches.

   for (int i=0; i < MAXM; i++) {
      MarginData *m = margins[i];
      if (strcasecmp(name, m->GetTypeStr()) == 0 ||
         (m->GetAlias() && (strcasecmp(name, m->GetAlias()) == 0))) {
         *id = i;
         return TCL_OK;
      }
   }
   char buf[100];
   sprintf(buf, "bad column name \"%s\"", name);
   _reportError(buf);

   return TCL_ERROR;
}

//==========================================================================
int MarginMgr::_validType(const char *type_s, MarginType *type) {

   static int lentypes = sizeof(MarginTypeNames) / sizeof(*MarginTypeNames);

   for (int i=0; i < lentypes; i++) {
      if (strcasecmp(type_s, MarginTypeNames[i]) == 0) {
         *type = static_cast<MarginType>(i);
         return TCL_OK;
      }
   }
   // compose the error message...

   Tcl_Obj *objbuf = Tcl_ObjPrintf("bad margin type \"%s\": must be ", type);

   for (int i=0; i < lentypes; i++) {
      if (i == (lentypes-1))
         Tcl_AppendToObj(objbuf, " or ", -1);
      else if (i != 0)
         Tcl_AppendToObj(objbuf, ", ", -1);
      Tcl_AppendToObj(objbuf, MarginTypeNames[i], -1);
   }
   _reportError(Tcl_GetString(objbuf));
   return TCL_ERROR;
}

/////////////////////////////////////////////////////////////////////////////////
// Tag configuration options
//
static const char *IndicatorTypeNames[] = {
    "none", "plain", "squiggle", "tt", "diagonal", "strike",
    "hidden", "box", "roundbox", "squarebox", "dash", "dots", 
    "squigglelow", "dotbox", "squigglepixmap", NULL
};

#define TOPT_BG          1
#define TOPT_FG          1 << 1
#define TOPT_UNDERLINE   1 << 2
#define TOPT_BOLD        1 << 3
#define TOPT_ITALIC      1 << 4
#define TOPT_INDIC_TYPE  1 << 5
#define TOPT_INDIC_ALPHA 1 << 6
#define TOPT_EOLFILLED   1 << 7

static Tk_OptionSpec tagOptionSpecs[] = {
   {TK_OPTION_COLOR,  "-background", NULL, NULL,
         NULL, -1, Tk_Offset(TagDataOptions, bgColor), TK_OPTION_NULL_OK, 0, TOPT_BG
   },
   {TK_OPTION_SYNONYM, "-bg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-background", 0},

   {TK_OPTION_BOOLEAN, "-bold", NULL, NULL,
         "0", -1, Tk_Offset(TagDataOptions, bold), 0, 0, TOPT_BOLD
   },
    // default for eolfilled is 1 (on) to be consistent with the Tk text widget
   {TK_OPTION_BOOLEAN, "-eolfilled", NULL, NULL,
         "1", -1, Tk_Offset(StyleOptions, eolfilled), 0, 0, TOPT_EOLFILLED
   },
   {TK_OPTION_COLOR,  "-foreground", NULL, NULL,
         NULL, -1, Tk_Offset(TagDataOptions, fgColor), TK_OPTION_NULL_OK, 0, TOPT_FG
   },
   {TK_OPTION_SYNONYM, "-fg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-foreground", 0},

   {TK_OPTION_STRING_TABLE, "-indicatortype", NULL, NULL,
	(char*)IndicatorTypeNames[indicNONE], -1, Tk_Offset(TagDataOptions, indicType), 0, (ClientData)IndicatorTypeNames, TOPT_INDIC_TYPE
   },
   {TK_OPTION_INT,     "-alpha", "alpha", "Alpha",
         "-1", -1, Tk_Offset(TagDataOptions, alpha), 0, 0, 0
   },
   {TK_OPTION_INT,     "-alphab", "alphab", "AlphaB",
         "-1", -1, Tk_Offset(TagDataOptions, alphab), 0, 0, 0
   },
   {TK_OPTION_BOOLEAN, "-italic", NULL, NULL,
         "0", -1, Tk_Offset(TagDataOptions, italic), 0, 0, TOPT_ITALIC
   },
   {TK_OPTION_BOOLEAN, "-underline", NULL, NULL,
         "0", -1, Tk_Offset(TagDataOptions, underline), 0, 0, TOPT_UNDERLINE
   },
   {TK_OPTION_END}
};


TagData::TagData(Tcl_Interp *interp, const char* n, int ttMask, int _priority)
{
    name        = new char[strlen(n)+1];
    strcpy(name, n);
    style_id    = -1;
    indic_id    = INDIC_PLAIN;
    tagType     = ttMask;
    options.bgColor     = NULL;
    options.fgColor     = NULL;
    options.bold        = 0;
    options.eolfilled   = 1;    // default is on to be consistent with Tk text widget
    options.italic      = 0;
    options.underline   = 0;
    options.indicType   = indicNONE;
    options.alpha       = -1;
    options.alphab      = -1;
    optionTable = Tk_CreateOptionTable(interp, tagOptionSpecs);
    // establish tag priority
    priority = _priority;
    flags = FLAG_NONE;
}

TagData::~TagData()
{
    delete [] name;
    Tk_DeleteOptionTable(optionTable);
}

void TagData::RegisterLocation(TagLocation *tl_p)
{
std::vector<TagLocation*>::iterator it;

#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"RegisterLocation %p %ld-%ld to tag %s\n", tl_p, tl_p->startpos(), tl_p->endpos(), Name());
    dumpLocations("TagData Locations", tagLocations);
#endif
    TagLocation *found_p = searchLocations("TagData Locations", tagLocations, SEARCH_IDX, tl_p->startpos(), tl_p->endpos(), &it);
    if (!found_p) {
        tagLocations.insert(it, tl_p);
#ifdef DEBUG_TAGLOCATIONS
        fprintf(stderr,"RegisterLocation added %p %ld-%ld to tag %s\n", tl_p, tl_p->startpos(), tl_p->endpos(), Name());
        dumpLocations("TagData Locations", tagLocations);
#endif
        checkLocations(tagLocations);
    }
}

void TagData::UnregisterLocation(TagLocation *tl_p)
{
std::vector<TagLocation*>::iterator it;

    // if we're in the process of deleting this tag, don't update the tagLocations vector or TagData::DeleteTag will have
    // invalid iterator issues
    if (flags & FLAG_DELETING)
        return;

#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"UnregisterLocation %p %ld-%ld from tag %s\n", tl_p, tl_p->startpos(), tl_p->endpos(), Name());
    dumpLocations("TagData Locations", tagLocations);
#endif
    // search tag location vector for desired tag
    TagLocation *found_p = searchLocations("TagData Locations", tagLocations, SEARCH_IDX, tl_p->startpos(), tl_p->endpos(), &it);
    if (found_p && found_p == tl_p) {
        // remove location from vector
        tagLocations.erase(it);
#ifdef DEBUG_TAGLOCATIONS
        fprintf(stderr,"UnregisterLocation erased %p %ld-%ld from tag %s\n", tl_p, tl_p->startpos(), tl_p->endpos(), Name());
        dumpLocations("TagData Locations", tagLocations);
#endif
    }
}

// implements tag nextrange command
void TagData::NextRange(Tcl_Interp *interp, ScintillaObject *obj, sptr_t pos1, sptr_t pos2, Tcl_Obj* indices)
{
std::vector<TagLocation*>::iterator it;
    // search tag location vector for desired tag
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"NextRange %ld-%ld tag %s\n", pos1, pos2, Name());
    dumpLocations("TagData Locations", tagLocations);
#endif
    TagLocation *found_p = searchLocations("TagData Locations", tagLocations, SEARCH_IDX, pos1, pos2, &it);
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"search %ld-%ld returned %p %ld-%ld\n", pos1, pos2, found_p, found_p?found_p->startpos():-1, found_p?found_p->endpos():-1);
    dumpLocations("TagData Locations", tagLocations);
#endif
    if (found_p) {
        sptr_t start = found_p->startpos();
        sptr_t end = found_p->endpos();
        // found location may begin before pos1, if so, go to the next location
        if ((start < pos1) && (++it < tagLocations.end())) {
            found_p = *it;
            if (found_p->startpos() > pos2) {
                return;
            }
            start = found_p->startpos();
            end = found_p->endpos();
        }
        // check following locations, if contiguous with previous, table may be fragmented due to tag add/remove operations
        while (++it < tagLocations.end()) {
            found_p = *it;
            if (found_p->startpos() != end+1)
                break;
            end = found_p->endpos();
        }
        PosToIndex(interp, obj, start, indices);
        PosToIndex(interp, obj, end+1, indices);
    }
}

// implements tag prevrange command
void TagData::PrevRange(Tcl_Interp *interp, ScintillaObject *obj, sptr_t pos1, sptr_t pos2, Tcl_Obj* indices)
{
std::vector<TagLocation*>::iterator it;
    // search tag location vector for desired tag
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"PrevRange %ld-%ld tag %s\n", pos1, pos2, Name());
    dumpLocations("TagData Locations", tagLocations);
#endif
    TagLocation *found_p = searchLocations("TagData Locations", tagLocations, SEARCH_IDX, pos1-1, pos1-1, &it);
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"search %ld-%ld returned %p %ld-%ld\n", pos1, pos1, found_p, found_p?found_p->startpos():-1, found_p?found_p->endpos():-1);
    dumpLocations("TagData Locations", tagLocations);
#endif
    if (found_p) {
        sptr_t start = found_p->startpos();
        sptr_t end = found_p->endpos();

        // found a location that began before pos1, make sure it does not begin before pos2
        // check previous locations, if contiguous, table may be fragmented due to tag add/remove operations
        if (start < pos2) {
            // tag starts too low
            return;
        }
        while (--it >= tagLocations.begin()) {
            found_p = *it;
            if  (found_p->endpos() != start-1) {
                break;
            }
            if (found_p->startpos() < pos2) {
                // tag starts too low
                return;
            }
            start = found_p->startpos();
        }
        PosToIndex(interp, obj, start, indices);
        PosToIndex(interp, obj, end+1, indices);
    }
}

// implements tag ranges command
void TagData::Ranges(Tcl_Interp *interp, ScintillaObject *obj, Tcl_Obj* indices)
{
sptr_t start = -1, end = -1;
    // generate start/end ranges for tag, adjacent Locations may have the same tag applied due to fragmentation caused by add/remove tag operations
    // concatenate adjacent locations if needed to span the full length of a tag range
    for (std::vector<TagLocation*>::iterator it = tagLocations.begin(); it != tagLocations.end(); ++it) {
        TagLocation *t_p = *it;
        if (start == -1) {
            start = t_p->startpos();
            end = t_p->endpos();
        } else if (end == t_p->endpos() - 1) {
            // logically concatenate locations
            end = t_p->endpos();
        } else {
            PosToIndex(interp, obj, start, indices);
            PosToIndex(interp, obj, end, indices);
            start = -1;
        }
    }
    // flush out last range
    if (start != -1) {
        PosToIndex(interp, obj, start, indices);
        PosToIndex(interp, obj, end, indices);
    }
}

void TagData::RemoveTag(TagMgr *mgr, sptr_t start, sptr_t end)
{
std::vector<TagLocation*>::iterator it;

    // see if tag is applied within start and end, remove it. If removed and display needs update, do so.
    // If this was the last tag on a location, notify TagMgr to remove the location.

    // search tag location vector for desired tag
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"RemoveTag %ld-%ld tag %s\n", start, end, Name());
    dumpLocations("TagData Locations", tagLocations);
#endif
    TagLocation *tl_p = searchLocations("TagData Locations", tagLocations, SEARCH_IDX, start, end, &it);
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"search %ld-%ld returned %p %ld-%ld\n", start, end, tl_p, tl_p?tl_p->startpos():-1, tl_p?tl_p->endpos():-1);
    dumpLocations("TagData Locations", tagLocations);
#endif
    while (tl_p) {
        TagData *newTd = tl_p->UnregisterTag(this);
        // if we're in the process of deleting this tag, don't update the tagLocations vector or TagData::DeleteTag will have
        // invalid iterator issues
        if (!(flags & FLAG_DELETING))
            tagLocations.erase(it);
        if (newTd) {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr, "   applying style for %s from %ld to %ld\n", newTd->name, tl_p->startpos(), tl_p->endpos());
#endif
            if (this->Type() & TT_INDIC) {
                // if we're replacing an indicator you have to remove the old value because they aren't mutually exclusive
                mgr->RemoveStyle(this, start, end);
            }
            mgr->ApplyStyle(newTd, tl_p->startpos(), tl_p->endpos());
        } else  if (tl_p->empty()) {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"      no more tags for location, setting to default\n");
#endif
            // remove all style, the tag location is empty, remove tag location too
            mgr->RemoveStyle(this, start, end);
            mgr->RemoveLocation(tl_p);
        }
        tl_p = nextLocation(tagLocations, start, end, &it);
    }
}

void TagData::DumpLocations()
{
char buf[2048];
    sprintf(buf, "%s TagData Locations:", Name());
    dumpLocations(buf, tagLocations);
}

void TagData::DeleteTag(TagMgr *mgr)
{
    flags |= FLAG_DELETING;
    // remove tag from tag manager's location list
    for (std::vector<TagLocation*>::iterator it = tagLocations.begin(); it != tagLocations.end(); ++it) {
        TagLocation *t_p = *it;
        mgr->RemoveTag(this, t_p->startpos(), t_p->endpos());
    }
}

/////////////////////////////////////////////////////////////////////////////////
// Implementation of TagMgr class
//
/////////////////////////////////////////////////////////////////////////////////

//### CTOR ###
TagMgr::TagMgr(SciWidget *w) :
   TAG_SEL("sel"),
   TAG_MARGIN("margin"),
   sciwidgetPtr(w),
   interp(w->interp),
   tkwin(w->tkwin),
   selTag(0),
   marginTag(0),
   numTags(0),
   numCurTags(0),
   curTagArrayPtr(0)
{
    // create the hash table
    Tcl_InitHashTable(&tagTable, TCL_STRING_KEYS);

    resetFreeIndicators();
    resetFreeTextStyles();
    resetFreeMarginStyles();
    resetFreeAnnotationStyles();

    pickEvent.type = LeaveNotify; // match initialization done in Tk Text widget code

    // allocate tag space for margins and annotations, maximum of 256 each
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    marginStyleOffset = sci_cmd(obj, SCI_ALLOCATEEXTENDEDSTYLES, 256, 0);
    sci_cmd(SCIOBJ(w), SCI_MARGINSETSTYLEOFFSET, marginStyleOffset);
    annotationStyleOffset = sci_cmd(obj, SCI_ALLOCATEEXTENDEDSTYLES, 256, 0);
    sci_cmd(SCIOBJ(w), SCI_ANNOTATIONSETSTYLEOFFSET, annotationStyleOffset);

    // create "special" built-in tags
    _createTag(TT_TEXT, TAG_SEL);
    _createTag(TT_MARG, TAG_MARGIN);
}

//### DTOR ###
TagMgr::~TagMgr() {

    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    TagData       *tagPtr;

    // first iterate through the objects in the hash table...
    for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
            hPtr != NULL; 
            hPtr = Tcl_NextHashEntry(&search)) {
        tagPtr = (TagData *) Tcl_GetHashValue(hPtr);

        _deleteTag(tagPtr, true/*ok to destroy built-in tags*/);
    }

    // then delete the has table itself
    Tcl_DeleteHashTable(&tagTable);

    // delete tagLocation vector
    for (std::vector<TagLocation*>::iterator it = tagLocations.begin(); it != tagLocations.end(); ++it) {
        delete *it;
    }

    // free tag space for margins and annotations
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    sci_cmd(obj, SCI_RELEASEALLEXTENDEDSTYLES, 0, 0);
}

void TagMgr::resetFreeTextStyles() {
    freeTextStyles.reset(); /* clear all bits */
    /* mark pre-allocated text styles */
    freeTextStyles.set(0); /* zero is the default, for unset style */
    freeTextStyles.set(32);
    freeTextStyles.set(33);
    freeTextStyles.set(34);
    freeTextStyles.set(35);
    freeTextStyles.set(36);
    freeTextStyles.set(37);
    freeTextStyles.set(38);
    freeTextStyles.set(39);
}

void TagMgr::resetFreeMarginStyles() {
    freeMarginStyles.reset(); /* clear all bits */
    freeMarginStyles.set(0); /* zero is the default, for unset style */
}

void TagMgr::resetFreeAnnotationStyles() {
    freeAnnotationStyles.reset(); /* clear all bits */
    freeAnnotationStyles.set(0); /* zero is the default, for unset style */
}

void TagMgr::resetFreeIndicators() {
   freeIndicators = 0xfffffff0; // first eight indicators are preallocated by Scinilla, 8-31 are free
}

/*
 *--------------------------------------------------------------
 *
 * PickCurrent --
 *
 *   Find the character containing the coordinates in an event
 *   and place the "current" mark on that character. If the 
 *   "current" mark has moved then generate a fake leave event
 *   on the old current character and a fake enter event on
 *   the new current character.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   The current mark for the widget may change.
 *
 * NOTE: This function mimics the implementation done in the
 *       Tk Text widget (see TkTextPickCurrent in tkTextTag.c).
 *       In other words, the code below borrows _HEAVILY_ from
 *       the Tk core.
 *--------------------------------------------------------------
 */
void TagMgr::PickCurrent(XEvent *eventPtr) {

    TagData **oldArrayPtr, **newArrayPtr;
    TagData **copyArrayPtr = NULL; // Initialization needed to prevent compiler warning
    int numOldTags, numNewTags;
    int pos=0, size;
    bool nearby = true; // assume we're not over any tags
    XEvent event;
    /*
     * If a button is down, then don't do anything at all; we'll be called
     * again when all buttons are up, and we can repick then. This implements
     * a form of mouse grabbing.
     */
    int et   = eventPtr->type;
    int mode = eventPtr->xcrossing.mode;

    if (sciwidgetPtr->flags & BUTTON_DOWN) {
        if (((et == EnterNotify) || (et == LeaveNotify))
            && ((mode == NotifyGrab) || (mode == NotifyUngrab))) {
            /*
             * Special case: the window is being entered or left because of a
             * grab or ungrab. In this case, repick after all. Furthermore,
             * clear BUTTON_DOWN to release the simulated grab.
             */
            sciwidgetPtr->flags &= ~BUTTON_DOWN;
        } else {
            return;
        }
    }
    /*
     * Save information about this event in the widget in case we have to
     * synthesize more enter and leave events later (e.g. because a character
     * was deleted, causing a new character to be underneath the mouse
     * cursor). Also translate MotionNotify events into EnterNotify events,
     * since that's what gets reported to event handlers when the current
     * character changes.
     */
    if (eventPtr != &pickEvent) {
        if ((et == MotionNotify) || (et == ButtonRelease)) {
            pickEvent.xcrossing.type        = EnterNotify;
            pickEvent.xcrossing.serial      = eventPtr->xmotion.serial;
            pickEvent.xcrossing.send_event  = eventPtr->xmotion.send_event;
            pickEvent.xcrossing.display     = eventPtr->xmotion.display;
            pickEvent.xcrossing.window      = eventPtr->xmotion.window;
            pickEvent.xcrossing.root        = eventPtr->xmotion.root;
            pickEvent.xcrossing.subwindow   = None;
            pickEvent.xcrossing.time        = eventPtr->xmotion.time;
            pickEvent.xcrossing.x           = eventPtr->xmotion.x;
            pickEvent.xcrossing.y           = eventPtr->xmotion.y;
            pickEvent.xcrossing.x_root      = eventPtr->xmotion.x_root;
            pickEvent.xcrossing.y_root      = eventPtr->xmotion.y_root;
            pickEvent.xcrossing.mode        = NotifyNormal;
            pickEvent.xcrossing.detail      = NotifyNonlinear;
            pickEvent.xcrossing.same_screen = eventPtr->xmotion.same_screen;
            pickEvent.xcrossing.focus       = False;
            pickEvent.xcrossing.state       = eventPtr->xmotion.state;

        } else  {
            pickEvent = *eventPtr;
        }
    }
    /*
     * Find the new current character, then find and sort all of the tags
     * associated with it.
     */
    if (pickEvent.type != LeaveNotify) {
        ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

        int x = pickEvent.xcrossing.x;
        int y = pickEvent.xcrossing.y;
        int _pos = sci_get_pos_from_xy(obj, x, y, true);
        if (_pos == -1) {
            _pos = sci_get_pos_from_xy(obj, x, y, false);
        } else {
            nearby = false; // << _NOT_ nearby, actually over
        }
        pos = _pos;
        if (nearby) {
            newArrayPtr = NULL;
            numNewTags = 0;
        } else {
            newArrayPtr = _getTagsAtPos(pos, &numNewTags);
            SortTags(numNewTags, newArrayPtr);
        }
    } else {
        newArrayPtr = NULL;
        numNewTags = 0;
    }
    /*
     * Resort the tags associated with the previous marked character (the
     * priorities might have changed), then make a copy of the new tags, and
     * compare the old tags to the copy, nullifying any tags that are present
     * in both groups (i.e. the tags that haven't changed).
     */
    SortTags(numCurTags, curTagArrayPtr);

    if (numNewTags > 0) {
        size = numNewTags * sizeof(TagData *);
        copyArrayPtr = (TagData **) Tcl_Alloc((unsigned) size);
        memcpy((void *) copyArrayPtr, (void *) newArrayPtr, (size_t) size);
        for (int i = 0; i < numCurTags; i++) {
            for (int j = 0; j < numNewTags; j++) {
                if (curTagArrayPtr[i] == copyArrayPtr[j]) {
                    curTagArrayPtr[i] = NULL;
                    copyArrayPtr[j] = NULL;
                    break;
                }
            }
        }
    }
    /*
     * Invoke the binding system with a LeaveNotify event for all of the tags
     * that have gone away. We have to be careful here, because it's possible
     * that the binding could do something (like calling tkwait) that
     * eventually modifies curTagArrayPtr. To avoid problems in situations
     * like this, update curTagArrayPtr to its new value before invoking any
     * bindings, and don't use it any more here.
     */
    numOldTags     = numCurTags;
    numCurTags     = numNewTags;
    oldArrayPtr    = curTagArrayPtr;
    curTagArrayPtr = newArrayPtr;
    if (numOldTags != 0) {
        if ((sciwidgetPtr->bindingTable != NULL) && (tkwin != NULL)) {

            event = pickEvent;
            event.type = LeaveNotify;
            /*
             * Always use a detail of NotifyAncestor. Besides being
             * consistent, this avoids problems where the binding code will
             * discard NotifyInferior events.
             */
            event.xcrossing.detail = NotifyAncestor;
            BindEvent(&event, numOldTags, oldArrayPtr);
        }
        Tcl_Free((char *) oldArrayPtr);
    }
    /*
     * Reset the "current" mark (be careful to recompute its location, since
     * it might have changed during an event binding). Then invoke the binding
     * system with an EnterNotify event for all of the tags that have just
     * appeared.
     */
    sciwidgetPtr->markMgr->AddMark("current", pos);

    if (numNewTags != 0) {
        if ((sciwidgetPtr->bindingTable != NULL)
            && (tkwin != NULL)
            && !(sciwidgetPtr->flags & DESTROYED) && !nearby) {
            event = pickEvent;
            event.type = EnterNotify;
            event.xcrossing.detail = NotifyAncestor;

            BindEvent(&event, numNewTags, copyArrayPtr);
        }
        Tcl_Free((char *) copyArrayPtr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * BindEvent --
 *
 *   Trigger given events for all tags that match the relevant
 *   bindings. To handle the "sel" tag correctly, we must use
 *   the name of the tags as the binding table element.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   Almost anything can be triggered by tag bindings, 
 *   including deletion of the text widget.
 *
 *--------------------------------------------------------------
 */
void TagMgr::BindEvent(XEvent *eventPtr, int nTags, TagData **tagArrayPtr) {

    if (nTags == 0 || 
        sciwidgetPtr->bindingTable == NULL) return;

    // NOTE: We use the same strategy of avoiding heap allocation
    //       as is found in TagBindEvent (tkTextTag.c)
    #define NUM_BIND_TAGS 10
    const char *nameArray[NUM_BIND_TAGS];
    const char **nameArrPtr;
    int i;
    /*
     * Try to avoid allocation unless there are lots of tags.
     */
    if (nTags > NUM_BIND_TAGS) {
        nameArrPtr = (const char**) Tcl_Alloc(nTags * sizeof(const char*));
    } else {
        nameArrPtr = nameArray;
    }
    /*
     * We use tag names as keys in the hash table to remain consistent
     * with how the Tk Text widget handles this
     */
    for (i = 0; i < nTags; i++) {
        TagData *tagPtr = tagArrayPtr[i];
        if (tagPtr != NULL) {
            nameArrPtr[i] = tagPtr->Name();
        } else {
            /*
             * Tag has been deleted elsewhere, and therefore nulled out in
             * this array. Tk_BindEvent is clever enough to cope with NULLs
             * being thrown at it.
             */
            nameArrPtr[i] = NULL;
        }
    }
    Tk_BindEvent(sciwidgetPtr->bindingTable, eventPtr,
        tkwin, nTags, (ClientData *) nameArrPtr);

    if (nTags > NUM_BIND_TAGS) {
        Tcl_Free((char*)nameArrPtr);
    }
}
// simple API, called by widget
void TagMgr::HandleBindEvent(XEvent *eventPtr) {

    BindEvent(eventPtr, numCurTags, curTagArrayPtr);
}

// if command == NULL, dump tags from pos1 to pos2 into dstr
// mask is a bit mask indicating what to dump, 1 == tag, 2 == style, 4 == indicator
// else call command with tag information as per Tk text widget description
void TagMgr::DumpText(int mask, Tcl_DString *dstr, int pos1, int pos2, const char *command)
{
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    Tcl_DString cmdStr;
    if (command) {
        Tcl_DStringInit(&cmdStr);
        Tcl_DStringAppendElement(&cmdStr, command);
        dstr = &cmdStr;
    }

    int i, ntags, isNew;
    TagData *tag, **tags;
    // create hash table of active tags in dump region
    Tcl_HashEntry *hPtr;
    Tcl_HashTable dumpTable;
    Tcl_HashSearch search;
    Tcl_InitHashTable(&dumpTable, TCL_STRING_KEYS);
    int last_sid = 0, last_indicator = 0;
    for (i = pos1; i < pos2; ++i) {
        // tags are only supported in text, not margins
        if (mask & 1) {
            tags = _getTagsAtPos(i, &ntags);
            for (int j = 0; j < ntags; ++j) {
                tag = tags[j];
                hPtr = Tcl_CreateHashEntry(&dumpTable, tag->name, &isNew);
                if (isNew) {
                    // newly found tag
                    Tcl_DStringAppendElement(dstr, "tagon");
                    Tcl_DStringAppendElement(dstr, tag->name);
                    PosToIndex(NULL, sciobj, i, NULL, dstr);
                    if (command) {
                        Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                        Tcl_DStringSetLength(dstr, 0);
                        Tcl_DStringAppendElement(dstr, command);
                    }
                }
            }
            // now go through dumpTable, if we find an entry that's not in the tag list, tags, then
            // issue a tagoff statement
            for (hPtr = Tcl_FirstHashEntry(&dumpTable, &search);
                hPtr != NULL; 
                hPtr = Tcl_NextHashEntry(&search)) {
                char *name = Tcl_GetHashKey(&dumpTable, hPtr);
                int found = 0;
                for (int j = 0; j < ntags; ++j) {
                    tag = tags[j];
                    if (!strcmp(name, tag->name)) {
                        found = 1;
                        break;
                    }
                }
                if (!found) {
                    // issue tagoff command, remove tag from table
                    Tcl_DStringAppendElement(dstr, "tagoff");
                    Tcl_DStringAppendElement(dstr, name);
                    PosToIndex(NULL, sciobj, i, NULL, dstr);
                    if (command) {
                        Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                        Tcl_DStringSetLength(dstr, 0);
                        Tcl_DStringAppendElement(dstr, command);
                    }
                    Tcl_DeleteHashEntry(hPtr);
                }
            }
        }
        if (mask & 6) {
            // done with user-defined tags, dump lexer-related styles next
            unsigned char full_sid = sci_cmd(sciobj, SCI_GETSTYLEAT, i);
            // full_sid may have indicator bits encoded within, strip them appart
            if (full_sid) {
                char sbuf[10];
                int sid = SID2STYLE(sciwidgetPtr, full_sid);
                int indicator = SID2INDICATOR(sciwidgetPtr, full_sid);
                if (sid) {
                    if (last_sid && (last_sid != sid)) {
                        sprintf(sbuf,"%d", last_sid);
                        Tcl_DStringAppendElement(dstr, "styleoff");
                        Tcl_DStringAppendElement(dstr, sbuf);
                        PosToIndex(NULL, sciobj, i, NULL, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                    }
                    if (last_sid != sid) {
                        // new style
                        sprintf(sbuf,"%d", sid);
                        Tcl_DStringAppendElement(dstr, "styleon");
                        Tcl_DStringAppendElement(dstr, sbuf);
                        PosToIndex(NULL, sciobj, i, NULL, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                        last_sid = sid;
                    }
                }
                if (indicator) {
                    if (last_indicator && (last_indicator != indicator)) {
                        Tcl_DStringAppendElement(dstr, "indicatoroff");
                        Tcl_DStringAppendElement(dstr, IndicatorName(last_indicator));
                        PosToIndex(NULL, sciobj, i, NULL, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                    }
                    if (last_indicator != indicator) {
                        // new indicator
                        Tcl_DStringAppendElement(dstr, "indicatoron");
                        Tcl_DStringAppendElement(dstr, IndicatorName(indicator));
                        PosToIndex(NULL, sciobj, i, NULL, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                        last_indicator = indicator;
                    }
                }
            }
        }
    }
    if (command) {
        Tcl_DStringFree(&cmdStr);
    }
    Tcl_DeleteHashTable(&dumpTable);
}

// if command == NULL, dump tags from line1,col1 to line2,col2 to dstr
// mask is a bit mask indicating what to dump, 1 == tag, 2 == style, 4 == indicator
// else call command with tag information as per Tk text widget description
void TagMgr::DumpMargin(int mask, Tcl_DString *dstr, int line1, int col1, int line2, int col2, const char *command)
{
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    Tcl_DString cmdStr;
    if (command) {
        Tcl_DStringInit(&cmdStr);
        Tcl_DStringAppendElement(&cmdStr, command);
        dstr = &cmdStr;
    }
    int len, offset = 0, textLen;
    int idx1 = col1, idx2;
    TagData *tagPtr;
    if (mask&1) {
        // get style id offset for margin styles, used for tag lookup
        offset = sci_cmd(sciobj, SCI_MARGINGETSTYLEOFFSET);
    }
    while (line1 <= line2) {
        // calling with NULL pointer gets the length of margin text string
        len = sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line1), 0);
        if (len == 0) {
            textLen = sci_cmd(sciobj, SCI_MARGINGETTEXT, LINE2SCI(line1), 0);
            // style/indicator encoding for entire line
            if (line1 < line2) {
                // set idx2 to end of current line if we haven't reached the end line yet
                idx2 = textLen;
            } else {
                // else set end position of last line
                idx2 = col2;
            }
            // no style tags, return style/tag for entire line
            int full_sid = sci_cmd(sciobj, SCI_MARGINGETSTYLE, LINE2SCI(line1));
            int sid = SID2STYLE(sciwidgetPtr, full_sid);
            int indicator = SID2INDICATOR(sciwidgetPtr, full_sid);
            if (full_sid) {
                if (sid && (mask & 1)) {
                    /* dump tag, if exists */
                    tagPtr = sciwidgetPtr->tagMgr->FindTag(sid+offset);
                    if (tagPtr && (tagPtr->Type() & TT_MARG)) {
                        Tcl_DStringAppendElement(dstr, "tagon");
                        Tcl_DStringAppendElement(dstr, tagPtr->name);
                        LineColumnToIndex(line1, idx1, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                        Tcl_DStringAppendElement(dstr, "tagoff");
                        Tcl_DStringAppendElement(dstr, tagPtr->name);
                        LineColumnToIndex(line1, idx2, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                    }
                }
                if (mask & 6) {
                    char sbuf[10];
                    if (sid) {
                        sprintf(sbuf,"%d", sid);
                        Tcl_DStringAppendElement(dstr, "styleon");
                        Tcl_DStringAppendElement(dstr, sbuf);
                        LineColumnToIndex(line1, idx1, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                        Tcl_DStringAppendElement(dstr, "styleoff");
                        Tcl_DStringAppendElement(dstr, sbuf);
                        LineColumnToIndex(line1, idx2, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                    }
                    if (indicator) {
                        Tcl_DStringAppendElement(dstr, "indicatoron");
                        Tcl_DStringAppendElement(dstr, IndicatorName(indicator));
                        LineColumnToIndex(line1, idx1, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                        Tcl_DStringAppendElement(dstr, "indicatoroff");
                        Tcl_DStringAppendElement(dstr, IndicatorName(indicator));
                        LineColumnToIndex(line1, idx2, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                    }
                }
            }
        } else {
            // style/indicator per character on line
            int last_sid = 0, last_indicator = 0;
            TagData *last_tagPtr = NULL;
            char *buf = new char[len+1];
            char sbuf[10];

            // get style codes for line
            sci_cmd(sciobj, SCI_MARGINGETSTYLES, LINE2SCI(line1), (sptr_t)buf);
            if (line1 < line2) {
                // set idx2 to end of current line if we haven't reached the end line yet
                idx2 = len;
            } else {
                // else set end position of last line
                idx2 = col2;
            }
            while (idx1 < idx2) {
                int full_sid = buf[idx1];
                if (full_sid) {
                    int sid = SID2STYLE(sciwidgetPtr, full_sid);
                    int indicator = SID2INDICATOR(sciwidgetPtr, full_sid);
                    if (sid && (mask & 1)) {
                        // dump tag, if exists
                        tagPtr = sciwidgetPtr->tagMgr->FindTag(sid+offset);
                        if (last_tagPtr && (last_tagPtr != tagPtr)) {
                            Tcl_DStringAppendElement(dstr, "tagoff");
                            Tcl_DStringAppendElement(dstr, last_tagPtr->name);
                            LineColumnToIndex(line1, idx1, dstr);
                            if (command) {
                                Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                Tcl_DStringSetLength(dstr, 0);
                                Tcl_DStringAppendElement(dstr, command);
                            }
                        }
                        if (tagPtr && (tagPtr->Type() & TT_MARG)) {
                            if (last_tagPtr != tagPtr) {
                                // newly found tag
                                Tcl_DStringAppendElement(dstr, "tagon");
                                Tcl_DStringAppendElement(dstr, tagPtr->name);
                                LineColumnToIndex(line1, idx1, dstr);
                                if (command) {
                                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                    Tcl_DStringSetLength(dstr, 0);
                                    Tcl_DStringAppendElement(dstr, command);
                                }
                            }
                            // set sid to zero to prevent styleon/styleoff for tags
                            sid = 0;
                        }
                        if (last_tagPtr != tagPtr) {
                            last_tagPtr = tagPtr;
                        }
                    }
                    if (mask & 6) {
                        if (sid) {
                            if (last_sid && (last_sid != sid)) {
                                sprintf(sbuf,"%d", last_sid);
                                Tcl_DStringAppendElement(dstr, "styleoff");
                                Tcl_DStringAppendElement(dstr, sbuf);
                                LineColumnToIndex(line1, idx1, dstr);
                                if (command) {
                                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                    Tcl_DStringSetLength(dstr, 0);
                                    Tcl_DStringAppendElement(dstr, command);
                                }
                            }
                            if (last_sid != sid) {
                                // new style
                                sprintf(sbuf,"%d", sid);
                                Tcl_DStringAppendElement(dstr, "styleon");
                                Tcl_DStringAppendElement(dstr, sbuf);
                                LineColumnToIndex(line1, idx1, dstr);
                                if (command) {
                                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                    Tcl_DStringSetLength(dstr, 0);
                                    Tcl_DStringAppendElement(dstr, command);
                                }
                            }
                        }
                        if (indicator) {
                            if (last_indicator && (last_indicator != indicator)) {
                                Tcl_DStringAppendElement(dstr, "indicatoroff");
                                Tcl_DStringAppendElement(dstr, IndicatorName(last_indicator));
                                LineColumnToIndex(line1, idx1, dstr);
                                if (command) {
                                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                    Tcl_DStringSetLength(dstr, 0);
                                    Tcl_DStringAppendElement(dstr, command);
                                }
                            }
                            if (last_indicator != indicator) {
                                // new indicator
                                Tcl_DStringAppendElement(dstr, "indicatoron");
                                Tcl_DStringAppendElement(dstr, IndicatorName(indicator));
                                LineColumnToIndex(line1, idx1, dstr);
                                if (command) {
                                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                                    Tcl_DStringSetLength(dstr, 0);
                                    Tcl_DStringAppendElement(dstr, command);
                                }
                                last_indicator = indicator;
                            }
                        }
                    }
                    if (last_sid != sid) {
                        last_sid = sid;
                    }
                }
                ++idx1;
            }
            if (last_tagPtr) {
                Tcl_DStringAppendElement(dstr, "tagoff");
                Tcl_DStringAppendElement(dstr, last_tagPtr->name);
                LineColumnToIndex(line1, idx1, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
            }
            if (last_sid) {
                // "finish" style at end of line
                sprintf(sbuf,"%d", last_sid);
                Tcl_DStringAppendElement(dstr, "styleoff");
                Tcl_DStringAppendElement(dstr, sbuf);
                LineColumnToIndex(line1, idx2, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
            }
            if (last_indicator) {
                // "finish" indicator at end of line
                Tcl_DStringAppendElement(dstr, "indicatoroff");
                Tcl_DStringAppendElement(dstr, IndicatorName(last_indicator));
                LineColumnToIndex(line1, idx2, dstr);
                if (command) {
                    Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                    Tcl_DStringSetLength(dstr, 0);
                    Tcl_DStringAppendElement(dstr, command);
                }
            }
            delete[] buf;
        }
        ++line1;
        idx1 = 0;
    }
    if (command) {
        Tcl_DStringFree(&cmdStr);
    }
}

/*----------------------------------------------------------------------
 * InstanceCmd --
 *
 *    Implements the TAG sub-command for the widget
 *
 * Results:
 *    A standard Tcl result
 *
 * Side affects:
 *    The display may get updated to reflect changes to an existing tag
 *
 *----------------------------------------------------------------------
 */
int TagMgr::InstanceCmd(int objc, Tcl_Obj *CONST objv[]) {

   int index;
   int result = TCL_OK;

   static CONST char *optionStrings[] = {
#ifdef DEBUG
      "dump",
#endif
      "add", "bind", "cget", "configure", "delete", "lower",
      "names", "nextrange", "prevrange", "raise", "ranges",
      "remove", NULL
   };
   enum options {
#ifdef DEBUG
CMD_DUMP,
#endif
      CMD_ADD,    CMD_BIND,   CMD_CGET,  CMD_CONFIGURE, 
      CMD_DELETE, CMD_LOWER,  CMD_NAMES, CMD_NEXTRANGE, CMD_PREVRANGE,
      CMD_RAISE,  CMD_RANGES, CMD_REMOVE
   };

   // at this point, we know we've got "path tag"
   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option ?arg arg ...?");
      return TCL_ERROR;
   }
   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings,
         "tag option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }

   switch ((enum options)index) {
#ifdef DEBUG
      case CMD_DUMP: {
         TagData *tagPtr = NULL; // this will dump all tags
         if (objc == 4)
            tagPtr = FindTag(Tcl_GetString(objv[3]));
         _dump(tagPtr);
         break;
      }
#endif
      case CMD_ADD: {
         result = CmdAdd(objc, objv);
         break;
      }
      case CMD_BIND: {
         result = CmdBind(objc, objv);
         break;
      }
      case CMD_CGET: {
         if (objc != 5) {
            Tcl_WrongNumArgs(interp, 3, objv, "tagName option");
            return TCL_ERROR;
         }
         TagData *tagPtr = FindTag(Tcl_GetString(objv[3]));
         if (tagPtr == NULL) {
            return TCL_ERROR;
         }
         Tcl_Obj *objPtr = Tk_GetOptionValue(interp, (char *)&tagPtr->options,
            tagPtr->optionTable, objv[4], tkwin);
         if (objPtr == NULL) {
            return TCL_ERROR;
         }
         Tcl_SetObjResult(interp, objPtr);
         break;
      }
      case CMD_CONFIGURE: {
         result = CmdConfig(objc, objv);
         break;
      }
      case CMD_DELETE: {
         result = CmdDelete(objc, objv);
         break;
      }
      case CMD_NAMES: {
         result = CmdNames(objc, objv);
         break;
      }
      case CMD_NEXTRANGE: {
         result = CmdNPRange(objc, objv, true);
         break;
      }
      case CMD_PREVRANGE: {
         result = CmdNPRange(objc, objv, false);
         break;
      }
      case CMD_RANGES: {
         result = CmdRanges(objc, objv);
         break;
      }
      case CMD_REMOVE: {
         result = CmdRemove(objc, objv);
         break;
      }
      case CMD_LOWER:
      case CMD_RAISE: {
          // The Scintilla widget doesn't have support for "raising" (or "lowering")
          // a tag's drawing priority. So for now we'll simply accept the client
          // request in order to minimize the need to change existing code that may
          // have been written using the Tk Text widget.
          break;
      }
   }
   return result;
}

//==========================================================================
// Implements the TAG ADD sub-command
//
// This function will associate a previously defined tag with a specified
// range of text, or if a tag doesn't already exist it will create a new
// one (this matches the Tk Text widget behavior).
//
int TagMgr::CmdAdd(int objc, Tcl_Obj *CONST objv[]) {

TagData *tagPtr;
char *index_str, *tagName;
sptr_t pos1 = -1;
sptr_t pos2 = -1;
int i;

    if (objc < 5) {
        Tcl_WrongNumArgs(interp, 3, objv,
            //ix: 3    4       5      6
            "tagName index1 ?index2 ?index1 index2 ...?");
        return TCL_ERROR;
    }
    tagName = Tcl_GetString(objv[3]);
    // first check if it's an existing tag...
    if ((tagPtr = FindTag(tagName)) == NULL) {
        // If not, let's mimic the Tk Text behavior where a tag will get 
        // created. And since we can only apply tags which are indicators,
        // that's the type we'll use.
        //
        Tcl_ResetResult(interp);
        int isNewTag;
        tagPtr = _createTag(TT_INDIC|TT_TEXT, tagName, &isNewTag);
        if (tagPtr == NULL) {
            return TCL_ERROR;
        }
        // Note that in order to support existing code that may have been converted
        // from use with the Tk Text widget, we default the indicator type to be a
        // a solid filled rectangle. An example usage would be:
        //    $widget tag add my_tag 2.5 2.10
        //    #...other code...
        //    $widget tag configure my_tag -background yellow
        //
        tagPtr->options.indicType = indicSquareBox;
        _configIndicator(tagPtr);
    }

#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"Tag %s, applying from ", tagPtr->name);
#endif
    for (i = 4; i < objc; ++i) {
        // decode first location
        index_str = Tcl_GetString(objv[i++]);
#ifdef DEBUG_TAGLOCATIONS
        fprintf(stderr,"%s to ", index_str);
#endif
        if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
            goto error;
        }
        if (pos1 == -1) {
            // position outside of document, ignore it
            continue;
        }
        if (i < objc) {
            index_str = Tcl_GetString(objv[i]);
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"%s\n", index_str);
#endif
            if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
                goto error;
            }
            if (pos2 == -1) {
                // position outside of document, ignore it
                continue;
            }
            --pos2; // character range ends at character before second index
        } else {
            pos2 = pos1;    // just tag one character
        }
        if ((pos1 > pos2) && (tagPtr != selTag)) {
            // according to the Tk documenation this is a no-op
        } else {
            ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
            if (tagPtr == selTag) {
                //-----------------
                //--- SELECTION ---
                //-----------------
                sci_select(obj, pos1, pos2+1, 0);
                if (!(sciwidgetPtr->flags & GOT_SELECTION)) {
                    Tk_OwnSelection(sciwidgetPtr->tkwin, XA_PRIMARY, ScintillaLostSelection, sciwidgetPtr);
                    sciwidgetPtr->flags |= GOT_SELECTION;
                }
                FireSelectionVirtual(); // indicate a change occurred
                // no need to register selection locations
            } else {
                ApplyTag(tagPtr, pos1, pos2);
            }
        }
    }
    return TCL_OK;
 error:
    return BadIndex(interp, index_str);
}

//==========================================================================
// Implements the TAG BIND sub-command
//
// This function manages the interface between client binding definitions
// for a tag and setting up the widget to respond to those events
//
int TagMgr::CmdBind(int objc, Tcl_Obj *CONST objv[]) {

    if ((objc < 4) || (objc > 6)) {
        Tcl_WrongNumArgs(interp, 3, objv, "tagName ?sequence? ?command?");
        return TCL_ERROR;
    }
    // check if the tag exists
    TagData *tagPtr = FindTag(Tcl_GetString(objv[3]));
    if (tagPtr == NULL) {
        return TCL_ERROR;
    }
    const char *tname = tagPtr->Name();

    // Make a binding table if the widget doesn't already have one.
    //
    if (sciwidgetPtr->bindingTable == NULL) {
        sciwidgetPtr->bindingTable = Tk_CreateBindingTable(interp);
    }
    Tk_BindingTable bt = sciwidgetPtr->bindingTable;
    if (objc == 6) {
        // define new command for a binding sequence
        
        int append = 0;
        unsigned long mask;
        char *bind  = Tcl_GetString(objv[4]); //<< BINDING
        char *fifth = Tcl_GetString(objv[5]); //<< COMMAND

        if (fifth[0] == 0) {
            // an empty command string means we remove the binding
            return Tk_DeleteBinding(interp, bt, (ClientData)tname, bind);
        }
        if (fifth[0] == '+') {
            // add this command to any existing binding
            fifth++;
            append = 1;
        }
        mask = Tk_CreateBinding(interp, bt, (ClientData)tname, bind, fifth, append);
        if (mask == 0) {
            return TCL_ERROR;
        }
        // NOTE: same mask values as Tk Text widget(see TkTextTagCmd in tkTextTag.c)

        if (mask & (unsigned) ~(ButtonMotionMask|Button1MotionMask
                |Button2MotionMask|Button3MotionMask|Button4MotionMask
                |Button5MotionMask|ButtonPressMask|ButtonReleaseMask
                |EnterWindowMask|LeaveWindowMask|KeyPressMask
                |KeyReleaseMask|PointerMotionMask|VirtualEventMask)) {

            Tk_DeleteBinding(interp, bt, (ClientData)tname, bind);
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "requested illegal events; ",
                "only key, button, motion, enter, leave, and virtual ",
                "events may be used", NULL);
            return TCL_ERROR;
        }
    } else if (objc == 5) {
        // return command for a specific binding sequence
        const char *command = Tk_GetBinding(interp,
            bt, (ClientData)tname, Tcl_GetString(objv[4]));

        if (!command) {
            const char *string = Tcl_GetStringResult(interp);
            /*
             * Ignore missing binding errors. This is a special hack that
             * relies on the error message returned by FindSequence in
             * tkBind.c.
             *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
             * NOTE: This code is taken directly from the TkTextTagCmd
             * function in tkTextTag.c
             *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
             */
            if (string[0] != '\0') {
                return TCL_ERROR;
            } else {
                Tcl_ResetResult(interp);
            }
        } else {
            Tcl_SetResult(interp, (char *) command, TCL_STATIC);
        }
    } else {
        Tk_GetAllBindings(interp, bt, (ClientData) tname);
    }
    return TCL_OK;
}

//==========================================================================
// Implements the TAG CONFIGURE sub-command
//
// This function will either make modifications to an existing tag
// or it will create a new one
//
int TagMgr::CmdConfig(int objc, Tcl_Obj *CONST objv[]) {

   TagData *tagPtr;
   int tt = TT_TEXT; // we default to text area
   int isNewTag;

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv,
         "tagName ?-margin? ?option? ?value? ?option value ...?");
      return TCL_ERROR;
   }
   // check for -margin option
   bool typeGiven = false;
   if (objc >= 5) {
      if (strcmp("-annotation", Tcl_GetString(objv[4])) == 0) {
         typeGiven = true;
         tt = TT_ANNO;
      } else if (strcmp("-margin", Tcl_GetString(objv[4])) == 0) {
         typeGiven = true;
         tt = TT_MARG;
      }
   }
   // This is kind of a hack, but we need to know when calling _createTag
   // if this is for an indicator since the identifiers for styles are
   // different than those for indicators
   for (int i=0; i<objc; i++) {
       if (strncmp(Tcl_GetString(objv[i]), "-in", 3) == 0) {
           tt |= TT_INDIC;
           if (tt & TT_MARG) {
              Tcl_AppendResult(interp, "Indicator tags not supported in margins", NULL);
              return TCL_ERROR;
           }
       }
   }

   if ( (typeGiven && (objc <= 6)) || (!typeGiven && (objc <= 5)) ) {
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // just return config info
      bool all = ( (typeGiven && (objc == 6)) || (!typeGiven && (objc == 5)) );
      int all_ix = 0;

      // look up tag, it's an error if tag does not exist
      tagPtr = FindTag(Tcl_GetString(objv[3]));
      if (tagPtr == NULL) {
          return TCL_ERROR;
      }
      if (all) {
         all_ix = typeGiven ? 5 : 4;
      }
      Tcl_Obj *objPtr = Tk_GetOptionInfo(
         interp, (char *) &tagPtr->options,
         tagPtr->optionTable,
         (all ? objv[all_ix] : NULL), tkwin);

      if (objPtr == NULL) {
         return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, objPtr);
      return TCL_OK;
   } else {
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // modify the config data
      Tk_SavedOptions savedOptions;
      int mask = 0;
      int adj = (typeGiven ? 5 : 4);

      tagPtr = _createTag(tt, Tcl_GetString(objv[3]), &isNewTag);
      if (tagPtr == NULL) {
          return TCL_ERROR;
      }
      if (Tk_SetOptions(interp, (char *)&tagPtr->options, tagPtr->optionTable,
            objc-adj, objv+adj, tkwin, &savedOptions, &mask) != TCL_OK) {
         deleteTag(Tcl_GetString(objv[3]));
         return TCL_ERROR;
      }
      ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

      if (tagPtr == selTag) {
         //-----------------
         //--- SELECTION ---
         //-----------------
         // selection style only allows setting foreground/background colors
         if ( (mask & ~(TOPT_FG|TOPT_BG)) != 0) {
            Tcl_AppendResult(interp, 
               "The \"sel\" tag only allows use of -background and/or -foreground options", (char *)NULL);
            return TCL_ERROR;
         }
         XColor *c = tagPtr->BG();
         if (c) {
            sci_cmd(sciobj, SCI_SETSELBACK, 1, _XRGB_(c));
         } else {
            sci_cmd(sciobj, SCI_SETSELBACK, 0, 0); // reset to using default
         }
         c = tagPtr->FG();
         if (c) {
            sci_cmd(sciobj, SCI_SETSELFORE, 1, _XRGB_(c));
         } else {
            sci_cmd(sciobj, SCI_SETSELFORE, 0, 0); // reset to using default
         }
         return TCL_OK;
      }
      // check if this is an indicator because they get setup separately
      if (tagPtr->IndicType() != indicNONE) {
          return _configIndicator(tagPtr);
      }
      int sid = tagPtr->StyleID();

      if (tagPtr == marginTag) {
         //--------------
         //--- MARGIN ---
         //--------------
         if (mask & TOPT_BG) {
            // UpdateMarginColor will sync w/ toplevel config value, so we make
            // sure it's updated with the new value _first_
            sciwidgetPtr->marginBg = tagPtr->BG();
            UpdateMarginColor(true, false);
         }
         if (mask & TOPT_FG) {
            sciwidgetPtr->marginFg = tagPtr->FG();
            UpdateMarginColor(false, true);
         }
      } else {
         // This is a user-defined(ie not a built-in) tag
         XColor *c;
         if (mask & TOPT_BG) {
            c = tagPtr->BG();
//printf("TAG CFG BG: RGB=%d,%d,%d  STYLE=%d\n",c->red,c->green,c->blue, tagPtr->StyleID());
            sci_set_style_bg(sciobj, sid, c);
         }
         if (mask & TOPT_FG) {
            c = tagPtr->FG();
//printf("TAG CFG FG: RGB=%d,%d,%d  STYLE=%d\n",c->red,c->green,c->blue, tagPtr->StyleID());
            sci_set_style_fg(sciobj, sid, c);
         }
      }

      if (mask & TOPT_BOLD) {
         sci_cmd(sciobj, SCI_STYLESETBOLD, sid, tagPtr->Bold());
         if ((tagPtr == marginTag) && (sid != STYLE_LINENUMBER)) {
            sci_cmd(sciobj, SCI_STYLESETBOLD, STYLE_LINENUMBER, tagPtr->Bold());
         }
      }
      if (tagPtr->EolFilled()) {
         sci_cmd(sciobj, SCI_STYLESETEOLFILLED, sid, tagPtr->EolFilled());
         // EolFilled is not supported in margins
      }
      if (mask & TOPT_ITALIC) {
         sci_cmd(sciobj, SCI_STYLESETITALIC, sid, tagPtr->Italic());
         if ((tagPtr == marginTag) && (sid != STYLE_LINENUMBER)) {
            sci_cmd(sciobj, SCI_STYLESETITALIC, STYLE_LINENUMBER, tagPtr->Italic());
         }
      }
      if (mask & TOPT_UNDERLINE) {
         sci_cmd(sciobj, SCI_STYLESETUNDERLINE, sid, tagPtr->Underline());
         if ((tagPtr == marginTag) && (sid != STYLE_LINENUMBER)) {
            sci_cmd(sciobj, SCI_STYLESETUNDERLINE, STYLE_LINENUMBER, tagPtr->Underline());
         }
      }
   }
   return TCL_OK;
}

//==========================================================================
// Implements the TAG DELETE sub-command
//
int TagMgr::CmdDelete(int objc, Tcl_Obj *CONST objv[]) {

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "tagName ?tagName ...?");
      return TCL_ERROR;
   }
   for (int i=3; i < objc; i++) {
      const char *name = Tcl_GetString(objv[i]);
      deleteTag(name);
   }
   return TCL_OK;
}

//==========================================================================
// Implements the TAG NAMES sub-command
//
// Returns the existing tag names
//
int TagMgr::CmdNames(int objc, Tcl_Obj *CONST objv[]) {

    if ((objc < 3) || (objc > 5)) {
        Tcl_WrongNumArgs(interp, 3, objv, "?index? ?tagName?");
        return TCL_ERROR;
    }
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    TagData       *tagPtr;
    Tcl_Obj *tmp = Tcl_NewListObj(0, NULL);
    if (objc == 3) {
        // show all existing tags in a sorted list

        // first we collect all of the names from the hash table; of
        // course these won't appear in sorted order.
        for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
             hPtr != NULL; 
             hPtr = Tcl_NextHashEntry(&search)) {
            tagPtr = (TagData *) Tcl_GetHashValue(hPtr);

            Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(tagPtr->Name(), -1));

        }
        // Then we add in our "special" tags
        Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(selTag->name, -1));
        Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(marginTag->name, -1));

        int len;
        if ( (Tcl_ListObjLength(interp, tmp, &len) == TCL_OK) && (len > 0)) {

            // To get the list sorted we take the lazy way out, and let the
            // Tcl "lsort" command handle the sorting
            //
            if (len > 1) {
                Tcl_Obj *cmd = Tcl_ObjPrintf("lsort {%s}", Tcl_GetStringFromObj(tmp,NULL));

                if (Tcl_Eval(interp, Tcl_GetStringFromObj(cmd,NULL)) != TCL_OK) {
                    fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(interp));
                    return TCL_ERROR;
                }
            } else {
                // just a single name, no need to sort
                Tcl_SetObjResult(interp, tmp);
            }
        }
    } else {
        // show only the tags at the specified location
        char *index_str = Tcl_GetString(objv[3]);
        char *tagName = (objc == 5 ? Tcl_GetString(objv[4]) : NULL);
        sptr_t pos = -1;
        if (GetPosition(sciwidgetPtr, index_str, &pos, false) != TCL_OK) {
            return BadIndex(interp, index_str);
        }
        for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
             hPtr != NULL; 
             hPtr = Tcl_NextHashEntry(&search)) {
            tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
            if (_isTagAtPos(tagPtr, pos, tagName)) {
                Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(tagPtr->Name(), -1));
            }
            Tcl_SetObjResult(interp, tmp);
        }
    }
    return TCL_OK;
}

//==========================================================================
// Implements the TAG NEXTRANGE/PREVRANGE sub-commands
//
int TagMgr::CmdNPRange(int objc, Tcl_Obj *CONST objv[], bool next) {

    char *index_str;
    sptr_t pos1, pos2;
    TagData *tagPtr;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    Tcl_Obj *indices = Tcl_NewListObj(0, NULL);

    if ((objc != 5) && (objc != 6)) {
        Tcl_WrongNumArgs(interp, 3, objv, "tagName index1 ?index2?");
        return TCL_ERROR;
    }
    tagPtr = FindTag(Tcl_GetString(objv[3]));
    if (tagPtr == NULL) {
        // The Tk Text widget doesn't treat this as an error
        Tcl_ResetResult(interp);
        return TCL_OK;
    }
    // figure out the search range
    index_str = Tcl_GetString(objv[4]);
    if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
        goto error;
    }
    if (pos1 == -1) goto done; // index is valid, but not within document
    if (objc == 6) {
        index_str = Tcl_GetString(objv[5]);
        if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
            goto error;
        }
        if (pos2 == -1) goto done;
        if (pos1 == pos2) goto done; // no-op
    } else {
        pos2 = (next ? sci_get_length(obj) : 0);
    }

    if (tagPtr == selTag) {
        // for code simplicity, we interpret the two indices to
        // represent the start and then the end of the target range.
        if (pos2 < pos1) {
            sptr_t tmp;
            tmp  = pos1;
            pos1 = pos2;
            pos2 = tmp;
        }
        //-----------------
        //--- SELECTION ---
        //-----------------
        if (sci_has_selection(obj)) {
            // need to handle both single and multiple selections
            int nsels = sci_cmd(obj,SCI_GETSELECTIONS);
            int dir = next ? 1 : -1;
            int first = next ? 0 : (nsels - 1);
            int last  = next ? nsels : -1;

            for (int sid = first; sid != last; sid += dir) {

                int beg = sci_cmd(obj, SCI_GETSELECTIONNSTART, sid);
                int end = sci_cmd(obj, SCI_GETSELECTIONNEND,   sid);

                // Note: "pos2" is interpretted as the position _after_ the
                //       last character in the target range. That's why the
                //       compare is done as "<" and not "<=".
                if (((beg >= pos1) && (beg < pos2)) || 
                    ((end >= pos1) && (end < pos2))) {

                    PosToIndex(interp, obj, beg, indices);
                    PosToIndex(interp, obj, end, indices);
                    break;
                }
            }
            Tcl_SetObjResult(interp, indices);
        }
    } else {
        // search registered locations for tag
        if (next)
            tagPtr->NextRange(interp, obj, pos1, pos2, indices);
        else
            tagPtr->PrevRange(interp, obj, pos1, pos2, indices);
        Tcl_SetObjResult(interp, indices);
    }
 done:
    return TCL_OK;
 error:
    return BadIndex(interp, index_str);
}

//==========================================================================
// Implements the TAG RANGES sub-command
//
int TagMgr::CmdRanges(int objc, Tcl_Obj *CONST objv[]) {

    TagData *tagPtr;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    if (objc != 4) {
        Tcl_WrongNumArgs(interp, 3, objv, "tagName");
        return TCL_ERROR;
    }
    tagPtr = FindTag(Tcl_GetString(objv[3]));
    if (tagPtr == NULL) {
        return TCL_ERROR;
    }
    Tcl_Obj *indices = Tcl_NewListObj(0, NULL);
    if (tagPtr == selTag) {
        //-----------------
        //--- SELECTION ---
        //-----------------
        if (sci_has_selection(obj)) {
            int nsels = sci_cmd(obj,SCI_GETSELECTIONS);
            // need to handle both single and multiple selections
            for (int i=0; i < nsels; i++) {
                int pos1 = sci_cmd(obj, SCI_GETSELECTIONNSTART, i);
                int pos2 = sci_cmd(obj, SCI_GETSELECTIONNEND, i);
                if (pos1 == pos2) continue;
                PosToIndex(interp, obj, pos1, indices);
                PosToIndex(interp, obj, pos2, indices);
            }
            Tcl_SetObjResult(interp, indices);
        }
    } else if (IsIndicator(tagPtr)) {
        //-----------------
        //--- INDICATOR ---
        //-----------------
        int id = tagPtr->IndicID();
        int len = sci_get_length(obj);
        for (int i=0; i < len; i++) {
            int beg = sci_cmd(obj, SCI_INDICATORSTART, id, i);
            int end = sci_cmd(obj, SCI_INDICATOREND,   id, beg);

            // check if the indicator exists at all within the document
            if ((beg == end) && (beg == 0)) break;
            i = end;

            // use the following call to know how to interpret the positions
            // we get from the above two calls; we want to verify that the
            // target indicator exists at the starting position.
            int mask = sci_cmd(obj, SCI_INDICATORALLONFOR, beg);

            if (mask & (1 << id)) {
                PosToIndex(interp, obj, beg, indices);
                PosToIndex(interp, obj, end, indices);
            } else {
                // when the indicator doesn't exist at the "beg" position
                // then we want the next search to begin at the "end"
                // position; so we need to decrement "i" since the next
                // iteration of the "for" loop will increment it first.
                i--;
            }
        }
        Tcl_SetObjResult(interp, indices);
    } else {
        //-----------------
        //--- STYLE ---
        //-----------------
        int i, id = tagPtr->StyleID();
        int beg = -1;
        int len = sci_get_length(obj);
        for (i=0; i < len; i++) {
            unsigned char style = sci_cmd(obj, SCI_GETSTYLEAT, i);
            if (style == id) {
                if (beg == -1) {
                    /* remember begining position */
                    beg = i;
                    PosToIndex(interp, obj, beg, indices);
                }
            } else {
                /* style is not what we looked for, see if we recorded a begining location */
                if (beg != -1) {
                    /* note end of style range */
                    PosToIndex(interp, obj, i, indices);
                    beg = -1;
                }
            }
        }
        if (beg != -1) {
            /* note end of style range */
            PosToIndex(interp, obj, i, indices);
        }
        Tcl_SetObjResult(interp, indices);
    }
    return TCL_OK;
}

//==========================================================================
// Implements the TAG REMOVE sub-command
//
// This function will remove the styling for a previously defined tag 
// from a specified range of text
//
int TagMgr::CmdRemove(int objc, Tcl_Obj *CONST objv[]) {
TagData *tagPtr;
char *index_str;
sptr_t pos1 = -1;
sptr_t pos2 = -1;
int i;

    if (objc < 5) {
        Tcl_WrongNumArgs(interp, 3, objv,
            "tagName index1 ?index2 index1 index2 ...?");
        return TCL_ERROR;
    }
    // check if the tag exists
    tagPtr = FindTag(Tcl_GetString(objv[3]));
    if (tagPtr == NULL) {
        // Tk Text widget doesn't issue an error for this, so we won't either
        Tcl_ResetResult(interp);
        return TCL_OK;
    }
    for (i = 4; i < objc; ++i) {
        // decode first location
        index_str = Tcl_GetString(objv[i++]);
        if (GetPosition(sciwidgetPtr, index_str, &pos1, false) != TCL_OK) {
            goto error;
        }
        if (pos1 == -1) {
            // position outside of document, ignore it
            continue;
        }
        if (i < objc) {
            index_str = Tcl_GetString(objv[i]);
            if (GetPosition(sciwidgetPtr, index_str, &pos2, false) != TCL_OK) {
                goto error;
            }
            if (pos2 == -1) {
                // position outside of document, ignore it
                continue;
            }
            --pos2; // character range ends at character before second index
        } else {
            pos2 = pos1;    // just process one character
        }
        if ((pos1 > pos2) && (tagPtr != selTag)) {
            // according to the Tk documenation this is a no-op
        } else {
            ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
            if (tagPtr == selTag) {
                //-----------------
                //--- SELECTION ---
                //-----------------
                int sel_beg = sci_cmd(obj, SCI_GETSELECTIONSTART);
                int sel_end = sci_cmd(obj, SCI_GETSELECTIONEND);
                if (sel_beg == sel_end) {
                    // there is no selection
                    goto done;
                }
                // check if the range to be removed overlaps with current selection
                if ((pos2 < sel_beg) || (pos1 > sel_end)) {
                    // no overlap
                    goto done;
                }
                if (pos1 <= sel_beg && pos2 >= sel_end) {
                    //CASE 1: range completely overlaps selection - remove it all
                    sci_cmd(obj, SCI_SETEMPTYSELECTION, sci_get_current_pos(obj));
                } else {
                    if (pos2 <= sel_end) {
                        //CASE 2: range overlaps selection start
                        sci_cmd(obj, SCI_SETSELECTIONSTART, pos2);
                    }
                    if (pos1 >= sel_beg) {
                        //CASE 3: range overlaps selection start
                        sci_cmd(obj, SCI_SETSELECTIONEND, pos1);
                    }
                }
                FireSelectionVirtual(); // indicate a change occurred
            } else {
                RemoveTag(tagPtr, pos1, pos2);
            }
        }
    }
 done:
   return TCL_OK;
 error:
   return BadIndex(interp, index_str);
}


//==========================================================================
// See if a tag with the specified name has been defined. If it is, a
// pointer to the tag structure is returned, else NULL is returned and
// an error is left in the interp's result.
//
TagData *TagMgr::FindTag(const char *tagname) {

   Tcl_HashEntry *hPtr;

   if (IS_SEL(tagname))    return selTag;
   if (IS_MARGIN(tagname)) return marginTag;

   hPtr = Tcl_FindHashEntry(&tagTable, tagname);
   if (hPtr) {
      return (TagData *)Tcl_GetHashValue(hPtr);
   }
   Tcl_AppendResult(interp, "there is no tag named \"", tagname, "\"", (char *)NULL);
   return NULL;
}

//==========================================================================
// Look for the tag with the specified style ID
//
TagData *TagMgr::FindTag(int style_id) {

   Tcl_HashEntry *hPtr;
   Tcl_HashSearch search;
   TagData *tagPtr;

   if (selTag->StyleID() == style_id)    return selTag;
   if (marginTag->StyleID() == style_id) return marginTag;

   for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
        hPtr != NULL; 
        hPtr = Tcl_NextHashEntry(&search)) {

      tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
      if (tagPtr->StyleID() == style_id) {
         return tagPtr;
      }
   }
   return NULL;
}

//==========================================================================
//
// Send an event that the selection changed. This is equivalent to:
// event generate $textWidget <<Selection>>
//
// NOTE: The code below is taken directly from the TkTextSelectionEvent
//       function tkText.c 
//
void TagMgr::FireSelectionVirtual() {

    XEvent e;
    memset((VOID *) &e, 0, sizeof(e));
    e.xany.type                  = VirtualEvent;
    e.xany.serial                = NextRequest(Tk_Display(tkwin));
    e.xany.send_event            = False;
    e.xany.window                = Tk_WindowId(tkwin);
    e.xany.display               = Tk_Display(tkwin);
    ((XVirtualEvent *) &e)->name = Tk_GetUid("Selection");
    Tk_HandleEvent(&e);
}

//==========================================================================
// Called when language set in widget and we need to reset all styles/indicators/tags
int TagMgr::Reset(Tcl_Interp *interp, int langType) {

    ColourDesired marginBG;
    ColourDesired marginFG;
    int type;
    TagData *tagPtr;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    // save bg/fg for default line number style because doing SCI_STYLECLEARALL
    // will reset it to the Scintilla default
    marginBG = ColourDesired(sci_cmd(obj, SCI_STYLEGETBACK, STYLE_LINENUMBER));
    marginFG = ColourDesired(sci_cmd(obj, SCI_STYLEGETFORE, STYLE_LINENUMBER));

    // clear all styling information
    SSM(SCI_CLEARDOCUMENTSTYLE, 0,0);
    SSM(SCI_STYLECLEARALL,0,0);

    // restore default line number style
    sci_cmd(obj, SCI_STYLESETBACK, STYLE_LINENUMBER,     marginBG.AsLong());
    sci_cmd(obj, SCI_STYLESETFORE, STYLE_LINENUMBER,     marginFG.AsLong());

    // re-allocate styles/indicators/tags
    for (hPtr = Tcl_FirstHashEntry(&tagTable, &search); hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
        tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
        type = tagPtr->Type();
        // indicators are unaffected by lexer change, only styles
        if (!IsIndicator(tagPtr)) {
            int sid = tagPtr->StyleID();
            if (tagPtr == selTag) {
                /* selection tag is special, you can only set foreground/background color */
                XColor *c = tagPtr->BG();
                if (c) {
                    sci_cmd(obj, SCI_SETSELBACK, 1, _XRGB_(c));
                } else {
                    sci_cmd(obj, SCI_SETSELBACK, 0, 0); // reset to using default
                }
                c = tagPtr->FG();
                if (c) {
                    sci_cmd(obj, SCI_SETSELFORE, 1, _XRGB_(c));
                } else {
                    sci_cmd(obj, SCI_SETSELFORE, 0, 0); // reset to using default
                }
                continue;
            } else if (tagPtr == marginTag) {
                /* margin tag is special, config value should already be set in widget */
                UpdateMarginColor(true, false);
                UpdateMarginColor(false, true);
                if (sid != STYLE_LINENUMBER) {
                    sci_cmd(obj, SCI_STYLESETBOLD, STYLE_LINENUMBER, tagPtr->Bold());
                    sci_cmd(obj, SCI_STYLESETITALIC, STYLE_LINENUMBER, tagPtr->Italic());
                    sci_cmd(obj, SCI_STYLESETUNDERLINE, STYLE_LINENUMBER, tagPtr->Underline());
                }
            } else {
                if ((tagPtr->Type() & TT_TEXT) && (langType != SCLEX_NULL)) {
                    /* setting language to something other than "none" */
                    Tcl_AppendResult(interp, "Tag \"", tagPtr->Name(), "\" invalid when language other than \"none\"", (char *)NULL);
                    return TCL_ERROR;
                }
                XColor *c = tagPtr->BG();
                if (c) {
                    sci_set_style_bg(obj, sid, c);
                }
                c = tagPtr->FG();
                if (c) {
                    sci_set_style_fg(obj, sid, c);
                }
                sci_cmd(obj, SCI_STYLESETBOLD, sid, tagPtr->Bold());
                sci_cmd(obj, SCI_STYLESETITALIC, sid, tagPtr->Italic());
                sci_cmd(obj, SCI_STYLESETUNDERLINE, sid, tagPtr->Underline());
            }
        }
    }
    return TCL_OK;
}

//==========================================================================
// Update the background/foreground colors for the margin tag
//
void TagMgr::UpdateMarginColor(bool bg, bool fg) {

   if (marginTag == NULL) return;

   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
   int sid = marginTag->StyleID();
   XColor *c;

   if (bg) {
      // keep the tag in sync w/ toplevel config option
      c = marginTag->options.bgColor = sciwidgetPtr->marginBg;

      sci_set_style_bg(sciobj, sid, c);
      if (sid != STYLE_LINENUMBER)
         sci_set_style_bg(sciobj, STYLE_LINENUMBER, c);
   }
   if (fg) {
      c = marginTag->options.fgColor = sciwidgetPtr->marginFg;

      sci_set_style_fg(sciobj, sid, c);
      if (sid != STYLE_LINENUMBER)
         sci_set_style_fg(sciobj, STYLE_LINENUMBER, c);
   }
   // now we need to iterate through all of the margin tags and check for
   // ones that have no explicit bg/fg values; for these we'll update them
   // so they correctly "inherit" from the margin
   TagData *tagPtr;
   Tcl_HashEntry *hPtr;
   Tcl_HashSearch search;
   for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
        hPtr != NULL; 
        hPtr = Tcl_NextHashEntry(&search)) {

      tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
      if (!(tagPtr->Type() & TT_MARG)) continue;

      if (bg && (tagPtr->BG() == NULL)) {
         sci_set_style_bg(sciobj, tagPtr->StyleID(), marginTag->BG());
      }
      if (fg && (tagPtr->FG() == NULL)) {
         sci_set_style_fg(sciobj, tagPtr->StyleID(), marginTag->FG());
      }
   }
}

//==========================================================================
// Verify the specified tag name is valid for the margin area
//
TagData *TagMgr::ValidateForMargin(const char *name) {

   TagData *tagPtr = FindTag(name);
   if (tagPtr == NULL) {
      // the previous call will leave an error message in the Tcl interp
      return NULL;
   }
   // if the tag isn't defined for the margin area, it's an error
   if (!(tagPtr->Type() & TT_MARG)) {
      Tcl_AppendResult(interp,
         "Tag \"", tagPtr->Name(), "\" has not been defined for the margin area", (char *)NULL);
      return NULL;
   }
   return tagPtr;
}

Tcl_Obj* TagMgr::GetMarginTags(Tcl_Interp *intrp)
{
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    TagData       *tagPtr;
    Tcl_Obj *tmp = Tcl_NewListObj(0, NULL);
    for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
        hPtr != NULL; 
        hPtr = Tcl_NextHashEntry(&search)) {
        tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
        if (tagPtr->Type() & TT_MARG) {
            Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(tagPtr->Name(), -1));
        }
    }
    // Then we add in our "special" margin tag
    Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(sciwidgetPtr->tagMgr->GetMarginTag()->name, -1));
    return tmp;
}

//---------------------------------------------------------
// Private functions
//---------------------------------------------------------

void TagMgr::deleteTag(const char* name) {
Tcl_HashEntry *hPtr;
TagData *tagPtr;

    if (IS_SEL(name)) return;    // 'sel' tag cannot be deleted by user
    if (IS_MARGIN(name)) return; // 'margin' tag cannot be deleted by user

    hPtr = Tcl_FindHashEntry(&tagTable, name);
    if (hPtr == NULL) return;

    tagPtr = (TagData *)Tcl_GetHashValue(hPtr);

    // now delete the tag itself
    _deleteTag(tagPtr);

    // and finally remove it from the hash table
    Tcl_DeleteHashEntry(hPtr);
}

//==========================================================================
// This handles configuration the indicator style info for a tag
int TagMgr::_configIndicator(TagData *tagPtr) {

    // sanity check
    if (!IsIndicator(tagPtr)) return TCL_ERROR;

    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    // assign the indicator style
    switch (tagPtr->IndicType()) {
        case indicNONE:
            break; // make compiler happy
        case indicPlain:
            _setIndicStyle(tagPtr, INDIC_PLAIN);
            break;
        case indicSquiggle:
            _setIndicStyle(tagPtr, INDIC_SQUIGGLE);
            break;
        case indicTt:
            _setIndicStyle(tagPtr, INDIC_TT);
            break;
        case indicDiag:
            _setIndicStyle(tagPtr, INDIC_DIAGONAL);
            break;
        case indicStrike:
            _setIndicStyle(tagPtr, INDIC_STRIKE);
            break;
        case indicHidden:
            _setIndicStyle(tagPtr, INDIC_HIDDEN);
            break;
        case indicBox:
            _setIndicStyle(tagPtr, INDIC_BOX);
            break;
        case indicRoundBox:
            _setIndicStyle(tagPtr, INDIC_ROUNDBOX);
#if USE_SOLID_INDICATORS
            // solid indicators, display under text so text isn't obscured
            sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
#else
            if (tagPtr->Alpha() != -1) {
                sci_cmd(obj, SCI_INDICSETALPHA, tagPtr->IndicID(), tagPtr->Alpha());
                if (tagPtr->Alpha() >= SC_ALPHA_OPAQUE) {
                    // need this, otherwise text will be obscured by opaqueness!
                    sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
                }
            }
            if (tagPtr->AlphaB() != -1)
                sci_cmd(obj, SCI_INDICSETOUTLINEALPHA, tagPtr->IndicID(), tagPtr->AlphaB());
#endif
            break;
        case indicSquareBox:
            _setIndicStyle(tagPtr, INDIC_STRAIGHTBOX);
#if USE_SOLID_INDICATORS
            // solid indicators, display under text so text isn't obscured
            sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
#else
            if (tagPtr->Alpha() != -1) {
                sci_cmd(obj, SCI_INDICSETALPHA, tagPtr->IndicID(), tagPtr->Alpha());
                if (tagPtr->Alpha() >= SC_ALPHA_OPAQUE) {
                    // need this, otherwise text will be obscured by opaqueness!
                    sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
                }
            }
            if (tagPtr->AlphaB() != -1)
                sci_cmd(obj, SCI_INDICSETOUTLINEALPHA, tagPtr->IndicID(), tagPtr->AlphaB());
#endif
            break;
        case indicDash:
            _setIndicStyle(tagPtr, INDIC_DASH);
            break;
        case indicDots:
            _setIndicStyle(tagPtr, INDIC_DOTS);
            break;
        case indicSquiggleLow:
            _setIndicStyle(tagPtr, INDIC_SQUIGGLELOW);
            break;
        case indicDotBox:
            _setIndicStyle(tagPtr, INDIC_DOTBOX);
#if USE_SOLID_INDICATORS
            // solid indicators, display under text so text isn't obscured
            sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
#else
            if (tagPtr->Alpha() != -1) {
                sci_cmd(obj, SCI_INDICSETALPHA, tagPtr->IndicID(), tagPtr->Alpha());
                if (tagPtr->Alpha() >= SC_ALPHA_OPAQUE) {
                    // need this, otherwise text will be obscured by opaqueness!
                    sci_cmd(obj, SCI_INDICSETUNDER, tagPtr->IndicID(), 1);
                }
            }
            if (tagPtr->AlphaB() != -1)
                sci_cmd(obj, SCI_INDICSETOUTLINEALPHA, tagPtr->IndicID(), tagPtr->AlphaB());
#endif
            break;
        case indicSquigglePixMap:
            _setIndicStyle(tagPtr, INDIC_SQUIGGLEPIXMAP);
            break;
    }
    // now assign colors
    if (tagPtr->FG()) {
        sci_cmd(obj, SCI_INDICSETFORE, tagPtr->IndicID(), _XRGB_(tagPtr->FG()));
    }
    return TCL_OK;
}

void TagMgr::_setIndicStyle(TagData *tagPtr, int id) {

    sci_cmd(SCIOBJ(sciwidgetPtr), SCI_INDICSETSTYLE, tagPtr->IndicID(), id);
}

static int spin = 1;
extern "C" void stopme() {
    while (spin);
}

//==========================================================================
// Creates a new tag or returns one matching the specfied name
//
TagData *TagMgr::_createTag(int ttMask, const char *tagname, int *newTag) {

   TagData *tagPtr;
   Tcl_HashEntry *hPtr = NULL;
   int isNew;
   const char *name;

   // check if it's one of the special tags
   if (IS_SEL(tagname)) {
      //--- SELECTION ---
      if (!(ttMask & TT_TEXT)) {
         Tcl_AppendResult(interp,
            "The \"", tagname, "\" tag can only be defined for the text area", (char *)NULL);
         return NULL;
      }
      if (selTag) {
         if (newTag != NULL) *newTag = 0;
         return selTag; // return the one we already have
      } else if (newTag) *newTag = 1;

      name = TAG_SEL;

   } else if (IS_MARGIN(tagname)) {
      //--- MARGIN ---
      if (!(ttMask & TT_MARG)) {
         Tcl_AppendResult(interp,
            "The \"", tagname, "\" tag can only be defined for the margin area", (char *)NULL);
         return NULL;
      }
      if (marginTag) {
         if (newTag != NULL) *newTag = 0;
         return marginTag; // return the one we already have
      } else if (newTag) *newTag = 1;

      name = TAG_MARGIN;

   } else {
      if (((ttMask & TT_TEXT|TT_INDIC) == TT_TEXT) && (sciwidgetPtr->langType != SCLEX_NULL)) {
          // you can't define a text tag when there's a lexer styling the text because the style tag will conflict with the lexer
          // indicators are ok
          Tcl_AppendResult(interp,
                "The \"", tagname, "\" tag can only be defined in the text area when language is \"none\"", (char *)NULL);
            return NULL;
      }
      // if not, create(if needed) in the hash table
      hPtr = Tcl_CreateHashEntry(&tagTable, tagname, &isNew);
      if (newTag) *newTag = isNew;
      // Make sure we have either an annotation, text or margin tag
      assert(((ttMask&(TT_MARG|TT_TEXT|TT_ANNO)) == TT_ANNO) || ((ttMask&(TT_MARG|TT_TEXT|TT_ANNO)) == TT_MARG) || ((ttMask&(TT_MARG|TT_TEXT|TT_ANNO)) == TT_TEXT));
      if (!isNew) {
         tagPtr = (TagData *)Tcl_GetHashValue(hPtr);
         // make sure tag type is the same
        if (((TT_MARG|TT_TEXT|TT_ANNO)&tagPtr->tagType) != ((TT_MARG|TT_TEXT|TT_ANNO)&ttMask)) {
            Tcl_AppendResult(interp,
                "The \"", tagname, "\" tag can only be defined for the annotation, margin, or text area, not more than one", (char *)NULL);
            return NULL;
        }
         return (TagData *)Tcl_GetHashValue(hPtr);
      }
      name = Tcl_GetHashKey(&tagTable, hPtr);
   }
   // No existing entry, so create a new one
   tagPtr = new TagData(interp, tagname, ttMask, numTags++);

   if (Tk_InitOptions(interp, (char *)&tagPtr->options, tagPtr->optionTable, tkwin) != TCL_OK) {
      _deleteTag(tagPtr);
      if (isNew)
        Tcl_DeleteHashEntry(hPtr);
      fprintf(stderr, "ERROR in Tk_InitOptions: %s\n", Tcl_GetStringResult(interp));
      return NULL;
   }

   if (IS_SEL(tagname)) {
      selTag = tagPtr;
      // set style info
      ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
      if (selTag->BG()) sci_cmd(obj, SCI_SETSELBACK, 1, _XRGB_(selTag->BG()));
      if (selTag->FG()) sci_cmd(obj, SCI_SETSELFORE, 1, _XRGB_(selTag->FG()));

   } else if (IS_MARGIN(tagname)) {
      marginTag = tagPtr;
      // assign it a Scintilla style id
      int id;
      if (_nextStyleID(&id, ttMask) != TCL_OK) {
          _deleteTag(tagPtr);
          if (isNew)
             Tcl_DeleteHashEntry(hPtr);
          Tcl_AppendResult(interp, "\nUnable to create tag ", tagname, (char*) NULL);
          return NULL;
      }
      tagPtr->style_id = id;
   } else {
      // assign it a Scintilla id
      int id;
      if (ttMask & TT_INDIC) {
          if (_nextIndicID(&id) != TCL_OK) {
              _deleteTag(tagPtr);
              if (isNew)
                 Tcl_DeleteHashEntry(hPtr);
              Tcl_AppendResult(interp, "\nUnable to create tag ", tagname, (char*) NULL);
              return NULL;
          }
          tagPtr->indic_id = id;
      } else {
          if (_nextStyleID(&id, ttMask) != TCL_OK) {
              _deleteTag(tagPtr);
              if (isNew)
                 Tcl_DeleteHashEntry(hPtr);
              Tcl_AppendResult(interp, "\nUnable to create tag ", tagname, (char*) NULL);
              return NULL;
          }
          tagPtr->style_id = id;
      }
      // store in hash table
      Tcl_SetHashValue(hPtr, tagPtr);
   }
   return tagPtr;
}

//==========================================================================
// Delete a tag
//

void TagMgr::_deleteTag(TagData *tagPtr, bool force) {

   if (tagPtr == NULL) return;

   // do cleanup of the options data
   Tk_FreeConfigOptions((char *)&tagPtr->options, tagPtr->optionTable, tkwin);

   // destroy the tag (possibly)
   bool is_special = ((tagPtr == selTag) || (tagPtr == marginTag));

   if (!is_special || force) {
       //
       // Make sure this tag isn't referenced from the 'current' tag array.
       //
       for (int i = 0; i < numCurTags; i++) {
           if (curTagArrayPtr[i] == tagPtr) {
               for (; i < numCurTags-1; i++) {
                   curTagArrayPtr[i] = curTagArrayPtr[i+1];
               }
               curTagArrayPtr[numCurTags-1] = NULL;
               numCurTags--;
               break;
           }
       }
       // remove tag from TagLocations, remove Location if no other tags associated with it
       tagPtr->DeleteTag(this);
       if (IsIndicator(tagPtr)) {
           if ((tagPtr->indic_id >= INDIC_CONTAINER) && (tagPtr->indic_id <= INDIC_MAX)) {
              // indicator is now "free"
              freeIndicators |= (1 << tagPtr->IndicID());
           }
       } else {
           if (tagPtr->Type() & TT_TEXT) {
              if (tagPtr->StyleID() > 0)
                 freeTextStyles.set(tagPtr->StyleID(), 0);
           }
           if (tagPtr->Type() & TT_MARG) {
              if (tagPtr->StyleID() > 0)
                 freeMarginStyles.set(tagPtr->StyleID() - GetMarginStyleOffset(), 0);
           }
           if (tagPtr->Type() & TT_ANNO) {
              if (tagPtr->StyleID() > 0)
                 freeAnnotationStyles.set(tagPtr->StyleID() - GetAnnotationStyleOffset(), 0);
           }
       }
       delete tagPtr;
       if (tagPtr == selTag) {
           selTag = NULL;
       } else if (tagPtr == marginTag) {
           marginTag = NULL;
       }
       numTags--;
   }
}

//==========================================================================
// Update the array used for maintaining the set of tags that exist at
// a specific location.
//
// The return value is a malloc-ed array containing pointers to
// information for each of the tags that is associated with the character
// at position "pos". The word at *numTags is filled in with the number of
// pointers in the array. It is up to the caller to free the array by 
// passing it to free. If there are no tags at the given character then a
// NULL pointer is returned and *numTagsPtr will be set to 0.
//
TagData **TagMgr::_getTagsAtPos(int pos, int *numTags) {

    // To find all of the tags at the requested position, we iterate 
    // through the list, skipping non-indicators since those are the only
    // Scintilla display objects that clients can use within the text area.
    // We do make a special case of including the selection tag since that
    // supports having bindings defined for it.

    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    TagData *tagPtr;
    TagData **tagPtrPtr = NULL;
    int ntags;

    std::vector<TagData *> tmp;

    for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
         hPtr != NULL; 
         hPtr = Tcl_NextHashEntry(&search)) {

        tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
        if (_isTagAtPos(tagPtr, pos)) {
            tmp.push_back(tagPtr);
        }
    }
    // add selection tag separately since it's not in the hash table
    if (_isTagAtPos(selTag, pos)) {
        tmp.push_back(selTag);
    }
    // update the storage array
    ntags = tmp.size();
    if (ntags > 0) {
        tagPtrPtr = (TagData **)Tcl_Alloc(ntags * sizeof(TagData *));        
    }
    // now fill in with the tags we found
    std::vector<TagData *>::const_iterator cii;
    int i=0;
    for (cii = tmp.begin(); cii != tmp.end(); cii++) {
        tagPtrPtr[i++] = *cii;
    }
    if (numTags) *numTags = ntags;

    return tagPtrPtr;
}

//==========================================================================
// Check if a tag exists at a specified position
//
bool TagMgr::_isTagAtPos(TagData *tagPtr, int pos, const char *tagName) {

    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    // If given a name to check, we do that first
    if (tagName && strcasecmp(tagName, tagPtr->Name())) {
        return false;
    }
    int mask = sci_cmd(obj, SCI_INDICATORALLONFOR, pos);
    if (mask == 0) {
        // easy case, no indicators at all
        return false;
    }
    if (tagPtr == selTag) {
        // handle selection tag special since it's not in the hash table
        int beg = sci_cmd(obj, SCI_GETSELECTIONSTART);
        int end = sci_cmd(obj, SCI_GETSELECTIONEND);
        return (
            (pos >= beg) && (pos <= end)
        );
    }
    // For now, the only non-lexer attributes that can be applied are
    // indicators - so that's what we'll check for
    if (!IsIndicator(tagPtr)) {
        return false;
    }
    if (mask & (1 << tagPtr->IndicID())) {
        return true;
    }
    return false;
}

//==========================================================================
// Return the next available Scintilla indicator id. This routine is
// responsible for managing a free-list so that IDs can be reused.
//
int TagMgr::_nextIndicID(int *idPtr) {
    int i;
    if (!idPtr) return TCL_ERROR;

    // find first bit in freeIndicators, no need to check first 8
    for (i = INDIC_CONTAINER; i <= INDIC_MAX; ++i) {
        int bitmask = 1 << i;
        if (freeIndicators & bitmask) {
            freeIndicators &= ~bitmask;
            *idPtr = i;
            return TCL_OK;
        }
    }
    // if all indicators allocated, report error
    if (i > INDIC_MAX) {
        char buf[33];
        sprintf(buf,"%d", (INDIC_MAX - INDIC_CONTAINER + 1));
        Tcl_AppendResult(interp,
            "cannot allocate new ID: max number of indicators(", buf, ") exceeded", (char *)NULL);
    }
    return TCL_ERROR;
}

//==========================================================================
// Return the next available Scintilla style id. This routine is responsible
// for managing a free-list so that IDs can be reused.
//
int TagMgr::_nextStyleID(int *idPtr, int ttMask) {
int max = (1 << sciwidgetPtr->styleBits) - 1;
char *type = "text";

    if (!idPtr) return TCL_ERROR;

    if (ttMask & TT_TEXT) {
        /* text style, allocate from max down to avoid lexer collisions */
        for (int id = max; id > 0; id--) {
            if (!freeTextStyles.test(id)) {
                /* found free style, mark it as allocated */
                freeTextStyles.set(id, 1);
                *idPtr = id;
                return TCL_OK;
            }
        }
    } else if (ttMask & TT_ANNO) {
        /* annotation style, this can be allocated starting at 0 because
         * the style ids won't collide with styles allocated by lexers
         */
        max = 255;
        type = "annotation";
        for (int id = 0; id < 255; id++) {
            if (!freeAnnotationStyles.test(id)) {
                /* found free style, mark it as allocated */
                freeAnnotationStyles.set(id, 1);
                *idPtr = id + GetAnnotationStyleOffset();
                return TCL_OK;
            }
        }
    } else if (ttMask & TT_MARG) {
        /* margin style, this can be allocated starting at 0 because
         * the style ids won't collide with styles allocated by lexers
         */
        max = 255;
        type = "margin";
        for (int id = 0; id < 255; id++) {
            if (!freeMarginStyles.test(id)) {
                /* found free style, mark it as allocated */
                freeMarginStyles.set(id, 1);
                *idPtr = id + GetMarginStyleOffset();
                return TCL_OK;
            }
        }
    }

    /* unable to allocate style, return error */
    {
    char buf[33];
        sprintf(buf,"%d", max);
        Tcl_AppendResult(interp, "cannot allocate new ", type, " ID: max number of styles(", buf, ") exceeded", (char *)NULL);
    }
    return TCL_ERROR;
}

//==========================================================================
void TagMgr::_dump(TagData *tagPtr) {

   if (tagPtr == NULL) {
      // dump all of the tags (ignore sorting for now)
      Tcl_HashEntry *hPtr;
      Tcl_HashSearch search;
      fprintf(stderr,    "--- DUMP all tags: --------------------------------\n");
      int n = 0;
      for (hPtr = Tcl_FirstHashEntry(&tagTable, &search);
           hPtr != NULL; 
           hPtr = Tcl_NextHashEntry(&search)) {

         tagPtr = (TagData *) Tcl_GetHashValue(hPtr);
         if (++n > 1)
         fprintf(stderr, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
         if (tagPtr) {
            _dumpIt(tagPtr);
         } else {
            const char *name = Tcl_GetHashKey(&tagTable, hPtr);
            fprintf(stderr,"%s: NULL tag pointer\n", name);
         }
      }
   } else {
      _dumpIt(tagPtr);
   }
}
void TagMgr::_dumpIt(TagData *tagPtr) {

   fprintf(stderr, "NAME : %s\n", tagPtr->Name());
   if (IsIndicator(tagPtr)) {
   fprintf(stderr, "ID: %d\n", tagPtr->IndicID());
   } else {
   fprintf(stderr, "ID: %d\n", tagPtr->StyleID());
   fprintf(stderr, "TYPE : %s\n", (tagPtr->Type() & TT_TEXT) ? "Text" : "Margin");
   }
   if (IsIndicator(tagPtr)) {
   fprintf(stderr, "INDIC: %s\n", IndicatorTypeNames[tagPtr->IndicType()]);
   }
   if (tagPtr->FG() == NULL)
   fprintf(stderr, "FG   : <null>\n");
   else
   fprintf(stderr, "FG   : %d,%d,%d\n", tagPtr->FG()->red,tagPtr->FG()->green,tagPtr->FG()->blue);
   if (tagPtr->BG() == NULL)
   fprintf(stderr, "BG   : <null>\n");
   else
   fprintf(stderr, "BG   : %d,%d,%d\n", tagPtr->BG()->red,tagPtr->BG()->green,tagPtr->BG()->blue);
   if (!IsIndicator(tagPtr)) {
   fprintf(stderr, "BOLD : %d\n", tagPtr->Bold());
   fprintf(stderr, "ITAL : %d\n", tagPtr->Italic());
   }
}


/////////////////////////////////////////////////////////////////////////////////
// Implementation of MarkerData class
//
/////////////////////////////////////////////////////////////////////////////////

enum MarkerType {
    MK_ARROW, MK_CIRCLE, MK_LEFTRECT, MK_LINEBG,
    MK_ROUNDRECT, MK_SHORTARROW, MK_SMALLRECT, MK_SOLID
};
static const char *MarkerTypeNames[] = {
    "arrow", "circle", "leftrect", "linebg",
    "roundrect", "shortarrow", "smallrect", "solid", NULL
};


// define IDs for options we either need to error check or respond to
#define MKOPT_BG   1
#define MKOPT_FG   1 << 1
#define MKOPT_TYPE 1 << 2
#define MKOPT_IMG  1 << 3

// define configuration options
static Tk_OptionSpec markerOptionSpecs[] = {
   {TK_OPTION_COLOR,  "-background", NULL, NULL,
         "white", -1, Tk_Offset(MarkerOptions, bgColor), TK_OPTION_NULL_OK, 0, MKOPT_BG
   },
   {TK_OPTION_SYNONYM, "-bg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-background", 0},

   {TK_OPTION_COLOR,  "-foreground", NULL, NULL,
         "black", -1, Tk_Offset(MarkerOptions, fgColor), TK_OPTION_NULL_OK, 0, MKOPT_FG
   },
   {TK_OPTION_SYNONYM, "-fg", NULL, NULL, NULL, 0, -1, 0, (ClientData) "-foreground", 0},

   {TK_OPTION_STRING,  "-image", NULL, NULL,
         "", -1, Tk_Offset(MarkerOptions, imgName), 0, 0, MKOPT_IMG
   },
   {TK_OPTION_STRING_TABLE, "-type", NULL, NULL,
	     (char*)MarkerTypeNames[MK_CIRCLE], -1, Tk_Offset(MarkerOptions, type), 0, (ClientData)MarkerTypeNames, MKOPT_TYPE
   },
   {TK_OPTION_END}
};
//### CTOR ###
MarkerData::MarkerData(MarkerMgr *mgr) :
   name(0),
   id(-1),
   handle(-1),
   typeID(-1),
   margin(-1),
   cloned_n(0),
   cloned_from(0),
   mgr(mgr)
{
   mopts = (MarkerOptions *) Tcl_Alloc(sizeof(MarkerOptions));

   // must initialize all pointer/token values as per the Tk doc page:
   //    "...is responsible for storing NULL or None in all pointer 
   //    and token fields before invoking Tk_InitOptions."
   mopts->bgColor = NULL;
   mopts->fgColor = NULL;
   mopts->imgName = NULL;
   mopts->img = NULL;
   mopts->type    = None;
   mopts->pixels = NULL;

   optionTable = Tk_CreateOptionTable(mgr->interp, markerOptionSpecs);

   if (Tk_InitOptions(mgr->interp, (char *)mopts, optionTable, mgr->tkwin) != TCL_OK) {
      fprintf(stderr, "ERROR in Tk_InitOptions: %s\n", Tcl_GetStringResult(mgr->interp));
   } else {
      // map the user-specified type to the Scintilla ID
      UpdateType();
   }
}
//### DTOR ###
MarkerData::~MarkerData() {
    if (cloned_from) {
        // update parent marker to let it know it just lost a clone
        cloned_from->cloned_n--;
    }
    // clean up objects used for marker options
    Tk_FreeConfigOptions((char *)mopts, optionTable, mgr->tkwin);
    Tk_DeleteOptionTable(optionTable);
    if (mopts->img) Tk_FreeImage(mopts->img);
    Tcl_Free((char *)mopts);
}

//==========================================================================
// This will assign the Scintilla marker ID corresponding to the
// user-visible type name
//
void MarkerData::UpdateType() {

    switch ((enum MarkerType)mopts->type) {
        case MK_CIRCLE:     typeID = SC_MARK_CIRCLE; break;
        case MK_ARROW:      typeID = SC_MARK_ARROW; break;
        case MK_SHORTARROW: typeID = SC_MARK_SHORTARROW; break;
        case MK_SOLID:      typeID = SC_MARK_FULLRECT; break;
        case MK_ROUNDRECT:  typeID = SC_MARK_ROUNDRECT; break;
        case MK_SMALLRECT:  typeID = SC_MARK_SMALLRECT; break;
        case MK_LEFTRECT:   typeID = SC_MARK_LEFTRECT; break;
        case MK_LINEBG:     typeID = SC_MARK_BACKGROUND; break;
        default:
            Tcl_AppendResult(mgr->interp,
                "Unsupported marker type: ", mopts->type, (char *)NULL);
   }
}

//==========================================================================
// Copy the attributes from one marker into the current marker, which by
// definition makes this marker a clone
//
void MarkerData::Copy(MarkerData *src) {

    mopts->type = src->mopts->type;
    UpdateType();

    // Ask the source/parent marker for its config settings and then copy 
    // those into the current marker. We really need to use the Tk options
    // functionality to allow that code to manage ownership of the data
    // associated with the settings.

    Tcl_Obj *objPtr = Tk_GetOptionInfo(mgr->interp, (char *)src->mopts, src->optionTable, NULL, mgr->tkwin);
    int len;
    Tcl_ListObjLength(mgr->interp, objPtr, &len);
    
    Tcl_Obj **objv_arr = (Tcl_Obj **)Tcl_Alloc(sizeof(Tcl_Obj*) * 2 * len);
    Tcl_Obj *arg0, *arg4;
    int n = 0;
    for (int i=0; i < len; i++) {
        Tcl_Obj *arg;
        Tcl_ListObjIndex(mgr->interp, objPtr, i, &arg);

        int _len;
        Tcl_ListObjLength(mgr->interp, arg, &_len);
        if (_len != 5) continue; // this is probably an alias

        // example format of each list: {-background {} {} white yellow}

        Tcl_ListObjIndex(mgr->interp, arg, 0, &arg0); // -background
        Tcl_ListObjIndex(mgr->interp, arg, 4, &arg4); // yellow
        objv_arr[n++] = arg0;
        objv_arr[n++] = arg4;
    }
    if (Tk_SetOptions(mgr->interp, (char *)mopts, optionTable, 
            n, objv_arr, mgr->tkwin, NULL, NULL) != TCL_OK) {
        // uh oh...
    }
    Tcl_Free((char *)objv_arr);

    ScintillaObject *sci = SCIOBJ(mgr->sciwidgetPtr);

    if (src->typeID == SC_MARK_RGBAIMAGE) {
        typeID = SC_MARK_RGBAIMAGE;
        
        sci_cmd(sci, SCI_MARKERDEFINERGBAIMAGE, id, (sptr_t)mopts->pixels);
    } else {
        sci_cmd(sci, SCI_MARKERDEFINE, id, typeID);
        if (mopts->bgColor) {
            sci_cmd(sci, SCI_MARKERSETBACK, id, _XRGB_(mopts->bgColor));
        }
        if (mopts->fgColor) {
            sci_cmd(sci, SCI_MARKERSETFORE, id, _XRGB_(mopts->fgColor));
        }
    }
    cloned_from = src;
}

//==========================================================================
// Check if the name of this marker matches that of the specified source
// marker. This will handle doing the match when the current marker
// is a clone, because in that case it will match against the original
// "parent" marker.
//
// NOTE: The "src_name" will always represent a name given from the
//       client code - so it will never be a cloned name.
//
bool MarkerData::MatchName(const char *src_name) {

    // check for an exact match
    if (strcmp(src_name, name) == 0) {
        return true;
    }
    // now check if the current marker is a clone, and if so, check
    // its parent marker to see if its name matches
    if (IsClone() && (strcmp(src_name, cloned_from->name) == 0)) {
        return true;
    }
    return false;
}


/////////////////////////////////////////////////////////////////////////////////
// Implementation of MarkerMgr class
//
/////////////////////////////////////////////////////////////////////////////////

//### CTOR ###
MarkerMgr::MarkerMgr(SciWidget *w) :
   ids(0),
   sciwidgetPtr(w),
   interp(w->interp),
   tkwin(w->tkwin)
{
   // create the hash table
   Tcl_InitHashTable(&markerTable, TCL_STRING_KEYS);
}

//### DTOR ###
MarkerMgr::~MarkerMgr() {

   Tcl_HashEntry *hPtr;
   Tcl_HashSearch search;
   MarkerData    *mPtr;

   // first iterate through the objects in the hash table...
   for (hPtr = Tcl_FirstHashEntry(&markerTable, &search);
        hPtr != NULL; 
        hPtr = Tcl_NextHashEntry(&search)) {
      mPtr = (MarkerData *) Tcl_GetHashValue(hPtr);

      // we skip clones because they get deleted when their "parent"
      // marker gets deleted
      if ( ! mPtr->IsClone() )
          _deleteMarker(mPtr);
   }
   // then delete the hash table itself
   Tcl_DeleteHashTable(&markerTable);
}

//==========================================================================
// Add a marker to a specific line for a specific margin. This function
// does no error check to verify the target margin supports symbols; that
// is left up to the calling code.
//
//*****************************************************************
// Here's how the mask value matches to the marker IDs that are
// used in the SCI_MARKERDEFINE call. Each mask bit corresponds to
// a position value of 0(lsb) to 31(msb). The marker ID should be
// in the range 0-31. If the added(via SCI_MARKERADD) marker ID is
// in the mask then it will appear in the target margin; if it isn't
// then the associated line will be shown in the marker's bgd color.
//
// For example, a mask value of "5" has a binary value of "101".
// This enables bit positions "2" and "0":
//   3322222222220000000000
//   1098765432109876543210
//                      101 << binary value is "5"
//
// Using SCI_MARKERADD with either marker ID "2" or "0" would then
// correctly add the marker into the target margin; if the marker ID
// had been "1" instead then no marker would appear in the margin 
// and instead would cause the line's background to be changed.
//
// To test if a marker already exists on a line, do this:
//   int state = SSM(SCI_MARKERGET, line, 0);
//   if (state & (1 << MY_MARKER))
//*****************************************************************

int MarkerMgr::AddMarker(int marginID, MarkerData *mPtr, int line) {

   ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);

   // check if the marker has been assigned to a margin yet; and if
   // it has, check if it has been assigned to THIS margin
   if (mPtr->margin == -1) {
       mPtr->margin = marginID;
   }
   else if (mPtr->margin != marginID) {
       // this is an error, because the calling code should have already
       // attempted to create a cloned marker before doing the assign
       Tcl_AppendResult(interp, "attempt to assign marker belonging to another margin", NULL);
       return TCL_ERROR;
   }
   int hndl = sci_cmd(sciobj, SCI_MARKERADD, LINE2SCI(line), mPtr->id);

   // We store the "handle", which is a unique id created by Scintilla.
   // This handle can be used for searching using SCI_MARKERNEXT/PREV
   mPtr->handle = hndl;

   // make sure the margin's mask is updated to allow the requested marker
   //
   int curr_mask = sci_get_margin_mask(sciobj, marginID);
   sci_set_margin_mask(sciobj, marginID, (curr_mask | (1 << mPtr->id)));

   return TCL_OK;
}

//==========================================================================
//
int MarkerMgr::InstanceCmd(int objc, Tcl_Obj *CONST objv[]) {

   int index;
   int result = TCL_OK;

   if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option arg ?arg arg ...?");
      return TCL_ERROR;
   }

   // define command names
   static const char *optionStrings[] = {
#ifdef DEBUG
      "dump", "dumpnames",
#endif
      "configure", "delete", "exists", "names", "types",
      (char *)NULL
   };
   enum optionIds {
#ifdef DEBUG
      CMD_DUMP, CMD_DUMPNAMES,
#endif
      CMD_CONFIG, CMD_DELETE, CMD_EXISTS, CMD_NAMES, CMD_TYPES
   };

   if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings, "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
   }
   switch ((enum optionIds)index) {
      case CMD_CONFIG: {
         result = CmdConfig(objc, objv);
         break;
      }
      case CMD_DELETE: {
         result = CmdDelete(objc, objv);
         break;
      }
#ifdef DEBUG
      case CMD_DUMP: {
         MarkerData *mPtr = NULL; // this will dump all markers
         if (objc == 4)
            mPtr = FindMarker(Tcl_GetString(objv[3]));
         _dump(mPtr);
         break;
      }
      case CMD_DUMPNAMES: {
         result = CmdNames(objc, objv, true);
         break;
      }
#endif
      case CMD_EXISTS: {
         result = CmdExists(objc, objv);
         break;
      }
      case CMD_NAMES: {
         result = CmdNames(objc, objv);
         break;
      }
      case CMD_TYPES: {
         Tcl_Obj *tmp = Tcl_NewListObj(0, NULL);
         const char *n;
         for (int i=0; (n = MarkerTypeNames[i]) != NULL; i++) {
            Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(n, -1));
         }
         Tcl_Obj *cmd = Tcl_ObjPrintf("lsort {%s}", Tcl_GetStringFromObj(tmp,NULL));

         if (Tcl_Eval(interp, Tcl_GetStringFromObj(cmd,NULL)) != TCL_OK) {
            fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(interp));
            return TCL_ERROR;
         }
         break;
      }
   }
   return result;
}

//==========================================================================
// Implements the MARKER CONFIGURE sub-command
//
// This function will either make modifications to an existing marker
// or it will create a new one
//
int MarkerMgr::CmdConfig(int objc, Tcl_Obj *CONST objv[]) {

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "markerName ?option? ?value option value ...?");
      return TCL_ERROR;
   }
   // either create a new marker or return an existing one
   int isNew;
   MarkerData *mPtr = _createMarker(Tcl_GetString(objv[3]), &isNew);
   if (mPtr == NULL) {
      return TCL_ERROR;
   }

   if (objc <= 5) {
      //--------------------------
      // return config info
      //--------------------------
      Tcl_Obj *objPtr = Tk_GetOptionInfo(interp, (char *) mPtr->mopts, mPtr->optionTable,
         ((objc == 5) ? objv[4] : NULL), tkwin);
      if (objPtr == NULL) {
         return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, objPtr);
      return TCL_OK;
   } else {
      //--------------------------
      // modify config info
      //--------------------------
      Tk_SavedOptions savedOptions;
      int mask = 0;
      if (Tk_SetOptions(interp, (char *)mPtr->mopts, mPtr->optionTable,
            objc-4, objv+4, tkwin, &savedOptions, &mask) != TCL_OK) {
         Tk_RestoreSavedOptions(&savedOptions);
         if (isNew) {
            // get rid of the marker we had to create
            _deleteMarker(mPtr);
         }
         return TCL_ERROR;
      }
      // do some error checking...

      if ((mask & MKOPT_TYPE) && ((mask & MKOPT_IMG))) {
         Tk_RestoreSavedOptions(&savedOptions);
         if (isNew) _deleteMarker(mPtr);
         Tcl_AppendResult(interp, "cannot use -type option with -image", (char *)NULL);
         return TCL_ERROR;
      }
      ScintillaObject *sci = SCIOBJ(sciwidgetPtr);
      bool do_define = true;

      if (mask & MKOPT_IMG) {
         // verify the image is legal
         if (!isNew) {
             // if marker already existed and had an image associated with it, free old image
             if (mPtr->mopts->img) {
                 Tk_FreeImage(mPtr->mopts->img);
             }
         }
         if (_validateImage(mPtr) != TCL_OK) {
            Tk_RestoreSavedOptions(&savedOptions);
            if (isNew) _deleteMarker(mPtr);
            return TCL_ERROR;
         }
         do_define = false;
      }
      if (mask & MKOPT_TYPE) {
         mPtr->UpdateType();
      }

      // now let Scintilla know about it
      if (do_define) {
         sci_cmd(sci, SCI_MARKERDEFINE, mPtr->id, mPtr->typeID);

         if (mPtr->mopts->bgColor) {
            sci_cmd(sci, SCI_MARKERSETBACK, mPtr->id, _XRGB_(mPtr->mopts->bgColor));
         }
         if (mPtr->mopts->fgColor) {
            sci_cmd(sci, SCI_MARKERSETFORE, mPtr->id, _XRGB_(mPtr->mopts->fgColor));
         }
      }
      // get the cloned markers (if any) updated also
      _cloneOp(cloneCopy, mPtr);
   }
   return TCL_OK;
}

//==========================================================================
// Implements the MARKER DELETE sub-command
//
int MarkerMgr::CmdDelete(int objc, Tcl_Obj *CONST objv[]) {

   Tcl_HashEntry *hPtr;
   MarkerData *mPtr;

   if (objc < 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "markerName ?markerName ...?");
      return TCL_ERROR;
   }
   for (int i=3; i < objc; i++) {
      const char *name = Tcl_GetString(objv[i]);

      hPtr = Tcl_FindHashEntry(&markerTable, name);
      if (hPtr == NULL) continue;

      mPtr = (MarkerData *)Tcl_GetHashValue(hPtr);

      // delete the marker
      _deleteMarker(mPtr);
   }
   return TCL_OK;
}

//==========================================================================
// Implements the MARKER EXISTS sub-command
//
// If the specified marker exists on the specified line then a value of "1"
// is set into the Tcl interpretter result, otherwise a value of "0" is set
//
int MarkerMgr::CmdExists(int objc, Tcl_Obj *CONST objv[]) {

   MarkerData *mPtr;
   int line;

   if (objc != 5 && objc!= 6) {
      Tcl_WrongNumArgs(interp, 3, objv, "markerName line ?column#|name|alias?");
      return TCL_ERROR;
   }
   if (objc == 5) {
       // find marker in any of the margins
       mPtr = FindMarker(Tcl_GetString(objv[3]));
   } else {
       // find in a specific margin

       // get the requested margin's id
       int id;
       if (sciwidgetPtr->marginMgr->Get(objv[5], &id) != TCL_OK) {
           return TCL_ERROR;
       }
       mPtr = FindMarker(id, Tcl_GetString(objv[3]), false/*do NOT create clone*/);
   }
   if (mPtr == NULL) {
      return TCL_ERROR;
   }
   // NOTE: "line" will come back adjusted to Scintilla scheme
   if (ValidateLine(interp, (ClientData)sciwidgetPtr, objv[4], &line) != TCL_OK) {
      return TCL_ERROR;
   }
   // check for the markers on this line

   int state = sci_get_line_markers(SCIOBJ(sciwidgetPtr), line);

   if (state & (1 << mPtr->id)) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(1));
   } else {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(0));
   }
   return TCL_OK;
}

//==========================================================================
// Implements the MARKER NAMES sub-command
//
int MarkerMgr::CmdNames(int objc, Tcl_Obj *CONST objv[], bool include_clones) {

   if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, NULL);
      return TCL_ERROR;
   }
   // show all existing markers in a sorted list

   // first we collect all of the names from the hash table; of
   // course these won't appear in sorted order.
   Tcl_HashEntry *hPtr;
   Tcl_HashSearch search;
   MarkerData *mPtr;
   Tcl_Obj *tmp = Tcl_NewListObj(0, NULL);
   for (hPtr = Tcl_FirstHashEntry(&markerTable, &search);
        hPtr != NULL; 
        hPtr = Tcl_NextHashEntry(&search)) {
      mPtr = (MarkerData *) Tcl_GetHashValue(hPtr);

      // ignore clones!!
      if (mPtr->IsClone() && !include_clones) continue;

      Tcl_ListObjAppendElement(interp, tmp, Tcl_NewStringObj(mPtr->name, -1));
   }
   int len;
   if ( (Tcl_ListObjLength(interp, tmp, &len) == TCL_OK) && (len > 0)) {

      // To get the list sorted we take the lazy way out, and let the
      // Tcl "lsort" command handle the sorting
      //
      if (len > 1) {
         Tcl_Obj *cmd = Tcl_ObjPrintf("lsort {%s}", Tcl_GetStringFromObj(tmp,NULL));

         if (Tcl_Eval(interp, Tcl_GetStringFromObj(cmd,NULL)) != TCL_OK) {
            fprintf(stderr, "Error in Tcl_Eval: %s\n", Tcl_GetStringResult(interp));
            return TCL_ERROR;
         }
      } else {
         // just a single name, no need to sort
         Tcl_SetObjResult(interp, tmp);
      }
   }
   return TCL_OK;
}

//==========================================================================
// See if a marker with the specified name has been defined. If it is, a
// pointer to the marker structure is returned, else NULL is returned and
// an error is left in the interp's result.
//
MarkerData *MarkerMgr::FindMarker(const char *name) {

   Tcl_HashEntry *hPtr;

   hPtr = Tcl_FindHashEntry(&markerTable, name);
   if (hPtr) {
      return (MarkerData *)Tcl_GetHashValue(hPtr);
   }
   Tcl_AppendResult(interp, "marker \"", name, "\" isn't defined", (char *)NULL);
   return NULL;
}

//==========================================================================
// Find a marker with the specified id
//
MarkerData *MarkerMgr::FindMarker(int marker_id) {

    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    MarkerData *mPtr = NULL;

    for (hPtr = Tcl_FirstHashEntry(&markerTable, &search);
         hPtr != NULL; 
         hPtr = Tcl_NextHashEntry(&search)) {
        mPtr = (MarkerData *) Tcl_GetHashValue(hPtr);
        if (mPtr->id == marker_id) 
            return mPtr;
    }
    Tcl_AppendResult(interp, "no marker with id ", marker_id, "was found", (char *)NULL);
    return NULL;
}

//==========================================================================
// This one will only match a marker if it's been assigned to the 
// specified margin, or one exists but hasn't yet been assigned to
// a margin. It will also handle getting a cloned marker created for
// the case where a marker has already been created but it got assigned
// to a different margin.
//
MarkerData *MarkerMgr::FindMarker(int marginID, const char *name, bool ok_to_clone) {

    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    MarkerData *mPtr = NULL;
    MarkerData *found = NULL;

    for (hPtr = Tcl_FirstHashEntry(&markerTable, &search);
         hPtr != NULL; 
         hPtr = Tcl_NextHashEntry(&search)) {
        mPtr = (MarkerData *) Tcl_GetHashValue(hPtr);

        // check for a name match, making sure to also consider
        // cloned names
        if ( ! mPtr->MatchName(name)) continue;

        // ok, the name matches, now check if the marker hasn't 
        // yet been assigned a margin, or if it has that it's
        // been assigned to the target margin
        if (mPtr->margin == -1) {
            // marker hasn't been assigned to a margin yet
            return mPtr;
        }
        if (mPtr->margin == marginID) {
            // marker has already been assigned to this margin
            return mPtr;
        }
        // marker exists but was assigned to another margin
        found = mPtr;
    }
    if (found) {
        if (ok_to_clone) {
            // a marker was already created by the client, but it got assigned
            // to a different margin - so let's try and create a new clone that
            // we can assign to the requested target margin. If the clone request
            // fails, an appropriate error will already be set so there's no need
            // to set another one.
            mPtr = _cloneMarker(name, found);
        }
        if (mPtr) return mPtr;
    } else {
        Tcl_AppendResult(interp, "marker \"", name, "\" isn't defined", (char *)NULL);
    }
    return NULL;
}


//---------------------------------------------------------
// Private functions
//---------------------------------------------------------
#define CLONE_STR "++CLONE"

static Tcl_Obj * _format_clone_name(const char *name, int i) {
    return Tcl_ObjPrintf("%s%s%d", name, CLONE_STR, i);
}

//==========================================================================
// Perform an operation on the cloned markers of a source marker
//
MarkerData *MarkerMgr::_cloneOp(CloneOpId op, MarkerData *src) {

    int n = src->CloneCount();
    for (int i=1; i <= n; i++) {
        Tcl_Obj *obj = _format_clone_name(src->name, i);

        MarkerData *c = FindMarker(Tcl_GetString(obj));
        if (!c) continue; // should NOT get here, but be safe

        switch (op) {
            case cloneCopy: {
                //--- COPY
                c->Copy(src);
                break;
            }
            case cloneDelete: {
                //--- DELETE
                _deleteMarkerImpl(c);
                break;
            }
        }
    }
    return NULL;
}

//==========================================================================
// Create a cloned copy of a another marker
//
MarkerData *MarkerMgr::_cloneMarker(const char *name, MarkerData *src) {

    // create a new marker
    Tcl_Obj *obj = _format_clone_name(name, src->CloneCount()+1);
    MarkerData *new_clone = _createMarker(Tcl_GetString(obj), NULL);
//fprintf(stderr,"CLONEMARKER  src=%s  new=%p\n",src->name,(void *)new_clone);
    if (new_clone == NULL) {
        return NULL;
    }
    // now have it clone the values from the source
    new_clone->Copy(src);
    src->cloned_n++;

    return new_clone;
}

//==========================================================================
// Create a marker
//
MarkerData *MarkerMgr::_createMarker(const char *markerName, int *isNew) {

   MarkerData *mPtr;
   Tcl_HashEntry *hPtr = NULL;
   int isNewMarker;

   // verify the name is "legal"
   if (strlen(markerName) == 0) {
      Tcl_AppendResult(interp, "empty marker name", (char *)NULL);
      return NULL;
   }
   if (strchr(markerName, ' ') != NULL) {
      Tcl_AppendResult(interp, "marker name cannot contain spaces", (char *)NULL);
      return NULL;
   }
   // look up the name in the hash table
   hPtr = Tcl_CreateHashEntry(&markerTable, markerName, &isNewMarker);
   if (isNew) *isNew = isNewMarker;
   if (!isNewMarker) {
      // Found it!
      mPtr = (MarkerData *)Tcl_GetHashValue(hPtr);
      return mPtr;
   }
   // No match, so create a new marker -- VERIFY WE CAN!!
   int id = _getNextId();
   if (id == -1) {
      Tcl_DeleteHashEntry(hPtr);
      Tcl_AppendResult(interp,
         "new marker could not be created: too many markers already exist", (char *)NULL);
      return NULL;
   }
   mPtr = new MarkerData(this);
   mPtr->name = Tcl_GetHashKey(&markerTable, hPtr);
   mPtr->id   = id;

   // Put it into the hash table
   Tcl_SetHashValue(hPtr, mPtr);

   return mPtr;
}

//==========================================================================
// Delete a marker. This not only removes the marker from the widget but
// also causes Scintilla to delete it internally.
//
void MarkerMgr::_deleteMarker(MarkerData *mPtr) {

    // do the clones as well(it's a no-op if there aren't any clones)
    _cloneOp(cloneDelete, mPtr);

    _deleteMarkerImpl(mPtr);
}
// this does the actual work
void MarkerMgr::_deleteMarkerImpl(MarkerData *mPtr) {

   // remove it from hash table
   Tcl_HashEntry *hPtr;
   hPtr = Tcl_FindHashEntry(&markerTable, mPtr->name);
   if (hPtr) {
      Tcl_DeleteHashEntry(hPtr);
   }
   // free its position id
   if (mPtr->id >= 0) {
      ids &= ~(1 << mPtr->id);

      // make sure the margins marker masks no
      // longer have this marker's id enabled
      sciwidgetPtr->marginMgr->RemoveMarker(mPtr->id);
   }

   // now delete the object
   delete mPtr;
}

//==========================================================================
// Returns the max ID value we can use
int MarkerMgr::GetMaxId() {

    // NOTE: The max ID will depend upon whether or not fold markers
    //       are being used, because if they are they occupy 7 positions
    //       in the ID mask(positions 25 thru 31, see Scintilla constant
    //       SC_MASK_FOLDERS in Scintilla.h).
    //
    return (sciwidgetPtr->foldstyle != foldDISABLED) ? 24 : 31;
}

//==========================================================================
// Return the next available marker ID (which is bit position in the mask
// used to control which markers a margin can show)
//
int MarkerMgr::_getNextId() {

   // Scan each bit position, looking for first one which is unset
   //
   for (int i=0; i <= GetMaxId(); i++) {
      if ( (ids & (1 << i)) == 0 ) {
         ids |= (1 << i);
         return i;
      }
   }
   return -1;
}

//==========================================================================
// Validate image information being requested for use as a marker
//

// need a callback defined if we want to use Tk_GetImage
static void
_imgCallbackProc(ClientData cd, int x, int y, int w, int h, int imgW, int imgH) 
{
    MarkerData *mPtr = (MarkerData*)cd;
    fprintf(stderr,"_imgCallbackProc(%p, x=%d, y=%d, w=%d, h=%d, imgW=%d, imgH=%d)\n"
            "imageName=%s\n", 
            (void*)cd, x, y, w, h, imgW, imgH, mPtr->mopts->imgName);
    if (mPtr->mopts->img) Tk_FreeImage(mPtr->mopts->img);
    mPtr->mopts->img = NULL;
    mPtr->mopts->pixels = NULL;
    mPtr->mgr->_validateImage(mPtr);
}

int MarkerMgr::_validateImage(MarkerData *mPtr) {
   const char *imageName = mPtr->mopts->imgName;
   Tk_Image tk_image = Tk_GetImage(interp, tkwin, imageName, _imgCallbackProc, (ClientData)mPtr);
   Tk_PhotoHandle photo = Tk_FindPhoto(interp, imageName);
   Tk_PhotoImageBlock photo_block;

   if (tk_image == NULL || photo == NULL) {
      return TCL_ERROR;
   }

   Tk_PhotoGetImage(photo, &photo_block);

   //fprintf(stderr,"pixelPtr=%p pixelSize=%d w=%d h=%d pitch=%d\n", 
   //        photo_block.pixelPtr, photo_block.pixelSize, 
   //        photo_block.width, photo_block.height, photo_block.pitch);
   if (photo_block.pixelSize != 4) {
       Tcl_SetResult(interp, "Unsupported photo image: pixelSize!=4\n", TCL_STATIC);
       Tk_FreeImage(tk_image);
       return TCL_ERROR;
   }

   mPtr->mopts->img = tk_image;
   mPtr->mopts->pixels = photo_block.pixelPtr;
   ScintillaObject *sci = SCIOBJ(sciwidgetPtr);
   sci_cmd(sci, SCI_RGBAIMAGESETWIDTH, photo_block.width);
   sci_cmd(sci, SCI_RGBAIMAGESETHEIGHT, photo_block.height);

   mPtr->typeID = SC_MARK_RGBAIMAGE;
   //fprintf(stderr, "sci_cmd(sci, SCI_MARKERDEFINERGBAIMAGE, %d, %p);\n", 
   //        mPtr->id, mPtr->mopts->pixels);
   sci_cmd(sci, SCI_MARKERDEFINERGBAIMAGE, mPtr->id, (sptr_t)mPtr->mopts->pixels);

   return TCL_OK;
}


//==========================================================================
void MarkerMgr::_dump(MarkerData *mPtr) {

   if (mPtr == NULL) {
      // dump all of the markers (ignore sorting for now)
      Tcl_HashEntry *hPtr;
      Tcl_HashSearch search;
      fprintf(stderr,    "--- DUMP all markers: -----------------------------\n");
      int n = 0;
      for (hPtr = Tcl_FirstHashEntry(&markerTable, &search);
           hPtr != NULL; 
           hPtr = Tcl_NextHashEntry(&search)) {

         mPtr = (MarkerData *) Tcl_GetHashValue(hPtr);
         if (++n > 1)
         fprintf(stderr, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
         _dumpIt(mPtr);
      }
   } else {
      _dumpIt(mPtr);
   }
}
void MarkerMgr::_dumpIt(MarkerData *mPtr) {

   MarkerOptions *mopts = mPtr->mopts;

   fprintf(stderr, "NAME   : %s\n", mPtr->name);
   fprintf(stderr, "ID     : %d\n", mPtr->id);
   fprintf(stderr, "HNDL   : %d\n", mPtr->handle);
   fprintf(stderr, "TYPE   : %s\n", MarkerTypeNames[mPtr->mopts->type]);
   if (mopts->imgName)
   fprintf(stderr, "IMAGE  : %s\n", mopts->imgName);
   fprintf(stderr, "S-TYPE : %d\n", mPtr->typeID);
   if (mopts->fgColor == NULL)
   fprintf(stderr, "FG     : <null>\n");
   else
   fprintf(stderr, "FG     : %d,%d,%d\n", mopts->fgColor->red,mopts->fgColor->green,mopts->fgColor->blue);
   if (mopts->bgColor == NULL)
   fprintf(stderr, "BG     : <null>\n");
   else
   fprintf(stderr, "BG     : %d,%d,%d\n", mopts->bgColor->red,mopts->bgColor->green,mopts->bgColor->blue);
}



/////////////////////////////////////////////////////////////////////////////////
// Implementation of MarkMgr class
//
/////////////////////////////////////////////////////////////////////////////////

//### CTOR ###
MarkMgr::MarkMgr(SciWidget *w) :
    sciwidgetPtr(w),
    insertMark(0),
    currentMark(0),
    interp(w->interp),
    tkwin(w->tkwin)
{
    // create the special marks
    AddMark("insert", 0);
    AddMark("current", 0);

    // create the hash table
    Tcl_InitHashTable(&markTable, TCL_STRING_KEYS);
}

//### DTOR ###
MarkMgr::~MarkMgr() {

    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    MarkData *mPtr;

    // first iterate through the objects in the hash table...
    for (hPtr = Tcl_FirstHashEntry(&markTable, &search);
         hPtr != NULL; 
         hPtr = Tcl_NextHashEntry(&search)) {
        mPtr = (MarkData *) Tcl_GetHashValue(hPtr);

        _deleteMark(mPtr);
    }
    // then delete the special marks
    _deleteMark(insertMark);

    // then delete the hash table itself
    Tcl_DeleteHashTable(&markTable);
}

//==========================================================================
//
// Implements the MARK widget command
//
int MarkMgr::InstanceCmd(int objc, Tcl_Obj *CONST objv[]) {

    int result = TCL_OK;
    int index;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    MarkData *markPtr;

    if (objc < 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "option ?arg arg ...?");
        return TCL_ERROR;
    }

    // define command names
    static const char *optionStrings[] = {
        "exists", "gravity", "names", "next", "previous", "set", "unset",
#ifdef DEBUG
        "dump",
#endif
        (char *)NULL
    };
    enum optionIds {
        CMD_EXISTS, CMD_GRAVITY, CMD_NAMES, CMD_NEXT, CMD_PREV, CMD_SET, CMD_UNSET
#ifdef DEBUG
        ,CMD_DUMP
#endif
    };
    if (Tcl_GetIndexFromObj(interp, objv[2], optionStrings, "mark option",
            0, &index) != TCL_OK) {
        return TCL_ERROR;
    }

    switch ((enum optionIds)index) {
#ifdef DEBUG
        case CMD_DUMP: _dump(); break;
#endif
        case CMD_EXISTS: {
            if (objc != 4) {
                Tcl_WrongNumArgs(interp, 3, objv, "markName");
                return TCL_ERROR;
            }
            if (FindMark(Tcl_GetString(objv[3])) == NULL) {
                Tcl_SetObjResult(interp, Tcl_NewIntObj(0));
            } else {
                Tcl_SetObjResult(interp, Tcl_NewIntObj(1));
            }
            break;
        }
        case CMD_GRAVITY: {
            result = CmdGravity(objc, objv);
            break;
        }
        case CMD_NAMES: {
            if (objc != 3) {
                Tcl_WrongNumArgs(interp, 3, objv, NULL);
                return TCL_ERROR;
            }
            // Tk Text widget doesn't do sorting, so we won't either
            Tcl_AppendElement(interp, "insert");

            for (hPtr = Tcl_FirstHashEntry(&markTable, &search);
                 hPtr != NULL;
                 hPtr = Tcl_NextHashEntry(&search)) {
                Tcl_AppendElement(interp, Tcl_GetHashKey(&markTable, hPtr));
            }
            break;
        }
        case CMD_NEXT:
        case CMD_PREV: {
            if (objc != 4) {
                Tcl_WrongNumArgs(interp, 3, objv, "index");
                return TCL_ERROR;
            }
            bool fwd = ((enum optionIds)index == CMD_NEXT);
            char *index_str = Tcl_GetString(objv[3]);
            sptr_t start_pos;
            if ((markPtr = FindMark(index_str)) != NULL) {
                if (fwd) {
                    start_pos = markPtr->pos + 1; // start AFTER the mark
                } else {
                    start_pos = markPtr->pos - 1; // start BEFORE the mark
                }
            } else {
                Tcl_ResetResult(interp);
                if (GetPosition(sciwidgetPtr, index_str, &start_pos, false) != TCL_OK) {
                    return TCL_ERROR;
                }
            }
            // search marks, looking for closest one at or AFTER/BEFORE our start position
            MarkData *closest = NULL;
            int delta = fwd ? INT_MAX : 0;

            for (hPtr = Tcl_FirstHashEntry(&markTable, &search);
                 hPtr != NULL;
                 hPtr = Tcl_NextHashEntry(&search)) {
                markPtr = (MarkData *)Tcl_GetHashValue(hPtr);
                if (fwd) {
                    //--- NEXT ---
                    if (markPtr->pos > start_pos) {
                        if ((markPtr->pos - start_pos) < delta) {
                            delta = markPtr->pos - start_pos;
                            closest = markPtr;
                        }
                    }
                } else {
                    //--- PREVIOUS ---
                    if (markPtr->pos < start_pos) {
                        if ((start_pos - markPtr->pos) > delta) {
                            delta = start_pos - markPtr->pos;
                            closest = markPtr;
                        }
                    }
                }
            }
            if (closest) {
                Tcl_SetObjResult(interp, Tcl_NewStringObj(closest->name, -1));
            }
            break;
        }
        case CMD_SET: {
            result = CmdSet(objc, objv);
            break;
        }
        case CMD_UNSET: {
            for (int i=3; i < objc; i++) {
                markPtr = FindMark(Tcl_GetString(objv[i]));
                if (markPtr) {
                    if (markPtr == insertMark) continue; // can't be removed

                    _deleteMark(markPtr);
                }
            }
            break;
        }
    }
    return result;
}

//==========================================================================
// Implements the MARK GRAVITY sub-command
//
int MarkMgr::CmdGravity(int objc, Tcl_Obj *CONST objv[]) {

    int length;
    char *str;
    MarkData *markPtr;

    if (objc < 4 || objc > 5) {
        Tcl_WrongNumArgs(interp, 3, objv, "markName ?direction?");
        return TCL_ERROR;
    }
    str = Tcl_GetStringFromObj(objv[3], &length); // markName
    markPtr = FindMark(str);
    if (markPtr == NULL) {
        return TCL_ERROR;
    }
    if (objc == 4) {
        if (markPtr->gravityLeft)
            Tcl_SetResult(interp, "left", TCL_STATIC);
        else
            Tcl_SetResult(interp, "right", TCL_STATIC);
        return TCL_OK;
    }
    str = Tcl_GetStringFromObj(objv[4], &length); // gravity

    if ((str[0] == 'l') && (strncmp(str, "left", (unsigned)length) == 0)) {
        markPtr->gravityLeft = true;
    } else if ((str[0] == 'r') && (strncmp(str, "right", (unsigned)length) == 0)) {
        markPtr->gravityLeft = false;
    } else {
        Tcl_AppendResult(interp, "bad mark gravity \"", str,
            "\": must be left or right", NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

//==========================================================================
// Implements the MARK SET sub-command
//
int MarkMgr::CmdSet(int objc, Tcl_Obj *CONST objv[]) {

    ScintillaObject *obj = SCIOBJ(sciwidgetPtr);
    MarkData *markPtr;
    char *name, *index_str;
    sptr_t pos;
    
    if (objc != 5) {
        Tcl_WrongNumArgs(interp, 3, objv, "markName index");
        return TCL_ERROR;
    }
    name      = Tcl_GetString(objv[3]);
    index_str = Tcl_GetString(objv[4]);

    // create a new mark or return an existing one
    markPtr = _createMark(name);

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //HACK: USE special index names to support arrow key navigation.
    //      This is a "cheap" way for us to implement the keyboard
    //      movement bindings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if (markPtr == insertMark) {
        // check the selection so we will know if it changed as a
        // result of moving the insertion point
        int sel1 = sci_cmd(obj, SCI_GETSELECTIONSTART);
        int sel2 = sci_cmd(obj, SCI_GETSELECTIONEND);

        if (       !strcmp(index_str,"charright")) {
            sci_cmd(obj, SCI_CHARRIGHT);
        } else if (!strcmp(index_str,"charleft")) {
            sci_cmd(obj, SCI_CHARLEFT);
        } else if (!strcmp(index_str,"linedown")) {
            sci_cmd(obj, SCI_LINEDOWN);
        } else if (!strcmp(index_str,"lineup")) {
            sci_cmd(obj, SCI_LINEUP);
        } else if (!strcmp(index_str,"lineend")) {
            sci_cmd(obj, SCI_LINEEND);
        } else if (!strcmp(index_str,"linehome")) {
            sci_cmd(obj, SCI_HOME);
        } else if (!strcmp(index_str,"top")) {
            sci_cmd(obj, SCI_DOCUMENTSTART);
        } else if (!strcmp(index_str,"bottom")) {
            sci_cmd(obj, SCI_DOCUMENTEND);
        } else {
            pos = -2; // let "-1" mean a valid position beyond end of doc

            if (GetPosition(sciwidgetPtr, index_str, &pos, true) != TCL_OK) {
                return BadIndex(interp, index_str);
            }
            if (pos == -2) {
                // indicates position was in a non-visible line -- do nothing
                return TCL_OK;
            }
            if (pos == -1) {
                // assume index was specified beyond end of doc
                pos = sci_get_length(obj);
            }
            if ((sel2 - sel1) == 0) {
                // no selection, just set new current position - we call
                // SETEMPTYSELECTION so both the Scintilla anchor and current
                // position are both set, which will avoid an inadvertant selection
                // being defined.
                //
                sci_cmd(obj, SCI_SETEMPTYSELECTION, pos);
            } else {
                if (pos != sel1 && pos != sel2) {
                    // moving the insert marker to a position outside of the selection
                    // requires the selection be cleared due to Scintilla restriction.
                    //
                    sci_cmd(obj, SCI_SETEMPTYSELECTION, pos); // this will clear the selection
                } else if (pos == sel1) {
                    sci_cmd(obj, SCI_SETSEL, sel2, sel1); // anchor at start of selection
                } else {
                    sci_cmd(obj, SCI_SETSEL, sel1, sel2); // anchor at end of selection
                }
            }
        }
        // now check if the change to the insert point also caused
        // a change in selection
        //
        int _sel1 = sci_cmd(obj, SCI_GETSELECTIONSTART);
        int _sel2 = sci_cmd(obj, SCI_GETSELECTIONEND);
        if (((sel1 == sel2) && (_sel1 != _sel2)) ||
            ((sel1 != sel2) && (_sel1 == _sel2))) {

            sciwidgetPtr->tagMgr->FireSelectionVirtual();
        }
    } else {
        // get the mark's position set
        pos = -2; // let "-1" mean a valid position beyond end of doc

        if (GetPosition(sciwidgetPtr, index_str, &pos, true) != TCL_OK) {
            return BadIndex(interp, index_str);
        }
        if (pos == -2) return TCL_OK;
        if (pos == -1) {
            // assume index was specified beyond end of doc
            pos = sci_get_length(obj);
        }
        markPtr->pos = pos;
    }
    return TCL_OK;
}

//==========================================================================
// Create a new marker and assign it the specified position
//
MarkData *MarkMgr::AddMark(const char *name, int pos) {

    MarkData *mPtr = _createMark(name);
    if (mPtr) {
        mPtr->pos = pos;
        return mPtr;
    }
    return NULL;
}

//==========================================================================
// See if a marker with the specified name has been defined. If it is, a
// pointer to the marker structure is returned, else NULL is returned and
// an error is left in the interp's result.
//
MarkData *MarkMgr::FindMark(const char *name, int *offset) {

    // check if we're supposed to handle a name that might have offset characters.
    // an example: "mymark linestart" or "mymark +5 chars"
    Tcl_DString dstr;
    Tcl_DStringInit(&dstr);
    if (offset) {
        // we need to extract out just the portion of the string we got
        // that appears to be a name(ie prior to any spaces and/or +/- char)
        *offset = 0;
        for (const char *p = name; *p != 0; p++) {
            if (isspace(*p) || (*p == '+') || (*p == '-')) {
                *offset = p - name;
                break;
            }
        }
        if (*offset > 0) {
            Tcl_DStringAppend(&dstr, name, *offset);
            name = Tcl_DStringValue(&dstr);
        }
    }
    Tcl_HashEntry *hPtr;
    if (name[0] == 'i' && !strcmp(name, "insert")) {
        insertMark->pos = sci_get_current_pos(SCIOBJ(sciwidgetPtr));
        Tcl_DStringFree(&dstr);
        return insertMark;
    }
    hPtr = Tcl_FindHashEntry(&markTable, name);
    if (hPtr) {
        Tcl_DStringFree(&dstr);
        return (MarkData *)Tcl_GetHashValue(hPtr);
    }
    Tcl_AppendResult(interp, "there is no mark named \"", name, "\"", (char *)NULL);
    Tcl_DStringFree(&dstr);
    return NULL;
}

void MarkMgr::Dump(Tcl_DString *dstr, int margin_id, int line1, int line2, const char *command)
{
    ScintillaObject *sciobj = SCIOBJ(sciwidgetPtr);
    Tcl_DString cmdStr;
    if (command) {
        Tcl_DStringInit(&cmdStr);
        Tcl_DStringAppendElement(&cmdStr, command);
        dstr = &cmdStr;
    }
    int mask = ~0;
    if (margin_id >= 0) {
        mask = sci_get_margin_mask(sciobj, margin_id);
    }
    while (line1 <= line2) {
        // only get the markers set for the specified line
        int markers_mask = (mask & sci_cmd(sciobj, SCI_MARKERGET, LINE2SCI(line1)));
        if (markers_mask) {
             for (int i=0; i < MARKER_MAX; i++) {
                 if (markers_mask & (1 << i)) {
                     MarkerData *mPtr = sciwidgetPtr->markerMgr->FindMarker(i);
                     if (mPtr) {
                        Tcl_DStringAppendElement(dstr, "mark");
                        Tcl_DStringAppendElement(dstr, mPtr->name);
                        // column doesn't matter here, just use zero
                        LineColumnToIndex(line1, 0, dstr);
                        if (command) {
                            Tcl_Eval(sciwidgetPtr->interp, Tcl_DStringValue(dstr));
                            Tcl_DStringSetLength(dstr, 0);
                            Tcl_DStringAppendElement(dstr, command);
                        }
                     }
                 }
             }
        }
        ++line1;
    }
    if (command) {
        Tcl_DStringFree(&cmdStr);
    }
}

//---------------------------------------------------------
// Private functions
//---------------------------------------------------------

//==========================================================================
// Create a new mark, or return an existing one
//
MarkData *MarkMgr::_createMark(const char *markname, int *newMark) {

    MarkData *markPtr;
    Tcl_HashEntry *hPtr = NULL;
    int isNew;
    bool isInsert = false;
    bool isCurrent = false;
    const char *name;

    // check if it's one of the special marks
    if (!strcmp(markname, "insert")) {
        isInsert = true;
        if (insertMark) {
            if (newMark != NULL) *newMark = 0;
            insertMark->pos = sci_get_current_pos(SCIOBJ(sciwidgetPtr));
            return insertMark; // return the one we already have
        } else if (newMark) *newMark = 1;

        name = "insert";

    } else if (!strcmp(markname, "current")) {
        isCurrent = true;
        if (currentMark) {
            if (newMark != NULL) *newMark = 0;
            return currentMark; // return the one we already have
        } else if (newMark) *newMark = 1;

        name = "current";

    } else {
        // if not, create(if needed) in the hash table
        hPtr = Tcl_CreateHashEntry(&markTable, markname, &isNew);
        if (newMark) *newMark = isNew;
        if (!isNew) {
            return (MarkData *)Tcl_GetHashValue(hPtr);
        }
        name = Tcl_GetHashKey(&markTable, hPtr);
    }
    int slen = strlen(name);
    char *buf = Tcl_Alloc(slen+1);
    strcpy(buf, name);

    // No existing entry, so create a new one
    markPtr = (MarkData *) Tcl_Alloc(sizeof(MarkData));
    markPtr->name        = buf;
    markPtr->gravityLeft = false; // Tk Text widget default is "right"
    markPtr->pos         = isInsert ? sci_get_current_pos(SCIOBJ(sciwidgetPtr)) : 0;

    if (isInsert) {
        insertMark = markPtr;
    } else if (isCurrent) {
        currentMark = markPtr;
    } else {
        // store in hash table
        Tcl_SetHashValue(hPtr, markPtr);
    }
    return markPtr;
}

//==========================================================================
// Delete a mark
//
void MarkMgr::_deleteMark(MarkData *mPtr) {

    // remove it from hash table
    Tcl_HashEntry *hPtr;
    hPtr = Tcl_FindHashEntry(&markTable, mPtr->name);
    if (hPtr) {
        Tcl_DeleteHashEntry(hPtr);
    }
    // get rid of the name field
    Tcl_Free((char *) mPtr->name);

    // now delete the mark object
    Tcl_Free((char *) mPtr);
}

//==========================================================================
// Dump a mark's data
//
void MarkMgr::_dump(MarkData *mPtr) {
    if (mPtr == NULL) {
        fprintf(stderr,     "--- DUMP all marks: --------------------------------\n");
        Tcl_HashEntry *hPtr;
        Tcl_HashSearch search;
        int n = 0;
        for (hPtr = Tcl_FirstHashEntry(&markTable, &search);
             hPtr != NULL;
             hPtr = Tcl_NextHashEntry(&search)) {
            if (++n > 1)
            fprintf(stderr, "::::::::::::::::::::::::::::::::::::::::::::::::::::\n");
            _dump((MarkData *)Tcl_GetHashValue(hPtr));
        }
        if (n > 0)
            fprintf(stderr, "::::::::::::::::::::::::::::::::::::::::::::::::::::\n");
        _dump(insertMark);
        return;
    }
    fprintf(stderr, "POINTER: %p\n", (void *)mPtr);
    fprintf(stderr, "NAME   : %s\n", mPtr->name);
    fprintf(stderr, "POS    : %d\n", mPtr->pos);
    fprintf(stderr, "GRAVITY: %s\n", mPtr->gravityLeft ? "Left" : "Right");
}

TagLocation::TagLocation(TagData *td, sptr_t start, sptr_t end)
{
    if (td) {
        tags.push_back(td);
    }
    location.startpos = start;
    location.endpos = end;
}

TagLocation::TagLocation(TagLocation *other, sptr_t end)
{
unsigned long i;

    tags.assign(other->tags.begin(), other->tags.end());
    location.startpos = other->startpos();
    location.endpos = end;
    for (i = 0; i < tags.size(); ++i) {
        tags[i]->UnregisterLocation(other); // location is about to change value, unregister it with the tag
        tags[i]->RegisterLocation(this);    // register new location with tag
    }
}

TagLocation::~TagLocation()
{
}

void TagLocation::setStartPos(sptr_t start)
{
unsigned long i;

    // remove registration of this location with tag & re-add it because it's location has been modified
    for (i = 0; i < tags.size(); ++i) {
        tags[i]->UnregisterLocation(this);    // unregister current location value
    }
    location.startpos = start;
    for (i = 0; i < tags.size(); ++i) {
        tags[i]->RegisterLocation(this);    // register new location value
    }
}

void TagLocation::setEndPos(sptr_t end)
{
unsigned long i;

    // remove registration of this location with tag & re-add it because it's location has been modified
    for (i = 0; i < tags.size(); ++i) {
        tags[i]->UnregisterLocation(this);    // unregister current location value
    }
    location.endpos = end;
    for (i = 0; i < tags.size(); ++i) {
        tags[i]->RegisterLocation(this);    // register new location value
    }
}

// Associate tag with location, return nonzero if new tag is highest priority and display update needed
int TagLocation::RegisterTag(TagData *td)
{
unsigned long i;
int highestPriority = -1;

    // the list of tags should be small, search tag vector and add data if not already in it
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"RegisterTag %p %ld-%ld tag %s\n", this, startpos(), endpos(), td->Name());
#endif
    for (i = 0; i < tags.size(); ++i) {
        if (tags[i]->priority > highestPriority)
            highestPriority = tags[i]->priority;
        if (tags[i] == td) {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"   RegisterTag %p %ld-%ld tag %s already registered\n", this, startpos(), endpos(), td->Name());
#endif
            td->RegisterLocation(this);
#ifdef ENABLE_TAG_PRIORITIES
            return 0;
#else
            return 1;
#endif
        }
    }
    tags.push_back(td);
    // notify tag of new location
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"   RegisterTag %p %ld-%ld tag %s registering location with tag\n", this, startpos(), endpos(), td->Name());
#endif
    td->RegisterLocation(this);
#ifdef ENABLE_TAG_PRIORITIES
    if (td->priority > highestPriority)
        return 1;
    else
        return 0;
#else
    return 1;
#endif
}

TagData *TagLocation::UnregisterTag(TagData *td)
{
int highestPriority = -1;
std::vector<TagData*>::iterator highest_it = tags.end();
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"UnregisterTag %s, priority %d\n", td->Name(), td->priority);
#endif
    for (std::vector<TagData*>::iterator it = tags.begin(); it != tags.end();) {
        if (*it == td) {
            // remove this tag from list
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"UnregisterTag, erasing %s from tags array\n", td->Name());
#endif
            it = tags.erase(it);
        } else {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr,"UT tag %s priority %d highestPriority %d\n", (*it)->Name(), (*it)->priority, highestPriority);
#endif
            if ((*it)->priority > highestPriority) {
                highestPriority = (*it)->priority;
                highest_it = it;
            }
            ++it;
        }
    }
    if (highestPriority != -1) {
        // we removed the highest priority tag, return the next highest
#ifdef DEBUG_TAGLOCATIONS
        fprintf(stderr,"UnregisterTag, returning %p %s\n", (void*) (*highest_it), (*highest_it)->Name());
#endif
        return *highest_it;
    }
    // either we removed the last tag or we removed a tag that doesn't affect styling
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"UnregisterTag, returning NULL\n");
#endif
    return NULL;
}

void TagLocation::dumpTags()
{
    fprintf(stderr,"TagLocation %ld-%ld tags:\n", startpos(), endpos());
    for (std::vector<TagData*>::iterator it = tags.begin(); it != tags.end();) {
        fprintf(stderr,"   %s\n", (*it)->Name());
        ++it;
    }
}

void TagMgr::RemoveStyle(TagData *td, sptr_t startpos, sptr_t endpos)
{
ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    // Called when tag removed from location and there's no lower priority tag to take it's place
    if (IsIndicator(td)) {
        //-----------------
        //--- INDICATOR ---
        //-----------------
        sci_cmd(obj, SCI_SETINDICATORCURRENT, td->IndicID());
        sci_cmd(obj, SCI_INDICATORCLEARRANGE, startpos, (endpos-startpos+1));
    } else {
        /* set style for selected positions */
        sci_cmd(obj, SCI_STARTSTYLING, startpos, STYLEMASK(sciwidgetPtr));
        sci_cmd(obj, SCI_SETSTYLING, (endpos-startpos+1), STYLE_DEFAULT);
    }
}

void TagMgr::ApplyStyle(TagData *td, sptr_t startpos, sptr_t endpos)
{
ScintillaObject *obj = SCIOBJ(sciwidgetPtr);

    // new tag was higher priority than previously applied tag, add styling for new tag to document
    if (IsIndicator(td)) {
        //-----------------
        //--- INDICATOR ---
        //-----------------
        sci_cmd(obj, SCI_SETINDICATORCURRENT, td->IndicID());
        sci_cmd(obj, SCI_INDICATORFILLRANGE, startpos, (endpos-startpos+1));
    } else {
        /* set style for selected positions */
        sci_cmd(obj, SCI_STARTSTYLING, startpos, STYLEMASK(sciwidgetPtr));
        sci_cmd(obj, SCI_SETSTYLING, (endpos-startpos+1), td->StyleID());
    }
}

void TagMgr::ApplyTag(TagData *td, sptr_t startpos, sptr_t endpos)
{
    // search location map to see if Location already registered
    std::vector<TagLocation*>::iterator it;
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"AppyTag: Tag %s, type 0x%x Searching for location %ld-%ld\n", td->name, td->Type(), startpos, endpos);
#endif
    TagLocation *tl_p = searchLocations("TagMgr Locations", tagLocations, SEARCH_ADD, startpos, endpos, &it);
#ifdef DEBUG_TAGLOCATIONS
    fprintf(stderr,"AppyTag: Searching returned %p\n", tl_p);
#endif
    while (tl_p) {
        if (tl_p->RegisterTag(td)) {
#ifdef DEBUG_TAGLOCATIONS
            fprintf(stderr, "   applying style for %s from %ld to %ld\n", td->name, tl_p->startpos(), tl_p->endpos());
#endif
            ApplyStyle(td, tl_p->startpos(), tl_p->endpos());
        }
#ifdef DEBUG_TAGLOCATIONS
        else {
            fprintf(stderr, "   not applying style for lower-priority tag %s from %ld to %ld\n", td->name, tl_p->startpos(), tl_p->endpos());
        }
#endif
        tl_p = nextLocation(tagLocations, startpos, endpos, &it);
    }
}

void TagMgr::RemoveLocation(TagLocation *tl_p)
{
    std::vector<TagLocation*>::iterator it;
    TagLocation *tl_p2 = searchLocations("TagMgr Locations", tagLocations, SEARCH_IDX, tl_p->startpos(), tl_p->endpos(), &it);
    assert(tl_p == tl_p2);  // we better find it!
    // remove location from vector
    tagLocations.erase(it);
}

void TagMgr::RemoveTag(TagData *td, sptr_t startpos, sptr_t endpos)
{
    // Notify the tag to remove itself and notify TagMgr if last tag removed from location
    td->RemoveTag(this, startpos, endpos);
}

}//--- end of namespace TkSci
